/*
 * ScrewDrive.h
 *
 *  Created on: Nov 18, 2008
 *      Author: winkler
 */

#ifndef SCREWDRIVE_H_
#define SCREWDRIVE_H_

// Robot3D files
#include <srCore/actuator/actuatorBase.h>
#include <srCore/simulationUtils.h>

// Delta3D files
#include <dtUtil/log.h>

#ifdef AUDIO_ON
	#include <dtAudio/sound.h>
#endif

//TODO: remodel the Screw Drive according to the test results of the real one (which needs to be done as well)
//TODO: provide current screw drive information such as velocity and acceleration. Currently only the values which have been set are returned


namespace srCore {


class ScrewDrive: public ActuatorBase {

public:
	ScrewDrive(const std::string name);

	virtual ~ScrewDrive();

	void registerActuator();

	void init(dtCore::RefPtr<ActuatorBase> front, dtCore::RefPtr<ActuatorBase> back,
			dtCore::RefPtr<ScrewDrive> second, bool isOnRightHandSide);

	/**
	 * \brief  Writes the parameter value in parameter
	 *
	 *		The available parameters can be retrieved by calling getParameterList()
	 * 		See also: ObjectBase::getParameter()
	 *
	 *
	 */
	inline virtual void getParameter(const std::string &name, ParameterReturnValue  parameter) {
		frontActuator->getParameter(name, parameter);
	}

	/**
	 * \brief  Sets a parameter
	 *
	 *		See also ObjectBase::getParameter()
	 *
	 */
	inline virtual void setParameter(const std::string &name, ParameterReturnValue  parameter) {

		frontActuator->setParameter(name, parameter);

	}

	/**
	 * Returns the jointID of the ODE joint in the front actuator, there are actually two of
	 * them.
	 */
	inline dJointID getJoint() {
		LOG_WARNING("The screw drive has actually two joints.");
		return frontActuator->getJoint();
	}

	/**
	 * The screw drive itself does not have a "hinge", but the two actuators it is made
	 * of have each a hinge.
	 */
	inline dJointID getJoint(int index) {
		return ((index==0) ? frontActuator->getJoint() : backActuator->getJoint());
	}

	//NULL;};

	// Maximum torque that can be applied by the motor.
	//virtual void setMaxTorque(float torque) = 0;

	// Maximum torque that the motor can stand(then switched off) to hold its position
	//virtual void setHoldingTorque(float torque) = 0;

	// Torque, that applies against the velocity vector, to slow down the joint.
	//virtual void setBrakeTorque(float torque) = 0;

	// Maximum speed the joint can reach.
	//virtual void setMaxSpeed(float v_max) = 0;

	//! Setting a torque directly to a joint, can't be greater than maxTorque.
	inline void setTorque(float torque, float deltaTime) {};

	//! Setting a velocity.
	/*! Motor tries to reach this velocity as fast as possible, but is constrained
	 * by maxTorque
	 */
	void setVel(float v, float deltaTime);

	//! Returns the current velocity of the joint.
	inline float getVel(float deltaTime) { return velocity; };

	//! Returns the current position of the joint.
	inline float getPos() {return 0;};

	//! Sets ode parameters for the joint. For more Information, see www.ode.org/ode-latest-user-guide.org.
	inline void setODEParameter(int parameter, float value) {};
	inline float getODEParameter(int parameter) { return 0.0; }; //Anne

	//! activate/deactivate the actuator
	inline void setActive(bool activate) {active = activate;}

	//! will be called in the TickLocal function of the robotActorBase class
	inline void update(float deltaSimTime) {setVel(getVel(deltaSimTime), deltaSimTime);}

	dtCore::RefPtr<ScrewDrive> secondScrewDrive;
	dtCore::RefPtr<ActuatorBase> frontActuator;
	dtCore::RefPtr<ActuatorBase> backActuator;

#ifdef	AUDIO_ON
	// Sound object corresponding to the sound made by screw drives
	dtAudio::Sound* screwSound;
#endif

protected:
#ifdef AUDIO_ON
	// Playing a sound with a pitch depending on the speed (called in setVel)
	void PlaySound(float speed);
#endif

	//! Set velocities, but they are semi-realistic (they don't consider torques etc.)
	void DriveSemiRealistic(float speedLeft, float speedRight, float speedRot, float speedTrans);

	//! Set forces, torques, etc. It doesn't work yet...
	void DriveRealistic(float speedLeft, float speedRight, float speedRot, float speedTrans);

	//! One of the screw drives is on the right-hand side
	bool isOnRightHandSide;

private:
  
	float d_m, my_Screw, l, F_div_M;

	// Stores last reported velocity
	float last_reported_velocity;

	// Reports new velocity when it deviates with the given granularity
	float velocity_report_granularity;

	// Print frequency of ODE parameters
	int print_freq_ode_params;

	// Counter used for printing frequency
	long int counter;
};

}

#endif /* SCREWDRIVE_H_ */
