/**
 * @file Thread.h
 *
 * @brief Threading for Robot3D components
 *
 * The Robot3D simulator is created within the European Replicator and Symbrion projects. The
 * Replicator project is funded by the European Commission within the work programme "Cognitive
 * Systems, Interaction, Robotics" under the grant agreement no. 216240. The Symbrion project
 * is funded within the work programme "Future and Emergent Technologies Proactive" under the
 * grant agreement no. 216342. Replicator and Symbrion concerns the design of modular robots
 * in the real world and in 3D simulation.
 *
 * This file is part of the 3D simulator Robot3D which is built on top of the open source game
 * engine Delta3D (copyright (C) 2005, BMH Associates, Inc). This library is free software; you
 * can redistribute it and/or modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version. This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * @author 	Lutz Winkler
 * @date	Apr 29, 2009
 * @project	Replicator FP7
 * @univ	Karlsruhe
 * @case	Simulation core
 */


#ifndef THREAD_H_
#define THREAD_H_

// General files
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>

namespace srCore {

/* **************************************************************************************
 * Interface of Thread
 * **************************************************************************************/

/**
 * A wrapper around the posix thread, cannot be used on embedded devices without porting
 * pthread. This is not likely on the Blackfin microcontroller which runs uClinux. The
 * latter has no memory management (no fork or malloc). It is much more likely to have a
 * cooperative scheduler than a preemptive one. Hence, do not use this class in code that
 * needs to run on the robot.
 */
class Thread {
private:
	pthread_t thread;
	static void *thread_func(void *d) {
		((Thread *)d)->task();
		//return NULL?
		return NULL;
	}
public:
	Thread(): thread(), mutex(), condition()
	{
		pthread_mutex_init(&mutex, NULL);
		pthread_cond_init(&condition, NULL);
	}

	virtual ~Thread()
	{
		printf("Thread Destructor \n");
		exit();
	}

	virtual void task() = 0;
	int start(){return pthread_create(&thread, NULL, Thread::thread_func, (void*)this);}
	int wait(){return pthread_join(thread, NULL);}
	void exit()
	{
		pthread_mutex_destroy(&mutex);
		pthread_cond_destroy(&condition);
		pthread_detach(thread);
	}
	void suspend(){pthread_cond_wait(&condition, &mutex);}
	void resume(){pthread_cond_signal(&condition);}

private:
	//! Forbid copying by copy constructor
	Thread(const Thread& thread);

	//! Forbid copying by assignment
	Thread& operator=(const Thread& thread);

	pthread_mutex_t mutex;
	pthread_cond_t condition;

};

}

#endif /* THREAD_H_ */
