//! Actuator base class.

/*!
 *
 * This is a basic actuator class. Actuators are joints, created by the Open Dynamics Engine (ODE) which connects two bodies
 *
 * @author Lutz Winkler
 *
 */

//TODO: 	Include an AMotor joint to simulate friction of the actuator

/* TODO:	What are the basic properties each actuator has in common --> adopt the interface accordingly
 *			Torque the motor can apply to the joint,
 * 			Max Velocity the joint can rotate/move
 *			(Friction of the motor)
 * 			Holding Torque of the motor
 *			Position / Velocity / Acceleration
 */

/* TODO: 	providing internal functions, such as joint distance to body 1 and 2, (kinematic) energy of the system,
 *   		oscillating movements of the bodies to provide a measurement of the stability of the simulation. These
 *   		functions should be private and a friend class which is determing the stability of the simulation is its
 *   		friend.
 */

// TODO:	Sound should be handled by the actuator classes and therefore an interface needs to be defined here.

#ifndef ACTUATOR_H_
#define ACTUATOR_H_

#include <srCore/robot/objectBase.h>

#include <dtCore/dt.h>
#include <dtCore/object.h>
#include <dtCore/globals.h>
#include <dtCore/transform.h>
#include <dtGame/gameactor.h>
#include <dtCore/refptr.h>


#include <osg/Node>
#include <osgDB/ReadFile>

#include <ode/ode.h>

#include <map>

namespace srCore {

/**
 * All actuators are derived from this class. The "hinge" of the robot, the
 * actuators that form the screw drive, etc.
 */
class ActuatorBase : public ObjectBase {

public:

	static std::map< dJointID, ActuatorBase*> jointActuatorMap;


//	static inline void registerActuator(ActuatorBase* actuator) {
//
//		jointActuatorMap.insert(std::pair<ActuatorBase*, dJointID>(actuator, actuator->getJoint()));
//
//
//	};

	//! Constructor
	ActuatorBase(const std::string name,  const std::string type) :
		ObjectBase(),
		velocity (0),
		name (name),
		type (type)
		{};

	/**
	  * \brief  Initializes the actuator
	  *
	  *
	  * \param     		bodyI   					First Body
	  * \param			bodyII 						Second Body
	  * \param			anchorX, anchorY, anchorZ	position of the joint
	  * \param			axisX, axisY, axisZ			orientation of the joint
	  *
	  *
	  */
	virtual void init(dtCore::RefPtr<BodyBase> myBody1, dtCore::RefPtr<BodyBase> myBody2,
			float anchorX, float anchorY, float anchorZ,
			float axisX, float axisY, float axisZ) {};



	virtual void registerActuator() = 0;

	// Returns the ODE jointID of the actuator (the hinge, or the wheel)
	virtual dJointID getJoint() = 0;

	// Maximum torque that can be applied by the motor.
	//virtual void setMaxTorque(float torque) = 0;

	// Maximum torque that the motor can stand(then switched off) to hold its position
	//virtual void setHoldingTorque(float torque) = 0;

	// Torque, that applies against the velocity vector, to slow down the joint.
	//virtual void setBrakeTorque(float torque) = 0;

	// Maximum speed the joint can reach.
	//virtual void setMaxSpeed(float v_max) = 0;

	//! Setting a torque directly to a joint, can't be greater than maxTorque.
	virtual void setTorque(float torque, float deltaTime) = 0;

	//! Setting a velocity.
	/*! Motor tries to reach this velocity as fast as possible, but is constrained
	 * by maxTorque
	 */
	virtual void setVel(float v, float deltaTime) = 0;

	//! Returns the current velocity of the joint.
	virtual float getVel(float deltaTime) {return velocity;};

	//! Returns the current position of the joint.
	virtual float getPos() = 0;

	//! Sets ODE parameters for the joint. For more information, see www.ode.org/ode-latest-user-guide.org.
	virtual void setODEParameter(int parameter, float value) = 0;

	//! Gets ODE parameters
	virtual float getODEParameter(int parameter) = 0;

	//! activate/deactivate the actuator
	virtual void setActive(bool activate) {}

	//! will be called in the TickLocal function of the robotActorBase class
	virtual void update(float deltaSimTme) {}

	//! returns the name of the actuator
	inline std::string getName() const {return name;}

	//!returns the type of the class
	inline std::string getType() const {return type;}

protected:
	virtual ~ActuatorBase() {};

	float velocity;

private:

	const std::string name;
	const std::string type;
};

typedef std::map< dJointID, ActuatorBase*> map_type;

typedef map_type::value_type pair_type;

}



#endif /*ACTUATOR_H_*/
