//! Basic connector class.

/*!
 *
 * This class simulates a connector in a very simple way. If two connectors
 * are close enough, have the right orientation and if both want to connect,
 * a fixed joint will be created that connects the two bodies, where the
 * connectors are attached to. If one of the connectors wants to disconnect,
 * the joint will be destroyed.
 *
 * @author Lutz Winkler
 *
 */

#ifndef CONNECTOR_BASE_H_
#define CONNECTOR_BASE_H_


#include <srCore/robot/objectBase.h>

#include <dtCore/dt.h>
//#include <dtCore/object.h>
#include <dtCore/globals.h>
#include <dtCore/transform.h>
#include <dtGame/gameactor.h>
#include <dtCore/refptr.h>
#include <dtCore/uniqueid.h>

#include <dtGame/gamemanager.h>


#include <osg/Node>
#include <osgDB/ReadFile>

#include <ode/ode.h>

#include <dtGame/gameactor.h>

#include <boost/function.hpp>

// Holds function with given robot returning void
typedef boost::function<void(srCore::RobotActorBase&)> ConnectCallback;

namespace srCore {

const float epsilon = 0.61;

/**
 * The ConnectorBase is derived from ObjectBase which comes with position, orientation,
 * etc.
 */
class ConnectorBase : public ObjectBase {

private:
	//! Function to be called when a connection event has been taken place
	ConnectCallback connection_event_function;

public:

	//! A callback function can be set to e.g. print state of connectors after an event
	void SetConnectionEventFunction(ConnectCallback func);

	//! The execution of the callback function
	void ExecuteConnectionEventFunction();

	/**
	 * The different states a connector can be in.
	 */
	enum ConnectorState {
		DISCONNECTED,												//not connected, do not wish to connect
		CHECK_CONNECTION,											//automatic connection, only during start to connect robots
		WANT_TO_CONNECT,											//not connected
		CONNECTING,													//connecting, but not connection not finished
		CONNECTED
	};



	//! String representation of connector state. Useful for debugging.
	void toString(ConnectorState state, std::string &str);

	//! Prints connectors states
	void PrintConnectorStates();

	//! Constructor
	ConnectorBase();

	/**
	* \brief Initializes the connector. Needs to be called, before using the connector.
	*
	* \param     		body   						Body, where the connector is attached to
	* \param			anchorX, anchorY, anchorZ	local position of the connector
	* \param			axisX, axisY, axisZ			local orientation of the connector
	* \param			gm							The GameManager, which handles the corresponding robot actor
	*
	*/
	void init(dtCore::RefPtr<BodyBase> body,
			float anchorX, float anchorY, float anchorZ,
			float axisX, float axisY, float axisZ);

	//! Sets scale of the display for the connector
	void setScale(osg::Vec3 scale);

	RobotActorBase* getRobot() const {return myBody->getRobot();};

	//! Returns the unique id of the connector.
	dtCore::UniqueId* getID();

	/**
	 * \brief  Writes the parameter value in parameter
	 *
	 *		The available parameters can be retrieved by calling getParameterList()
	 *		See also: ObjectBase::getParameter()
	 *
	 *
	 */
	virtual void getParameter(const std::string &name, ParameterReturnValue  parameter);

	/**
	 * \brief  Sets a parameter
	 *
	 *		See also getParameter()
	 *
	 */
	virtual void setParameter(const std::string &name, ParameterReturnValue  parameter);

	//! returns the state of the connector
	ConnectorState getState() const;

	//! returns the state of the connector
	void setState(ConnectorState state);

	//! returns the connecte robot if connected, otherwise NULL
	RobotActorBase* getConnectedRobot() const;

	//! returns the connector of the robot that is connected to. If not connected, it return NULL
	dtCore::RefPtr<ConnectorBase> getOtherConnector() const;

	void sendLocalMessage(const std::string& message);

	/** sets the connector to be ready for connection. If another connector has the right transformation towards this connector
	 * 	and is ready for connection, the will be connected together.
	 */
	void connect();


	//! Disconnects from the other connector. The joint will be destroyed.
	void disconnect();

	//! Returns the ODE body ID of the main body the connector is attached to.
	dBodyID getBodyID();

	//! Returns the Robot3D body
	BodyBase* getSRBody() {return myBody;}

	//! Returns the global position of the connector
	osg::Vec3f getTranslation();

	//! Returns the global look-at vector of the connector
	osg::Vec3f getRotation();

	//! Returns the up-vector of the connector
	inline osg::Vec3f getUpVector();

	//! Returns the look-at-vector of the connector
	inline osg::Vec3f getLookAtVector();

	//! Sets ODE-joint. Should only be called within this class while connecting.
	void setConnector(dJointID *connector);

	//! Returns the ODE-joint.
	inline dJointID getConnector() const {return this->connector;}

	//! Activate/Deactivate the joint
	virtual void setActive(bool activate) {}

	//! Will be called within RobotActorBase::TickLocal
	virtual void update(float deltaSimTme);

	//! Returns the name of the connector
	inline std::string getName() {return name;}

	//! Returns the type of the connector
	inline std::string getType() {return type;}

protected:
	virtual ~ConnectorBase();

	//! Connects with the other connector. Establish a fixed connection
	void connect(dtCore::RefPtr<ConnectorBase> theOtherConnector);

	//! Helper function for connect, if a joint can be used
	void connectByJoint(dtCore::RefPtr<ConnectorBase> theOtherConnector);

	//! Helper function for connect, if two bodies need to be actually merged
	void connectByFusing(dtCore::RefPtr<ConnectorBase> theOtherConnector);

	//! Move robots to the final position.
	void connecting(dtCore::RefPtr<ConnectorBase> theOtherConnector);

	//! Get distance to the closest connector
	osg::Vec3 GetClosestConnector(int &robotId, int &connectorId);

	std::string name;
	std::string type;
	dtCore::UniqueId* myID;
	dJointID connector;										// will be type of fixedJoint, if connected
	dtCore::RefPtr<ConnectorBase> theOtherConnector;		// NULL: if not connected, otherwise the
															// pointer to the other connector
	dtCore::RefPtr<dtGame::GameManager> gm;

	//! function to check if connectors have the right alignment to connect to each other
	virtual inline bool checkAngle(float angle, float epsilon) const
	{
		if ((angle<(-1+epsilon)) || (angle>1-epsilon) || ((angle>-epsilon) && (angle<epsilon))) return true;
		else return false;

	}

private:
	//! Position and orientation of the connector
	osg::Vec3f paramAbsTranslation, paramAbsOrientation, paramUpVector;

	//! Connection state (disconnected to connected and intermediate stages)
	ConnectorState connectorState;

	//! Initialisation boolean
	bool firstRun;

	//! Forbid copying by copy constructor
	ConnectorBase(const ConnectorBase& conn);

	//! Forbid copying by assignment
	ConnectorBase& operator=(const ConnectorBase& conn);

};

}


#endif /*CONNECTOR_BASE_H_*/
