
//! General body class, which all other body classes have to inherit.

/*!
 *
 * Describes general properties of a body in the simulation environment.
 *
 * @author Lutz Winkler
 *
 */


#ifndef BODY_H_
#define BODY_H_

#include <iostream>

#include <dtCore/dt.h>
#include <dtCore/object.h>
#include <dtCore/globals.h>
#include <dtCore/transform.h>
#include <dtGame/gameactor.h>
#include <dtCore/refptr.h>


#include <osg/Node>
#include <osg/Geode>
#include <osg/Group>
#include <osg/PositionAttitudeTransform>
#include <osgDB/ReadFile>
#include <osg/Material>
#include <osg/NodeVisitor>

#include <ode/ode.h>

#include <srCore/simulationUtils.h>

#define _I(i,j) I[(i)*4+(j)]

namespace srCore {



class BodyBase;
class osgBodyNodeVisitor;

class osgBodyTransform : public osg::PositionAttitudeTransform {

public:
	osgBodyTransform(dtCore::RefPtr<BodyBase> srBody);

	dtCore::RefPtr<BodyBase> getMyBody();
	virtual void accept(osgBodyNodeVisitor& bnv);
	virtual void traverse (osgBodyNodeVisitor &bnv);

protected:
	~osgBodyTransform();

private:
	dtCore::RefPtr<BodyBase> myBody;

};

typedef std::vector< dtCore::RefPtr<osgBodyTransform> > BodyTransformNodeList;

class osgBodyNodeVisitor : public osg::NodeVisitor
{
public:
	osgBodyNodeVisitor(BodyBase* root, TraversalMode tm=TRAVERSE_NONE);
	osgBodyNodeVisitor (TraversalMode tm=TRAVERSE_NONE);
 	osgBodyNodeVisitor (VisitorType type, TraversalMode tm=TRAVERSE_NONE);

 	virtual void apply(osgBodyTransform &node);
// 	virtual void accept(osgBodyNodeVisitor& bnv);

 	void addChildToBodyRoot (BodyBase* child);
 	BodyBase* returnBodyRoot ();

 	virtual	~osgBodyNodeVisitor ();

private:
 	BodyBase* parent;

	//! Forbid copying by copy constructor
	osgBodyNodeVisitor(const osgBodyNodeVisitor& visitor);

	//! Forbid copying by assignment
	osgBodyNodeVisitor& operator=(const osgBodyNodeVisitor& visitor);


};

class RobotActorBase;

class BodyBase : public dtCore::Object
{

	friend class ConnectorBase;
	friend class SimulationUtils_;

public:

	static std::map< dBodyID, BodyBase*> odeBodyMap;

	//! Constuctor
	BodyBase();
	
	//! Register the body to the simulation, i.e. map the BodyBase object to the ODE body ID
	void registerBody();

	//! Merges the collision shape and the masses of two bodies together. theBody will become child of this body and the physics will be disabled of theBody
	void mergeBody(BodyBase* theBody);

	//! Adds the body to the scene, makes it visible and activates the physical behaviour
	void addToScene(); //dtCore::Scene*& scene,  osg::Group*& osgGroup);

	//! Scaling of the model to Robot3D environment
	void setScale(float scaleX, float scaleY, float scaleZ);

	//! Show the collision geometry of the body.

	/*! Activate this to check manually/optically if collision geometry fits to the model.
	 *  Will also be used to indicate which robot is under manual control.
	 */
	void renderCollisionGeometry(bool renderIt);
	void renderCollisionGeometry2(bool renderIt);
	void colorBody(osg::Vec4 color);
	void colorCollisionShape(osg::Vec4 color);

	/*! Initializes the body. Needs to be called before using the body
	 *
	 *	@param name 					The name of the body
	 *	@param robot 					The robot the body belongs to
	 *	@param x,y,z 					The relative position of the body within the robot
	 *	@param heading, roll, pitch		The relative orientation of the body within the robot
	 *
	 */
	void init(	std::string name, RobotActorBase* robot,
				float x, float y, float z,
				float heading, float roll, float pitch);

	//! Return the body Id, which represents the geomCollide and geomCategory bits
	long getID();

	//! returns the global position of the body
	osg::Vec3f getTranslation() const;

	//! returns the global Rotation of the body ()
	osg::Vec3f getRotation() const;

	//! returns the scaling factor between the physical and the geometrical world
	osg::Vec3f getScale() const;

	//! returns the Transform of the Body
	dtCore::Transform getTransform() const {return this->xform;}

	//! sets the Transform of the Body
	virtual void setTransform(dtCore::Transform &transform);

	//! sets the position of the Body. Only use that while physics is deactivated.
	virtual void setTranslation(osg::Vec3f pos);

	//! sets the rotation of the Body. Only use that while physics is deactivated.
	virtual void setRotation(osg::Vec3f hrp);

	//! Returns the Robot the body belongs to
	RobotActorBase* getRobot() const {return myRobot;};

	//! Returns the current collision model of the body
	dtCore::RefPtr<osg::Node> getCollisionModel() const {return collisionModel;};

	//! Sets the collision model
	void setCollisionModel(dtCore::RefPtr<osg::Node> collModel);
	void setCollisionModel(osgBodyTransform* collModel);
	osgBodyTransform* getCollisionModel();


	unsigned long int getCollisionBits() const {return collCategoryBit;};
	void setCollisionBits(unsigned long int bit);

	//! Set the mass to this subOrgansim, consisting of this body and all its children
	void setSubOrgBodyMass(dMass *mass) {dMassSetParameters(&subOrgBodyMass, mass->mass,
															mass->c[0], mass->c[1], mass->c[2],
															mass->_I(0,0), mass->_I(1,1), mass->_I(2,2),
															mass->_I(1,2), mass->_I(1,3), mass->_I(2,3));};

	//! Returns the mass to this subOrgansim, consisting of this body and all its children
	void getSubOrgBodyMass(dMass &mass) {dMassSetParameters(&mass, subOrgBodyMass.mass,
															subOrgBodyMass.c[0], subOrgBodyMass.c[1], subOrgBodyMass.c[2],
															subOrgBodyMass._I(0,0), subOrgBodyMass._I(1,1), subOrgBodyMass._I(2,2),
															subOrgBodyMass._I(1,2), subOrgBodyMass._I(1,3), subOrgBodyMass._I(2,3));};
	
	//! Returns true, when the body has a contact with the floot or other obstacles
	bool hasContact() {return bodyHasContact;};

protected:
	virtual ~BodyBase();

	//! To be set true, when that body has collided with an obstacle or the floor (currently by the simulationUtils class)
	void setContact(bool contact) {bodyHasContact = true;};

	long int myID;
	RobotActorBase* myRobot;
	float scaleX, scaleY, scaleZ;

	//! collision model of merged bodies (e.g. the through the coupling devices merged bodies)
	//dtCore::RefPtr<osg::PositionAttitudeTransform>
	dtCore::RefPtr<osgBodyTransform> collisionModel;
	dMass subOrgBodyMass;

private:

	void setModelTranslation(osg::Vec3 pos);
	void setAllModelTranslations(dtCore::Object &object, std::string theOtherObjectName, osg::Vec3 pos);

	// transformation of the body
	dtCore::Transform xform;

	//! If true, collision geometry will be rendered
	bool renderGeom;

	bool bodyHasContact;

	//!collision model of the single body
	dtCore::RefPtr<dtCore::Object> bodyCollisionModel;

	unsigned long int collCategoryBit;

	//! body color
	osg::Vec4 body_color;

	//! Forbid copying by copy constructor
	BodyBase(const BodyBase& body);

	//! Forbid copying by assignment
	BodyBase& operator=(const BodyBase& body);

};

typedef std::map< dBodyID, BodyBase*> body_map_type;
typedef body_map_type::value_type body_pair_type;

}

//#include "../robotActorBase.h"


#endif /*BODY_H_*/
