/*
 * @file export.h
 *
 * The Robot3D simulator is created within the European Replicator and Symbrion projects. The
 * Replicator project is funded by the European Commission within the work programme "Cognitive
 * Systems, Interaction, Robotics" under the grant agreement no. 216240. The Symbrion project
 * is funded within the work programme "Future and Emergent Technologies Proactive" under the
 * grant agreement no. 216342. Replicator and Symbrion concerns the design of modular robots
 * in the real world and in 3D simulation.
 *
 * This file is part of the 3D simulator Robot3D which is built on top of the open source game
 * engine Delta3D (copyright (C) 2005, BMH Associates, Inc). This library is free software; you
 * can redistribute it and/or modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version. This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * @author    Lutz Winkler
 * @date      2008
 */
#ifndef EXPORT_H
#define EXPORT_H

/************************************************************************************************
 *                      Defines to run the simulator differently
 ***********************************************************************************************/

/*
 * Use spaces instead of tabs in define statements. Text editors and html browsers (with an online
 * SVN viewer) can then display proper alignment.
 */

/**
 * The robots can be put in the same scene as the GUI components, which subsequently need to be
 * filtered out in the view of the robot cameras. Set USE_SEPARATE_GUI to decouple the view of the
 * user from the view of the cameras on the robots.
 */
#define USE_SEPARATE_GUI                        0

/**
 * The Delta3D simulator allows for distribution of the game actor logic (the software that runs on
 * an individual robot) to be distributed over machines by HLA. This introduces remote ticks besides
 * local ticks. With the separation of robot-specific software from the simulation code by YARP,
 * the use of HLA is rendered obsolete.
 */
#define WITH_HLA                                0

/**
 * FreeRTOS is the operating system that runs on the Cortex microcontrollers in the robot. Currently,
 * a dual-core BlackFin microprocessor is integrated as the main processing unit. It is not clear if
 * a FreeRTOS interface is still necessary. It will do nothing.
 */
#define WITH_FREE_RTOS                          0

/**
 * The robots can propagate by a screwdrive. Those screws can be simulated by wheels. To use wheels
 * instead of the screwdrive class, turn the NO_WHEELS macro off.
 */
#define NO_WHEELS                               1

/**
 * In certain cases it might not be necessary to move the robots in a physically realistic way. By
 * turning physics off, no forces will be calculated within the robots or between the robots. It will
 * reduce jitter. It will not remove the effect of gravity or collisions with the walls or the ground.
 * It is recommended however to reduce the time scale instead, if the robots exhibit awkward behaviour.
 */
#define NO_PHYSICS                              0

/**
 * @todo What does this macro?
 */
#define WITH_CONFIG_PARAMS                      0

/**
 * @todo What does this macro?
 */
#define WITH_ORGANISM_STRING_CHECK              0

/**
 * This macro enables reading an organism layout from an XML file. However, because there is a bug
 * it is now disabled by default:
    #0  0x014a3785 in ?? () from /lib/tls/i686/cmov/libc.so.6
    #1  0x0138dcdb in std::basic_string<char, std::char_traits<char>,
    std::allocator<char> >::basic_string(char const*, std::allocator<char>
    const&) () from /usr/lib/libstdc++.so.6
    #2  0x02dcd782 in GetConfig::processOrganismString() ()
    from /usr/lib/libsrCore.so.0
    #3  0x02dcd63c in GetConfig::readConfigFile(std::string) ()
    from /usr/lib/libsrCore.so.0
    #4  0x02de0f5e in
    SimulationEntryPoint::OnStartup(dtGame::GameApplication&) ()
      from /usr/lib/libsrCore.so.0
    #5  0x00400915 in dtGame::GameApplication::Config() ()
    from /usr/lib/libdtGame.so
    #6  0x080495c0 in main ()
 *
 * It is not a problem on my system, so I couldn't debug it for now. Please, if you use this code
 * consider debugging it and enabling it again.
 */
#define WITH_ORGANISM_STRING					0


/**
 * Stack smashing error when adding LED

#6  0x025e7c94 in ?? () from /usr/lib/libsrBody.so.0
#7  0x025e6884 in srCore::Led::Led(dtCore::RefPtr<srCore::BodyBase>, osg::Vec3f, osg::Vec3f, std::string, std::string, bool) () from /usr/lib/libsrBody.so.0
#8  0x01a91bf2 in srRobot::RobotKIT::Initialize() () from /usr/lib/libsrRobot.so.0
 *
 */
#define WITH_LED								0

/************************************************************************************************
 *                      Macro interpretation
 ***********************************************************************************************/

#if WITH_CONFIG_PARAMS == 0
#undef WITH_CONFIG_PARAMS
#endif

#if USE_SEPARATE_GUI == 0
//#warning "USE_SEPARATE_GUI off is deprecated!" cannot select robot elseway!
#undef USE_SEPARATE_GUI
#endif

#if WITH_HLA == 0
#undef WITH_HLA
#endif

#if WITH_FREE_RTOS == 0
#undef WITH_FREE_RTOS
#endif

#if NO_WHEELS == 0
#undef NO_WHEELS
#endif

#if NO_PHYSICS == 0
#undef NO_PHYSICS
#endif

#if WITH_ORGANISM_STRING_CHECK == 0
#undef WITH_ORGANISM_STRING_CHECK
#endif

#if WITH_ORGANISM_STRING == 0
#undef WITH_ORGANISM_STRING
#endif

#if WITH_LED == 0
#undef WITH_LED
#endif

/************************************************************************************************
 *                      Others
 ***********************************************************************************************/

#define DELTA3D_2_4
#define DELTA3D_2_3

#include <iostream>

/**
 * This is modelled from the DT_EXPORT macro found in dtCore/export.h.
 * We define another due to conflicts with using the DT_EXPORT while
 * trying to import Delta3D symbols.  The VFST_GAME_LIBRARY macro should be used
 * in front of any classes that are to be exported from the terrain library.
 * Also note that VFST_GAME_LIBRARY should be defined in the compiler
 * preprocessor #defines.
 */
#if defined(_MSC_VER) || defined(__CYGWIN__) || defined(__MINGW32__) || defined( __BCPLUSPLUS__)  || defined( __MWERKS__)
#  ifdef ROBOT_LIBRARY
#    define ROBOT_EXPORT __declspec(dllexport)
#  else
#    define ROBOT_EXPORT __declspec(dllimport)
#  endif
#else
#  define ROBOT_EXPORT
#endif

#endif
