/* -*-c++-*-
 * Delta3D Open Source Game and Simulation Engine
 * Copyright (C) 2005, BMH Associates, Inc.
 *
 * GUI Component
 *
 * @author Lutz Winkler
 * @author Adrian Friebel
 */

#ifndef SIMULATION_GUI_COMPONENT_H_
#define SIMULATION_GUI_COMPONENT_H_

// Robot3D files
#include <srCore/export.h>
#include <srCore/sensor/cameraSensorBase.h>
#include <srCore/robot/robotActorBase.h>
#include <srCore/robotSelectionComponent.h>
#include <srCore/arenaComponent.h>
#include <srCore/simulationEvents.h>

// Delta3D files
#include <dtGame/gmcomponent.h>
#include <dtDAL/gameeventmanager.h>
#include <dtGUI/ceuidrawable.h>

// CEGUI
#include <CEGUI/CEGUI.h>
#include <CEGUI/CEGUIWindow.h>

// TODO: cleaning up the source and divide into several classes.
// TODO: Another gui window (maybe Qt?), where more information is visible


#define GUICONTROLMAXTEXTSIZE 100

namespace srCore {

enum SelectionMenu
{
	NONE,
	SENSOR,
	ACTUATOR,
	CONNECTOR
};

/**
 * Class: SimulationGUIComponent
 *
 * SimulationGUIComponent is one of the core component of the Robot3D simulator. It is built on top
 * of a Delta3D game component. It processes user events like mouse clicks in menus and in the scene. It
 * also organizes all CEGUI windows, etc.
 */
class ROBOT_EXPORT SimulationGUIComponent : public dtGame::GMComponent
{
public:
	SimulationGUIComponent(const std::string &name, dtCore::DeltaWin &win, dtCore::Keyboard &keyboard, dtCore::Mouse &mouse);
	virtual void ProcessMessage(const dtGame::Message& message);
	void TickGUI();

	void SetupGUI(dtCore::DeltaWin &win, dtCore::Keyboard &keyboard, dtCore::Mouse &mouse);

	dtGUI::CEUIDrawable* GetGUIDrawable() { return mGUI.get(); };

	void setCameraMode(bool camstate);

	dtCore::Mouse* GetMouse() { return mMouse.get(); }

	dtCore::Keyboard* GetKeyboard() { return mKeyboard.get(); }

	void SetMouse(dtCore::Mouse* mouse) { mMouse = mouse; }

	void SetKeyboard(dtCore::Keyboard* keyboard) { mKeyboard = keyboard; }

protected:
	virtual ~SimulationGUIComponent();

private:
	//! Disable copying
	SimulationGUIComponent(const SimulationGUIComponent& comp);

	//! Disable copying by assignment
	void operator=(const SimulationGUIComponent& comp);

	void sendMessageToAll(dtDAL::GameEvent &event);
	bool robotMenuHandler(const CEGUI::EventArgs& eventArgs);		//handles click on menuitem and build up robot selection menu
	bool robotSelectionHandler(const CEGUI::EventArgs& eventArgs);	//handles robot selection via drop down menu
	bool addRobotHandler(const CEGUI::EventArgs& eventArgs);
	bool sensorMenuHandler(const CEGUI::EventArgs& eventArgs);		//handles click on button and build up sensor selection menu
	bool sensorSelectionHandler(const CEGUI::EventArgs& eventArgs);	//handles sensor selection via drop down menu
	bool connectorMenuHandler(const CEGUI::EventArgs& eventArgs);		//handles click on button and build up connector selection menu
	bool connectorSelectionHandler(const CEGUI::EventArgs& eventArgs);	//handles connector selection via drow down menu
	bool actuatorMenuHandler(const CEGUI::EventArgs& eventArgs);		//handles click on menuitem and build up actuator selection menu
	bool actuatorSelectionHandler(const CEGUI::EventArgs& eventArgs);	//handles actuator selection via drow down menu
	bool mouseInGUI(const CEGUI::EventArgs& eventArgs);				//stops motionmodel if mouse over GUI element
	bool mouseInSim(const CEGUI::EventArgs& eventArgs);				//starts motionmodel if mouse not over GUI element
	bool changeTimeScale(const CEGUI::EventArgs& eventArgs);			//handles Time Scale changes
	bool changeObservationMode(const CEGUI::EventArgs& eventArgs);	//handles observation mode changes (free cam vs. follow robot mode)
	bool changeSensorActivity(const CEGUI::EventArgs& eventArgs);	//handles changes of Sensor activity
	bool changeSensorVisibility(const CEGUI::EventArgs& eventArgs);	//handles changes of Sensor visibility
	bool actuatorStopActivity(const CEGUI::EventArgs& eventArgs);	//set velocity of selected actuator to zero
	bool actuatorFwdActivity(const CEGUI::EventArgs& eventArgs);	//increase velocity of selected actuator
	bool actuatorBwdActivity(const CEGUI::EventArgs& eventArgs);	//decrease velocity of selected actuator
	void buildRobotPropertyMap();									//builds up robot property display in mRPanel
	void buildSensorPropertyMap();									//builds up sensor property display in mRPanel
	void buildConnectorPropertyMap();								//builds up connector property display in mRPanel
	void buildActuatorPropertyMap();									//builds up actuator property display in mRPanel

	bool freeCamState;
	unsigned int sensorID;
	unsigned int connectorID;
	unsigned int actuatorID;
	unsigned int robotID;

	// Devices
	dtCore::RefPtr<dtCore::DeltaWin> mDeltaWin;					//the default window
	dtCore::RefPtr<dtCore::Mouse> mMouse;						//the mouse for the gui
	dtCore::RefPtr<dtCore::Keyboard> mKeyboard;					//the keyboard for the gui

	// Events
	dtCore::RefPtr<dtDAL::GameEvent> mSelectRobot;
	dtCore::RefPtr<dtDAL::GameEvent> mMouseInGUI;
	dtCore::RefPtr<dtDAL::GameEvent> mMouseInSim;
	dtCore::RefPtr<dtDAL::GameEvent> mChangeCameraMode;

	// GUI Elements
	dtCore::RefPtr<dtGUI::CEUIDrawable> mGUI;
	dtCore::DeltaWin *mWin;
	CEGUI::WindowManager *wm;

	CEGUI::Window *mSheet;					//CEGUI sheet
	CEGUI::Window *mPanel;					//invisible overall panel
	CEGUI::Window *mBckgr;					//invisible background panel
	CEGUI::Window *mRPanel;					//right sided panel dedicated to show robot information

	CEGUI::Window *mSensorSelectorIt;		//item, click event opens robots sensor selecting popup menu
	CEGUI::Window *mSensorSelectorPM;		//robots sensor selecting popup menu
	std::vector<CEGUI::Window*> mSensor;		//items in menubar, click event selects sensor

	CEGUI::Window *mActuatorSelectorIt;		//item, click event opens robots actuator selecting popup menu
	CEGUI::Window *mActuatorSelectorPM;		//robots actuator selecting popup menu
	std::vector<CEGUI::Window*> mActuator;	//items in menubar, click event selects actuator

	CEGUI::Window *mConnectorSelectorIt;		//item, click event opens robots connector selecting popup menu
	CEGUI::Window *mConnectorSelectorPM;		//robots connector selecting popup menu
	std::vector<CEGUI::Window*> mConnector;	//items in menubar, click event selects connector

	CEGUI::Window *mBar;						//menubar on top dedicated to selecting miscellaneous
	CEGUI::Window *mRobotSelectorIt;			//item in menubar, click event opens robot selecting popup menu
	CEGUI::Window *mRobotSelectorPM;			//robot selecting popup menu
	std::vector<CEGUI::Window*> mRobot;		//items in menubar, click event selects robot
	//	   CEGUI::Window *mAddRobotDialog;			//dialog for collecting information for robot creation
	CEGUI::Window *mButtonSensorActivity;    //Button for changing activity state of sensor
	CEGUI::Window *mButtonSensorVisibility;  //Button for switching on/off visibility of the sensor
	CEGUI::Window *mButtonActuatorFwd;		//Button for increasing speed of actuator
	CEGUI::Window *mButtonActuatorBwd;		//Button for decreasing speed of actuator
	CEGUI::Window *mButtonActuatorStop;		//Button for stopping speed of actuator
	CEGUI::Window *mButtonTimeScaleUp;		//Button for increasing the Time scale
	CEGUI::Window *mButtonTimeScaleDown;		//Button for decreasing the Time scale
	CEGUI::Window *mButtonObservationMode;	//Button for switching between camera robot follow mode & free camera movement mode


	// Time Scale Value
	std::string tSValue;

	// Info text
	CEGUI::Window *mTimeScaleText;
	CEGUI::Window *mFrameRateText;
	CEGUI::Window *mSensorValueText;
	CEGUI::Window *mRobotText;
	CEGUI::Window *mRobotPosRotText;
	CEGUI::Window *mInfoText;
	CEGUI::Window *mActuatorInfoText;

	// Camera/sensor window
	osg::ref_ptr<osg::Image> image;
	bool hideRenderTexture;

	float simTime;
	bool saveScreenShot;

	// Selection Menu
	SelectionMenu selectedMenuItem;

	bool destroying;
};

}

#endif /*SIMULATION_GUI_COMPONENT_H_*/
