/**
 * @file parameterReturnValue.h
 * @brief 
 *
 * This file is created at Almende B.V. It is open-source software and part of the Common 
 * Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools, ranging from 
 * thread pools and TCP/IP components to control architectures and learning algorithms. 
 * This software is published under the GNU Lesser General Public license (LGPL).
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author 	Anne C. van Rossum
 * @date	Nov 22, 2010
 * @project	Replicator FP7
 * @company	Almende B.V.
 * @case	
 */


#ifndef PARAMETERRETURNVALUE_H_
#define PARAMETERRETURNVALUE_H_

// General files
#include <string>
#include <typeinfo>

// Robot3D files
#include <srCore/singleton.h>

namespace srCore {

/* **************************************************************************************
 * Interface of parameterReturnValue
 * **************************************************************************************/

/**
 * ParameterReturnValue is a small helper class for type-value pairs. In this case it is
 * represented by a "string"-"void pointer" pair. Methods that operate on this class can
 * be found in PRVUtils.
 */
class ParameterReturnValue {
public:
	ParameterReturnValue(const std::string typeName, void* returnValue):
		type (typeName),
		pValue (returnValue) {}

	//! Copy by copy constructor
	ParameterReturnValue(const ParameterReturnValue& value);

	//! Destructor
	~ParameterReturnValue() {}

	//! A type, like typeid(osg::Vec4*).name()
	const std::string type;

	//! A void pointer, needs to be cast to type
	void* pValue;
private:
	//! Forbid copying by assignment
	ParameterReturnValue& operator=(const ParameterReturnValue& prv);
};

/*
 * Function: MAKE_PARAMETER_VALUE
 *
 * Returns a default ParameterReturnValue class, in which the string gets the C++ name
 * of the given typename.
 *
 * Returns:
 *   ParameterReturnValue
 */
template<class T>
inline ParameterReturnValue MAKE_PARAMETER_VALUE(T *t) {
	return ParameterReturnValue(typeid(t).name(), t);
};

template<class T>
inline ParameterReturnValue * MAKE_PRV(T *t) {
	return new ParameterReturnValue(typeid(t).name(), t);
};


/**
 * A collection of different functions for ParameterReturnValue%s.
 */
class PRVUtils_ {
public:
	// Default constructor
	PRVUtils_();

	// Default destructor
	virtual ~PRVUtils_();

	/**
	 * Is the parameter of the given type?
	 * Example:
	 * 	 if (!PRVUtils::Instance()->isCorrectParameterType(p, typeid(float*).name())) return;
	 */
	bool isCorrectParameterType(const ParameterReturnValue &parameter,
			const std::string &type);

private:
	//! Forbid copying by copy constructor
	PRVUtils_(const PRVUtils_& prvutils);

	//! Forbid copying by assignment
	PRVUtils_& operator=(const PRVUtils_& prvutils);
};


/**
 * PRVUtils is a singleton of PRVUtils_ and stores all kind of
 * parameters for the Robot3D simulator. It is not possible to conceal its singleton
 * implementation entirely, it needs to be called by PRVUtils::Instance()
 */
typedef srCore::Singleton<srCore::PRVUtils_> PRVUtils;

} // end of namespace

#endif /* PARAMETERRETURNVALUE_H_ */
