/**
 * @file pluginEntryPointBase.cpp
 *
 * This file pluginEntryPointBase.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * @author  Anne C. van Rossum
 * @date    Dec 2, 2009
 * @project Replicator FP7
 * @company Almende B.V.
 * @case
 */

#ifndef PLUGINENTRYPOINTBASE_H_
#define PLUGINENTRYPOINTBASE_H_

// Delta3D files
#include <dtGame/gameapplication.h>
#include <dtGame/gmcomponent.h>

// Robot3D files
#include <srCore/export.h>

namespace srCore {

/**
 * This class functions exactly the same as the Delta3D class dtGame::GameEntryPoint. It is meant
 * to be inherited by classes that want to add additional functionality to the simulator, for
 * example in the form of components. Examples of such components:
 * <ul>
 * <li>A component that adds coloured traces to the trajectories of the robots.
 * <li>A component that adds colours to the robots.
 * <li>A component that adds an UART interface to each robot in the simulator.
 * </ul>
 * For example, inherit PluginEntryPointBase by TraceComponent.
 *
 * The two functions that are needed (you define them in TraceComponent.cpp):
 * <ul>
 * <li>extern "C" ROBOT_EXPORT PluginEntryPoint* CreatePluginEntryPoint();
 * <li>extern "C" ROBOT_EXPORT void DestroyPluginEntryPoint();
 * </ul>
 * They need to have the code to create a TraceComponent and destroy it again. Of course, it is
 * possible to have a Singleton of the derived TraceComponent class. However, PluginEntryPointBase
 * is not a singleton, it can be inherited in multiple libraries.
 */
class ROBOT_EXPORT PluginEntryPointBase {
public:
	/**
	 * Constructor
	 */
	PluginEntryPointBase() {;}

	/**
	 * Destructor
	 */
	virtual ~PluginEntryPointBase() {;}

	/**
	 * A virtual function that makes this class abstract (like a Java interface). This function
	 * needs to be implemented by your derived class. It will be the function that is called
	 * by Robot3D right after CreatePluginEntryPoint. It will have a reference to the Delta3D
	 * dtGame::GameApplication class, so it is possible to get hands on all possible components
	 * and controllers via that class.
	 *
	 * As example, take a look how the srInterface library is organised.
	 *
	 * @param application A Delta3D GameApplication class through which everything can be obtained
	 */
	virtual void StartPlugin(dtGame::GameApplication &application) = 0;
};


}

#endif /* PLUGINENTRYPOINTBASE_H_ */
