#ifndef CAMERASENSORBASE_H_
#define CAMERASENSORBASE_H_

// Open scene graph files
#include <osgDB/WriteFile>
#include <osg/Camera>
#include <osg/Texture2D>
#include <osg/Material>
#include <osg/ShapeDrawable>
#include <osgGA/StateSetManipulator>
#include <osgViewer/Viewer>
#include <osgViewer/ViewerEventHandlers>
#include <osgShadow/ShadowedScene>
#include <osgShadow/SoftShadowMap>

// Delta3D files
#include <dtGame/gameapplication.h>
#include <dtGame/gamemanager.h>
#include <dtABC/application.h>
#include <dtCore/deltawin.h>

// Robot3D files
#include <srCore/parameterReturnValue.h>
#include <srCore/sensor/sensorBase.h>

//TODO: camera based sensors need to be simulated completly on the GPU, as the data is already there and this would bring a huge speed up.

namespace srCore {

//! screen shot call back class
class CameraScreenShotCallback : public osg::Camera::DrawCallback
{
public:
	CameraScreenShotCallback(glType type) :
		image (new osg::Image),
		mTakeScreenShotNextFrame(false),
		type (type)

	{}

	void setRequestDataToTrue()
	{
		mTakeScreenShotNextFrame  = true;
	}

	// overwrite operator() of DrawCallback to retrieve the right image
	// image will be retrieved after the simulation step, right after it is written
	// to the viewport. Only then, it is possible to get the corresponding image, as
	// many cameras use the same viewport.
	virtual void operator()(const osg::Camera &camera) const
	{

		if(mTakeScreenShotNextFrame)
		{
			mTakeScreenShotNextFrame = false;


			int x = static_cast<int>(camera.getViewport()->x());
			int y = static_cast<int>(camera.getViewport()->y());
			unsigned int width = static_cast<unsigned int>(camera.getViewport()->width());
			unsigned int height = static_cast<unsigned int>(camera.getViewport()->height());

			//	std::cout << "cam viewport: " << x << ", " << y << ", " << width << ", " << height << std::endl;

			switch (type) {

			case DEPTH_CAMERA:
				image->allocateImage(width, height, 1, GL_DEPTH_COMPONENT, GL_FLOAT);		//GL_RGB, GL_UNSIGNED_BYTE
				image->readPixels(x, y, width, height, GL_DEPTH_COMPONENT, GL_FLOAT);		//GL_DEPTH_COMPONENT
				break;

			case RGB_CAMERA:
				image->allocateImage(width, height, 1, GL_RGB, GL_UNSIGNED_BYTE);		//GL_RGB, GL_UNSIGNED_BYTE
				image->readPixels(x, y, width, height, GL_RGB, GL_UNSIGNED_BYTE);		//GL_DEPTH_COMPONENT
				break;

			default:
				break;

			}
			//			unsigned char *dst;
			//			glReadPixels(0, 0, width, height, GL_DEPTH_COMPONENT, GL_FLOAT, dst);

		}
	}

	osg::ref_ptr<osg::Image> getOSGImage() const {
		return image;
	}

	void requestData()
	{
		setRequestDataToTrue();
	}

protected:
	//! The internal image
	osg::ref_ptr<osg::Image> image;

	//! Control boolean for when it becomes to time to take a picture again
	mutable bool  mTakeScreenShotNextFrame;

	//! The type of camera (i.e. RGB)
	glType type;
};



//! Camera sensor base class.

/*!
 *
 * Used for all sensor devices, which need to retrieve information via osg::camera.
 *
 * @author Lutz Winkler
 *
 */
class CameraSensorBase : public SensorBase
{

	friend class SimulationGUIComponent;

public:

	/**
	 * @brief  Initialises the camera sensor base. This is the parent class of most of the sensors.
	 *
	 * @param			attachedTo					Body, where the camera is attached to
	 * @param     		pos			   				Local position of the camera
	 * @param			orient 						Local orientation of the camera
	 * @param			name						Name of the camera
	 * @param			type						Type of camera (i.e. RGB) in string format
	 * @param			glType						OpenGL type of camera
	 * @param			active						Activate it directly or not
	 */
	CameraSensorBase(dtCore::RefPtr<BodyBase> attachedTo,
			osg::Vec3f pos, osg::Vec3f orient, std::string name, std::string type, glType cameraType, bool active);

	//! Actually retrieve data
	virtual void evaluate(ParameterReturnValue srv) = 0;

	//! Returns the last image taken by mScreenShotTaker
	osg::ref_ptr<osg::Image> getRawData();

	//! Request to take a new image after the simulation step
	void requestData();

	//! Will be called in RobotActorBase::TickLocal()
	void update(float deltaSimTme);

	virtual void getNewSensorValue() = 0;

	/**
	 * @brief  Sets a parameter by name in the given field
	 */
	virtual void setParameter(const std::string &name, ParameterReturnValue parameter);


	/**
	 * @brief  Writes the parameter value in parameter
	 *
	 * The available parameters can be retrieved by calling getParameterList()
	 * See also: ObjectBase::getParameter()
	 */
	void getParameter(const std::string &name, ParameterReturnValue parameter);

	//! Retrieves parameter by pointer, type does not need to be set beforehand
	ParameterReturnValue* getParameter(const std::string &name);

	//! sets up the camera sensor and put it into the scene
	void setup();

	//! Set rendering order
	void setRenderOrder(int renderOrder) {view->SetRenderOrder(renderOrder);}

	int getRenderOrder() {return view->GetRenderOrder();}

protected:
	//! Destructor deallocates all pointer fields
	virtual ~CameraSensorBase();

	//! Minimum and maximum distance
	float minDistance, maxDistance;

	//! Restrictions of viewing box etc.
	float frustumLeft, frustumRight, frustumUp, frustumDown;

	//! The "width" angle
	float flareAngle;

	//! Default color
	osg::Vec4 clearColor;

	//! OpenGL cameraType
	glType cameraType;

	//! The actual sensor as Transformable
	dtCore::RefPtr<dtCore::Transformable> sensor;

	//! Resolution (how many beams are shot i.e.)
	osg::Vec2 resolution;

	//! Reference to window
	dtCore::DeltaWin* win;

	//! Reference to view (there might be multiple in a window)
	dtCore::View* view;

	//! The camera
	osg::ref_ptr<osg::Camera> mOsgCamera;

	//! The image to be operating upon internally
	osg::ref_ptr<osg::Image> sensorImage;

	//! The thing that gets an image for us
	osg::ref_ptr<CameraScreenShotCallback> mScreenShotTaker;

private:
	//! Disable copying
	CameraSensorBase(const CameraSensorBase& cam);

	//! Disable copying by assignment
	void operator=(const CameraSensorBase& cam);

};

}

#endif /*CAMERASENSORBASE_H_*/
