//! Ray sensor base class.

/*!
 *
 * Used for all sensor devices, which need to retrieve information via Ray Trace.
 *
 * @author Lutz Winkler
 *
 */

#ifndef RAYSENSORBASE_H_
#define RAYSENSORBASE_H_

#include "sensorBase.h"
#include <dtGame/gamemanager.h>
#include <dtABC/application.h>
#include <dtCore/deltawin.h>

#include <osgDB/WriteFile>
#include <osg/Camera>

#include <dtCore/isector.h>
#include <dtCore/scene.h>
#include <dtCore/uniqueid.h>

#include <dtUtil/noiseutility.h>

namespace srCore {

class RaySensorBase : public SensorBase

{
public:

	/**
	 * \brief  Initializes the sensor
	 *
	 * \param			gm							GameManager, which handles the corresponding robot
	 * \param			attachedTo					Body, where the sensor is attached to
	 * \param     		pos			   				Local position of the sensor
	 * \param			orient 						Local orientation of the sensor
	 * \param			flareAngle					flare angle of the sensor
	 * \param			resolution					resolution of the sensor
	 * \param			name						name of the sensor
	 *
	 *
	 */
	RaySensorBase(dtCore::RefPtr<BodyBase> attachedTo,
			osg::Vec3f pos, osg::Vec3f orient, const std::string &name, std::string type, bool active);




	//! sends command to retrieve data
	virtual void evaluate(ParameterReturnValue srv) = 0;

	//! returns the last image taken by mScreenShotTaker
	inline osg::ref_ptr<osg::Image> getRawData() {return sensorImage;}

	//! calculates sensor values
	void requestData();

	/**
	 * \brief  Sets a parameter
	 *
	 *		See also getParameter()
	 *
	 */
	virtual void setParameter(const std::string &name, ParameterReturnValue  parameter);


	/**
	 * \brief  Writes the parameter value in parameter
	 *
	 *		The available parameters can be retrieved by calling getParameterList()
	 *		See also: ObjectBase::getParameter()
	 *
	 *
	 */
	void getParameter(const std::string &name, ParameterReturnValue parameter);

	//sets up the sensor
	void setup();

	//! Will be called in RobotActorBase::TickLocal()
	void update(float deltaSimTme);

	virtual void getNewSensorValue() {};

	void DisplayRays(bool flag = true) {displayRays = flag;};


protected:
	virtual ~RaySensorBase();

	/**
	 *  Generates a ray, given by start and end position and calculates a value, depending on the distance
	 *  of the object that reflected the ray and its orientation towards the ray.
	 */
	virtual osg::Vec3f getNextIntersection(osg::Vec3f startPos, osg::Vec3f endPos, float *z);

	float minDistance, maxDistance;
	float frustumLeft, frustumRight, frustumUp, frustumDown;
	float flareAngle;

	osg::ref_ptr<osg::Image> sensorImage;
	osg::Vec2d resolution;

	int currentSensorValue;

	unsigned int pixelPos;
	unsigned int x, y;

	dtUtil::Noise2f* mNoise;



    /** OSG node that will contain debug geometry */
	osg::Geode* mGeode;
	osg::Geometry* lineGeometry;
	osg::Vec3Array* vArray;
	osg::Vec4Array* colors;
	bool displayRays;
	bool renderedRays;

	int timer;

};





}

#endif /*RAYSENSORBASE_H_*/
