
//! Sensor base class.

/*!
 *
 * Used for all sensor devices.
 *
 * @author Lutz Winkler
 *
 */
#ifndef BASICSENSOR_H_
#define BASICSENSOR_H_

// Open scene graph
#include <osgViewer/Viewer>
#include <osgDB/WriteFile>
#include <osg/Camera>

// Delta3D
#include <dtCore/refptr.h>
#include <dtCore/camera.h>
#include <dtCore/tripod.h>
#include <dtCore/view.h>
#include <dtCore/deltawin.h>
#include <dtCore/transform.h>
#include <dtCore/scene.h>
#include <dtGame/gameactor.h>
#include <dtGame/gamemanager.h>
#include <dtABC/application.h>

// Robot3D
#include <srCore/robot/objectBase.h>
#include <srCore/robot/robotActorBase.h>
#include <srCore/parameterReturnValue.h>

namespace srCore {


class SensorBase : public ObjectBase		//, public osg::Referenced
{
public:

	/**
	 * \brief  Initializes the sensor
	 *
	 * \param			gm							GameManager, which handles the corresponding robot
	 * \param			attachedTo					Body, where the sensor is attached to
	 * \param     		translation	   				Local position of the sensor
	 * \param			orientation					Local orientation of the sensor
	 * \param			name						name of the sensor
	 * \param			type						type of the sensor
	 * \param			active						initial active state of the sensor
	 *
	 */
	SensorBase(dtCore::RefPtr<BodyBase> attachedBody = NULL,
			osg::Vec3f translation = osg::Vec3f(2,0,0), osg::Vec3f orientation= osg::Vec3f(0,0,0),
			std::string name = "noName", std::string type = "noType", glType cameraType = NO_CAMERA,
			bool active = true);

	void makeVisible(bool visible);

	inline glType getCameraType() const {return cameraType;}


	/**
	 * \brief  Writes the parameter value in parameter
	 *
	 *		The available parameters can be retrieved by calling getParameterList()
	 *		See also: ObjectBase::getParameter()
	 *
	 *
	 */
	virtual inline void getParameter(const std::string &name, ParameterReturnValue parameter) {
		ObjectBase::getParameter(name, parameter);
	}

	virtual inline ParameterReturnValue* getParameter(const std::string &name) {
		return ObjectBase::getParameter(name);
	}

	/**
	 * \brief  Sets a parameter
	 *
	 *		See also getParameter()
	 *
	 */
	virtual inline void setParameter(const std::string &name, ParameterReturnValue  parameter) {

		ObjectBase::setParameter(name, parameter);
	}

	friend std::ostream& operator<< (std::ostream& theStream, const SensorBase& sensor);

	//! sends command to retrieve data
	virtual void evaluate(ParameterReturnValue srv) = 0;

	//! Request to get new data
	virtual void requestData() = 0;

	//! Returns type of sensor data
	inline std::string getReturnValueType() const {return sensorReturnValueType;}

	//! sets up sensor
	virtual void setup() = 0;


	inline RobotActorBase* getRobot() const {return myBody->getRobot();};
	dtGame::GameManager* getGameManager() const;

protected:

	virtual ~SensorBase();

	virtual void getSensorValue(std::ostream& theStream) const = 0;


	glType cameraType;

	dtCore::RefPtr<dtCore::Scene> mScene;
	std::string sensorReturnValueType;

	bool isActiveSensor;


};

}

//------------------------------------------------------------------------------------------



//
//
//class FunctionSensorBase: public SensorBase
//{
//public:
//	FunctionSensorBase(){}
//
//	virtual void Evaluate(SensorReturnValue srv) = 0;
//
//protected:
//	~FunctionSensorBase(){}
//
//};

#endif /*BASICSENSOR_H_*/

