/*
 * simulationUtils.h
 *
 *  Created on: Nov 21, 2008
 *      Author: winkler
 */

#ifndef SIMULATIONUTILS_H_
#define SIMULATIONUTILS_H_

#include <dtGame/messagetype.h>
#include <dtGame/messagefactory.h>
#include <dtGame/gameapplication.h>
#include <dtCore/scene.h>
#include <dtCore/view.h>
#include <dtCore/light.h>

#include <dtCore/uniqueid.h>
#include <typeinfo>

#include <osgShadow/ShadowedScene>
#include <osgShadow/SoftShadowMap>
#include <osgShadow/ShadowMap>
#include <osgViewer/View>

#include <osgDB/WriteFile>
#include <osg/Camera>

#include <srCore/export.h>
#include <srCore/singleton.h>

#include <ode/ode.h>

// TODO: Everything I did not know where to put, ended up here. Restructuring here the code might be necessary.

namespace srCore {

/*
 * Enumeration: glType
 *
 * OpenGL type of camera (sensor) of basically two types: DEPTH_CAMERA and RGB_CAMERA.
 */
enum glType
{
	NO_CAMERA,
	DEPTH_CAMERA,
	RGB_CAMERA

};

/**
 * Helper class for screen shot call back derived from DrawCallback from open scene
 * graph.
 */
class ScreenShotCallback : public osg::Camera::DrawCallback {
public:
	ScreenShotCallback(glType type, osg::Viewport* viewport) :
		image (new osg::Image),
		filename (""),
		viewport (viewport),
		mTakeScreenShotNextFrame(false),
		type (type)
	{}

	/**
	 * Enables a screenshot for the next frame.
	 */
	void setRequestDataToTrue();

	/**
	 * Overwrite operator() of DrawCallback to retrieve the right image. The image will
	 * be retrieved after the simulation step, right after it is written to the viewport.
	 * Only then, it is possible to get the corresponding image, as many cameras use the
	 * same viewport.
	 */
	void operator()(const osg::Camera &camera) const;

	//! Takes screenshot
	void takeScreenShot(const std::string filename);

protected:
	//! Internal image
	osg::ref_ptr<osg::Image> image;

	//! File name to write to
	std::string filename;

	//! Viewport
	osg::Viewport* viewport;

	//! When to take a next screenshot
	mutable bool  mTakeScreenShotNextFrame;

	//! Type of camera (i.e. RGB)
	glType type;

private:
	//! Forbid copying by copy constructor
	ScreenShotCallback(const ScreenShotCallback& cback);

	//! Forbid copying by assignment
	ScreenShotCallback& operator=(const ScreenShotCallback& cback);
};

class ROBOT_EXPORT SimulationMessageType : public dtGame::MessageType
{
	DECLARE_ENUM(SimulationMessageType);

	enum commType
	{
		NONE,
		GLOBAL,
		ORGANISM,
		LOCAL
	};

public:

	static const SimulationMessageType COMM_BUS;

	static void RegisterMessageTypes(dtGame::MessageFactory& factory);

protected:

	SimulationMessageType( const std::string &name,
			const std::string &category,
			const std::string &description,
			const unsigned short messageId) :
				dtGame::MessageType(name, category, description, messageId)
	{
		AddInstance(this);
	}

	virtual ~SimulationMessageType() { }
};

class CommMessage : public dtGame::Message
{
public:
	CommMessage();
	CommMessage(SimulationMessageType::commType type, dtCore::UniqueId* id,
			const std::string& message);
protected:
	virtual ~CommMessage();
public:
	void setBusID(dtCore::UniqueId* uniqueId );
	dtCore::UniqueId* getBusID() const;

	void setMessage(const std::string &message );
	const std::string& getMessage() const;

	void setCommType(SimulationMessageType::commType type);
	const SimulationMessageType::commType& getCommType() const;

private:
	dtCore::UniqueId* busID;
	std::string message;
	SimulationMessageType::commType type;

	//! Forbid copying by copy constructor
	CommMessage(const CommMessage& msg);

	//! Forbid copying by assignment
	CommMessage& operator=(const CommMessage& msg);
};


/*
 * Function: minAbs
 *
 * Calculates min(abs(t1), abs(t2))
 *
 * Returns:
 *   minimum of two values, both taken absolute
 */
template<class T>
inline T minAbs(T t1, T t2){
	if (fabs(t1)<fabs(t2)) return fabs(t1);
	else return fabs(t2);
}

//! Returns max(abs(t1), abs(t2))	--	used for robotUniKarl movement principle
template<class T>
inline T maxAbs(T t1, T t2){
	if (fabs(t1)>fabs(t2)) return fabs(t1);
	else return fabs(t2);
}

/**
 * A collection of different functions, mainly to set global parameters in the Robot3D
 * simulator.
 */
class SimulationUtils_ {
public:

	SimulationUtils_();
	virtual ~SimulationUtils_();

	void NearCallback(void* data, dGeomID o1, dGeomID o2);

	void setSceneFile(std::string file) {sceneFile = file;};
	std::string getSceneFile() {return sceneFile;};

	inline bool getDetailedGraphics() {return detailedGraphics;};
	inline void setDetailedGraphics(bool set) {detailedGraphics = set;};
	inline bool getWithShadows() {return withShadows;};
	inline void setWithShadows(bool set) {withShadows = set;};
	inline std::string getStageProjectPath() {return stageProjectPath;};
	inline void setStageProjectPath(std::string path) {stageProjectPath = path;};
	inline bool getConnectorsAsJoints() {return connectorsAsJoints;};
	inline void setConnectorsAsJoints(bool asJoints) { connectorsAsJoints = asJoints;};
	inline void setWithWheels(bool robotsWithWheels) { withWheels = robotsWithWheels;};
	inline bool getWithWheels() {return withWheels;};
	inline int getCollisionGeometryDetails() {return collisionGeometryDetails;};
	inline void setCollisionGeometryDetails(int details) {collisionGeometryDetails = details;};

	inline dJointGroupID getContactJointGroupID() { return mContactJointGroupID; };

	inline int getLaserRayNodeMask() { return laserRayNodeMask; };

	inline int getCollisionNodeMask() { return collisionNodeMask; };

	inline int getGeometryDetailsNodeMask() { return geometryDetailsNodeMask; };

	/**
	 * Return path where CEGUI resides. This is the parent path, so it should have a "gui"
	 * directory as subdirectory. It should not point to the "gui" directory itself.
	 */
	inline std::string getCEGUIPath() {return ceguiPath;};

	/**
	 * Set the path where the CEGUI configuration files reside. This is the parent path, so it
	 * should have a "gui" directory as subdirectory. It should not point to the "gui"
	 * directory itself.
	 */
	inline void setCEGUIPath(std::string path) {ceguiPath = path;};

	/**
	 * Returns the CEGUI scheme file.
	 */
	inline std::string getCEGUISchemeFile() {return ceguiSchemeFile;};

	/**
	 * Sets the CEGUI scheme file
	 */
	inline void setCEGUISchemeFile(std::string path) {ceguiSchemeFile = path;};

	osg::ref_ptr<osgShadow::ShadowedScene> getShadowScene();

	bool fileExists(std::string fileName);

private:
	//! Forbid copying by copy constructor
	SimulationUtils_(const SimulationUtils_& utils);

	//! Forbid copying by assignment
	SimulationUtils_& operator=(const SimulationUtils_& utils);

	// Masks
	int geometryDetailsNodeMask;

	int collisionNodeMask;

	int laserRayNodeMask;

	// Graphics
	bool detailedGraphics;
	bool withShadows;
	osg::ref_ptr<osgShadow::ShadowedScene> shadowedScene;
	int collisionGeometryDetails;

	// Physical constitution
	bool connectorsAsJoints;
	bool withWheels;
	dJointGroupID mContactJointGroupID;

	// Paths / file names
	std::string sceneFile;
	std::string stageProjectPath;
	std::string ceguiPath;
	std::string ceguiSchemeFile;

	// Initialisation
	bool startup;

};

class SimulationUtils_;

/**
 * SimulationUtils is a singleton of SimulationUtils_ and stores all kind of
 * parameters for the Robot3D simulator. It is not possible to conceal its singleton
 * implementation entirely, it needs to be called by SimulationUtils::Instance()
 */
typedef srCore::Singleton<srCore::SimulationUtils_> SimulationUtils;

} // end of srCore namespace

#endif /* SIMULATIONUTILS_H_ */
