/*
 * xmlConfig.h
 *
 * Parser for <config.xml> file which contains the configuration for the Robot3D
 * robot simulator. In the WITH_MOTES modus it can be used for simulating a WSN.
 *
 *  Created on: Nov 25, 2008
 *      Author: winkler
 */

#ifndef XMLPARSER_H_
#define XMLPARSER_H_
#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMDocumentType.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMNodeIterator.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMText.hpp>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>

#include <osg/Geometry>
#include <osg/Matrix>

#include <dtUtil/matrixutil.h>
//#include <srCore/organismTree.h>

#include <map>

#include <string>
#include <stdexcept>
#include <vector>


#include "export.h"

// Error codes

enum {
	ERROR_ARGS = 1,
			ERROR_XERCES_INIT,
			ERROR_PARSE,
			ERROR_EMPTY_DOCUMENT
};

struct SensorInitData
{
	char* type;
	char* name;
	osg::Vec3f position;
	osg::Vec3f orientation;
	float apertureAngle;
	osg::Vec2f resolution;
	int bodyPart;
};


//make class for that to make sure releasing all the stuff:
class RobotInitData {
public:
	RobotInitData(): name(NULL), type(NULL), pos(osg::Vec3f(0,0,0)), quat(osg::Quat(0,0,0,1)),
			connectorStates(0), stateTestFlags(0), jointAngles(std::vector<float>()),
			sensors(){};
	char* name;
	char* type;
	osg::Vec3f pos;
	osg::Quat quat;
	int connectorStates;
	short stateTestFlags;
	std::vector<float> jointAngles;
	std::vector<SensorInitData> sensors;

	//! Copy by copy constructor
	RobotInitData(const RobotInitData& data);

	//! Copy by assignment
	RobotInitData& operator=(const RobotInitData& data);

};

struct ControllerInitData
{
	ControllerInitData(): name(NULL), robot(NULL), type(NULL), parameter(NULL),
				pos() {};
	char* name;
	char* robot;
	char* type;
	char* parameter;
	osg::Vec3f pos;
#ifdef WITH_CONFIG_PARAMS
	std::map<std::string, std::string> paramMap;
#endif
};


struct Terminal
{
	char* id;
	RobotInitData robot;
	ControllerInitData controller;
};

struct OrganismInitData
{
	char* structure;
	osg::Vec3f pos;
	osg::Quat quat;
	std::vector<Terminal> terminals;
};


/**
 * A struct that stores name and path of plugins that can be loaded by Robot3D.
 */
struct PluginInitData
{
	char* name;
	char* path;
};

//! Class for reading and parsing the scene file
class ROBOT_EXPORT GetConfig
{
public:
	GetConfig();
	~GetConfig();
	void readConfigFile(std::string) throw(std::runtime_error);

	std::vector<RobotInitData> getRobots() const {return robots;};
	std::vector<ControllerInitData> getControllers() const {return controllers;};
	std::vector<PluginInitData> getPlugins() const {return plugins;};

	char* getMap() const {return map;};

	bool getWithShadows() const {return withShadows;};
	bool getWithDetailedModels() const {return withDetailedModels;};
	bool getConnectorsAsJoints() const {return connectorsAsJoints;};
	bool getRobotsWithWheels() const  {return withWheels;};
	int  getCollisionGeometryDetails() const {return collisionGeometryDetails;};

	//   char *getOptionA() { return m_OptionA; };
	//   char *getOptionB() { return m_OptionB; };

private:
	//! Forbid copying by copy constructor
	GetConfig(const GetConfig& config);

	//! Forbid copying by assignment
	GetConfig& operator=(const GetConfig& config);
	
	void structureParser(char* structure);

#ifdef WITH_ORGANISM_STRING
	void processOrganismString(); // throw(std::runtime_error);
#endif

	void computePosQuat(std::string type, std::string lastType, int robot, int si, int lastSide, int rot, osg::Vec3f &pos, osg::Quat &quat);
	std::vector<osg::Matrix> getSideMatrix(float d);
	
	xercesc::XercesDOMParser *m_ConfigFileParser;
	bool withShadows, withDetailedModels, connectorsAsJoints, withWheels;
	int collisionGeometryDetails;

	char* map;
	OrganismInitData organism;
	std::vector<RobotInitData> robots;
	std::vector<ControllerInitData> controllers;
	std::vector<PluginInitData> plugins;

	// Internal class use only. Hold Xerces data in UTF-16 SMLCh type.

	XMLCh* TAG_scene;
	XMLCh* ATTR_activateShadows;
	XMLCh* ATTR_loadDetailedModels;
	XMLCh* ATTR_connectorsAsJoints;
	XMLCh* ATTR_withWheels;
	XMLCh* ATTR_collisionDetails;
	XMLCh* ATTR_map;
	XMLCh* ATTR_enableAudio;

#ifdef WITH_MOTES
	XMLCh* TAG_mote;
	XMLCh* ATTR_moteName;
	XMLCh* ATTR_moteType;
#endif

	XMLCh* TAG_organism;
	XMLCh* ATTR_org_structure;
	XMLCh* ATTR_org_x;
	XMLCh* ATTR_org_y;
	XMLCh* ATTR_org_z;
	XMLCh* ATTR_org_h;
	XMLCh* ATTR_org_p;
	XMLCh* ATTR_org_r;
	XMLCh* ATTR_org_quat_x;
	XMLCh* ATTR_org_quat_y;
	XMLCh* ATTR_org_quat_z;
	XMLCh* ATTR_org_quat_w;
	
	XMLCh* TAG_terminal;
	XMLCh* ATTR_termId;
	
	
	XMLCh* TAG_robot;
	XMLCh* ATTR_robotName;
	XMLCh* ATTR_robotType;
	XMLCh* ATTR_x;
	XMLCh* ATTR_y;
	XMLCh* ATTR_z;
	XMLCh* ATTR_h;
	XMLCh* ATTR_p;
	XMLCh* ATTR_r;
	XMLCh* ATTR_connected;
	XMLCh* ATTR_q0;
	XMLCh* ATTR_q1;

	XMLCh* TAG_sensor;
	XMLCh* ATTR_sensorName;
	XMLCh* ATTR_sensorType;
	XMLCh* ATTR_position_x;
	XMLCh* ATTR_position_y;
	XMLCh* ATTR_position_z;
	XMLCh* ATTR_orientation_x;
	XMLCh* ATTR_orientation_y;
	XMLCh* ATTR_orientation_z;
	XMLCh* ATTR_apertureAngle;
	XMLCh* ATTR_resolution_x;
	XMLCh* ATTR_resolution_y;
	XMLCh* ATTR_bodyPart;

	XMLCh* TAG_controller;
	XMLCh* ATTR_controllerName;
	XMLCh* ATTR_controllerType;
	XMLCh* ATTR_robotToControl;
	XMLCh* ATTR_controllerParameter;

	XMLCh* TAG_plugin;
	XMLCh* ATTR_pluginName;
	XMLCh* ATTR_pluginPath;

	XMLCh* TAG_param;
	XMLCh* ATTR_name;
	XMLCh* ATTR_value;
};
#endif /* XMLPARSER_H_ */

