/**
 * file: Agent.h
 *
 *
 */
#ifndef AGENT
#define AGENT

// General files
#include <map>

// Delta3D files
#include <dtCore/refptr.h>
#include <dtDAL/gameevent.h>
#include <dtDAL/gameeventmanager.h>
#include <dtGame/gmcomponent.h>
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/robot/robotActorBase.h>
#include <srCore/export.h>
#include <srCore/simulationEvents.h>
#include <srCore/simulationUtils.h>

using namespace srCore;

namespace srInterface {

/*
 * Class: Agent
 *
 * An agent descends from a Delta3D game component. This can be seen as a wrapper that exposes
 * a robot to the world outside of the Robot3D simulator.
 */
class ROBOT_EXPORT Agent : public dtGame::GMComponent {

	friend class SimulationArenaComponent;

public:
	// Inner class for simulation messages
	// Todo: Check if this is still needed
	struct SimMessage {

		SimMessage(): data(), type(), port(-1) {};

		// String with data
		std::string data;
		// Type
		SimulationMessageType::commType type;
		// Port (what port?)
		int port;

	};

	//! Constructor for a Controller that controls one specific robot
	Agent(const std::string &name, RobotActorBase* robotActor);

	//! colors the collisionMesh of the Robot. This function will not be available on the real robot!
	void __colorRobot(osg::Vec4 color);

	/** moves the robot forward in 2D mode - physics and collision hadling still activated
	 *@param name velocity: velocity vector, robot will be moved by ds = dt * velocity in its own coordinate system, only for this time step
	 */
	void move2D(osg::Vec2 velocity);

	/** rotates the robot in 2D mode - physics and collision hadling still activated
	 *@param name angular_v: angular velocity, robot will be rotate by delta alpha = dt * angular_v.
	 */
	void rotate2D(float angular_v);

	//! returns the collision ID
	int getRobotID() {return robot->getID(); };

	virtual void ProcessMessage(const dtGame::Message &message);

	inline void setParamMap(std::map<std::string, std::string> paramMap) { this->paramMap = paramMap; } ;

protected:

	/**
	 * Temporary function to retrieve the robot's position. A robot should only have access
	 * to its position by an appropriate sensors. In the case of absolute positions, there
	 * should be a GPS sensor on board (with the proper noise etc.) Relative positions should
	 * be derived from the wheels/servos.
	 */
	osg::Vec3f getRobotPosition() const {
		return robot->getTranslation();
	};

	/**
	 * Temporary function to retrieve the robot's orientation. A robot should only have access
	 * to its orientation by an appropriate sensors. This should be an accelerometer for example
	 * (with the proper noise etc.)
	 */
	osg::Vec3f getRobotRotation() const {
		return robot->getRotation();
	};

	/**
	 * Temporary function to set the robot's position. See getRobotPosition.
	 */
	void setRobotTranslation(const osg::Vec3& pos);

	/**
	 * Temporary function to set the robot's orientation. See getRobotRotation.
	 */
	void setRobotRotation(const osg::Vec3& rot);



	//get robot state:

	ParameterReturnValue getSensorValue(dtCore::RefPtr<SensorBase> sensor);
	float getActuatorPosition(dtCore::RefPtr<ActuatorBase> actuator);
	float getActuatorVelocity(dtCore::RefPtr<ActuatorBase> actuator);
	//	   float getActuatorAcceleration(dtCore::RefPtr<ActuatorBase> actuator);

	int getConnectorState(dtCore::RefPtr<ConnectorBase> connector);


	/**
	 * Connect a robot to another robot. If there is no connector on a robot (except for this
	 * very one) close enough, only WANTS_TO_CONNECT is set.
	 */
	void connect(dtCore::RefPtr<ConnectorBase> connector);

	/**
	 * Disconnect a robot from another robot.
	 */
	void disconnect(dtCore::RefPtr<ConnectorBase> connector);

	//	void setActuatorPosition(dtCore::RefPtr<ActuatorBase> actuator, float value);

	void setActuatorVelocity(dtCore::RefPtr<ActuatorBase> actuator, float value);
	//	   void setActuatorAcceleration(ActuatorBase* actuator, float value);

	//! Returns the amount of sensors
	inline size_t getSensorSize() const {return robot->getSensors().size();};

	//! Returns the amount of actuators (excluding the connectors)
	inline size_t getActuatorSize() const {return robot->getActuators().size();};

	//! Returns the amount of connectors
	inline size_t getConnectorSize() const {return robot->getConnectors().size();};

	//! Returns the amount of objects
	inline size_t getObjectSize() const {return robot->getObjects().size();};

	//! Returns sensor with the given id that is attached to the robot
	inline dtCore::RefPtr<SensorBase> getSensor(unsigned int id) const {
		if (id < robot->getSensors().size()) return robot->getSensors()[id];
		else {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "Sensor with id " << id << " doesn't exist!";
			LOG_WARNING(msg.str());
			return NULL;
		}
	}

	//! Returns sensor with the given name that is attached to the robot
	inline dtCore::RefPtr<SensorBase> getSensor(const std::string name) const {
		for (unsigned int i=0; i<robot->getSensors().size(); i++)
			if (robot->getSensors()[i]->getName() == name) return robot->getSensors()[i];

		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Sensor with name " << name << " doesn't exist";
		LOG_WARNING(msg.str());
		return NULL;
	}

	//! Returns actuator with the given id that is attached to the robot
	inline dtCore::RefPtr<ActuatorBase> getActuator(unsigned int id) const {
		if (id < robot->getActuators().size()) return robot->getActuators()[id];
		else {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "Actuator with id " << id << " doesn't exist!";
			LOG_WARNING(msg.str());
			return NULL;
		}
	}

	//! Returns actuator with the given name that is attached to the robot
	inline dtCore::RefPtr<ActuatorBase> getActuator(const std::string name) const {

		for (unsigned int i=0; i<robot->getActuators().size(); i++)
			if (robot->getActuators()[i]->getName() == name) return robot->getActuators()[i];

		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Actuator with name " << name << " doesn't exist!";
		LOG_WARNING(msg.str());
		return NULL;
	}

	//! Returns actuator with the given id that is attached to the robot
	inline dtCore::RefPtr<ConnectorBase> getConnector(unsigned int id) const {
		if (id < robot->getConnectors().size()) return robot->getConnectors()[id];
		else {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "Connector with id " << id << " doesn't exist!";
			LOG_WARNING(msg.str());
			return NULL;
		}
	}

	//! Returns actuator with the given name that is attached to the robot
	inline dtCore::RefPtr<ConnectorBase> getConnector(const std::string name) const {

		for (unsigned int i=0; i<robot->getConnectors().size(); i++)
			if (robot->getConnectors()[i]->getName() == name) return robot->getConnectors()[i];

		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Connector with name " << name << " doesn't exist!";
		LOG_WARNING(msg.str());
		return NULL;
	}

	//! Returns object with the given id that is attached to the robot
	inline dtCore::RefPtr<ObjectBase> getObject(unsigned int id) const {
		if (id < robot->getObjects().size()) return robot->getObjects()[id];
		else {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "Object with id " << id << " doesn't exist!";
			LOG_WARNING(msg.str());
			return NULL;
		}
	}

	//! Returns object with the given name that is attached to the robot
	inline dtCore::RefPtr<ObjectBase> getObject(const std::string name) const {

		for (unsigned int i=0; i<robot->getObjects().size(); i++)
			if (robot->getObjects()[i]->getName() == name) return robot->getObjects()[i];

		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Object with name " << name << " doesn't exist!";
		LOG_WARNING(msg.str());
		return NULL;
	}

	inline float getDeltaSimTime() const {return deltaSimTime;};


	/**
	 * Simple helper method to fire a game event.  This method creates the game event
	 * message and sends it on to the Game Manager for processing.
	 * @param event The game event to fire.
	 */
	//	   void sendMessageToAll(dtDAL::GameEvent &event);
	//	   void sendMessageToRobot(dtDAL::GameEvent &event, unsigned int robotID);

	// message stuff:
	void sendGlobalMessage(const std::string& msg);

	/**
	 * Sends a local message, local in the sense that it will only be directed towards the
	 * robot that is attached at the current robot with the given connectorID
	 * @param msg 			The message to be sent
	 * @param connectorID 	The identifier of the connector on this side
	 */
	void sendLocalMessage(const std::string& msg, unsigned int connectorID);

	/// @see sendLocalMessage(msg, connectorID)
	void sendLocalMessage(const std::string& msg, const std::string& connectorName);

	/**
	 * Sends a "kind of" local message, namely to all the robot modules that are somehow
	 * attached to the organism.
	 */
	void sendOrganismMessage(const std::string& msg);

	bool receivedNewMsg() const {return !lastMessages.empty();};
	std::vector<SimMessage> getLastMessages() const {return lastMessages;};
	inline void clearMsgList() { lastMessages.clear(); };

	//	//! Destructor
	~Agent();

	void OnRemovedFromGM();

	inline dtCore::UniqueId* getBusID() const {return robot->getBusID();}


	std::map<std::string, std::string> paramMap;

private:

	/**
	 * The robot field is private. A controller should not have direct access to the position
	 * of the robot. This would mean that it can magically move a robot, not by using the wheels,
	 * but by simply setting its x,y coordinates. There might be reasons that you nevertheless
	 * want to have access to a robot from a controller. This needs to be done via YarpUtil, this
	 * allows Robot3D to initiate a warning and for example remove that possibility in a
	 * contest setting.
	 */
	dtCore::RefPtr<RobotActorBase> robot;

	void sendMessage(const std::string& msg, SimulationMessageType::commType type, dtCore::UniqueId* commID);

	float deltaSimTime;

	std::vector<SimMessage> lastMessages;

	bool manuallySetAngularVel;

};

} // end of namespace

#endif

