/**
 * @file InterfaceUtil.h
 * @brief Utility class for starters
 *
 * This file is created at Almende B.V. It is open-source software and part of the Common
 * Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools. Ranging from
 * thread pools, and TCP/IP components to control architectures and learning algorithms.
 * This software is published under the GNU Lesser General Public license (LGPL).
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author  Anne C. van Rossum
 * @date    Mar 26, 2010
 * @project Replicator FP7
 * @company Almende B.V.
 */


#ifndef INTERFACEUTIL_H_
#define INTERFACEUTIL_H_

#include <cassert>

// Robot3D files
#include <srCore/arenaComponent.h>
#include <srCore/robot3dComponent.h>

// Delta3D inline files
#include <dtGame/gamemanager.inl>

#include <srInterface/Agent.h>

/**
 * \addtogroup Utilities
 * @{
 */

/* **************************************************************************************
 * Interface
 * **************************************************************************************/

/**
 * Returns true if the actor can be cast as a RobotActorBase class. First it has to be cast from ActorProxy
 * to a GameActorProxy. Then that what GetGameActor on the proxy returns is cast into a RobotActorBase. If
 * all this succeeds the operator() returns true.
 */
class CheckForRobots {
public:
	bool operator()(dtDAL::ActorProxy& proxy) {
		dtGame::GameActorProxy *robot_proxy = dynamic_cast<dtGame::GameActorProxy*>(&proxy);
		if (robot_proxy == NULL) return false;
		srCore::RobotActorBase *actor = dynamic_cast<srCore::RobotActorBase*>(&robot_proxy->GetGameActor());
		bool result = (actor != NULL);
		return result;
	}
};

/**
 * Routines in this class can be called directly. See this class as a wrapper of all
 * important classes that you might need.
 */
class InterfaceUtil {
public:
	/**
	 * When the plugin is called, there is a SimulationArenaComponent constructed. This component
	 * provides an easy way of adding and deleting actors to the scene. It does - of course - not
	 * recognize actors that are known only by the plugin. It can however also be used to reposition
	 * actors.
	 */
	static srCore::SimulationArenaComponent* GetArena() {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		srCore::SimulationArenaComponent* arena = static_cast<srCore::SimulationArenaComponent*>(
				GA->GetGameManager()->GetComponentByName(srCore::SimulationArenaComponent::DEFAULT_NAME));
		return arena;
	}

	static srCore::Robot3DComponent* GetRobot3DComponent() {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		srCore::Robot3DComponent* comp = static_cast<srCore::Robot3DComponent*>(
				GA->GetGameManager()->GetComponentByName(srCore::Robot3DComponent::DEFAULT_NAME));
		return comp;
	}

	/**
	 * Returns a robot. Can for example be used to know the sender of a message.
	 */
	static srCore::RobotActorBase* GetRobot(dtCore::UniqueId &id) {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		dtGame::GameActorProxy *proxy = GA->GetGameManager()->FindGameActorById(id);
		assert (proxy != NULL);
		srCore::RobotActorBase *actor = dynamic_cast<srCore::RobotActorBase*>(&proxy->GetGameActor());
		assert (actor != NULL);
		return actor;
	}

	/**
	 * Returns a robot by name
	 */
	static srCore::RobotActorBase* GetRobot(const std::string &name) {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		dtGame::GameActorProxy *proxy;
		GA->GetGameManager()->FindActorByName(name, proxy);
		if (proxy == NULL) {
			LOG_WARNING("Actor proxy cannot be found");
		}
		srCore::RobotActorBase *actor = dynamic_cast<srCore::RobotActorBase*>(&proxy->GetGameActor());
		if (actor == NULL) {
			LOG_WARNING("Proxy found, but cannot be casted properly");
		}
		return actor;
	}

	/**
	 * Returns all the robots (derived classes from RobotActorBase).
	 */
	static void GetAllRobots(std::vector<srCore::RobotActorBase*> &robots) {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		std::vector<dtDAL::ActorProxy*> actors;
		CheckForRobots functor;
		GA->GetGameManager()->FindActorsIf<CheckForRobots>(functor, actors);

		// It would be nicer if we could have defined the return type directly
		std::vector<dtDAL::ActorProxy*>::iterator iter;
		for (iter = actors.begin(); iter != actors.end(); iter++) {
			dtGame::GameActorProxy *robot_proxy = dynamic_cast<dtGame::GameActorProxy*>(*iter);
			srCore::RobotActorBase *robot = dynamic_cast<srCore::RobotActorBase*>(&robot_proxy->GetGameActor());
//			assert (robot != NULL);
			robots.push_back(robot);
		}
	}

	/**
	 * Return robot by identifier.
	 */
	static srCore::RobotActorBase* GetRobot(long int id) {
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		std::vector<dtDAL::ActorProxy*> actors;
		CheckForRobots functor;
		GA->GetGameManager()->FindActorsIf<CheckForRobots>(functor, actors);

		// It would be nicer if we could have defined the return type directly
		std::vector<dtDAL::ActorProxy*>::iterator iter;
		for (iter = actors.begin(); iter != actors.end(); iter++) {
			dtGame::GameActorProxy *robot_proxy = dynamic_cast<dtGame::GameActorProxy*>(*iter);
			// if the asserts fail, the FindActorsIf routine is incorrect
			assert (robot_proxy != NULL);
			srCore::RobotActorBase *robot = dynamic_cast<srCore::RobotActorBase*>(&robot_proxy->GetGameActor());
			assert (robot != NULL);
//			std::ostringstream msg; msg.clear(); msg.str("");
//			msg << "Check id " << id << " against robot with id " << robot->getID();
//			LOG_INFO(msg.str());
			if (robot->getID() == id) return robot;
		}
		return NULL;
	}

	/**
	 * Returns the Agent that belongs to the robot.
	 */
	static srInterface::Agent* GetAgent(srCore::RobotActorBase *robot) {
		return GetAgent(robot->getID());
#ifdef CORRECTED
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		srCore::SimulationArenaComponent* arena = static_cast<srCore::SimulationArenaComponent*>(
				GA->GetGameManager()->GetComponentByName(srCore::SimulationArenaComponent::DEFAULT_NAME));
		std::vector<dtCore::RefPtr<srInterface::Agent> > controllers = arena->getRobotsController(robot);
		assert (controllers.size() > 0);
		return controllers.front();
#endif
	}

	/**
	 * Returns a agent that belongs to a robot with the given id. This is the first one that is encountered
	 * if there are multiple agents per robot.
	 */
	static dtCore::RefPtr<srInterface::Agent> GetAgent(int robotId) {
		std::vector< dtGame::GMComponent* > allComponents;
		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		GA->GetGameManager()->GetAllComponents(allComponents);

		for(unsigned int i=0; i < allComponents.size(); i++) {
			dtCore::RefPtr<srInterface::Agent> controller = dynamic_cast<srInterface::Agent*>(allComponents[i]);
			if(controller != NULL) {
				if(controller->getRobotID() == robotId)	{
					return controller;
				}
			}
		}
		return NULL;
	}

	/**
	 * Gets all agents, adds it to the given vector.
	 */
	static void GetAllAgents(std::vector<srInterface::Agent*>& toFill) {
		std::vector<dtGame::GMComponent*> comps;
		std::vector<dtGame::GMComponent*>::iterator iter;

		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		GA->GetGameManager()->GetAllComponents(comps);

		for (iter = comps.begin(); iter != comps.end(); iter++) {
			srInterface::Agent *a = dynamic_cast<srInterface::Agent*>(*iter);
			if (a != NULL)
				toFill.push_back(a);
		}
	}

	/**
	 * Print the position of a robot.
	 */
//	static void PrintPosition(dtUtil::Log::LogMessageType type, const srCore::RobotActorBase &robot) {
//		osg::Vec3f pos = robot.getPosition();
//		dtUtil::Log::GetInstance().LogMessage(type, __func__, __LINE__,
//				"Get position of robot %li (%f,%f,%f)", robot.getID(), pos.x(), pos.y(), pos.z());
//	}

	/**
	 * Print orientation of a robot
	 */
//	static void PrintOrientation(dtUtil::Log::LogMessageType type, const srCore::RobotActorBase &robot) {
//		osg::Vec3f ori = robot.getRotation();
//		dtUtil::Log::GetInstance().LogMessage(type, __func__, __LINE__,
//				"Get orientation of robot %li (%f,%f,%f)", robot.getID(), ori.x(), ori.y(), ori.z());
//	}

};

///@}


#endif /* INTERFACEUTIL_H_ */
