/**
 * @file YarpRobotNavigation.h
 * @brief Robot navigation by commands over a YARP port
 *
 * This file YarpRobotNavigation.h is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Andrei Rusu <andrei@almende.org>
 *
 * @author 	Andrei A. Rusu
 * @date	Jun 15, 2010
 * @project	Replicator FP7
 * @company	Almende B.V.
 * @case	YARP
 */

#ifndef YARPROBOTNAVIGATION_H_
#define YARPROBOTNAVIGATION_H_

// General files
#include <srInterface/Agent.h>
#include <srCore/Thread.h>

// YARP
#include <yarp/sig/all.h>
#include <yarp/os/all.h>
#include <yarp/dev/all.h>
#include <yarp/os/RpcServer.h>

namespace srInterface {

/* **************************************************************************************
 * Forward declarations
 * **************************************************************************************/

class YarpRobotNavigationRpcThread;

/* **************************************************************************************
 * Interface of YarpRobotNavigation
 * **************************************************************************************/

class YarpRobotNavigation: public srInterface::Agent, srCore::Thread {
public:
	/**
	 * The constructor sets the default queue limit and motion units.
	 */
	YarpRobotNavigation(const std::string &name,
			srCore::RobotActorBase *robotActor);

	virtual ~YarpRobotNavigation();

	void SetPort(const std::string &portPrefix);

	void task();

	void ProcessMessage(const dtGame::Message& message);

	void setSpeed(float speedL, float speedR);

	void advanceX(float X);

	void strafeX(float X);

	void rotateX(float X);

	void OnAddedToGM();

	void OnRemovedFromGM();
private:
	//! Forbid copying by copy constructor
	YarpRobotNavigation(const YarpRobotNavigation& nav);

	//! Forbid copying by assignment
	YarpRobotNavigation& operator=(const YarpRobotNavigation& nav);


	// needed to use YARP
	yarp::os::Network yarp;
	// port for Remote Procedure Calls via YARP
	yarp::os::RpcServer rpcPort;
	// needed so that calls to the respond() handler do not overlap
	yarp::os::Semaphore rpcMutex;
	// sets up a custom "/robotName/rpc/command" type name uniqueue to
	// every robot controlled by this controller
	std::string createRpcPortName();
	// will hold a thread which monitors the rpcPort and executes
	// appropriate commands
	YarpRobotNavigationRpcThread *rpcThread;

	// the external module is free to issue commands at a different rate
	// than the controller can process, so a maximum number of N commands

	// are buffered in this command queue
	std::deque<yarp::os::Bottle> mainCommandQueue;
	// synchronous command queues; commands in these
	// queues are weaved with the ones in "main" commandQueue
	std::list<std::deque<yarp::os::Bottle> > syncCommandQueues;

	// modifications to the commandQueue are not thread safe
	// so a mutex is needed
	yarp::os::Semaphore commandQueueMutex;

	// method used to respond to remote procedure calls via YARP RPC
	bool respond(const yarp::os::Bottle &command, yarp::os::Bottle &reply);

	// method used to respond to remote procedure calls via YARP RPC,
	// here aggregating a list of synchronous commands
	bool respond(const std::list<yarp::os::Bottle> &commandList,
			yarp::os::Bottle &reply);

	//// IN PROCESS MESSAGE COMMANDS WITH ONE UNIT ARE EXECUTED
	// unit for the advance (adv) command
	double RPC_ADV_UNIT;
	// unit for the rotate (rot) command
	double RPC_ROT_UNIT;
	// the queue size limit; when more than these commands are
	// still valid, the oldest are removed to fit this limit;
	unsigned int RPC_QLMT;

	friend class YarpRobotNavigationRpcThread;
};

/**
 * A specific thread.
 */
class YarpRobotNavigationRpcThread: public yarp::os::Thread {
	YarpRobotNavigation *ctrl;
public:
	YarpRobotNavigationRpcThread(YarpRobotNavigation *_ctrl) :
		ctrl(_ctrl) {
	}
	virtual ~YarpRobotNavigationRpcThread() {
	}

	virtual void run();

	inline void sequencer(yarp::os::Bottle &command, yarp::os::Bottle &reply);

private:
	//! Forbid copying by copy constructor
	YarpRobotNavigationRpcThread(const YarpRobotNavigationRpcThread&);

	//! Forbid copying by assignment
	YarpRobotNavigationRpcThread& operator=(const YarpRobotNavigationRpcThread&);

};
}


#endif /* YARPROBOTNAVIGATION_H_ */
