/**
 * @file YarpedRobot.h
 * @brief 
 *
 * This file YarpedRobot.h is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license.
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author  Anne C. van Rossum
 * @date    Jul 20, 2010
 * @project Replicator FP7
 * @company Almende B.V.
 * @case	YARP
 */


#ifndef YARPEDROBOT_H_
#define YARPEDROBOT_H_

// General files
#include <boost/tuple/tuple.hpp>
#include <boost/variant.hpp>

#define WITH_OPENCV

// YARP files
#include <yarp/os/all.h>
#include <yarp/sig/Image.h>

// OpenCV
#ifdef WITH_OPENCV
#include <opencv/cv.h> //IplImage & osg::Image
#endif

// Robot3D files
#include <srInterface/Agent.h>

using namespace yarp::os;
using namespace yarp::sig;

namespace srInterface {

/**
 * \addtogroup Default
 * @{
 */

/* **************************************************************************************
 * Interface of YarpedRobot
 * **************************************************************************************/

class YarpedRobotThread;

/*
 * Enumeration: IMAGE_TYPE
 *
 * Two types, blue-green-red, and red-green-blue.
 */
enum IMAGE_TYPE {
	BGR,
	RGB
};

enum PORT_TYPE {
	SENSOR,
	ACTUATOR,
	ALL
};

/*
 * Class: YarpedRobot
 *
 * A YARP interface for a robot. It exposes its sensors and its cameras, such as a camera,
 * as well as connector states and eventually additional controls like the robot's colour.
 */
class YarpedRobot: public Agent {
public:
	//! Enables handling RPC commands while in the main module e.g. images are read and written
	friend class YarpedRobotThread;

	// Constructor
	YarpedRobot(const std::string &name, RobotActorBase *robot);

	// Destructor
	virtual ~YarpedRobot();

	// ProcessMessage implementation of Agent
	void ProcessMessage(const dtGame::Message& message);

	// Set camera frequency in Hz
	void SetCameraFrequency(const int &freq);

	// Sets port name (and opens it)
	void SetPort(const std::string &prefixName);

	//! Sets port name (and opens it) for the camera
	void SetCameraPort(const std::string &prefixName);

	// Gets yarp (bottles) over cmd port
	void ProcessYARPCommands();

	//! Callback about adding to game manager
	void OnAddedToGM();
protected:
	//! List all ports
	bool ListPorts(const Bottle& command, Bottle& reply);

	//! Add ports to port_names of given type
	bool getPorts(const PORT_TYPE &portType, std::vector<std::string> &portNames);

	//! Get one port, use getPorts with vector to know the port order
	bool getPort(const PORT_TYPE &portType, int index, std::string &portName);

	//! Add ports to port_names with given key
	bool getPorts(const std::string &key, std::vector<std::string> &portNames);

	//! Add port to portName of given key and index
	bool getPort(const std::string &key, int index, std::string &portName);

	//! Configure camera
	bool ConfigureCamera(const Bottle& command, Bottle& reply);

	//! Constructor by copy
	YarpedRobot(const YarpedRobot& yarpedRobot);

	//! Constructor by copy
	void operator=(const YarpedRobot& yarpedRobot);

	//! Get reference to actual sensor class by given name
	bool GetSensor(const std::string &name, dtCore::RefPtr<srCore::SensorBase> &sensor);

	//! Select camera (turn it on)
	void SelectCamera(const int &vocab_robotside);

	//! Return camera name for side
	std::string GetCameraName(const int &vocab_robotside);
private:
	//! Initialisation finished or not
	bool initialised;

	//! Tick counter (used for frequencies)
	long long int tick_counter;

	//! Frequency of camera polling
	int camera_frequency;

	//! Port to receive commands over (the other party is the initiator)
	yarp::os::RpcServer rpcServerPort;

	//! Create selected sets of possible ports
	typedef boost::variant<BufferedPort<Bottle>*, BufferedPort<Image>* > PortVariant;

	//! Port iterator
	typedef std::map<std::string, PortVariant >::iterator PortMapIter;

	//! List of ports, strings here are not port names but keys
	std::multimap<std::string, PortVariant > sensorPorts;

	//! List of ports, strings here are not port names but keys.
	std::multimap<std::string, PortVariant > actuatorPorts;

#ifdef WITH_OPENCV
	// Convert OpenSceneGraph image to an OpenCV image
	IplImage* convertToOpenCV(osg::Image *source, IMAGE_TYPE type);
#endif

	// Specific camera in use
	int camera_selected; //vocab

	// Camera on/off
	bool camera_enabled;

	//! Thread to process commands
	YarpedRobotThread *yarpedRobotThread;
};

/**
 * Helper class for returning port names for different types of YARP ports.
 */
struct GetPortName: public boost::static_visitor<ConstString> {
  ConstString operator()(BufferedPort<Image>* port) const {
	  return port->getName();
  }

  ConstString operator()(BufferedPort<Bottle>* port) const {
	  return port->getName();
  }
};

/**
 * We need a separate thread for the RPC commands. The main thread can retrieve for instance camera images,
 * while the module listens in the meanwhile to RPC commands. The thread contains the RFModuleRPC as public
 * field and is able to access its field because it is declared as a friend class.
 */
class YarpedRobotThread: public yarp::os::Thread {
private:
	//! Pointer to the robot
	YarpedRobot *yarpedRobot;

protected:
	//! Disable copy by assignment
	void operator=(const YarpedRobotThread& thread);

	//! Disable copy by construction
	YarpedRobotThread(const YarpedRobotThread &thread);
public:
	//! Constructor requires pointer to a RFModuleRPC instance
	YarpedRobotThread(YarpedRobot *yarpedRobot): yarpedRobot(yarpedRobot) {	}

	//! Empty destructor
	inline ~YarpedRobotThread() {}

	//! Default "run" command for this Thread derived class
	void run();
};

///@}

}

#endif /* YARPEDROBOT_H_ */
