/*
 * @file YarpedSim.h
 * @brief Simulation-wide class respecting YARP messages
 *
 * This file YarpedSim.h is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license.
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author		Anne C. van Rossum
 * @date		Aug 2, 2010
 * @project		Replicator FP7
 * @company		Almende B.V.
 * @case		YARP
 */

#ifndef YARPEDSIM_H_
#define YARPEDSIM_H_

// General files
#include <string>
#include <iostream>

// Delta3D files
#include <dtGame/gmcomponent.h>
#include <dtCore/refptr.h>

// YARP files
#include <yarp/os/BufferedPort.h>
#include <yarp/os/Bottle.h>
#include <yarp/os/Vocab.h>
#include <yarp/os/Network.h>

// Robot3D files
#include <srInterface/YarpUtil.h>
#include <srInterface/Agent.h>

using namespace yarp::os;

namespace srInterface {

// Forward declaration
class YarpedSimThread;

/* **************************************************************************************
 * Interface of YarpedSim
 * **************************************************************************************/

/*
 * The simulator can be interfaced over YARP via this class. It makes it possible to add
 * and remove robots. It is a Delta3D component and a YARP module at once. The first parent
 * class describes a ProcessMessage routine that needs to be implemented to receive game
 * messages. The latter defines runModule, updateModule, etc. This leads to conflicts: which
 * entity should be able to tick YarpedSim? We choose to call updateModule from the
 * ProcessMessage routine ourselves. So, runModule will not be called at all.
 */
class YarpedSim: public dtGame::GMComponent {
public:

	friend class YarpedSimThread;

	/*
	 * Function: YarpedSim
	 */
	YarpedSim();

	/*
	 * Function: ~YarpedSim
	 */
	virtual ~YarpedSim();

	/*
	 * Function: ProcessMessage ticks
	 */
	void ProcessMessage(const dtGame::Message &message);

	//! Read YARP message from port
	Bottle *ReadPort();

	void ReadBottle(Bottle *b);

	/**
	 * Get AddRobot information from bottle
	 */
	bool ReadBottleAddRobot(Bottle &bottle, std::string &name, std::string &type,
			int &id, osg::Vec3f &position);

	// Add controller
	void AddController(const ControllerInitData &cid);

protected:

	// Add robot
	void AddRobot(const std::string &name, const std::string &type, const int &id, const osg::Vec3f &position);

	// Vector of agents
	std::vector<dtCore::RefPtr <srInterface::Agent> > *agents;
private:
	//! Forbid copying by copy constructor
	YarpedSim(const YarpedSim& sim);

	//! Forbid copying by assignment
	YarpedSim& operator=(const YarpedSim& sim);

	//! Global port for YARP commands to Robot3D
	BufferedPort<Bottle> globalYarpPort;

	//! Reference to YARP will be created when this object is created
	yarp::os::Network network;

	//! Handle add robot to thread
	YarpedSimThread *yarpedSimThread;

	//! Debug purposes mainly, a counter incremented each tick
	long int time_counter;

	//! Received a COMMAND_PAUSE message
	bool pauseMessageProcessed;
};

/**
 * We need a separate thread for adding components to the Delta3D simulator. There is a "bug" in
 * GameManager::DoSendMessageToComponents(const Message& message) in the sense that an iterator
 * is used, which becomes invalid if you add a component in the ProcessMessage routine.
 */
class YarpedSimThread: public yarp::os::Thread {
private:
	//! Pointer to YarpSim
	YarpedSim *yarpedSim;

	//!
	long int time_counter;
protected:
	//! Disable copy by assignment
	void operator=(const YarpedSimThread& thread);

	//! Disable copy by construction
	YarpedSimThread(const YarpedSimThread &thread);
public:
	//! Constructor requires pointer to a RFModuleRPC instance
	YarpedSimThread(YarpedSim *yarpedSim): yarpedSim(yarpedSim), time_counter(0) {	}

	//! Empty destructor
	inline ~YarpedSimThread() {}

	//! Default "run" command for this Thread derived class
	void run();
};

}

#endif /* YARPEDSIM_H_ */
