/*
 * ScrewDrive.cpp
 *
 *  Created on: Nov 18, 2008
 *      Author: winkler
 */

// General files
#include <cmath>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <stdexcept>

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srActuator/screwDrive.h>
#include <srCore/actuator/hingeActuator.h>

#define MAX(X,Y) (X>Y?X:Y)
#define MIN(X,Y) (X<Y?X:Y)

using namespace std;

namespace srCore {

/**
 * This is an old version of the screw drive and will be replaced later on by ScrewDrive2.
 */
ScrewDrive::ScrewDrive(const std::string name):
									ActuatorBase(name, "ScrewDrive")
{
	// Physical parameters
	d_m = 0.0155; 			//15.5 mm
	my_Screw = 0.1;
	l = 0.0099;				// 9.9 mm

	// From ScrewDrive2
	//	d_m = 0.31;					//15.5mm
	//	my_Screw = 0.1;
	//	l = 4.95;					//9.9mm

	velocity = 0.0;
	last_reported_velocity = 0.0;
	velocity_report_granularity = 0.2;
	counter = 0;
	print_freq_ode_params = 500;

	// Set private fields to null to get clear error messages
	secondScrewDrive = NULL;
	frontActuator = NULL;
	backActuator = NULL;

	//speed is here the torque *gulp*
	F_div_M = d_m/2.0 * ((l + M_PI*my_Screw*d_m) / (M_PI*d_m - my_Screw*l));

}

ScrewDrive::~ScrewDrive() {
	frontActuator = NULL;
	backActuator = NULL;
}

void ScrewDrive::registerActuator() {
	ActuatorBase::jointActuatorMap.insert(std::pair<dJointID, ActuatorBase*>(
			frontActuator->getJoint(), frontActuator));
	ActuatorBase::jointActuatorMap.insert(std::pair<dJointID, ActuatorBase*>(
			backActuator->getJoint(), backActuator));
}

/**
 * Initialisation of the screw drive. It needs both sides of the robot to be attached at.
 *
 * Oct 27, 2010: Solved bug about isOnRightHandSide which caused recursive loop on SetVel.
 */
void ScrewDrive::init(dtCore::RefPtr<ActuatorBase> front, dtCore::RefPtr<ActuatorBase> back,
		dtCore::RefPtr<ScrewDrive> second, bool isOnRightHandSide) {

	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Initialise screw drive: " << GetUniqueId();
	LOG_INFO(msg.str());

	this->isOnRightHandSide = isOnRightHandSide;
	secondScrewDrive = second;
	frontActuator = front;
	backActuator = back;

	velocity = 0;
	myParameters = front->getParameterList();
#ifdef	AUDIO_ON
	screwSound = NULL;

	for(unsigned int i = 0; i < this->GetNumChildren(); ++i)
		if((screwSound = dynamic_cast<dtAudio::Sound*>(this->GetChild(i))) != NULL)
			break;
#endif
}

/**
 * Sets the velocity of the screw drive.
 */
void ScrewDrive::setVel(float v, float deltaTime)
{
	velocity = v;
	float speedLeft = 0.0, speedRight = 0.0;

	if (secondScrewDrive == NULL) {
		LOG_DEBUG("Second screw drive is NULL. Did you actually initialise the screw drives in the robot class?");
		return;
	}

	assert (secondScrewDrive->isOnRightHandSide != isOnRightHandSide);

	counter++;

	if (isOnRightHandSide) {
		speedRight = v;
		speedLeft = secondScrewDrive->getVel(deltaTime);
	}
	else {
		speedLeft = v;
		speedRight = secondScrewDrive->getVel(deltaTime);
	}

	if (std::fabs(velocity - last_reported_velocity) >= velocity_report_granularity) {
		last_reported_velocity = velocity;
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << std::fixed << std::setprecision(2) << "Speed [left,right] = [" << speedLeft << "," << speedRight << "]";
		LOG_INFO(msg.str());
	}

	// Translation from 0..1
	float speedTrans = std::min(std::fabs(speedLeft), std::fabs(speedRight));
	// Rotation from 0..1
	float speedRot = std::fabs(speedLeft - speedRight);

	// If translation == 1, then rotation part == 0
	float sum = speedTrans + speedRot;
	if (sum != 0) {
		speedTrans = speedTrans / sum;
		speedRot = speedRot / sum;
	}

//	if (!(counter % print_freq_ode_params)) {
//		std::ostringstream msg; msg.clear(); msg.str("");
//		msg << "Drive robot with speed [" << speedLeft << "," << speedRight << "], which is translation+rot=[";
//		msg << speedTrans << "," << speedRot << "]";
//		LOG_INFO(msg.str());
//	}

	DriveSemiRealistic(speedLeft, speedRight, speedRot, speedTrans);

//	DriveRealistic(speedLeft, speedRight, speedRot, speedTrans);
//	float temp = speedLeft;
//	speedLeft = speedRight;
//	speedRight = temp;
//	secondScrewDrive->DriveRealistic(speedLeft, speedRight, speedRot, speedTrans);

#ifdef AUDIO_ON
	PlaySound(v);
#endif
}

/**
 * The actual forces etc are calculated and applied at the right positions on the robot's body.
 */
void ScrewDrive::DriveRealistic(float speedLeft, float speedRight, float speedRot, float speedTrans) {
//	if (isOnRightHandSide) {

		speedRight *= 0.01;
		speedLeft *= 0.01;

		// The ODE joints corresponding to the four actuators
		dJointID f0 = getJoint(0);
		dJointID f1 = getJoint(1);
		dJointID b0 = secondScrewDrive->getJoint(0);
		dJointID b1 = secondScrewDrive->getJoint(1);

		dBodyID frontBody = dJointGetBody( f0, 1);
		//notused			dBodyID backBody = dJointGetBody( b0, 1);

		dVector3 pos0;
		dJointGetHingeAnchor (f0, pos0);
		dVector3 pos1;
		dJointGetHingeAnchor (f1, pos1);

		dVector3 pos2;
		dJointGetHingeAnchor (b0, pos2);
		dVector3 pos3;
		dJointGetHingeAnchor (b1, pos3);


		float F_Screw_left = speedLeft * F_div_M;
		float F_Screw_right = -speedRight * F_div_M;

		float F_Roll_left = 2 * speedLeft / d_m;
		float F_Roll_right = 2 * speedRight / d_m;

		float F_Screw = 0;
		if (F_Screw_left * F_Screw_right > 0) F_Screw = 2 * MIN(F_Screw_left, F_Screw_right);

		float F_Roll = 0;
		if (F_Roll_left * F_Roll_right > 0) F_Roll = 2 * MIN(F_Roll_left, F_Roll_right);


		if (!(counter % print_freq_ode_params)) {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "F_Roll = " << F_Roll << ", F_Screw = " << F_Screw;
			msg << " (" << F_Screw_left << "," << F_Screw_right << ")";
			//			msg << "Position = " << osg::Vec3((pos0[0] + pos1[0]) / 2, (pos0[1] + pos1[1]) / 2, (pos0[2] + pos1[2]) / 2);
			LOG_INFO(msg.str());
		}

		//dBodyAddRelForceAtPos    (frontBody, 0, F_Roll, F_Screw*100, (pos0[0] + pos1[0]) / 2, (pos0[1] + pos1[1]) / 2, (pos0[2] + pos1[2]) / 2);

		//			dBodyAddRelForceAtRelPos (frontBody, 0, 0, F_Roll, 0,0,0);//, F_Screw*100);

		/**
		 * A force is added to the body, let's use the Y-coordinate. Okay, then the robot rolls over when
		 * enough force is applied. The X-coordinate however, it starts to move... The only problem is that
		 * it moves in one direction, a kind of circle... That is probably because the force is applied
		 * relative to the position of this screw drive if it obeys isOnRightHandSide. Moreover, it can not
		 * drive backwards (key "j" instead of "u").
		 */
		dBodyAddRelForceAtRelPos (frontBody, F_Roll, 0, 0, 0, 0, 0);//, F_Screw*100);

		//dBodyAddRelForceAtPos    (backBody, 0, F_Roll, -F_Screw*100, (pos2[0] + pos3[0]) / 2, (pos2[1] + pos3[1]) / 2, (pos2[2] + pos3[2]) / 2);
		//dBodyAddRelForce    (backBody, 0, 0, -F_Roll);//, F_Screw*100);

//	}
	//	else {
	//		secondScrewDrive->setVel(secondScrewDrive->getVel(deltaTime), deltaTime);
	//	}

}

/**
 * Direct commands about velocity are sent to the ODE engine. It doesn't seem to do much
 * on the moment. Are the wheels in RobotKIT properly placed for this setting?
 */
void ScrewDrive::DriveSemiRealistic(float speedLeft, float speedRight, float speedRot, float speedTrans) {
	speedLeft = speedLeft * 2;
	speedRight = speedRight * 2;
	if (isOnRightHandSide) {
		frontActuator->setODEParameter(dParamVel, speedRight);
		backActuator->setODEParameter(dParamVel, -speedRight * speedTrans + speedRot * speedLeft);
		secondScrewDrive->frontActuator->setODEParameter(dParamVel,  speedLeft);
		secondScrewDrive->backActuator->setODEParameter(dParamVel, -speedLeft * speedTrans + speedRot * speedRight);
	} else {
		frontActuator->setODEParameter(dParamVel,  speedLeft);
		backActuator->setODEParameter(dParamVel, -speedLeft * speedTrans + speedRot * speedRight);
		secondScrewDrive->frontActuator->setODEParameter(dParamVel,  speedRight);
		secondScrewDrive->backActuator->setODEParameter(dParamVel, -speedRight * speedTrans + speedRot * speedLeft);
	}
}

/**
 * Play the sound with a pitch relative to the speed.
 */
#ifdef AUDIO_ON
void ScrewDrive::PlaySound(float speed) {
	if(screwSound != NULL)
	{
		if(speed > 0.0)
		{
			screwSound->SetGain(0.1f+speed/10);
			screwSound->SetPitch(0.5f+speed/100);
			if (!screwSound->IsPlaying())
				screwSound->Play();
		}
		else if(screwSound->IsPlaying())
			screwSound->Stop();
	}
}
#endif

}
