//! Specialized body class.

/*!
 * Predefined body class. Ready to use in a RobotActor.
 * This body represents the front part of the robot.
 *
 * @author Lutz Winkler
 *
 */

#include <srBody/kitBody1.h>

namespace srCore {

// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
#define SCALE_3DS   			0.02

#define COLLISION_SIDE 			(64*SCALE_3DS)					//60		// cm
#define COLLISION_LENGTH 		(64*SCALE_3DS)					//85

// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
#define BODY_LENGTH 			1.7					// equals 85 mm
#define BODY_HEIGHT 			1.2
#define BODY_WIDTH  			1.2
#define BODY_MASS   			0.3					//0.13

using namespace srCore;

//! Constructor
/*! @param id a 16bit id, used to set the colliding bit parameter.
 * 	@param x, y, z the position of the hinge in the world coordinate system.
 *  @param heading, pitch, roll the orientation of the axis of the hinge in the world coordinate system.
 */
KITBody1::KITBody1(	RobotActorBase* robot,
		float x, float y, float z,
		float heading, float pitch, float roll)
{
	init("KITBody1", robot, x, y, z, heading, pitch, roll);

	// Loading .3ds model
	std::ostringstream bodymesh; bodymesh.clear(); bodymesh.str("");
	bodymesh << SimulationUtils::Instance()->getStageProjectPath() << "/StaticMeshes/";

	// Use detailed graphics if you've a graphics card from say a NVidia 8800GT and up
	if (SimulationUtils::Instance()->getDetailedGraphics())
		bodymesh << "newKITBody1.3ds";
	else
		bodymesh << "KITcollBody1.3ds";

	std::ostringstream msg; msg.clear();
	msg << "Loading file " << bodymesh.str() << " for body visualisation";
	LOG_DEBUG(msg.str());

	LoadFile(bodymesh.str().c_str());

	// scaling the model
	this->setScale(SCALE_3DS,SCALE_3DS,SCALE_3DS);
	// defining the collision box
	this->setCollisionBox(COLLISION_SIDE, COLLISION_SIDE, COLLISION_SIDE);

	// our Object just needs a mass now ...
	this->setMass(BODY_MASS, COLLISION_SIDE, COLLISION_SIDE, COLLISION_SIDE);

	this->GetOSGNode()->setNodeMask(SimulationUtils::Instance()->getGeometryDetailsNodeMask());

	dtCore::Object* collModel = new dtCore::Object();
	dtCore::Transform xform;

	std::ostringstream collisionmesh; collisionmesh.clear(); collisionmesh.str("");
	collisionmesh << SimulationUtils::Instance()->getStageProjectPath() << "/StaticMeshes/";

	switch (SimulationUtils::Instance()->getCollisionGeometryDetails()) {
	case 0: collisionmesh << "KITcollBodySquare1.3ds"; break;
	case 1: collisionmesh << "KITcollBodyPolygone1.3ds"; break;
	case 2: collisionmesh << "KITcollBody.3ds"; break;
	case 15: collisionmesh << "newKITBody1.3ds"; break;
	}

	msg.clear(); msg.str("");
	msg << "Loading file " << collisionmesh.str() << " for collision shape";
	LOG_DEBUG(msg.str());

	collModel->LoadFile(collisionmesh.str().c_str());
	collModel->SetScale(osg::Vec3(0.021,0.021,0.021));//this->GetScale());
	collModel->GetOSGNode()->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());
	SetCollisionMesh(collModel->GetOSGNode());
	AddChild(collModel);
	setCollisionModel(collModel->GetOSGNode());

	osg::StateSet* stateset;
	stateset = collModel->GetOSGNode()->getOrCreateStateSet();
	osg::Material *mm = dynamic_cast<osg::Material*>(stateset->getAttribute(osg::StateAttribute::MATERIAL));
	if (!mm) {
		mm = new osg::Material;
	}
	mm->setAlpha(osg::Material::FRONT, 0.2);
	mm->setAmbient(osg::Material::FRONT, osg::Vec4(0,0,0,0));
	stateset->setAttributeAndModes( mm, osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON);
	stateset->setMode(GL_BLEND,   osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON );
	stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON );
	stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	// Transformable:Can't render unhandled geometry class:8 (when set to true)
	// This happens if there is no collision geometry set, but that seems to be done properly...
	renderCollisionGeometry(false);

	// ... and then, it is ready to place it in the scene
	addToScene();
}


KITBody1::~KITBody1()
{
}

}
