//! Specialized body class.

/*!
 * Predefined body class. Ready to use in a RobotActor.
 * This body represents the front part of the robot.
 *
 * @author Lutz Winkler
 *
 */

#include <srBody/kitBody2.h>

namespace srCore {

// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
#define SCALE_3DS   			0.02

#define COLLISION_SIDE 			(64*SCALE_3DS)					//60		// cm
#define COLLISION_LENGTH 		(64*SCALE_3DS)					//85

// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
#define BODY_LENGTH 			1.7					// equals 85 mm
#define BODY_HEIGHT 			1.2
#define BODY_WIDTH  			1.2
#define BODY_MASS   			0.3					//0.13

using namespace srCore;

//! Constructor
/*! @param id a 16bit id, used to set the colliding bit parameter.
 * 	@param x, y, z the position of the hinge in the world coordinate system.
 *  @param heading, pitch, roll the orientation of the axis of the hinge in the world coordinate system.
 */
KITBody2::KITBody2(	RobotActorBase* robot,
		float x, float y, float z,
		float heading, float pitch, float roll)

{
	init("KITBody2", robot, x, y, z, heading, pitch, roll);

	// Loading .3ds model
	std::ostringstream bodymesh; bodymesh.clear(); bodymesh.str("");
	bodymesh << SimulationUtils::Instance()->getStageProjectPath();

	// Use detailed graphics if you've a graphics card from say a NVidia 8800GT and up
	if (SimulationUtils::Instance()->getDetailedGraphics())
		bodymesh << "/StaticMeshes/newKITBody2.3ds";
	else
		bodymesh << "/StaticMeshes/KITcollBody2.3ds";

	std::ostringstream msg; msg.clear();
	msg << "Loading file " << bodymesh.str() << " for body visualisation";
	LOG_DEBUG(msg.str());

	LoadFile(bodymesh.str());
	//	SetModelTranslation(osg::Vec3(0,0,COLLISION_SIDE*25));
	SetModelRotation(osg::Vec3(0,0,0));

	// scaling the model
	this->setScale(SCALE_3DS,SCALE_3DS,SCALE_3DS);

	// defining the collision box
	this->setCollisionBox(COLLISION_SIDE, COLLISION_SIDE, COLLISION_SIDE);

	// our Object just needs a mass now ...
	this->setMass(BODY_MASS, COLLISION_SIDE, COLLISION_SIDE, COLLISION_SIDE);

	dtCore::Transform xform;
	dtCore::Object* collModel = new dtCore::Object();

	std::ostringstream collisionmesh; collisionmesh.clear(); collisionmesh.str("");
	collisionmesh << SimulationUtils::Instance()->getStageProjectPath() << "/StaticMeshes/";

	switch (SimulationUtils::Instance()->getCollisionGeometryDetails()) {
	case 0: collisionmesh << "KITcollBodySquare2.3ds"; break;
	case 1: collisionmesh << "KITcollBodyPolygone2.3ds"; break;
	case 2: collisionmesh << "KITcollBody.3ds"; break;
	case 15: collisionmesh << "newKITBody2.3ds"; break;
	}

	msg.clear(); msg.str("");
	msg << "Loading file " << collisionmesh.str() << " for collision shape";
	LOG_DEBUG(msg.str());

	collModel->LoadFile(collisionmesh.str().c_str());

	collModel->GetTransform(xform);
	xform.SetRotation(0,180,0);
	collModel->SetTransform(xform);

	collModel->SetScale(this->GetScale());

	osg::StateSet* stateset;
	stateset = collModel->GetOSGNode()->getOrCreateStateSet();
	osg::Material * mm = dynamic_cast<osg::Material*>(stateset->getAttribute
			(osg::StateAttribute::MATERIAL));

	if (!mm) {
		mm = new osg::Material;
	}
	mm->setAlpha(osg::Material::FRONT, 0.2);
	mm->setAmbient(osg::Material::FRONT, osg::Vec4(0,0,0,0));
	stateset->setAttributeAndModes( mm, osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON);
	stateset->setMode(GL_BLEND,   osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON );
	stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE |
			osg::StateAttribute::ON );
	stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

	renderCollisionGeometry(false);

	SetCollisionMesh(collModel->GetOSGNode());

	AddChild(collModel);

	setCollisionModel(collModel->GetOSGNode());

	renderCollisionGeometry(false);

	// ... and then, it is ready to place it in the scene
	addToScene();
}


KITBody2::~KITBody2() {
	//Todo: Remove attributes from stateset? Check valgrind...
}

}
