//! Specialized body class.

/*!
 * Predefined body class. Ready to use in a RobotActor.
 * This body represents the wheels of the robot.
 *
 * @author Lutz Winkler
 *
 */

#include <srBody/led.h>
#include <srCore/robot/robotActorBase.h>

namespace srCore {

//! Constructor
/*! @param id a 16bit id, used to set the colliding bit parameter.
 * 	@param x, y, z the position of the hinge in the world coordinate system.
 *  @param heading, pitch, roll the orientation of the axis of the hinge in the world coordinate system.
 *  @param Scene the scene in which the body should be placed.
 *
 */
Led::Led(dtCore::RefPtr<BodyBase> attachedBody,
		osg::Vec3f translation, osg::Vec3f orientation,
		std::string name, std::string type,	bool active): ObjectBase(
				attachedBody, translation, orientation, name, type, active),
				myColor (new osg::Vec4(1.0f, 1.0f, 0.0f, 0.5f)),
				cv (new ColorVisitor( *myColor )),
				ledSize (0.2)
{

	// loading 3ds model
	char file[50];
	sprintf(file, "%s/StaticMeshes/sphere.3ds", SimulationUtils::Instance()->getStageProjectPath().c_str());

	LoadFile(file, false);
	//	this->mesh->SetModelRotation(osg::Vec3(90,90,0));//relOrientation.x(), relOrientation.y()+90, relOrientation.z()));
	//	this->mesh->SetModelTranslation(osg::Vec3(relTranslation.x() * scale.x(), relTranslation.y() * scale.y(), relTranslation.z() * scale.z()));

	// Adds the led as a child to the OSG tree
	myBody->AddChild(this);

	dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(), relOrientation.x()+180, relOrientation.y(), relOrientation.z());
	SetTransform( offset, dtCore::Transformable::REL_CS );

	// scaling the model
	SetScale(osg::Vec3(ledSize, ledSize, ledSize));

	std::ostringstream msg;	msg.clear(); msg.str("");
	msg << "Add LED to robot \"" << attachedBody->getRobot()->getID() << "\" to the scene";
	LOG_INFO(msg.str());

	GetOSGNode()->traverse(*cv);
}

/**
 *
 * See: http://www.fredosaurus.com/notes-cpp/oop-condestructors/copyconstructors.html
 */
Led::Led(const srCore::Led &led): ObjectBase(led.myBody, led.absTranslation, led.absOrientation, name, type, active),
		myColor (led.myColor),
		cv (led.cv),
		ledSize (led.ledSize) {
}

/**
 * From the copied object, delete existing fields and (re)populate it with the given
 * object data.
 * See: http://www.fredosaurus.com/notes-cpp/oop-overloading/overloadassign.html
 */
srCore::Led& Led::operator=(const srCore::Led& led) {
	if (this != &led) {
		delete myColor;
		myColor = new osg::Vec4(led.myColor->x(), led.myColor->y(), led.myColor->z(), led.myColor->w());
		delete cv;
		cv = new ColorVisitor( *myColor );
		ledSize = led.ledSize;
	}
	return *this;
}

/**
 * Size of the LED. This may make the LED detection task simpler or harder.
 */
float Led::getSize() {
	return ledSize;
}

/**
 * Returns current LED color.
 */
osg::Vec4* Led::getColor() {
	return myColor;
}

/**
 * Sets the color of the LEDS. The last field concerns transparency.
 */
void Led::setColor(osg::Vec4 color) {
	cv = new ColorVisitor(color);
	GetOSGNode()->traverse(*cv);
}

/**
 * Destructor of LED.
 */
Led::~Led() {
	delete myColor;
	myColor = NULL;
	delete cv;
	cv = NULL;
}

}
