//! Specialized body class.

/*!
 * Predefined body class. Ready to use in a RobotActor.
 * This body represents the wheels of the robot.
 *
 * @author Lutz Winkler
 *
 */

#include <iostream>

#include <srBody/wheel.h>

namespace srCore {

// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
#define SCALE_3DS   			0.2				//0.2
#define WHEEL_SIZE  			SCALE_3DS
#define WHEEL_MASS  			0.01
#define WHEEL_RADIUS			0.6 * SCALE_3DS //4.5 //0.5
#define WHEEL_LENGTH			1 * SCALE_3DS	//2

using namespace srCore;

//! Constructor
/*! @param id a 16bit id, used to set the colliding bit parameter.
 * 	@param x, y, z the position of the hinge in the world coordinate system.
 *  @param heading, pitch, roll the orientation of the axis of the hinge in the world coordinate system.
 *  @param Scene the scene in which the body should be placed.
 *
 */
Wheel::Wheel(	RobotActorBase* robot,
				float x, float y, float z,
				float heading, float pitch, float roll)
{
	// initializing. parameters: name, id (of the robot), position, orientation and the scene
	// needs to be done first
	init("wheel", robot, x, y, z, heading, pitch, roll);

	// Loads 3ds model, requires the name wheel.3ds and has to be found in StaticMeshes.
	std::ostringstream file; file.clear(); file.str("");
	file << SimulationUtils::Instance()->getStageProjectPath() << "/StaticMeshes/wheel.3ds";
	LoadFile(file.str(), false);

	// Scale the model, this actually depends on the size of the original 3D model
	SetScale(osg::Vec3(WHEEL_SIZE, WHEEL_SIZE, WHEEL_SIZE));

	// Our object just needs a mass now ...
	setMass(WHEEL_MASS, WHEEL_RADIUS, WHEEL_LENGTH);

	// defining the collision box
	setCollisionCylinder(WHEEL_RADIUS, WHEEL_LENGTH);

	// ... and then, it is ready to place it in the scene
	addToScene();

	// to check if the collision geometry fits to the model:
	renderCollisionGeometry(true);
	colorCollisionShape(osg::Vec4(255,0,0,0));

	// need to be set to let the wheels rotate
    dBodySetFiniteRotationMode(GetBodyID(), 1);

    setCollisionBits(2);
}

Wheel::~Wheel()
{
}

}
