//! Basic connector class.

/*!
 *
 * This class simulates a connector in a very simple way. If two connectors
 * are close enough, have the right orientation and if both want to connect,
 * a fixed joint will be created that connects the two bodies, where the
 * connectors are attached to. If one of the connectors wants to disconnect,
 * the joint will be destroyed.
 *
 * @author Lutz Winkler
 *
 */

//TODO: (re)assigning the geomCollide and geomCategory bits of each body so that bodies
//TODO: Disconnecting for the new coupling mechanism needs to be implemented
//TODO: transform the new coupling mechanism to the actuators as well to reduce degrees of freedom  -->  enable larger organisms

#include <dtABC/application.h>

#include <srCore/actuator/connectorBase.h>
#include <srCore/robot/robotActorBase.h>
#include <srCore/actuator/hingeActuator.h>

#include <osg/PositionAttitudeTransform>
#include <osg/Group>

#include <dtCore/odegeomwrap.h>
#include <dtUtil/log.h>

using namespace srCore;

ConnectorBase::ConnectorBase() :
					ObjectBase(),
					connectorState (DISCONNECTED)
{
	connection_event_function = NULL;

	ObjectParameter sp;
	//
	//	sp.valueName = "AbsPosition";
	//	sp.valueType = typeid(osg::Vec3f*).name();
	//	myParameters.push_back(sp);
	//
	//	sp.valueName = "AbsOrientation";
	//	sp.valueType = typeid(osg::Vec3f*).name();
	//	myParameters.push_back(sp);

	sp.valueName = "UpVector";
	sp.valueType = typeid(osg::Vec3f*).name();
	myParameters.push_back(sp);

	sp.valueName = "State";
	sp.valueType = typeid(int*).name();
	myParameters.push_back(sp);

}

ConnectorBase::~ConnectorBase(){
	connection_event_function = NULL;
	disconnect();
	myParameters.clear();
	delete myID;
}

/**
 * Just sets the internal callback function. Take care not to use it Callback
 * because that is used by OpenSceneGraph too.
 */
void ConnectorBase::SetConnectionEventFunction(ConnectCallback func) {
	connection_event_function = func;
}

/**
 * Calls the function with the proper parameter, namely the robot on which the
 * connection / disconnection event takes place.
 */
void ConnectorBase::ExecuteConnectionEventFunction() {
	if (connection_event_function != NULL) {
		connection_event_function(*getRobot());
	}
}

//! initializing the connector. Needs to be called before using the connector.
void ConnectorBase::init(dtCore::RefPtr<BodyBase> body,
		float anchorX, float anchorY, float anchorZ,
		float axisX, float axisY, float axisZ)
{

	this->relTranslation = osg::Vec3(anchorX, anchorY, anchorZ);
	this->relOrientation = osg::Vec3(axisX, axisY, axisZ);

	this->myBody = body;

	// pointer to the fixed joint. Only set then
	// there is a connection and the id of this
	// joint is smaller than the id of the other joint.
	this->connector = NULL;

	// set the mesh. The mesh is a drawable Object in the scene to show the position and orientation
	// of the connector.
	std::ostringstream file;
	file << SimulationUtils::Instance()->getStageProjectPath();
	file << "/StaticMeshes/cone2.3ds";
	LoadFile(file.str());
	file.clear();

	SetScale(osg::Vec3(0.01,0.01,0.01));
	this->GetOSGNode()->setNodeMask(SimulationUtils::Instance()->getLaserRayNodeMask());

	// 	Activate this to see the position of the Connector
	//	this->RenderProxyNode(true);

	osg::Vec3 scale = this->myBody->getScale();

	this->myBody->AddChild(this);
	dtCore::Transform xform;
	xform.SetTranslation(relTranslation);
	xform.SetRotation(relOrientation);
	SetTransform(xform, dtCore::Transformable::REL_CS );

	this->theOtherConnector = NULL;						// if set to NULL, then there is no connection.
	this->myID = new dtCore::UniqueId();				// create a unique id


}

void ConnectorBase::getParameter(const std::string &name, ParameterReturnValue  parameter) {

	if (name == "RelPosition") {

		//check parameter type
		if (parameter.type != typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");

		// cast void* to Vec3f*
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);

		// get parameter
		*data = relTranslation;
		parameter.pValue = data;


	}

	else if (name == "RelOrientation") {

		if (parameter.type != typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		*data = relOrientation;
		parameter.pValue = data;
	}

	else if (name == "AbsPosition") {

		if (parameter.type != typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		paramAbsTranslation = getTranslation();
		*data = paramAbsTranslation;
		parameter.pValue = data;


	}

	else if (name == "AbsOrientation") {

		if (parameter.type != typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		paramAbsOrientation = getRotation();
		*data = paramAbsOrientation;
		parameter.pValue = data;
	}


	else if (name == "UpVector") {

		if (parameter.type != typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		paramUpVector = getUpVector();
		*data = paramUpVector;
		parameter.pValue = data;
	}

	else if (name == "State") {

		if (parameter.type != typeid(int*).name()) LOG_WARNING("Parameter type needs to be int*. Casting could lead to errors.");
		int* data = reinterpret_cast<int*> (parameter.pValue);
		*data = (int) getState();
		parameter.pValue = data;
	}

	else ObjectBase::getParameter(name, parameter);

}


void ConnectorBase::setParameter(const std::string &name, ParameterReturnValue  parameter) {

	if (name == "RelPosition") {

		// parameter type check
		if (parameter.type!=typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");

		// cast void* to Vec3f*
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		relTranslation = *data;

		// set parameter

		dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(), relOrientation.x(), relOrientation.y(), relOrientation.z());
		SetTransform( offset, dtCore::Transformable::REL_CS );


	}


	else if (name == "RelOrientation") {

		if (parameter.type!=typeid(osg::Vec3f*).name()) LOG_WARNING("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		relOrientation = *data;

		dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(), relOrientation.x(), relOrientation.y(), relOrientation.z());
		SetTransform( offset, dtCore::Transformable::REL_CS );


	}

	else ObjectBase::setParameter(name, parameter);

}


void ConnectorBase::setScale(osg::Vec3 scale) {

	dtCore::Transform offset;
	GetTransform(offset, dtCore::Transformable::REL_CS);
	offset.MakeScale(scale.x(), scale.y(), scale.z());
	SetTransform( offset, dtCore::Transformable::REL_CS );

}

dtCore::UniqueId* ConnectorBase::getID()
{
	return this->myID;
}


//! returns the BodyId of the body the connector is attached to.
dBodyID ConnectorBase::getBodyID()
{
	return this->myBody->GetBodyID();
}


/**
 * The two connectors have to be close enough and need to have the right orientation,
 * as well as both need to be set the connectionWanted to be true. If so, there are two
 * possibilities:
 *   (1) a fixedJoint will be created that links the two bodies together
 *   (2) the bodies will actually be merged for computational reasons
 * The function connect does not influence the position or orientation of the robot. It
 * basically sets the state of the connector to WANT_TO_CONNECT and then checks if there
 * is another, foreign, connector nearby given a look-at vector and an up-vector. If
 * that is the case, connecting(theOtherConnector) is called.
 *
 * @see connecting(theOtherConnector)
 * @see connect(theOtherConnector)
 *
 * @remark Step if other connector is in state CONNECTED
 */
void ConnectorBase::connect()
{
	// check if connector is already connected
	if (getState() == CONNECTED) {
//		LOG_INFO("Connector is already connected.");
		return;
	}

	setState(WANT_TO_CONNECT);

	// get all game actors
	std::vector< dtGame::GameActorProxy* > allGameActors;
	getRobot()->GetGameActorProxy().GetGameManager()->GetAllGameActors(allGameActors);

	if (!allGameActors.empty()) {
		for (unsigned int i=0; i<allGameActors.size(); i++) {

			RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[i]->GetGameActor());

			for (unsigned int j=0; j != actor.getConnectors().size(); j++) {

				bool success = false;
				// Connectors attached to the same body can never connect with each other
				if (actor.getConnectors()[j]->getBodyID() != getBodyID()) {

					// A robot is made up out of two bodies, so skip also when it concerns the same robot
					if (actor.getConnectors()[j]->getRobot() == getRobot()) continue;

					// Get distance between the connectors
					osg::Vec3 minDist = actor.getConnectors()[j]->getTranslation() - getTranslation();

					// When distance is large, do not even bother to check other parameters
					if (minDist.length() > 10 * epsilon) continue;

					// get angle of the two look-at vectors
					osg::Vec3f rot2 = actor.getConnectors()[j]->getLookAtVector();
					osg::Vec3f rot1 = this->getLookAtVector();
					rot2.normalize();
					rot1.normalize();
					float angle = rot2 * rot1;

					// get angle of the two up-vectors
					rot2 = actor.getConnectors()[j]->getUpVector();
					rot1 = this->getUpVector();
					float angle2 = rot2 * rot1;

					// check condition for connection. prev values checkAngle = 0.1, epsilon = 0.5
					if ((minDist.length() < epsilon) && (angle < -0.9) && (checkAngle(angle2, 0.1)) &&
							(actor.getConnectors()[j]->getState() != DISCONNECTED) &&
							(actor.getConnectors()[j]->getState() != CONNECTED)) {
						success = true;
					}

//#define VERBOSE

#ifdef VERBOSE
					std::ostringstream msg;
					if (success) {
						msg.clear(); msg.str("");
						msg << "Conditions for connecting robot " << getRobot()->getID();
						msg << " to connector " << j << " on robot " << actor.getID() << " are indeed satisfied! :-) :-D : ";
						msg << " distance < " << epsilon << ", angle < -0.9";
						LOG_INFO(msg.str());
					} else {
						msg.clear(); msg.str("");
						msg << "Conditions for connecting robot " << getRobot()->getID();
						msg << " to connector " << j << " on robot " << actor.getID() << " are not satisfied..: ";
						msg << " distance < " << epsilon << ", angle < -0.9";
						LOG_INFO(msg.str());
					}
					msg.clear(); msg.str("");
					msg << "Parameters: ";
					msg << " distance = " << minDist.length();
					if (fabs(angle) < 0.01) angle = 0.0;
					msg << ", angle = " << angle;
					msg << ", checkAngle = " << checkAngle(angle2, 0.1);
					std::string str_state;
					toString(actor.getConnectors()[j]->getState(), str_state);
					msg << ", other\'s state = " << str_state;
					LOG_INFO(msg.str());
#endif

				}

				// Actually connect
				if (success) {
					connecting(actor.getConnectors()[j]);
				}
			}
		}
	}
}

osg::Vec3f ConnectorBase::getTranslation() {

	dtCore::Transform xform;
	GetTransform(xform, dtCore::Transformable::ABS_CS);
	return xform.GetTranslation();

}

//! Returns the global look-at vector of the connector
osg::Vec3f ConnectorBase::getRotation() {

	dtCore::Transform xform;
	GetTransform(xform);
	osg::Vec3 hpr;
	xform.GetRotation(hpr);
	return hpr;

}

//! Returns the look-at-vector of the connector
inline osg::Vec3f ConnectorBase::getLookAtVector() {

	dtCore::Transform xform;
	this->GetTransform(xform);
	osg::Vec3 r,u,f;
	xform.GetOrientation(r,u,f);
	return r;
}

//! Returns the up-vector of the connector
inline osg::Vec3f ConnectorBase::getUpVector() {

	dtCore::Transform xform;
	this->GetTransform(xform);
	osg::Vec3 r,u,f;
	xform.GetOrientation(r,u,f);
	return f;

}

/**
 * Precondition is set by ConnectorBase::connect which only calls connecting when the two
 * connectors are in proximity. The state changes from WANT_TO_CONNECT to CONNECTING. The
 * proximity condition is potentially valid for both connectors, so it needs to be decided
 * which one will be responsible for actually connecting. This is done by comparing the
 * IDs.
 */
void ConnectorBase::connecting(dtCore::RefPtr<ConnectorBase> theOtherConnector) {
	// Do not connect to connectors on the same robot
	if (getRobot()->GetUniqueId() == theOtherConnector->getRobot()->GetUniqueId()) return;

	float k = 30;				//100

	// Set state of both connectors
	setState(CONNECTING);
	theOtherConnector->setState(CONNECTING);

	// Let both connectors know with which connector they are connected
	this->theOtherConnector = theOtherConnector;
	theOtherConnector->theOtherConnector = this;

	// Calculate forces depending on the distance
	osg::Vec3f force = theOtherConnector->getTranslation() - this->getTranslation();
	force.normalize();
	force = osg::Vec3f(k * force.x(), k * force.y(), k * force.z());

	// Decide which connector is responsible for connection
	if (getID() >= theOtherConnector->getID()) {
		// Break out of this routine and start it again, now with the other connector as argument
		theOtherConnector->connecting(this);

	} else {

		dBodyAddForceAtPos (this->getBodyID(), force.x(), force.y(), force.z(),
				this->getTranslation().x(), this->getTranslation().y(), this->getTranslation().z());

		dBodyAddForceAtPos (theOtherConnector->getBodyID(), -force.x(), -force.y(), -force.z(),
				theOtherConnector->getTranslation().x(), theOtherConnector->getTranslation().y(),
				theOtherConnector->getTranslation().z());

		//get distance
		osg::Vec3 minDist = theOtherConnector->getTranslation() - this->getTranslation();
		//myRot.normalize();

		// get angle of the two look-at vectors
		osg::Vec3f rot2 = theOtherConnector->getLookAtVector();
		osg::Vec3f rot1 = this->getLookAtVector();
		rot2.normalize();
		rot1.normalize();
		float angle = rot2 * rot1;

		// get angle of the two up-vectors
		rot2 = theOtherConnector->getUpVector();
		rot1 = this->getUpVector();
		float angle2 = rot2 * rot1;

		// check condition for fixed connection, was epsilon / 4
		if ((minDist.length() < epsilon/1.5) && (angle < -0.99) && (checkAngle(angle2, 0.01)) && (theOtherConnector->getState() != DISCONNECTED)) {

			connect(theOtherConnector);
//			std::ostringstream msg; msg.clear(); msg.str("");
//			msg << "Connected " << getRobot()->getID() << " with " << theOtherConnector->getRobot()->getID();
//			LOG_INFO(msg.str());
		} else if (theOtherConnector->getState() == DISCONNECTED) {
			std::ostringstream msg; msg.clear(); msg.str("");
			msg << "Connection state of " << theOtherConnector->getRobot()->getID() << " is set to disconnected";
			LOG_INFO(msg.str());
		} else {
			//			std::ostringstream msg; msg.clear(); msg.str("");
			//			msg << "Distance: " << minDist.length() << " < " << epsilon / 4 << ", ";
			//			msg << (int)checkAngle(angle2, 0.01) << " to robot " << theOtherConnector->getRobot()->getID();
			//			LOG_INFO(msg.str());
		}

		//check condition for disconnection
		if (!((minDist.length() < epsilon) && (angle < -0.9) && (checkAngle(angle2, 0.1)) && (theOtherConnector->getState() != DISCONNECTED))) {
			disconnect();
		}
	}
	LOG_DEBUG("Connecting process is finished.");
}

/**
 * Scale a vector with a constant @p a.
 */
osg::Vec3 _scale(const float a, const osg::Vec3& vec) {
	return osg::Vec3(a*vec[0], a*vec[1], a*vec[2]);
}

//! subtracts two ode masses, including their center point and inertia tensor
void subMass(dMass* a, dMass* b) {

	int i;
	a->mass = a->mass - b->mass;
	dReal mass_cb = a->mass + b->mass;
	for (i=0; i<3; i++) a->c[i] = (a->c[i]*mass_cb - b->c[i]*b->mass) / a->mass;
	for (i=0; i<12; i++) a->I[i] = a->I[i] - b->I[i];
}

/**
 * Updates the state of the connector, and if different from before, a callback function will
 * be called (if it is set beforehand).
 */
void ConnectorBase::setState(ConnectorState state) {
	if (connectorState != state) {
		connectorState = state;
		ExecuteConnectionEventFunction();
	}
}

/**
 * Returns the state of the connector. Use the getter and setter (instead of writing directly
 * to connectorState), so that eventual callback functions are actually evoked, like the
 * ExecutedConnectionEventFunction.
 */
ConnectorBase::ConnectorState ConnectorBase::getState() const {
	return connectorState;
}

/**
 * Return connected robot if the state of the connector is set to CONNECTED, in any other
 * case this function returns NULL.
 */
RobotActorBase* ConnectorBase::getConnectedRobot() const {
	if (getState() == CONNECTED) return theOtherConnector->getRobot();
	else return NULL;
}

/**
 * Return the other connector (on the other robot).
 */
dtCore::RefPtr<ConnectorBase> ConnectorBase::getOtherConnector() const {
	if (getState() == CONNECTED) return theOtherConnector;
	else return NULL;
}

/**
 * Attaches an ODE joint to the connector. It actually sets the state from whatever state to
 * CONNECTED if the argument is non-NULL.
 */
void ConnectorBase::setConnector(dJointID *connector) {
	if (connector == NULL) {
		setState(DISCONNECTED);
	} else {
		this->connector = *connector;
		setState(CONNECTED);
	}
}

/**
 * Update the state if we are in CONNECTING or in CHECK_CONNECTION.
 */
void ConnectorBase::update(float deltaSimTme) {
	if (getState() == CONNECTING)
		connecting(theOtherConnector);

	if (getState() == CHECK_CONNECTION) {
		connect();
	}
}

/**
 * Finally we end up in the actually connect routine. Previously we have encountered connect(),
 * and connecting(dtCore::RefPtr<ConnectorBase>). The state transitions from CONNECTING to
 * CONNECTED over here.
 *
 * The type of connection is read from: SimulationUtils::Instance()->getConnectorsAsJoints().
 * This variable stems from the scene.xml file as:
 * 		<Scene ... ConnectorsAsJoints="1" ... >
 * The version more likely to have bugs is ConnectorsAsJoints="0" because actually merging bodies
 * and splitting bodies depends on their collision shapes etc. It is not a trivial procedure.
 * After the connector has actually been connected, the function
 */
void ConnectorBase::connect(dtCore::RefPtr<ConnectorBase> theOtherConnector)
{
	if (this->getRobot()->GetUniqueId() == theOtherConnector->getRobot()->GetUniqueId()) return;

	setState(CONNECTED);
	theOtherConnector->setState(CONNECTED);

	// Bus connection, to communicate through the (connected) robot organism
	if ((this->getID()) < (this->theOtherConnector->getID())) {
		dtCore::UniqueId* newBusID = theOtherConnector->getRobot()->getBusID();
		std::vector<RobotActorBase* > robots = getRobot()->getConnectedRobots();
		for (unsigned int i=0; i < robots.size(); i++) robots[i]->setBusID(newBusID);
	}
	else {
		dtCore::UniqueId* newBusID = getRobot()->getBusID();
		std::vector<RobotActorBase* > robots = theOtherConnector->getRobot()->getConnectedRobots();
		for (unsigned int i=0; i < robots.size(); i++) robots[i]->setBusID(newBusID);
	}

	// Let both connectors know with which connector they are connected
	// @todo Already done in connecting?
	this->theOtherConnector = theOtherConnector;
	theOtherConnector->theOtherConnector = this;

	// decide which connector is responsible for connection
	if (getID() >= theOtherConnector->getID()) {

		theOtherConnector->connect(this);

	} else {

		if (SimulationUtils::Instance()->getConnectorsAsJoints()) {
			connectByJoint(theOtherConnector);
		}
		else {
			connectByFusing(theOtherConnector);
		}
	}
}

/**
 * 	This connecting mechanism merges the two bodies together. Therefore:
 *    1. The physics of the other body will be disabled,
 *    2. its collision mesh will be merged to this body,
 *    3. its mass matrix will be added to this one,
 *    4. all its joints will be attached to this body and
 *    5. the other body (and therefore all its children) will be added as child to this one.
 */
void ConnectorBase::connectByFusing(dtCore::RefPtr<ConnectorBase> theOtherConnector) {

	// get parent of my robot body
	dtCore::RefPtr<BodyBase> myParentBody = myBody;
	while (dynamic_cast<BodyBase*>(myParentBody->GetParent())) {
		myParentBody = dynamic_cast<BodyBase*>(myParentBody->GetParent());
	}

	dtCore::RefPtr<BodyBase> theOtherBody = this->theOtherConnector->getSRBody();

	dtCore::RefPtr<BodyBase> theOtherParentBody = theOtherBody;
	while (dynamic_cast<BodyBase*>(theOtherParentBody->GetParent())) {
		theOtherParentBody = dynamic_cast<BodyBase*>(theOtherParentBody->GetParent());
	}


	dtCore::Transform myParentTransform;
	myParentBody->GetTransform(myParentTransform);

	osg::Vec3 parentTranslation = myParentTransform.GetTranslation();
	osg::Vec3 oldCPM = parentTranslation;

	//---- mass stuff --------------------------------------------------------
	// CPM = center point of mass
	// theOtherBody: the Body, what will be attached to myBody
	// myBody: this Body;
	// myParent: root of myBody
	// theOtherParent: root of the other Body


	// add mass of the body to my parent
	dMass theOtherMass;

	dReal *parentPos, *thePos;
	//			dVector3 dist;
	parentPos = const_cast<dReal*>(dBodyGetPosition(myParentBody->GetBodyID()));

	thePos = const_cast<dReal*>(dBodyGetPosition(theOtherParentBody->GetBodyID()));

	osg::Vec3 theParentPosVec = osg::Vec3(thePos[0],thePos[1],thePos[2]);
	osg::Matrix theParentRot;
	dtCore::Transform xform;
	theOtherParentBody->GetTransform(xform, ABS_CS);
	xform.GetRotation(theParentRot);
	osg::Matrix theOtherParentInv;
	theOtherParentInv.invert(theParentRot);
	osg::Vec3 theOtherCPM = xform.GetTranslation();

	osg::Vec3 myParentPosVec = osg::Vec3(parentPos[0],parentPos[1],parentPos[2]);
	osg::Matrix myParentRot;
	theOtherParentBody->GetTransform(xform);
	xform.GetRotation(myParentRot);
	osg::Matrix myParentInv;
	myParentInv.invert(myParentRot);


	BodyBase* allParentBodies = myBody;

	theOtherBody->GetTransform(xform, ABS_CS);
	osg::Vec3 theOldCPM = xform.GetTranslation();

	theOtherBody->GetTransform(xform, ABS_CS);
	osg::Vec3 theNewCPM = xform.GetTranslation();

	// new center point of gravity
	osg::Vec3 newCPM;

	theOtherParentBody->getSubOrgBodyMass(theOtherMass);

	//calculate global position of the cpm of the other body
	osg::Vec3 theMassVec = theOtherParentInv * (osg::Vec3(theOtherMass.c[0], theOtherMass.c[1], theOtherMass.c[2]) + theOtherParentBody->GetModelTranslation()) + theOtherCPM;

	theOtherMass.c[0] = theMassVec[0];
	theOtherMass.c[1] = theMassVec[1];
	theOtherMass.c[2] = theMassVec[2];

	// add mass of the new organism to all suborganisms, beginning with myBody, its parent and so on up to the root
	while (dynamic_cast<BodyBase*>(allParentBodies)) {

		dMass allParentMass;
		allParentBodies->getSubOrgBodyMass(allParentMass);

		dReal *allParentPos;
		allParentPos = const_cast<dReal*>(dBodyGetPosition(allParentBodies->GetBodyID()));

		dtCore::Transform allParentTrans;
		allParentBodies->GetTransform(allParentTrans);
		osg::Matrix allParentRot;
		allParentTrans.GetRotation(allParentRot);
		osg::Matrix allParentInv;
		allParentInv.invert(allParentRot);

		//calculate global position of the cpm of the current parent
		theMassVec = allParentInv * (osg::Vec3(allParentMass.c[0], allParentMass.c[1], allParentMass.c[2]) + allParentBodies->GetModelTranslation()) + allParentTrans.GetTranslation();

		allParentMass.c[0] = theMassVec[0];
		allParentMass.c[1] = theMassVec[1];
		allParentMass.c[2] = theMassVec[2];

		allParentMass.add(&theOtherMass);

		//new Center of Gravity (global), last CPM will be the one of the root = the new CPM
		newCPM = osg::Vec3(allParentMass.c[0], allParentMass.c[1], allParentMass.c[2]);

		//calculate local position of the CPM
		osg::Vec3 allParentMassVec = osg::Vec3(allParentMass.c[0], allParentMass.c[1], allParentMass.c[2]) - allParentTrans.GetTranslation();//myParentPosVec;
		osg::Vec3 allParentMassVecLocal = allParentRot * allParentMassVec - allParentBodies->GetModelTranslation();

		for (int i=0; i < 3; i++)
			allParentMass.c[i] = allParentMassVecLocal[i];

		// assign new mass of the sub-organism
		allParentBodies->setSubOrgBodyMass(&allParentMass);

		//ODE needs the mass center at the origin
		allParentMass.c[0] = 0;
		allParentMass.c[1] = 0;
		allParentMass.c[2] = 0;

		allParentBodies->SetMass(&allParentMass);

		//go upwards
		allParentBodies = dynamic_cast<BodyBase*>(allParentBodies->GetParent());
	}

	BodyBase* allOtherParentBodies = dynamic_cast<BodyBase*>(theOtherBody.get());
	std::vector<dMass> allOtherMasses;

	// Calculate the new masses for the other parents as their child-parent relationship between theOtherBody and theOtherParentBody will be switched
	//

	// Push all the other masses into a std::vector
	while (allOtherParentBodies) {
		dMass m;
		allOtherParentBodies->getSubOrgBodyMass(m);

		osg::Matrix allOtherMatrices;
		dtCore::Transform xform;
		allOtherParentBodies->GetTransform(xform);
		xform.GetRotation(allOtherMatrices);
		osg::Matrix allOtherInverse;
		allOtherInverse.invert(allOtherMatrices);

		osg::Vec3 allParentMassVecGlobal = allOtherInverse * (osg::Vec3(m.c[0], m.c[1], m.c[2]) + allOtherParentBodies->GetModelTranslation()) + xform.GetTranslation();

		for (int i=0; i < 3; i++)
			m.c[i] = allParentMassVecGlobal[i];


		allOtherMasses.push_back(m);
		allOtherParentBodies = dynamic_cast<BodyBase*>(allOtherParentBodies->GetParent());
	}

	//switch child-parent relation for all the "former" parents of the other body
	allOtherParentBodies = dynamic_cast<BodyBase*>(theOtherBody.get());


	for (unsigned int i=0; i<allOtherMasses.size(); i++) {

		dMass m;
		m = allOtherMasses[allOtherMasses.size() - 1];

		dMass m_x;
		if (i < 1) m_x.setZero();
		else m_x = allOtherMasses[i - 1];

		//m_0:...... mass of the other body
		//m_n:...... mass of root body
		//m_x' = m_n - m_(x-1)

		subMass(&m, &m_x);

		dtCore::Transform xform;
		allOtherParentBodies->GetTransform(xform);
		osg::Matrix allParentRot;
		xform.GetRotation(allParentRot);

		osg::Vec3 allParentMassVec = osg::Vec3(m.c[0], m.c[1], m.c[2]) - xform.GetTranslation();
		osg::Vec3 allParentMassVecLocal = allParentRot * allParentMassVec - allOtherParentBodies->GetModelTranslation();
		for (int i=0; i < 3; i++)
			m.c[i] = allParentMassVecLocal[i];

		allOtherParentBodies->setSubOrgBodyMass(&m);

		m.c[0] = 0;
		m.c[1] = 0;
		m.c[2] = 0;


		allOtherParentBodies->SetMass(&m);
		allOtherParentBodies = dynamic_cast<BodyBase*>(allOtherParentBodies->GetParent());
	}

	dtCore::RefPtr<BodyBase> currentBody;
	BodyBase* currentParentBody = theOtherBody;
	BodyBase* newParentBody = myBody;

	//			osgDB::writeNodeFile(*myParentBody->getCollisionModel(), "myParentNode.osg");
	//			osgDB::writeNodeFile(*theOtherParentBody->getCollisionModel(), "theOtherParentNode.osg");

	// merge organism
	// switch parent-child relation of theOtherParent and its parent, do also for the parent's parent up to the root
	// myBody will be parent of theOtherParent
	while (dynamic_cast<BodyBase*>(currentParentBody)) {

		currentBody = currentParentBody;
		currentParentBody = dynamic_cast<BodyBase*>(currentParentBody->GetParent());
		newParentBody->mergeBody(currentBody);

		newParentBody = currentBody;

	}

	parentTranslation = newCPM;

	// move root to the CPM
	myParentTransform.SetTranslation(parentTranslation);
	myParentBody->SetTransform(myParentTransform);

	dtCore::Transform myParentBodyTransform;
	myParentBody->GetTransform(myParentBodyTransform);
	osg::Matrix myParentBodyMatrix;
	osg::Matrix myParentBodyInverse;
	myParentBodyTransform.GetRotation(myParentBodyMatrix);
	myParentBodyInverse.invert(myParentBodyMatrix);

	// move collision model in such a way that it is at the same global position as it was before moving the Delta-3D body
	osgBodyTransform* myParentCollisionNode = myParentBody->getCollisionModel();
	myParentCollisionNode->setPosition(myParentCollisionNode->getPosition() + (myParentBodyMatrix * (oldCPM - newCPM)));

	myParentBody->setCollisionModel(myParentCollisionNode);
	myParentBody->SetCollisionMesh(myParentCollisionNode);

	dtCore::Transform childTransform;
	osg::Matrix childRotation;
	theOtherBody->GetTransform(childTransform);
	childTransform.GetRotation(childRotation);
	osg::Matrix childInverse;
	childInverse.invert(childRotation);

	dtCore::Transform oldRootTrans;
	theOtherParentBody->GetTransform(oldRootTrans, ABS_CS);

	// set all model transaltion inside the organism in such a way that they are at the same global position as they have been before
	myParentBody->setAllModelTranslations(*myParentBody, "", oldCPM - newCPM );


	osg::Vec3 tempTrans = theOtherBody->GetModelTranslation();
	theOtherBody->RenderProxyNode(false);

	//			osgDB::writeNodeFile(*myParentCollisionNode, "myParentCollisionNode.osg");

	//reset the joints, so that they keep their position
	dBodyID myParentBodyID = myParentBody->GetBodyID();
	for (int i=0; i<dBodyGetNumJoints (myParentBodyID); i++) {

		dJointID joint = dBodyGetJoint (myParentBodyID, i);
		float angle = dJointGetHingeAngle (joint);

		dVector3 pos;
		dVector3 axis;
		dJointGetHingeAxis(joint, axis);

		dBodyID b0,b1;
		b0 = dJointGetBody(joint,0);
		b1 = dJointGetBody(joint,1);

		if (b1 == myParentBodyID) dJointGetHingeAnchor(joint, pos);
		else dJointGetHingeAnchor2(joint, pos);

		dJointAttach(joint, b0, b1);

		dJointSetHingeAnchor(joint, pos[0], pos[1], pos[2]);
		dJointSetHingeAxis(joint, axis[0], axis[1], axis[2]);

		std::map< dJointID, ActuatorBase*>::iterator iter = ActuatorBase::jointActuatorMap.begin();
		iter = ActuatorBase::jointActuatorMap.find(joint);

		if (iter != ActuatorBase::jointActuatorMap.end()) {
			HingeActuator* theHinge = dynamic_cast<HingeActuator*>(iter->second);
			if (theHinge) theHinge->setZeroAngle(angle);
		}

	}

	myParentBody->RenderCollisionGeometry(true);
	myParentBody->RenderProxyNode(true);
}

/**
 * Connect two bodies by adding an (ODE) joint between them. Of course all joints need to be
 * calculated. For that reason connectByFusing is implemented that merges two bodies into one.
 */
void ConnectorBase::connectByJoint(dtCore::RefPtr<ConnectorBase> theOtherConnector) {

	// create joint that physically connects the two bodies, where the connectors are attached to
	connector = dJointCreateFixed(GetSceneParent()->GetWorldID(), 0);

	dJointAttach(connector, myBody->GetBodyID(), theOtherConnector->getBodyID());

	dJointSetFixed (connector);
}

//! Auxiliary function to reset actuators
void resetAllActuators(BodyBase& newRootBody, BodyBase* oldRootBody,  BodyBase* theBody, osg::Vec3 rootVel ) {
	std::vector<dtCore::RefPtr<ActuatorBase> > actuators = theBody->getRobot()->getActuators();
	for (unsigned int i=0; i<actuators.size(); i++) {

		dJointID joint = actuators[i]->getJoint();
		dBodyID b0,b1;

		float angle = dJointGetHingeAngle (joint);

		b0 = dJointGetBody(joint,0);
		b1 = dJointGetBody(joint,1);

		dVector3 axis;
		dJointGetHingeAxis(joint, axis);

		dReal* pos0;
		dReal* pos1;

		dBodyID newRootBodyID = newRootBody.GetBodyID();
		dBodyID rootBodyID = oldRootBody->GetBodyID();
		dBodyID theBodyID = theBody->GetBodyID();

		if (b0 == rootBodyID) {
			pos1 = const_cast<dReal*>(dBodyGetPosition   (theBodyID));
			pos0 = const_cast<dReal*>(dBodyGetPosition   (b1));

			//			dBodySetAngularVel (b1, 0,0,0);
			dBodySetLinearVel (b1, rootVel[0], rootVel[1], rootVel[2]);
			dJointAttach(joint, newRootBodyID, b1);

		}
		else if (b1 == rootBodyID) {
			pos0 = const_cast<dReal*>(dBodyGetPosition   (b0));
			pos1 = const_cast<dReal*>(dBodyGetPosition   (theBodyID));

			//			dBodySetAngularVel (b0, 0,0,0);
			dBodySetLinearVel (b0, rootVel[0], rootVel[1], rootVel[2]);
			dJointAttach(joint, b0, newRootBodyID);

		}

		dJointSetHingeAnchor(joint, pos0[0], pos0[1], pos0[2]);
		dJointSetHingeAxis(joint, axis[0], axis[1], axis[2]);
		dynamic_cast<HingeActuator*>(actuators[i].get())->setZeroAngle(angle);
	}

	for (unsigned int i=0; i<theBody->GetNumChildren(); i++) {
		BodyBase* child = dynamic_cast<BodyBase*>(theBody->GetChild(i));
		if (child)
			resetAllActuators(newRootBody, oldRootBody, child, rootVel);
	}


}

//! Auxiliary function to print out the mass tree
void printAllBodyMasses(BodyBase* body) {

	dMass mass;
	body->getSubOrgBodyMass(mass);
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "suborg mass of " << body->GetName() << std::endl;
	msg << "mass = " << mass.mass << std::endl;
	msg << "mass center = " << osg::Vec3(mass.c[0], mass.c[1], mass.c[2]) << std::endl;
	msg << "mass inertia tensor = " << std::endl;
	msg << osg::Vec3(mass.I[0],mass.I[1],mass.I[2]) << std::endl;
	msg << osg::Vec3(mass.I[3],mass.I[4],mass.I[5]) << std::endl;
	msg << osg::Vec3(mass.I[6],mass.I[7],mass.I[8]) << std::endl;
	msg << osg::Vec3(mass.I[9],mass.I[10],mass.I[11]) << std::endl;
	LOG_INFO(msg.str());

	msg.clear(); msg.str("");
	msg << "children of " << body->GetName() << " ------------------------------------------------------------------" << std::endl;
	for (unsigned int i=0; i<body->GetNumChildren(); i++)
		if (dynamic_cast<BodyBase*>(body->GetChild(i)))
			printAllBodyMasses(dynamic_cast<BodyBase*>(body->GetChild(i)));
	msg << " ------------------------------------------------------------------------------------------------------" << std::endl;
	LOG_INFO(msg.str());
}

/**
 * The actual disconnection. There seems to be something specific about the order in which a
 * disconnect needs to be done. There is a statement here that checks for getID() versus
 * theOtherConnector->getID(), and decides which connection should disconnect first. This
 * seems to be buggy.
 */
void ConnectorBase::disconnect() {

	// Check if connector is already disconnected (or in the process of disconnection)
	if (getState() != CONNECTED) {
//		std::ostringstream msg; msg.clear(); msg.str("");
//		msg << "Connector state is ";
//		std::string str_state;
//		toString(getState(), str_state);
//		msg << str_state << " of connector " << getID()->ToString() << " (so break out of disconnect)";
//		LOG_INFO(msg.str());
		setState(DISCONNECTED);
		return;
	}

	setState(DISCONNECTED);

	dtCore::RefPtr<ConnectorBase> disConnector = this->theOtherConnector;

	//the one who owns the connector, destroys the connector and disconnect (virtually) on the other side as well
	if (theOtherConnector != NULL) {

//		std::ostringstream msg; msg.clear(); msg.str("");
//		msg << "Disconnect connector " << getID()->ToString() << " for real";
//		LOG_DEBUG(msg.str());

		if ((this->getID()) >= (this->theOtherConnector->getID())) {
			// there is no other connector anymore, this connector is attached to
			this->setConnector(NULL);

			// the other connector needs to be disconnected as well
			if (this->theOtherConnector->getState() != DISCONNECTED) theOtherConnector->disconnect();

		} else {
			//LOG_INFO("disconnect");

			if (SimulationUtils::Instance()->getConnectorsAsJoints()) {
				// destroy joint
				dJointDestroy(this->connector);
			}

			//TODO: the new connecting principle only works for organisms which have a topology which can be represented with a graph without cycles.
			// When disconnecting the organism can be divided into two, therefore:
			// 1. Disconnect the two robots from each other.
			// 2. Activate the physics of the child and set it to the current position and orientation (quat)
			// 3. add all actuator joints of the child's children to the child
			// 4. set the new collision geometries to the root-parent body and to the new parent.

			// Build two new organisms when disconnecting.
			//
			else {

				//get the parent, destroy its organism collision mesh and give back its robot collision mesh.
				dtCore::RefPtr<BodyBase> rootBody = myBody;
				while (dynamic_cast<BodyBase*>(rootBody->GetParent())) {
					rootBody = dynamic_cast<BodyBase*>(rootBody->GetParent());
				}


				osgBodyTransform* rootBodyNode = rootBody->getCollisionModel();

				osgBodyTransform* myCollisionNode = myBody->getCollisionModel();
				osgBodyTransform* theOtherCollisionNode = theOtherConnector->getSRBody()->getCollisionModel();

				dVector3 vel;

				dtCore::RefPtr<BodyBase> newRootBody;
				dtCore::RefPtr<BodyBase> oldParentBody;				//the former parent of our new root


				//retrieve who is parent and remove the connection
				if (!myCollisionNode->removeChild(theOtherCollisionNode)) {

					//theOtherCollisionNode was parent
					theOtherCollisionNode->removeChild(myCollisionNode);
					oldParentBody = theOtherConnector->getSRBody();
					newRootBody = myBody;

				}

				else {

					oldParentBody = myBody;
					newRootBody = theOtherConnector->getSRBody();

				}

				osgBodyTransform* newRootBodyNode = newRootBody->getCollisionModel();
				//				osgBodyTransform* oldParentBodyNode = oldParentBody->getCollisionModel();

				dtCore::Transform xform;
				newRootBody->GetTransform(xform, dtCore::Transformable::ABS_CS);
				oldParentBody->RemoveChild(newRootBody);

				//Add newRootBody to the scene, as it has been removed and reposition it
				dtCore::Scene* scene = oldParentBody->GetSceneParent ();
				scene->AddDrawable(newRootBody);
				newRootBody->SetTransform(xform, dtCore::Transformable::ABS_CS);

				newRootBody->RenderCollisionGeometry(false);

				//set collision models to the both new organisms
				rootBody->setCollisionModel(rootBodyNode);
				newRootBody->setCollisionModel(newRootBodyNode);
				rootBody->RenderCollisionGeometry(true);

				dBodyGetPointVel (rootBody->GetBodyID(), newRootBody->getTranslation().x(), newRootBody->getTranslation().y(), newRootBody->getTranslation().z(), vel);

				dBodyID rootBodyID = rootBody->GetBodyID();
				dBodyID newRootBodyID = newRootBody->GetBodyID();

				dVector3 rootVel;
				dBodyGetPointVel (rootBodyID, 0, 0, 0, rootVel);




				//---- begin mass stuff --------------------------------------------------------
				//1. Retrieve masses for rootBody and newRootBody
				dMass newRootMass;
				newRootBody->getSubOrgBodyMass(newRootMass);

				//2. update all parent masses p.mass -= newOrg.mass from oldParentBody up to rootBody
				//new center point of Gravity of old organism
				osg::Vec3 rootCPM;

				BodyBase* allParentBodies = oldParentBody;

				dReal *rootPos, *newRootPos;
				rootPos = const_cast<dReal*>(dBodyGetPosition(rootBody->GetBodyID()));
				newRootPos = const_cast<dReal*>(dBodyGetPosition(newRootBody->GetBodyID()));

				dtCore::Transform  newRootTransform;
				newRootBody->GetTransform(newRootTransform);
				osg::Matrix newRootRot;
				newRootTransform.GetRotation(newRootRot);
				osg::Matrix newRootInv;
				newRootInv.invert(newRootRot);

				//calculate global center of mass
				osg::Vec3 massVec = newRootTransform.GetTranslation() + newRootInv * (newRootBody->GetModelTranslation() + osg::Vec3(newRootMass.c[0],newRootMass.c[1],newRootMass.c[2]));
				for (int i=0; i<3; i++) {
					newRootMass.c[i] =  massVec[i];
				}

				//for all parents between oldParent (=new leaf) and root, subtract the mass of the newRootBody
				while (allParentBodies) {

					dMass parentMass;
					allParentBodies->getSubOrgBodyMass(parentMass);
					rootCPM = osg::Vec3(parentMass.c[0],parentMass.c[1],parentMass.c[2]);

					dtCore::Transform allParentTransform;
					allParentBodies->GetTransform(allParentTransform);
					newRootBody->GetTransform(newRootTransform);
					osg::Matrix allParentRot;
					osg::Vec3 allParentTrans = allParentTransform.GetTranslation();
					allParentTransform.GetRotation(allParentRot);
					osg::Matrix allParentInv;
					allParentInv.invert(allParentRot);

					osg::Vec3 massVec = allParentInv * (osg::Vec3(parentMass.c[0], parentMass.c[1], parentMass.c[2]) + allParentBodies->GetModelTranslation()) + allParentTrans;

					for (int i=0; i<3; i++) {
						parentMass.c[i] = massVec[i];
					}
					osg::Vec3 pM = osg::Vec3(parentMass.c[0], parentMass.c[1], parentMass.c[2]);
					osg::Vec3 nRW = osg::Vec3(newRootMass.c[0],newRootMass.c[1],newRootMass.c[2]);

					subMass(&parentMass, &newRootMass);

					//will become the new center point of mass of the root
					rootCPM = osg::Vec3(parentMass.c[0],parentMass.c[1],parentMass.c[2]);

					pM = rootCPM - allParentTrans;
					pM = (allParentRot * pM) - allParentBodies->GetModelTranslation();

					for (int i=0; i<3; i++) {
						parentMass.c[i] = pM[i];
					}

					allParentBodies->setSubOrgBodyMass(&parentMass);

					//ode needs center point of mass at the origin
					parentMass.c[0] = 0;
					parentMass.c[1] = 0;
					parentMass.c[2] = 0;

					allParentBodies->SetMass(&parentMass);
					allParentBodies = dynamic_cast<BodyBase*>(allParentBodies->GetParent());
				}

				newRootBody->getSubOrgBodyMass(newRootMass);
				osg::Vec3 newRootCPMLocal = osg::Vec3(newRootMass.c[0],newRootMass.c[1],newRootMass.c[2]);
				newRootMass.c[0] = 0;
				newRootMass.c[1] = 0;
				newRootMass.c[2] = 0;
				newRootBody->SetMass(&newRootMass);

				//---- end mass stuff --------------------------------------------------------

				//3. Reset position of old organism in such a way, that the center point of mass is in the origin
				dtCore::Transform rootTransform;
				rootBody->GetTransform(rootTransform, dtCore::Transformable::ABS_CS);

				osg::Matrix rootMatrix;
				osg::Matrix rootInverse;
				rootTransform.GetRotation(rootMatrix);
				rootInverse.invert(rootMatrix);

				//global position of the model
				osg::Vec3 rootOsgPos = rootTransform.GetTranslation() + rootInverse * rootBody->GetModelTranslation();

				//reset models so that they appear at the same position as before the disconnect
				rootBody->setAllModelTranslations(*rootBody, newRootBody->GetName(), (-rootCPM + rootOsgPos) - rootInverse * rootBody->GetModelTranslation());
				rootTransform.SetTranslation(rootCPM);
				rootBody->SetTransform(rootTransform, dtCore::Transformable::ABS_CS);

				//reset the collision geometry in such a way that they are at the same pos as before the disconnect
				rootBodyNode->setPosition(rootMatrix * (-rootCPM + rootOsgPos));
				rootBody->setCollisionModel(rootBodyNode);
				rootBody->SetCollisionMesh(rootBodyNode);

				//4. ... and the new organism
				newRootBody->GetTransform(newRootTransform, dtCore::Transformable::ABS_CS);
				osg::Matrix newRootBodyMatrix;
				osg::Matrix newRootBodyInverse;
				newRootTransform.GetRotation(newRootBodyMatrix);
				newRootBodyInverse.invert(newRootBodyMatrix);

				//global position of the model
				osg::Vec3 newRootOsgPos = newRootTransform.GetTranslation() + newRootBodyInverse * newRootBody->GetModelTranslation();

				//new Delta3D position of the newRootBody
				newRootTransform.SetTranslation(newRootTransform.GetTranslation() + newRootBodyInverse * (newRootCPMLocal + newRootBody->GetModelTranslation()));
				newRootBody->SetTransform(newRootTransform, dtCore::Transformable::ABS_CS);

				osg::Matrix newRootMat;
				osg::Quat newRootQuat;
				newRootMat.makeIdentity();
				newRootMat.get(newRootQuat);

				//set position and attitude (=orientation) of the collision geometry
				newRootBodyNode->setPosition(-newRootCPMLocal);
				newRootBodyNode->setAttitude(newRootQuat);

				//add all (collision geometry) bodies that belong to the new root and remove them from the old root
				osgBodyNodeVisitor* bnv = new osgBodyNodeVisitor(newRootBody);
				if (newRootBodyNode)
					newRootBodyNode->traverse(*bnv);

				newRootBody = bnv->returnBodyRoot();

				newRootBody->EnableDynamics(true);
				newRootBody->SetCollisionDetection(true);

				//assign new collision mesh and model to the new root
				newRootBody->SetCollisionMesh(newRootBodyNode);
				newRootBody->setCollisionModel(newRootBodyNode);

				//reset all model translations in such a way that the appear at the same (global) position as before the disconnect
				newRootBody->setAllModelTranslations(*newRootBody, "", newRootBodyInverse * (-newRootCPMLocal - newRootBody->GetModelTranslation()));

				//set position of the ODE body (=Delta3D body) in such a way that the CPM is at the origin
				newRootBody->GetTransform(xform, dtCore::Transformable::ABS_CS);
				dBodySetPosition(newRootBody->GetBodyID(), xform.GetTranslation().x(), xform.GetTranslation().y(), xform.GetTranslation().z());
				osg::Matrix m;
				xform.GetRotation(m);
				osg::Quat q = m.getRotate();
				dQuaternion odeQuat;
				odeQuat[0] = q.w();
				odeQuat[1] = q.x();
				odeQuat[2] = q.y();
				odeQuat[3] = q.z();

				//reset origin in such a way that it is at the same global position as it has been before the disconnect
				dBodySetQuaternion(newRootBody->GetBodyID(), odeQuat);

				dBodyGetPointVel (rootBodyID, xform.GetTranslation().x(), xform.GetTranslation().y(), xform.GetTranslation().z(), rootVel);

				//assign actuators to the new root
				resetAllActuators(*newRootBody, rootBody.get(), newRootBody.get(), osg::Vec3(rootVel[0],rootVel[1],rootVel[2]));

				//				osgDB::writeNodeFile(*(newRootBody->GetOSGNode()), "myBodyBNV.osg");

				//					dBodySetAngularVel (newRootBody->GetBodyID(), 0,0,0);
				dBodySetLinearVel  (newRootBody->GetBodyID(), rootVel[0], rootVel[1], rootVel[2]);

				rootBody->ClearCollisionGeometry();
				rootBody->SetCollisionMesh(rootBodyNode);
				rootBody->RenderCollisionGeometry(true);

				//reset/reposition all actuators of the new root in such a way that they are at the same global position
				for (int i=0; i<dBodyGetNumJoints (newRootBodyID); i++) {

					dJointID joint = dBodyGetJoint (newRootBodyID, i);

					dVector3 pos;
					dVector3 axis;
					dReal angle = 0;

					if (dJointGetType(joint) == dJointTypeHinge)
						angle = dJointGetHingeAngle (joint);

					dJointGetHingeAxis(joint, axis);

					dBodyID b0,b1;
					b0 = dJointGetBody(joint,0);
					b1 = dJointGetBody(joint,1);

					if (b1 == newRootBodyID) dJointGetHingeAnchor(joint, pos);
					else dJointGetHingeAnchor2(joint, pos);

					dJointAttach(joint, b0, b1);

					dJointSetHingeAnchor(joint, pos[0], pos[1], pos[2]);
					dJointSetHingeAxis(joint, axis[0], axis[1], axis[2]);

					std::map< dJointID, ActuatorBase*>::iterator iter = ActuatorBase::jointActuatorMap.begin();
					iter = ActuatorBase::jointActuatorMap.find(joint);

					if (iter != ActuatorBase::jointActuatorMap.end()) {
						HingeActuator* theHinge = dynamic_cast<HingeActuator*>(iter->second);
						if (theHinge) theHinge->setZeroAngle(angle);
					}


				}

				//reset/reposition all actuators of the old root in such a way that they are at the same global position
				for (int i=0; i<dBodyGetNumJoints (rootBodyID); i++) {

					dJointID joint = dBodyGetJoint (rootBodyID, i);

					dVector3 pos;
					dVector3 axis;

					dReal angle = 0;

					if (dJointGetType(joint) == dJointTypeHinge)
						angle = dJointGetHingeAngle (joint);

					dJointGetHingeAxis(joint, axis);

					dBodyID b0,b1;
					b0 = dJointGetBody(joint,0);
					b1 = dJointGetBody(joint,1);

					if (b1 == rootBodyID) dJointGetHingeAnchor(joint, pos);
					else dJointGetHingeAnchor2(joint, pos);


					dJointAttach(joint, b0, b1);
					dJointSetHingeAnchor(joint, pos[0], pos[1], pos[2]);
					dJointSetHingeAxis(joint, axis[0], axis[1], axis[2]);

					std::map< dJointID, ActuatorBase*>::iterator iter = ActuatorBase::jointActuatorMap.begin();
					iter = ActuatorBase::jointActuatorMap.find(joint);

					if (iter != ActuatorBase::jointActuatorMap.end()) {
						HingeActuator* theHinge = dynamic_cast<HingeActuator*>(iter->second);
						if (theHinge) theHinge->setZeroAngle(angle);
					}
				}

				rootBody->RenderProxyNode(true);
				newRootBody->RenderProxyNode(true);

				rootBody->RenderCollisionGeometry(true);
				newRootBody->RenderCollisionGeometry(true);

			}
		}
	} else {

		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Connector " << getID()->ToString();
		msg << " is owned by other module. It will not be disconnected (yet)";
		LOG_INFO(msg.str());

	}
}

/**
 * Create string representation of given connector state.
 */
void ConnectorBase::toString(ConnectorState state, std::string &str) {
	switch(state) {
	case CONNECTED: str.append("CONNECTED"); break;
	case DISCONNECTED: str.append("DISCONNECTED"); break;
	case CONNECTING: str.append("CONNECTING"); break;
	case CHECK_CONNECTION: str.append("CHECK_CONNECTION"); break;
	case WANT_TO_CONNECT: str.append("WANT_TO_CONNECT"); break;
	default:
		str.append("UNDEFINED");
	}
}

