/*
 * Robot3D Physical realistic 3D simulator for robots
 *
 * @license GNU Lesser General Public License
 *
 * @author Adrian Friebel
 * @author Anne C. van Rossum
 */

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/arenaComponent.h>
#include <srCore/body/bodyBase.h>
#include <srCore/cameraComponent.h>
#include <srCore/simulationEvents.h>
#include <srCore/robotCreator.h>

namespace srCore {

const std::string SimulationArenaComponent::DEFAULT_NAME("DefaultArenaComponent");

////////////////////////////////////////////////////////////////////
SimulationArenaComponent::SimulationArenaComponent(const std::string &name)
 : dtGame::GMComponent(name),
   isResettingTheScene (false),
   paused(false)
{
	numRobots 	= 0;

	numRobotsAtBeginOfFrame = 0;
	robotFocus 	= 0;
	robotFocusIndex = 0;

	mFocusChanged = new dtDAL::GameEvent( ROBOT_SELECTION_CHANGED );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mFocusChanged);

	mActorDestroyed = new dtDAL::GameEvent( ACTOR_DESTROYED );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mActorDestroyed);
}

void SimulationArenaComponent::init()
{
	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);

	for(int i=0; i<numRobots; i++)
	{
		RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[i]->GetGameActor());

		if(robotFocus == actor.getID())
		{
			robotFocusIndex = i;
			actor.setSelectionState(true);
			actor.renderCollisionGeometry(true);
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
void SimulationArenaComponent::sendMessageToAll(dtDAL::GameEvent &event)
{
   dtCore::RefPtr<dtGame::GameEventMessage> eventMsg;
   GetGameManager()->GetMessageFactory().CreateMessage(dtGame::MessageType::INFO_GAME_EVENT, eventMsg);

   eventMsg->SetGameEvent(event);
   GetGameManager()->SendMessage(*eventMsg);
}

//////////////////////////////////////////////////////////////////////////
void SimulationArenaComponent::ProcessMessage(const dtGame::Message& message)
{

	if (message.GetMessageType() == dtGame::MessageType::TICK_END_OF_FRAME) {

		if (isResettingTheScene && (numRobots == 0)) {
			LOG_ERROR("Not possible, see GetConfig twice BUG");
			assert (false);
//			createRobotsAndControllersFromSceneFile(srCore::SimulationUtils::Instance()->getSceneFile());
			isResettingTheScene = false;
		}
	}


   if ((!isResettingTheScene) && (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL)) {
		std::vector< dtGame::GameActorProxy* > allGameActors;
		GetGameManager()->GetAllGameActors(allGameActors);
		numRobots =  robotsInScene.size();
		dJointGroupEmpty(SimulationUtils::Instance()->getContactJointGroupID());


	}

	if (message.GetMessageType() == dtGame::MessageType::INFO_ACTOR_CREATED) {

		srCore::RobotActorBase *actor = dynamic_cast<srCore::RobotActorBase *>(GetGameManager()->FindActorById(message.GetAboutActorId())->GetActor());

		if (actor) {
			this->robotsInScene.push_back(actor);
			selectRobot(actor);
			sendMessageToAll(*mFocusChanged);
		}

	}


	if (message.GetMessageType() == dtGame::MessageType::INFO_ACTOR_DELETED) {

			if (robotsInScene.size() > 0)
				selectRobot(robotsInScene[ (robotFocusIndex) % robotsInScene.size() ]);
			else
				selectRobot(NULL);
	}

	if ((!isResettingTheScene) && (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT))
	{
		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);

		if (eventMsg.GetGameEvent() != NULL)
		{
			if (eventMsg.GetGameEvent()->GetName() == PAUSE_SIM)
			{
				if(paused)		start();
				else			pause();

			}
			else if (eventMsg.GetGameEvent()->GetName() == RESET_SIM) {
				reset();
			}
			else if (eventMsg.GetGameEvent()->GetName() == DESTROY_ACTOR)
				destroyRobot();
			else if (eventMsg.GetGameEvent()->GetName() == DESTROY_ALL_ACTOR)
				clear();
			else if (eventMsg.GetGameEvent()->GetName() == CREATE_ACTOR) {
				LOG_ERROR("Signature of file changed, GetConfig was created twice. We don't want to load everything twice...");
				LOG_ERROR("Fix it and then you can use this Game event again...");
				assert (false);
//				createRobotsAndControllersFromSceneFile("../scenarios/scene.xml");
			}
			else if(eventMsg.GetGameEvent()->GetName() == CHANGE_ROBOT_SELECTION )
			{
				if(numRobots>0)
				{
					std::vector< dtGame::GameActorProxy* > allGameActors;
					GetGameManager()->GetAllGameActors(allGameActors);

					//RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[robotFocusIndex]->GetGameActor());
					RobotActorBase *selectedActor = robotsInScene[robotFocusIndex];
					selectedActor->setSelectionState(false);
					selectedActor->renderCollisionGeometry(false);

					std::string number = eventMsg.GetGameEvent()->GetDescription();
					if(number.empty())
					{
//						robotFocusIndex = (robotFocusIndex+1) % numRobots;
//						RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[robotFocusIndex]->GetGameActor());
//						robotFocus = actor.getID();
//						actor.setSelectionState(true);
//						actor.renderCollisionGeometry(true);

						if (robotsInScene.size() > 0) {
							robotFocusIndex = (robotFocusIndex+1) % robotsInScene.size();
							RobotActorBase *actor = robotsInScene[robotFocusIndex];
							robotFocus = actor->getID();
							actor->setSelectionState(true);
							actor->renderCollisionGeometry(true);
						}

					}
					else
					{
						robotFocus = atoi(number.c_str());
						//init();
					}
					sendMessageToAll(*mFocusChanged);
				}
			}
		}
	}
}

void SimulationArenaComponent::loadScene(std::string file)
{
	srCore::SimulationUtils::Instance()->setSceneFile(file);
	reset();

}

void SimulationArenaComponent::reset()
{
	clear();
	isResettingTheScene = true;
}

void SimulationArenaComponent::start()
{
	paused = false;
	GetGameManager()->SetPaused(false);
	GetGameManager()->ChangeTimeSettings(GetGameManager()->GetSimulationTime(), timeScale,
			GetGameManager()->GetSimulationClockTime());
}

void SimulationArenaComponent::pause()
{
	paused = true;
	GetGameManager()->SetPaused(true);
	timeScale = GetGameManager()->GetTimeScale();
	GetGameManager()->ChangeTimeSettings(GetGameManager()->GetSimulationTime(), 0,
				GetGameManager()->GetSimulationClockTime());
}

/**
 * Toggles the creation of screenshots. Please, notice that this does not guarantee that the shots
 * are actually stored. The directory might not exist for example.
 */
void SimulationArenaComponent::toggleScreenShots(bool toggle)
{
	SimulationCameraController* guiCamera = dynamic_cast<SimulationCameraController*>(
			GetGameManager()->GetComponentByName("SimulationCameraController"));
	assert (guiCamera != NULL);
	guiCamera->toggleScreenShots(toggle);
}

/**
 * Create robot from RobotInitData, the struct set from the XML configuration file. This function
 * adds a robot to the game manager in the predefined position, but corrected in the z-dimension.
 * This is because the default Stage files seem to have the floor surface hovering above the actual
 * z=0 plane. Originally this contained also SensorBase->setup() calls for each sensor, but this is
 * moved to RobotActorBase.
 * @todo Include a check for the ground level on z = 0, and remove getZPos.
 */
void SimulationArenaComponent::createRobot(RobotInitData rid) {
	static int robIdx = 0;
	osg::Vec3f pos = rid.pos;
	float z = getZPos(pos.x(), pos.y(), GetGameManager()->GetScene());

	LOG_ALWAYS("Raise robot (temporarily) slightly above ground, setting z-axis");
	if(!freedInd.empty()) {
		SimulationRobotCreator::Instance()->addRobotToGM(freedInd[0], rid, z);
		freedInd.erase(freedInd.begin());
	} else {
		SimulationRobotCreator::Instance()->addRobotToGM(robIdx++, rid, z);
	}
}

/**
 * Retrieve a robot by its name. It uses the predefined functions in the GameManager. There are even
 * function templates that can search for other attributes than name and type. If the robot is not found,
 * the function quits aggressively.
 * @name The name of the robot returns a smart pointer to the RobotActorBase
 */
dtCore::RefPtr<RobotActorBase> SimulationArenaComponent::getRobot(const std::string &name) {
	RobotActorBaseProxy *proxy = NULL;
	GetGameManager()->FindActorByName(name, proxy);
	if (proxy != NULL) {
		dtCore::RefPtr<RobotActorBase> actor;
		actor = &static_cast<RobotActorBase&>(proxy->GetGameActor());
		return actor;
	} else {
		LOG_ERROR("Error, an actor with this name does not exist!");
		assert (false);
	}
	return NULL;
}

/**
 * The main routine loading all robots from the predefined XML file. Be careful not to load things twice.
 */
void SimulationArenaComponent::createRobotsFromSceneFile(const GetConfig &appConfig)
{
	LOG_INFO("Create robots from scene file");

	int numRobOld = numRobots;
	numRobots = numRobots + appConfig.getRobots().size();

	// setup robots
	for (unsigned int i=0; i<appConfig.getRobots().size(); i++) {
		createRobot(appConfig.getRobots()[i]);
	}
	std::vector<dtGame::GameActorProxy*> allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);

	std::ostringstream printmsg;
	printmsg.clear(); printmsg.str("");
	printmsg << "Load a total amount of " << allGameActors.size() << " actors";
	LOG_INFO(printmsg.str());

	if (!allGameActors.empty()) {
		for(unsigned int i=0; i<appConfig.getRobots().size(); i++) {
			RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[numRobOld+i]->GetGameActor());
			std::vector<dtCore::RefPtr<SensorBase> > allSensors = actor.getSensors();
			if (!allSensors.empty()) {
				for (unsigned int j=0; j<allSensors.size(); j++)
					allSensors[j]->setup();
			}
		}
	}
}

/**
 * Destroys the robot that is currently selected.
 */
void SimulationArenaComponent::destroyRobot()
{
	if(robotsInScene.size()>0)
	{
		std::vector< dtGame::GameActorProxy* > allGameActors;
		GetGameManager()->GetAllGameActors(allGameActors);

		if (robotFocusIndex<(int)robotsInScene.size()) {
			dtCore::RefPtr<RobotActorBase> selectedActor = robotsInScene[robotFocusIndex];

			this->robotsInScene.erase(robotsInScene.begin() + robotFocusIndex);
			if (robotsInScene.size() > 0)
				robotFocusIndex = robotFocusIndex % robotsInScene.size();
			else robotFocusIndex = 0;

			//@remark Doesn't remove controllers / components
			numRobots--;
			freedInd.push_back(selectedActor->getID());
			GetGameManager()->DeleteActor((selectedActor->GetGameActorProxy()));
		}
	}
}

/**
 * Destroys the given robot
 */
void SimulationArenaComponent::destroyRobot(RobotActorBase &robotActor) {

	std::vector< dtGame::GMComponent* > allComponents;
	GetGameManager()->GetAllComponents(allComponents);

	//@remark Does not destroy controllers anymore

	freedInd.push_back(robotActor.getID());
	GetGameManager()->DeleteActor(robotActor.GetGameActorProxy());

}


/**
 * Calls destroyRobot() for all robots in the scene.
 */
void SimulationArenaComponent::clear()
{
	while (!robotsInScene.empty()) destroyRobot();
	LOG_INFO("Scene has been cleared.");
}

//needed to set the robots a little bit over the position of the surface.
float SimulationArenaComponent::getZPos(float x, float y, dtCore::Scene &scene)
{
	osg::Vec3 intersection;
	osg::Vec3 p;
	dtCore::RefPtr<dtCore::Isector> query = new dtCore::Isector((&scene));
	query->Reset();
	query->SetStartPosition(osg::Vec3(x, y, -10000));
	query->SetEndPosition(osg::Vec3(x, y, 10000));
	query->SetDirection(osg::Vec3(0,0,1));

	query->Update();
//	query->GetHitPoint(p);


	if (query.get()->Update())
	{
		osgUtil::IntersectVisitor &iv = query.get()->GetIntersectVisitor();
		p = iv.getHitList(query.get()->GetLineSegment())[0].getWorldIntersectPoint();
		// make it hover
		p.z() = p.z() + 1.0f;
	}

	return p.z();
}

dtCore::RefPtr<RobotActorBase> SimulationArenaComponent::getSelectedRobot() {

	if (robotFocusIndex < (int)robotsInScene.size())
		return robotsInScene[robotFocusIndex];
	else return NULL;

}

std::vector<dtCore::RefPtr<RobotActorBase> > SimulationArenaComponent::getRobotsInScene(){
	return robotsInScene;
}

void SimulationArenaComponent::selectNextRobot() {
	if (robotsInScene.size()>0) {
		if (robotFocusIndex < (int)robotsInScene.size()) {
			dtCore::RefPtr<RobotActorBase> selectedActor = robotsInScene[robotFocusIndex];
			selectedActor->setSelectionState(false);
			selectedActor->renderCollisionGeometry(false);
		}
		robotFocusIndex = (robotFocusIndex + 1) % robotsInScene.size();

		dtCore::RefPtr<RobotActorBase> selectedActor = robotsInScene[robotFocusIndex];
		selectedActor->setSelectionState(true);
		selectedActor->renderCollisionGeometry(true);

		sendMessageToAll(*mFocusChanged);
	}
}

void SimulationArenaComponent::selectRobot(dtCore::RefPtr<srCore::RobotActorBase> robotToSelect) {

	//deselect previous robot
	if ((robotFocusIndex < (int)robotsInScene.size()) && (robotFocusIndex >= 0)) {
		dtCore::RefPtr<RobotActorBase> selectedActor = robotsInScene[robotFocusIndex];
		selectedActor->setSelectionState(false);
		selectedActor->renderCollisionGeometry(false);
	}

	robotFocusIndex = -1;

	if (robotToSelect) {
		for (unsigned int i=0; i<robotsInScene.size(); i++)
			if (robotsInScene[i] == robotToSelect) {

				robotFocusIndex = i;
				dtCore::RefPtr<RobotActorBase> selectedActor = robotsInScene[robotFocusIndex];
				robotFocus = selectedActor->getID();
				selectedActor->setSelectionState(true);
				selectedActor->renderCollisionGeometry(true);
			}
	}

	sendMessageToAll(*mFocusChanged);
}

}
