//! General body class, which all other body classes have to inherit.

/*!
 *
 * Describes general properties of a body in the simulation environment.
 *
 * @author Lutz Winkler
 *
 */

//TODO: make the robot (and therefore the bodies) scalable in size


#include <srCore/body/bodyBase.h>

#include <osg/Material>
#include <osg/StateSet>
#include <osg/PositionAttitudeTransform>
#include <osg/Group>
#include <osg/Node>

#include <srCore/robot/robotActorBase.h>
#include <srCore/actuator/hingeActuator.h>

#include <dtUtil/log.h>

#include <map>



namespace srCore {

std::map< dBodyID, BodyBase*> BodyBase::odeBodyMap;

osg::Vec3 scale(const float a, const osg::Vec3& vec) {
	return osg::Vec3(a*vec[0], a*vec[1], a*vec[2]);
}

osgBodyNodeVisitor::osgBodyNodeVisitor (BodyBase* root, TraversalMode tm) :
			osg::NodeVisitor(tm),
			parent (root)
{
	setNodeMaskOverride(0xffffffff);
	setTraversalMode(osg::NodeVisitor::TRAVERSE_ALL_CHILDREN);
}

osgBodyNodeVisitor::osgBodyNodeVisitor (TraversalMode tm) :
			osg::NodeVisitor(tm) {}

osgBodyNodeVisitor::osgBodyNodeVisitor (VisitorType type, TraversalMode tm) :
			osg::NodeVisitor(type, tm) {}

void osgBodyNodeVisitor::apply(osgBodyTransform &node) {

	addChildToBodyRoot(node.getMyBody());
	traverse(node);

}

//! Add children's joints to the root body
void  osgBodyNodeVisitor::addChildToBodyRoot (BodyBase* child) {

	if (parent) {

		dtCore::RefPtr<BodyBase> oldParentBody = child;
		while (dynamic_cast<BodyBase*>(oldParentBody->GetParent())) {
			oldParentBody = dynamic_cast<BodyBase*>(oldParentBody->GetParent());
		}

		dBodyID oldParentBodyID = oldParentBody->GetBodyID();

		//child->Emancipate();

		dBodyID parentBodyID = parent->GetBodyID();
		//		dBodyID childBodyID = child->GetBodyID();

		std::vector<dtCore::RefPtr<ActuatorBase> > actuators = child->getRobot()->getActuators();

		for (unsigned int i=0; i < actuators.size(); i++) {

			dJointID joint = actuators[i]->getJoint();
			float angle = dJointGetHingeAngle (joint);

			dBodyID b0,b1;

			b0 = dJointGetBody(joint,0);
			b1 = dJointGetBody(joint,1);

			dVector3 axis, anchor;
			dJointGetHingeAxis(joint, axis);
			dJointGetHingeAnchor(joint, anchor);

			dReal* pos0;
			dReal* pos1;

			dJointGetHingeAxis(joint, axis);
			dJointGetHingeAnchor(joint, anchor);

			dVector3 parentVel;
			dQuaternion parentQuat;
			for (unsigned int j=0; j<4; j++) parentQuat[j] = dBodyGetQuaternion (parentBodyID)[j];
			dBodyGetPointVel (parentBodyID, 0, 0, 0, parentVel);

			if (b0 == oldParentBodyID) {

				pos1 = const_cast<dReal*>(dBodyGetPosition   (b1));
				pos0 = const_cast<dReal*>(dBodyGetPosition   (parentBodyID));

				//				dBodySetAngularVel (b1, 0,0,0);
				dBodySetLinearVel (b1, parentVel[0], parentVel[1], parentVel[2]);
				dJointAttach(joint, parentBodyID, b1);

			}
			else if (b1 == oldParentBodyID) {

				pos0 = const_cast<dReal*>(dBodyGetPosition   (b0));
				pos1= const_cast<dReal*>(dBodyGetPosition   (parentBodyID));

				//				dBodySetAngularVel (b0, 0,0,0);
				dBodySetLinearVel (b0, parentVel[0], parentVel[1], parentVel[2]);
				dJointAttach(joint, b0, parentBodyID);

			}

			else {
				std::cout << "WARNING: could not reattach every joint. b0 = " << b0 << ", b1 = " << b1 <<  "parentBodyID = " <<  parentBodyID << ", oldParentBodyID = " <<  oldParentBodyID << std::endl;
				return;
			}


			dJointSetHingeAnchor(joint, anchor[0],  anchor[1],  anchor[2]);//pos1[0] - pos0[0], pos1[1] - pos0[1], pos1[2] - pos0[2]);
			dJointSetHingeAxis(joint, axis[0], axis[1], axis[2]);
			dynamic_cast<HingeActuator*>(actuators[i].get())->setZeroAngle(angle);

		}

	}
}

//! Returns the root body
BodyBase*  osgBodyNodeVisitor::returnBodyRoot () {
	return parent;
}

osgBodyNodeVisitor::~osgBodyNodeVisitor () {

}


osgBodyTransform::osgBodyTransform(dtCore::RefPtr<BodyBase> srBody) :
			myBody (srBody)
{

}

dtCore::RefPtr<BodyBase> osgBodyTransform::getMyBody() {
	return myBody;
}

void osgBodyTransform::accept(osgBodyNodeVisitor& bnv) {

	bnv.addChildToBodyRoot(myBody);
	traverse(bnv);

}

void osgBodyTransform::traverse (osgBodyNodeVisitor& bnv) {

	for(osg::NodeList::iterator itr=_children.begin();
			itr!=_children.end();
			++itr)
	{
		if (dynamic_cast<osgBodyTransform*>((*itr).get()) )
			dynamic_cast<osgBodyTransform*>((*itr).get())->accept(bnv);
	}

}


osgBodyTransform::~osgBodyTransform() {

}


class RobotActorBase ;

/************************************************************************************************
 *                      BodyBase implementation
 ***********************************************************************************************/

/**
 * The body is constructed from a Delta3D object. It is decorated with a collision model and an
 * OpenSceneGraph node. Entities that derive from BodyBase are for example Wheel through Cylinder,
 * or KITScrew through Cuboid.
 */
BodyBase::BodyBase() : dtCore::Object()
{
	renderGeom = true;
	GetOSGNode()->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());

	//initialise collision model
	collisionModel = new osgBodyTransform(this);
	collisionModel->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());
	collisionModel->setPosition(osg::Vec3(0,0,0));

	osg::Node* osgNode = new osg::Node();
	osgNode->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());

	collisionModel->addChild(osgNode);

	registerBody();
}

/**
 * Remove the body from the scene.
 */
BodyBase::~BodyBase()
{
	while(this->GetNumChildren()>0)
		this->RemoveChild(this->GetChild(0));

	this->ClearCollisionGeometry();

	myRobot = NULL;
}

//! Register Body to the simulation, physically
void BodyBase::registerBody() {
	LOG_DEBUG("Register body to the physical engine");
	BodyBase::odeBodyMap.insert(std::pair<dBodyID, BodyBase*>(this->GetBodyID(), this));

};

/**
 * Sets the collision category and collide bits. For example, a cloud does not collide with another
 * cloud. A robot does collide with another robot, or with the floor. This is why collision
 * categories have been introduced.
 */
void BodyBase::setCollisionBits(unsigned long int bit) {
	SetCollisionCollideBits(~bit & ~UNSIGNED_BIT(2));
	SetCollisionCategoryBits(bit);
	collCategoryBit = bit;
	std::ostringstream msg; msg.clear();
	msg << "The collision category of object " << GetName() << " in bits is " << (long int)GetCollisionCategoryBits();
	msg << " and its collide bits are " << (long int)GetCollisionCollideBits();
	LOG_DEBUG(msg.str());
}

//! Moves the model translation by pos
void BodyBase::setModelTranslation(osg::Vec3 pos) {

	dtCore::Transform myTransform;
	GetTransform(myTransform);
	osg::Matrix myMatrix;
	osg::Matrix myInverse;
	myTransform.GetRotation(myMatrix);
	myInverse.invert(myMatrix);

	osg::Vec3 absObjectPos = myInverse * GetModelTranslation();// + objectTransform.GetTranslation();
	osg::Vec3 absObjectTranslation = absObjectPos - scale(50, pos);

	SetModelTranslation(myMatrix * absObjectTranslation);

	for (unsigned int i=0; i<GetNumChildren(); i++) {

		dtCore::Object* child = dynamic_cast<dtCore::Object*>(GetChild(i));

		if ((dynamic_cast<ConnectorBase*>(child)) || (dynamic_cast<SensorBase*>(child)))
		{
			dtCore::Transform childTransform;
			child->GetTransform(childTransform);
			childTransform.SetTranslation(childTransform.GetTranslation() + pos);
			child->SetTransform(childTransform, dtCore::Transformable::ABS_CS);
		}

	}

}

//! moves all model translations of parent and its children by pos, stops when the object named 'theOtherObjectName'
void BodyBase::setAllModelTranslations(dtCore::Object &object, std::string theOtherObjectName, osg::Vec3 pos) {
	//	dtCore::Object* parent = &object;
	osg::Vec3 trans = object.GetModelTranslation();

	dtCore::Transform objectTransform;
	object.GetTransform(objectTransform, dtCore::Transformable::ABS_CS);
	osg::Matrix objectMatrix;
	objectTransform.GetRotation(objectMatrix);

	osg::Matrix inverseObjectMatrix;
	inverseObjectMatrix.invert(objectMatrix);

	osg::Vec3 absObjectPos = inverseObjectMatrix * object.GetModelTranslation() + objectTransform.GetTranslation();
	osg::Vec3 absObjectTranslation = absObjectPos - pos;

	if (dynamic_cast<BodyBase*>(&object)) {
		dtCore::Transform xform;
		object.GetTransform(xform);
		object.SetModelTranslation((scale(50,object.GetModelTranslation()) + objectMatrix *scale(50,pos)));
		object.RenderProxyNode(false);
		object.RenderCollisionGeometry(false);

		dynamic_cast<BodyBase*>(&object)->colorCollisionShape(osg::Vec4(0,0,0,0));
	}
	else if ((dynamic_cast<ConnectorBase*>(&object)) || (dynamic_cast<SensorBase*>(&object)))
	{
		objectTransform.SetTranslation(objectTransform.GetTranslation() + pos);
		object.SetTransform(objectTransform, dtCore::Transformable::ABS_CS);
		//object.RenderProxyNode(true);
	}

	for (unsigned int i=0; i<object.GetNumChildren(); i++) {
		dtCore::Object* childObject = dynamic_cast<dtCore::Object*>(object.GetChild(i));
		if (childObject) {

			dtCore::Transform childTransform;
			osg::Matrix childRotation;
			osg::Matrix childRotInv;

			childRotInv.invert(childRotation);
			childObject->GetTransform(childTransform);
			childTransform.GetRotation(childRotation);

			if (childObject->GetName() != theOtherObjectName)
				setAllModelTranslations(*childObject, theOtherObjectName,  pos);

		}

	}

}


//merges the collision shape and the masses of two bodies together. theBody will become child of this body and the physics will be disabled of theBody
void BodyBase::mergeBody(BodyBase* theBody) {


	// get parent of my robot body
	dtCore::RefPtr<BodyBase> myParentBody = this;
	while (dynamic_cast<BodyBase*>(myParentBody->GetParent())) {
		myParentBody = dynamic_cast<BodyBase*>(myParentBody->GetParent());
	}

	//---- actuator stuff ---------------------------------------------------

	//get all joints attached to that body
	dBodyID theBodyID = theBody->GetBodyID();

	if (theBody->GetBodyID() == GetBodyID()) {
		LOG_WARNING("BodyBase: Merging with myself not possible");
		return;
	}

	while (dBodyGetNumJoints (theBodyID) > 0) {

		dJointID theJoint = dBodyGetJoint (theBodyID, 0);
		float angle = dJointGetHingeAngle (theJoint);

		dBodyID b0,b1;
		b0 = dJointGetBody(theJoint,0);
		b1 = dJointGetBody(theJoint,1);

		dVector3 axis;
		dVector3 pos;
		dJointGetHingeAxis(theJoint, axis);
		dJointGetHingeAnchor(theJoint, pos);

		dReal* pos0;
		dReal* pos1;

		dVector3 parentVel;
		dQuaternion parentQuat;
		for (int i=0; i<4; i++) parentQuat[i] = dBodyGetQuaternion (myParentBody->GetBodyID())[i];
		dBodyGetPointVel (myParentBody->GetBodyID(), 0, 0, 0, parentVel);


		if (b0 == theBodyID) {

			pos0 = const_cast<dReal*>(dBodyGetPosition   (b1));
			pos1 = const_cast<dReal*>(dBodyGetPosition   (myParentBody->GetBodyID()));

			//				dBodySetAngularVel (b1, 0,0,0);
			dBodySetLinearVel (b1, parentVel[0], parentVel[1], parentVel[2]);

			if (myParentBody->GetBodyID() != b1)
				dJointAttach(theJoint, myParentBody->GetBodyID(), b1);
			else dJointAttach(theJoint, myParentBody->GetBodyID(), 0);

		}
		else if (b1 == theBodyID) {

			pos0 = const_cast<dReal*>(dBodyGetPosition   (b0));
			pos1= const_cast<dReal*>(dBodyGetPosition   (myParentBody->GetBodyID()));

			//				dBodySetAngularVel (b0, 0,0,0);
			dBodySetLinearVel (b0,  parentVel[0], parentVel[1], parentVel[2]);

			if (myParentBody->GetBodyID() != b0)
				dJointAttach(theJoint, b0, myParentBody->GetBodyID());
			else dJointAttach(theJoint, 0, myParentBody->GetBodyID());

		}

		dJointSetHingeAnchor(theJoint, pos[0], pos[1], pos[2]);
		dJointSetHingeAxis(theJoint, axis[0], axis[1], axis[2]);

		std::map< dJointID, ActuatorBase*>::iterator iter = ActuatorBase::jointActuatorMap.begin();

		iter = ActuatorBase::jointActuatorMap.find(theJoint);

		if (iter != ActuatorBase::jointActuatorMap.end()) {
			HingeActuator* theHinge = dynamic_cast<HingeActuator*>(iter->second);
			if (theHinge) theHinge->setZeroAngle(angle);
		}
	}

	//---- body stuff --------------------------------------------------------
	//	RobotActorBase* theRobot = theBody->getRobot();

	// get robot properties:

	// collision shape:
	osgBodyTransform* myParentCollisionNode = myParentBody->getCollisionModel();
	osgBodyTransform* myCollisionNode = this->getCollisionModel();

	dMass mass;
	theBody->GetMass(&mass);

	theBody->EnableDynamics(false);
	theBody->SetCollisionDetection(false);

	//calculate relative position
	osg::Vec3 pos = theBody->getTranslation();// - myParentBody->getTranslation();
	osg::Vec3 myTranslation = myParentBody->getTranslation();

	osg::Matrix myParentMatrix;
	dtCore::Transform xform, myTransform;
	theBody->GetTransform(xform);
	myParentBody->GetTransform(myTransform);

	osg::Matrix theMatrix;

	myTransform.GetRotation(myParentMatrix);
	xform.GetRotation(theMatrix);

	osg::Matrix inverse;
	inverse.invert(theMatrix);

	//calculate relative transform matrix
	theMatrix = inverse * myParentMatrix;

	osg::Vec3 hpr; osg::Quat quat;
	xform.GetRotation(hpr);

	xform.Set(theMatrix);
	xform.GetRotation(quat);

	dtCore::Transform myParentTransform;
	dtCore::RefPtr<BodyBase> theOtherBody = theBody;
	myParentBody->GetTransform(myParentTransform);
	theOtherBody->GetTransform(xform);
	this->GetTransform(myTransform);
	osg::Vec3 theOtherTranslation = theOtherBody->getTranslation();
	myTranslation = this->getTranslation();

	myParentCollisionNode = myParentBody->getCollisionModel();
	myCollisionNode = this->getCollisionModel();
	osgBodyTransform* theCollisionNode = theBody->getCollisionModel();

	if (theCollisionNode->getNumParents() > 0)
		theCollisionNode->getParent(0)->removeChild(theCollisionNode);
	myCollisionNode->addChild(theCollisionNode);

	osg::Matrix myMatrix;

	myParentTransform.GetRotation(myParentMatrix);
	myTransform.GetRotation(myMatrix);
	xform.GetRotation(theMatrix);

	theBody->GetTransform(xform, ABS_CS);

	//add the other parent as child to my parent and set transform
	if (theBody->GetParent())
		theBody->GetParent()->RemoveChild(theBody);


	osg::Matrix theInverse;
	theInverse.invert(theMatrix);


	this->AddChild(theBody);
	theBody->SetTransform(xform, ABS_CS);

	inverse.invert(myMatrix);

	//calculate relative transform matrix
	osg::Matrix theMatrixI = theMatrix * inverse;


	quat = theMatrixI.getRotate();

	//merge collision models and set transform
	theCollisionNode->setPosition( myMatrix * ((theOtherTranslation + theInverse * theOtherBody->GetModelTranslation()) - (myTranslation + inverse * this->GetModelTranslation())));

	theCollisionNode->setAttitude(quat);

	osg::Quat parentQuat = myParentCollisionNode->getAttitude();
	osg::Matrix parentMatrix;
	parentQuat.get(parentMatrix);

}

/**
 * Initializes the body. Needs to be called before actually using the body.
 */
void BodyBase::init(	std::string name, RobotActorBase* robot,
		float x, float y, float z,
		float heading, float roll, float pitch)
{
	std::ostringstream msg;	msg.clear(); msg.str("");
	if (robot->GetName().empty()) {
		msg << "Body " << name << " will be attached to a nameless robot";
		LOG_WARNING(msg.str());
	}

	myRobot = robot;
	//	myRobot->AddChild(this);
	char idString[20];
	renderGeom = true;

	long int id = myRobot->getID();
	sprintf(idString, "%ld", (1+id));
	SetName(idString + name);

	//	if (SimulationUtils::getConnectorsAsJoints())
	this->myID = 1 + id;							//need to be not zero, as it is used for the CollideBits as well
	//	else
	//		this->myID = 1;

	this->xform.Set(x, y, z, heading, roll, pitch);

	msg.clear(); msg.str("");
	msg << "Initialised BodyBase: " << myID << " which is a " << name;
	LOG_DEBUG(msg.str());

}

//! Assigns an OSG collision geometry to our body
void BodyBase::setCollisionModel(dtCore::RefPtr<osg::Node> collModel) {

	collModel->setName(GetName());
	collModel->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());
	osg::Group* osgNode = collisionModel->asGroup();
	if (osgNode->getChild(0)) osgNode->removeChild(osgNode->getChild(0));
	osgNode->addChild(collModel);

}

//! Assigns an OSG collision geometry to our body
void BodyBase::setCollisionModel(osgBodyTransform* collModel) {

	collModel->setName(GetName());
	collModel->setNodeMask(SimulationUtils::Instance()->getCollisionNodeMask());
	collisionModel = collModel;

}

//! Returns the current collision model
osgBodyTransform* BodyBase::getCollisionModel() {

	return collisionModel.get();

}

//! shows the collision geometry of the object
void BodyBase::renderCollisionGeometry(bool renderIt)
{
	renderGeom = renderIt;
	//	RenderCollisionGeometry(renderIt);
	float alpha = 0.0;
	if (renderIt) alpha = 0.8;

	osg::StateSet* stateset;
	if (collisionModel->getChild(0)) {
		stateset = collisionModel->getChild(0)->getOrCreateStateSet();
		osg::Material * mm = dynamic_cast<osg::Material*>(stateset->getAttribute(osg::StateAttribute::MATERIAL));

		if (!mm)
		{
			mm = new osg::Material;
		}
		mm->setAlpha(osg::Material::FRONT, alpha);
		mm->setAmbient(osg::Material::FRONT, osg::Vec4(body_color.r(), body_color.g(), body_color.b(), 0));
//		mm->setAmbient(osg::Material::FRONT, osg::Vec4(1,0,0,0));
		stateset->setAttributeAndModes( mm, osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON);
		stateset->setMode(GL_BLEND,   osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	}
}

//! Set the color and alpha value to the collision shape
void BodyBase::colorCollisionShape(osg::Vec4 color) {

	osg::StateSet* stateset;
	if (collisionModel->asGroup()->getChild(0)) {
		stateset = collisionModel->asGroup()->getChild(0)->getOrCreateStateSet();
		osg::Material * mm = dynamic_cast<osg::Material*>(stateset->getAttribute(osg::StateAttribute::MATERIAL));

		if (!mm)
		{
			mm = new osg::Material;
		}
		mm->setAlpha(osg::Material::FRONT, color.a());
		mm->setAmbient(osg::Material::FRONT, osg::Vec4(color.r(), color.g(), color.b(),0));
		stateset->setAttributeAndModes( mm, osg::StateAttribute::PROTECTED | osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON);
		stateset->setMode(GL_BLEND,   osg::StateAttribute::PROTECTED | osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setMode(GL_LIGHTING,osg::StateAttribute::PROTECTED | osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	}

}

//! colors the collision mesh of the body. parameter is in rgba.
void BodyBase::colorBody(osg::Vec4 color) {
	body_color = color;
	osg::StateSet* stateset;
	if (collisionModel->asGroup()->getChild(0)) {
		//stateset = collisionModel->asGroup()->getChild(0)->getOrCreateStateSet();
		stateset = this->GetOSGNode()->getOrCreateStateSet();
		osg::Material * mm = dynamic_cast<osg::Material*>(stateset->getAttribute(osg::StateAttribute::MATERIAL));

		if (!mm)
		{
			mm = new osg::Material;
		}
		mm->setAlpha(osg::Material::FRONT, color.a());
		mm->setAmbient(osg::Material::FRONT, osg::Vec4(color.r(), color.g(), color.b(),0));
		stateset->setAttributeAndModes( mm, osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON);
		stateset->setMode(GL_BLEND,   osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setRenderingHint(osg::StateSet::DEFAULT_BIN);//:TRANSPARENT_BIN);
	}

}

//! shows the collision geometry of the object
void BodyBase::renderCollisionGeometry2(bool renderIt)
{

	renderGeom = renderIt;
	//	RenderCollisionGeometry(renderIt);
	float alpha;
	if (renderIt) alpha = 0.8;
	else alpha = 0.2;
	//alpha = 0;

	for (int i=0; i< (int) collisionModel->asGroup()->getNumChildren(); i++) {

		osg::StateSet* stateset;
		if (dynamic_cast<osg::PositionAttitudeTransform*>(collisionModel->asGroup()->getChild(i))) {

			for (int j=0; j< (int) collisionModel->asGroup()->getChild(i)->asGroup()->getNumChildren(); j++)
				stateset = collisionModel->asGroup()->getChild(i)->asGroup()->getChild(j)->getOrCreateStateSet();
		}

		else
			stateset = collisionModel->asGroup()->getChild(i)->getOrCreateStateSet();

		osg::Material * mm = dynamic_cast<osg::Material*>(stateset->getAttribute(osg::StateAttribute::MATERIAL));

		if (!mm)
		{
			mm = new osg::Material;
		}
		mm->setAlpha(osg::Material::FRONT, alpha);
		mm->setAmbient(osg::Material::FRONT, osg::Vec4(1,0,0,0));
		stateset->setAttributeAndModes( mm, osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON);
		stateset->setMode(GL_BLEND,   osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setMode(GL_LIGHTING,osg::StateAttribute::OVERRIDE | osg::StateAttribute::ON );
		stateset->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);
	}


}

/**
 * Adds the body to the scene, makes it visible and activates the physical behaviour.
 */
void BodyBase::addToScene()
{
	dtCore::Scene &scene  = myRobot->GetGameActorProxy().GetGameManager()->GetScene();

	//add object to the scene
	scene.AddDrawable(this);

	//render collision geometry if desired
	if (renderGeom) RenderCollisionGeometry();

	std::ostringstream msg;	msg.clear(); msg.str("");
	msg << "Add body part of robot \"" << myRobot->GetName() << "\" to the scene";
	LOG_DEBUG(msg.str());

#ifdef NO_PHYSICS
	//sets category and collide bits for the collision detection
	dGeomSetCategoryBits(GetGeomID(), 0);
	dGeomSetCollideBits(GetGeomID(), 0);


	//transform the object to position/orientation/scaling
	SetTransform(xform);

	//activate physics
	EnableDynamics(false);

#else

	//transform the object to position/orientation/scaling
	SetTransform(xform);

	//activate physics
	EnableDynamics();
	SetCollisionDetection(true);
	PrePhysicsStepUpdate(); //must do *before* attaching joint
#endif


}

void BodyBase::setScale(float scaleX, float scaleY, float scaleZ)
{

	this->scaleX = scaleX;
	this->scaleY = scaleY;
	this->scaleZ = scaleZ;
	SetScale(osg::Vec3(scaleX, scaleY, scaleZ));
}



long BodyBase::getID()
{
	return this->myID;
}

//! Returns the translation of the Delta3D body. Its true translation is getTranslation() + Rotation^(-1) * getModelTranslation(), whereas Roation is the Rotation of the Body
osg::Vec3 BodyBase::getTranslation() const
{

	dtCore::Transform xform;
	GetTransform(xform,dtCore::Transformable::ABS_CS);
	return xform.GetTranslation();

}

osg::Vec3 BodyBase::getRotation() const
{
	dtCore::Transform xform;
	osg::Vec3 hpr;

	GetTransform(xform,dtCore::Transformable::ABS_CS);

	//return osg::Vec3(1,0,0) * xform.GetRotation();
	xform.GetRotation(hpr);

	return hpr;
}

osg::Vec3 BodyBase::getScale() const
{
	return osg::Vec3(this->scaleX, this->scaleY, this->scaleZ);
}


void BodyBase::setTransform(dtCore::Transform &transform)
{
	xform = transform;
	float x,y,z,h,r,p;
	xform.Get(x,y,z,h,r,p);

	SetTransform(xform);
}

/**
 * Sets the translation of the Delta3D body. Its true translation is:
 *    getTranslation() + Rotation^(-1) * getModelTranslation(),
 * whereas Rotation is the rotation of the body.
 */
void BodyBase::setTranslation(osg::Vec3f pos)
{
	this->GetTransform(xform);
	this->xform.SetTranslation(pos);

	//transform the object to position/orientation/scaling
	SetTransform(xform, dtCore::Transformable::ABS_CS);
}

void BodyBase::setRotation(osg::Vec3f hrp)
{
	this->xform.SetRotation(hrp);

	//transform the object to position/orientation/scaling
	SetTransform(this->xform, dtCore::Transformable::ABS_CS);
}


}
