/*
 * Robot3D Physical realistic 3D simulator for robots
 *
 * @license GNU Lesser General Public License
 *
 * @author Adrian Friebel
 * @author Anne van Rossum
 */

#include <iomanip>

// Robot3D files
#include <srCore/cameraComponent.h>
#include <srCore/simulationEvents.h>

// Delta3D files
#include <dtUtil/log.h>

// Boost
#include <boost/filesystem.hpp>

#define CAM_SCALE								0.6			//1=direct following, 0=no following

using namespace srCore;

/**
 * The camera controller is a Delta3D Game Component. It concerns the "user" camera, through which the
 * scene is seen. It is not the camera on the robot.
 */
SimulationCameraController::SimulationCameraController(const std::string &name, osg::Vec3f camPos,
		bool camState): dtGame::GMComponent(name),
		camPos(camPos),
		freeCamState(camState),
		counter(0),
		mScreenShotTaker(),
		screenShotToggle(false),
		mFMM(NULL),
		mTrp(NULL),
		screenShotPath() {
	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
	mScreenShotTaker = new ScreenShotCallback(RGB_CAMERA, GA->GetCamera()->GetOSGCamera()->getViewport());
}

/**
 * Destroys the camera.
 */
SimulationCameraController::~SimulationCameraController() {
	mTrp->SetAttachToTransformable(NULL);
	delete mScreenShotTaker;
}

/**
 * Initialises the camera and where it looks at. If there are no actors at all, it actually
 * should look at the position indicated in the XML configuration file. It is possible to
 * set the viewport on the generic camera, like this:
 *   camera->GetOSGCamera()->setViewport(new osg::Viewport(0.0, 10.0, 480, 480));
 * This has the effect that the entire thing plus the GUI elements becomes limited to this
 * part of the screen.
 * The function:
 *   GetWindow()->GetOsgViewerGraphicsWindow()->getWindowRectangle(x, y, width, height);
 * does not return the proper x and y, it starts at x=35 and y=50 instead of 0,0.
 */
void SimulationCameraController::init(dtCore::Mouse *mouse, dtCore::Keyboard *keyboard)
{
	bool noRobots = false;
	SimulationArenaComponent *comp = static_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	mFMM = new dtCore::FlyMotionModel(keyboard, mouse);
	/*
	mFMM->SetMaximumFlySpeed(1000 / GetGameManager()->GetTimeScale());
	mFMM->SetMaximumTurnSpeed(10000 / GetGameManager()->GetTimeScale());
	 */
	mFMM->SetMaximumFlySpeed(100);
	mFMM->SetMaximumTurnSpeed(1000);

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);

	if (index < 0) noRobots = true;
	if (index >= (int)allGameActors.size()) noRobots = true;

	//Set Camera position and drop out of here if there are no robots
	if (noRobots) {
		dtCore::Transform xform;
		osg::Vec3 origin = osg::Vec3(0,0,0);
		osg::Vec3 up = osg::Vec3(0,0,1);
		xform.Set(camPos, origin, up);
		GetGameManager()->GetApplication().GetCamera()->SetTransform(xform);
		mFMM->SetTarget(GetGameManager()->GetApplication().GetCamera());
		mTrp = new dtCore::Tripod(GetGameManager()->GetApplication().GetCamera());
		mTrp->SetTetherMode(dtCore::Tripod::TETHER_WORLD_REL);
		mTrp->SetScale(CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE);
		mTrp->SetCamera(NULL);
		return;
	}

	//Create a Tripod and attach it to the robot which is specified
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());
	mTrp = new dtCore::Tripod(GetGameManager()->GetApplication().GetCamera(), actor.getBodies().at(0).get());
	mTrp->SetTetherMode(dtCore::Tripod::TETHER_WORLD_REL);
	mTrp->SetScale(CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE,CAM_SCALE);

	//	setCamera();
	//}

	//void SimulationCameraController::setCamera() { };
	if(freeCamState)
	{
		mTrp->SetCamera(NULL);
		mFMM->SetTarget(GetGameManager()->GetApplication().GetCamera());
	}
	else
	{
		//Set initial Tripod Offset
		dtCore::Transform tx(-40.0f, 0.0f, 40.0f, 0.0f, 0.0f, 0.0f);
		osg::Vec3 hpr;
		tx.GetRotation(hpr);
		mTrp->SetOffset(camPos, hpr);
		mTrp->SetLookAtTarget(actor.getBodies().at(0).get());
		mTrp->SetCamera(GetGameManager()->GetApplication().GetCamera());
		mFMM->SetTarget(NULL);
	}
}

/**
 * The size of the viewport depends on the existence of the GUI. If there is a GUI then
 * the viewport becomes smaller. It starts then namely at (0,0) and it ends at 3/4th of
 * the width, and 95% of the height.
 *
 * @param fullwindow true indicates that there is no GUI and the viewport fill the
 * window, false means that it will be limited to 3/4 of the width and 95% of the height
 */
void SimulationCameraController::setViewport(bool fullwindow=true) {
	dtCore::Transform xform;
	osg::Vec3 robotPos = osg::Vec3(0,0,0); //actor.getTranslation();
	osg::Vec3 up  		= osg::Vec3(0,0,1);
	xform.Set(camPos, robotPos, up);
	GetGameManager()->GetApplication().GetCamera()->SetTransform(xform);
	if (!fullwindow) {
		dtCore::Camera *camera = GetGameManager()->GetApplication().GetCamera();
		dtCore::RefPtr<osg::Viewport> viewport = camera->GetOSGCamera()->getViewport();

		float x = viewport->x();
		float y = viewport->y();
		float width = viewport->width() * 0.75;
		float height = viewport->height() * 0.95;

		std::ostringstream msg;	msg.clear(); msg.str("");
		msg.clear(); msg.str("");
		msg << "Window [x,y,width,height] = [" << x << "," << y << "," << width << "," << height << "]";
		LOG_INFO(msg.str());
		camera->GetOSGCamera()->setViewport(new osg::Viewport(x, y, width, height));
	}
}

/**
 * Set path for screenshots. If it does not exist it will be created, and an informational
 * message will be sent to the user.
 */
void SimulationCameraController::setScreenshotPath(std::string &path) {
	if (!boost::filesystem::exists(path)) {
		dtUtil::Log::GetInstance().LogMessage(dtUtil::Log::LOG_INFO, __func__, __LINE__,
				"Path doesn't exist \"%s\", it will be created.", path.c_str());
		boost::filesystem::create_directories(path);
	}
	screenShotPath = path;
}

/**
 * Retrieve path for screenshots
 */
std::string SimulationCameraController::getScreenshotPath() {
	return screenShotPath;
}

//////////////////////////////////////////////////////////////////////////
void SimulationCameraController::sendMessageToAll(dtDAL::GameEvent &event)
{
	dtCore::RefPtr<dtGame::GameEventMessage> eventMsg;
	GetGameManager()->GetMessageFactory().CreateMessage(dtGame::MessageType::INFO_GAME_EVENT, eventMsg);

	eventMsg->SetGameEvent(event);
	GetGameManager()->SendMessage(*eventMsg);
}

/**
 * Turn on screenshot taking.
 */
void SimulationCameraController::toggleScreenShots(bool toggle) {
	screenShotToggle = toggle;
}

/**
 * Saves a screenshot to file.
 */
void SimulationCameraController::TakeScreenShot() {

	if (!screenShotToggle) return;

	counter++;

	std::ostringstream t2;
	t2 << screenShotPath << "/srCore_";

	if (screenShotPath.empty()) {
		if (counter == 0) {
			LOG_ERROR("Screenshot path is not set");
		}
	} else {
		t2 << std::setfill('0') << std::setw(5) << counter - 1 << ".jpeg";

		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);

		//find the camera which will be drawn the latest
		int topView = 0;
		for (int i=0; i<(int)GA->GetNumberOfViews(); i++) {
			if (GA->GetView(topView) < GA->GetView(i))
				topView = i;
		}

		dtUtil::Log::GetInstance().LogMessage(dtUtil::Log::LOG_DEBUG, __func__, __LINE__,
				"Saving screenshot %s", t2.str().c_str());

		GA->GetView(topView)->GetCamera()->GetOSGCamera()->setFinalDrawCallback(mScreenShotTaker);
		mScreenShotTaker->takeScreenShot(t2.str());
	}
}

/**
 * Controls the "camera" that represents the users. This routine does focus
 * on a robot if it gets a tick from the Delta3D engine. If the boolean
 * screenshotToggle is true, it saves a screenshot in a predefined directory.
 */
void SimulationCameraController::ProcessMessage(const dtGame::Message& message) {

	if (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL) {

		SimulationArenaComponent *comp = static_cast<SimulationArenaComponent*>(
				GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
		int index = comp->getRobFocIndex();
		if(index!=-1) {
			std::vector< dtGame::GameActorProxy* > allGameActors;
			GetGameManager()->GetAllGameActors(allGameActors);
			dtCore::RefPtr<RobotActorBase> selectedActor = comp->getSelectedRobot();
			if (selectedActor)
				if(selectedActor->getBodies().size()>0)
					mTrp->SetLookAtTarget(selectedActor->getBodies().at(0).get());
		}

		TakeScreenShot();
	}

	if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT)
	{
		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);

		if(eventMsg.GetGameEvent() != NULL)
		{

			if(eventMsg.GetGameEvent()->GetName() == TIME_SCALE_CHANGED ) {

				mFMM->SetMaximumFlySpeed(10 / GetGameManager()->GetTimeScale());
				mFMM->SetMaximumTurnSpeed(10 / GetGameManager()->GetTimeScale());

			}
			if(eventMsg.GetGameEvent()->GetName() == MOUSE_IN_GUI )
			{
				mFMM->SetEnabled(false);
			}
			else if(eventMsg.GetGameEvent()->GetName() == MOUSE_IN_SIM )
			{
				mFMM->SetEnabled(true);
			}
			//else if(eventMsg.GetGameEvent()->GetName() == CHANGE_ROBOT_SELECTION )
			else if(eventMsg.GetGameEvent()->GetName() == ROBOT_SELECTION_CHANGED )
			{
				//			   SimulationRobotSelectionComponent *comp = static_cast<SimulationRobotSelectionComponent*>(GetGameManager()->GetComponentByName("SimulationRobotSelectionComponent"));
				//			   int index = comp->getRobFocIndex();
				SimulationArenaComponent *comp = static_cast<SimulationArenaComponent*>(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
				int index = comp->getRobFocIndex();

				if(index!=-1)
				{
					std::vector< dtGame::GameActorProxy* > allGameActors;
					GetGameManager()->GetAllGameActors(allGameActors);
					// RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

					dtCore::RefPtr<RobotActorBase> selectedActor = comp->getSelectedRobot();

					if (selectedActor)
						if((!freeCamState) && (selectedActor->getBodies().size()>0))
						{
							dtCore::Object *body = selectedActor->getBodies().at(0).get();
							mTrp->SetAttachToTransformable(body);						//change the target of the tripod
							mTrp->SetLookAtTarget(body);
						}
				}
				else
				{
					mTrp->SetAttachToTransformable(NULL);
					mTrp->SetLookAtTarget(NULL);
				}
			}
			else if(eventMsg.GetGameEvent()->GetName() == CHANGE_CAMERA_MODE )
			{
				if(!freeCamState)
				{
					LOG_INFO("Switch from Tripod to FreeCam mode");
					mTrp->SetCamera(NULL);
					mFMM->SetTarget(GetGameManager()->GetApplication().GetCamera());
					freeCamState = true;
					//					setCamera();
				}
				else
				{
					LOG_INFO("Switch from FreeCam to Tripod mode");
					dtCore::Transform trfTrp;
					dtCore::Transform trfAct;
					osg::Vec3 xyzTrp;
					osg::Vec3 xyzAct;
					osg::Vec3 hpr;

					mFMM->SetTarget(NULL);

					if (mTrp->GetAttachedTransformable() != NULL) {
						GetGameManager()->GetApplication().GetCamera()->GetTransform(trfTrp, dtCore::Transformable::ABS_CS);
						trfTrp.Get(xyzTrp, hpr);
						mTrp->GetAttachedTransformable()->GetTransform(trfAct, dtCore::Transformable::ABS_CS);
						trfAct.Get(xyzAct, hpr);
						mTrp->SetOffset(xyzTrp.x() - xyzAct.x(), xyzTrp.y() - xyzAct.y(), xyzTrp.z() - xyzAct.z(), 0, 0, 0);
					}
					mTrp->SetCamera(GetGameManager()->GetApplication().GetCamera());

					freeCamState = false;
					//					setCamera();
				}
			}
		}
	}
}

