/* -*-c++-*-
 * Delta3D Open Source Game and Simulation Engine
 * Copyright (C) 2005, BMH Associates, Inc.
 *
 * GUI Component
 *
 * @author Lutz Winkler
 * @author Adrian Friebel
 */

// General files
#include <iostream>
#include <cassert>

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/simulationUtils.h>
#include <srCore/guiComponent.h>
#include <srCore/arenaComponent.h>
#include <CEGUIDefaultResourceProvider.h>

using namespace srCore;

/*
 * Function: SimulationGUIComponent
 *
 * The constructor that wraps around a GUI, either a CEGUI window, or a Qt window. On the moment
 * CEGUI is used.
 *
 * ObjectParameter:
 *   name		- A fixed name for this component
 *   win		- A Delta3D window
 *   keyboard	- A Delta3D keyboard
 *   mouse		- A Delta3D mouse
 *
 * Returns:
 *
 * The SimulationGUIComponent component itself. It is also possible to retrieve it via the Delta3D
 * game manager by searching for its name.
 */
SimulationGUIComponent::SimulationGUIComponent(const std::string &name, dtCore::DeltaWin &win,
		dtCore::Keyboard &keyboard, dtCore::Mouse &mouse)
: dtGame::GMComponent(name),
  freeCamState(true),
  sensorID (0),
  robotID (0),
  mGUI(NULL),
  mWin(&win),
  mSheet(NULL),
  mRPanel(NULL),
  mSensorSelectorIt(NULL),
  mSensorSelectorPM(NULL),
  mActuatorSelectorIt(NULL),
  mActuatorSelectorPM(NULL),
  mBar(NULL),
  mRobotSelectorIt(NULL),
  mRobotSelectorPM(NULL),
  image (new osg::Image),
  hideRenderTexture(false),
  simTime(0),
  saveScreenShot(false),
  selectedMenuItem(NONE)
{
	mSelectRobot = new dtDAL::GameEvent( CHANGE_ROBOT_SELECTION );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mSelectRobot);

	mMouseInGUI = new dtDAL::GameEvent( MOUSE_IN_GUI );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mMouseInGUI);

	mMouseInSim = new dtDAL::GameEvent( MOUSE_IN_SIM );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mMouseInSim);

	mChangeCameraMode = new dtDAL::GameEvent( CHANGE_CAMERA_MODE );
	dtDAL::GameEventManager::GetInstance().AddEvent(*mChangeCameraMode);

	destroying = false;

	SetupGUI(win, keyboard, mouse);

	//	mDeltaWin = win;
	LOG_ALWAYS("Set mouse");
	SetMouse(&mouse);
	SetKeyboard(&keyboard);
	LOG_ALWAYS("Initialized");

}

//////////////////////////////////////////////////////////////////////////
/*
 * Function: ~SimulationGUIComponent
 *
 * Destructor of the SimulationGUIComponent. If there are still game messages going
 * around, this might lead to a segmentation fault on leaving the simulator. The game
 * manager is already gone, so it doesn't help to call pause.
 *
 * Returns:
 * Deallocated object
 */
SimulationGUIComponent::~SimulationGUIComponent()
{
	destroying = true;
	LOG_INFO("Clear events");
	dtDAL::GameEventManager::GetInstance().ClearAllEvents();
	CEGUI::System::getSingleton().removeAllEvents();
	LOG_INFO("Destroy SimulationGUIComponent");
	try {
		wm->destroyAllWindows();
	} catch (CEGUI::Exception exp) {
		std::ostringstream msg;	msg.clear(); msg.str("");
		msg << exp.getMessage();
		LOG_ERROR(msg.str());
	}
	LOG_INFO("Shut down the rest...");
	wm->cleanDeadPool();
	mGUI->ShutdownGUI();
	mGUI.release();
	LOG_INFO("GUI destroyed");
}

/**
 * Every game component process Delta3D messages. This component understands the
 * following messages: TICK_END_OF_FRAME, TICK_LOCAL, INFO_MAP_LOADED, and
 * INFO_GAME_EVENT. In the latter case it is a user event which can be further
 * subdivided into: TIME_SCALE_CHANGED, ROBOT_SELECTION_CHANGED, and
 * MOUSE_LEFT_CLICK.
 *
 * @param message	Delta3D game message
 */
void SimulationGUIComponent::ProcessMessage(const dtGame::Message& message) {

	// Update GUI and frame rate box
	if (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL) {
		TickGUI();

		const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(message);
		float deltaSimTime = tick.GetDeltaSimTime();
		std::ostringstream t;
		float framerate = 1.0 / (deltaSimTime / GetGameManager()->GetTimeScale());
		t << "Frame Rate = " << (int)(framerate) << " Hz";
		mFrameRateText->setText(t.str());
		simTime += deltaSimTime;
	}

	else if (message.GetMessageType() == dtGame::MessageType::INFO_MAP_LOADED) {
		LOG_WARNING("GUI should not be assigned to default scene!");
	}

	else if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT)
	{
		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);

		if(eventMsg.GetGameEvent() != NULL)
		{
			if (eventMsg.GetGameEvent()->GetName() == TIME_SCALE_CHANGED)
			{
				std::ostringstream t;
				t << (float)((int)(GetGameManager()->GetTimeScale()*100))/100;
				tSValue = t.str();
				mTimeScaleText->setText("Time Scale: " + tSValue);
			}

			else if (eventMsg.GetGameEvent()->GetName() == ROBOT_SELECTION_CHANGED)
			{
				SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
						GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
				int index = comp->getRobFocIndex();
				if (index != -1) {
					std::vector< dtGame::GameActorProxy* > allGameActors;
					GetGameManager()->GetAllGameActors(allGameActors);

					if (index > (int)allGameActors.size()) {
						LOG_INFO("Not so many game actors"); return;
					}

					RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[robotID]->GetGameActor());

					if (sensorID < actor.getSensors().size())
						if (dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[sensorID].get())) {
							CameraSensorBase* mySensor = dynamic_cast<srCore::CameraSensorBase*>(
									actor.getSensors()[sensorID].get());
							mySensor->setRenderOrder(0);
						}

					robotID = index;

					//show new robots camera view
					RobotActorBase &newActor = static_cast<RobotActorBase&>(allGameActors[robotID]->GetGameActor());
					if (sensorID < newActor.getSensors().size())
						if (dynamic_cast<srCore::CameraSensorBase*>(newActor.getSensors()[sensorID].get())) {
							CameraSensorBase* mySensor = dynamic_cast<srCore::CameraSensorBase*>(
									newActor.getSensors()[sensorID].get());
							mySensor->setRenderOrder(1000);
						}

					buildRobotPropertyMap();

				}
			}

			else if (eventMsg.GetGameEvent()->GetName() == MOUSE_LEFT_CLICK)
			{
				std::ostringstream msg;	msg.clear(); msg.str("");
				msg << "Event received: " << eventMsg.GetGameEvent()->GetName();
				LOG_DEBUG(msg.str());

				dtCore::DeltaDrawable *objHit = GetGameManager()->GetApplication().GetView()->GetMousePickedObject(0xffffffff);
				if(objHit != NULL)
				{
					if (dynamic_cast<BodyBase*>(objHit)) {
						BodyBase* body = dynamic_cast<BodyBase*>(objHit);
						std::ostringstream oss;
						oss << "Clicked on robot " << body->getRobot()->GetName();
						LOG_INFO(oss.str());

						SimulationArenaComponent* sac = dynamic_cast<SimulationArenaComponent*>
						(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
						if (sac) sac->selectRobot(body->getRobot());
					}

					//				   TODO: if something is in line of sight to robot make it transparent
					//				   TODO: if wanted deselecting of robots when click subject is a non-robot.
					//				   TODO: if wanted context menu on right mouse click for options like kill, allSensorsOff, etc.
				}
			}

			//			else if (eventMsg.GetGameEvent()->GetName() == TAKE_SCREEN_SHOT_TOGGLE) saveScreenShot ^= 1;
		}
	}
}

/**
 * Creates a dtGUI::CEGUIDrawable from the standard dtCore::DeltaWin window, keyboard
 * and mouse. It is recommended not to hand-over the app.GetWindow() handle, but a
 * separate window for only the GUI components. This means, that the GUI components will
 * not be visible to all cameras defined in the scene. If you don't do that, you will need
 * all kind of hacks to make sure the cameras on the robots do not see the GUI components.
 *
 * Todo: dtGUI::CEUIDrawable and FindFileInPathList are deprecated.
 *
 * @param win		A Delta3D window
 * @param keyboard	A Delta3D keyboard
 * @param mouse		A Delta3D mouse
 */
void SimulationGUIComponent::SetupGUI(dtCore::DeltaWin &win, dtCore::Keyboard &keyboard,
		dtCore::Mouse &mouse)
{
	LOG_ALWAYS("Initialize GUI component");
	try
	{	
		mGUI = new dtGUI::CEUIDrawable(&win, &keyboard, &mouse);
		//		mGUI->SetAutoResizing(false);
		//		mGUI->SetRenderingSize(20,480);

		assert (mGUI != NULL);

		CEGUI::DefaultResourceProvider* rp = static_cast<CEGUI::DefaultResourceProvider*>(
				CEGUI::System::getSingleton().getResourceProvider());
		rp->setResourceGroupDirectory("imagesets", SimulationUtils::Instance()->getCEGUIPath());

		std::string guifilespath(  dtCore::GetDeltaDataPathList() + ";" +
				SimulationUtils::Instance()->getCEGUIPath() );

		dtCore::SetDataFilePathList(guifilespath);

		std::string schemeFileName = dtCore::FindFileInPathList(SimulationUtils::Instance()->getCEGUISchemeFile());
		std::ostringstream msg;	msg.clear(); msg.str("");
		msg << "Use the following CEGUI scheme file: " << schemeFileName;
		LOG_INFO(msg.str());

		CEGUI::SchemeManager::getSingleton().loadScheme(schemeFileName);
		CEGUI::System::getSingleton().setDefaultMouseCursor("WindowsLook", "MouseArrow");
		CEGUI::System::getSingleton().setDefaultFont("DejaVuSans-10");
		CEGUI::System::getSingleton().getDefaultFont()->setProperty("PointSize", "8");

		wm = CEGUI::WindowManager::getSingletonPtr();

		/*
		 * The main / root window. All others are attached to it as child windows.
		 */
		//TODO: Groessen und abstaende dynamisch regeln und auf verzerrung durch maximierung achten
		mSheet = wm->createWindow("DefaultGUISheet", "root_wnd");
		CEGUI::System::getSingleton().setGUISheet(mSheet);

		//		CEGUI::System::getSingleton().setMouseMoveScaling(2.0f);
		//		if (mSheet) mSheet->setMousePassThroughEnabled(true);
		//				mSheet->subscribeEvent(CEGUI::Window::EventMouseEnters,
		//						CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		//				mSheet->subscribeEvent(CEGUI::Window::EventMouseMove,
		//						CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));

		mPanel = mSheet;
		//		mPanel = wm->createWindow("WindowsLook/StaticImage", "Panel 1");
		//		mSheet->addChildWindow(mPanel);
		//		mPanel->setPosition(CEGUI::UVector2(cegui_reldim(0.f), cegui_reldim(0.f)));
		//		mPanel->setSize(CEGUI::UVector2(cegui_reldim(1.f), cegui_reldim(1.f)));
		if (mPanel) mPanel->setMousePassThroughEnabled(true);

		/*
		 * Invisible background panel eventually for robot selection click handling in
		 * scene or context menus. It starts a bit from the top (0.05f) so there is
		 * place for a menu.
		 */
		mBckgr = wm->createWindow("WindowsLook/StaticImage", "Background layer");
		mPanel->addChildWindow(mBckgr);
		mBckgr->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.05f)));
		mBckgr->setSize(CEGUI::UVector2(cegui_reldim(0.75f), cegui_reldim(0.95f)));
		mBckgr->setProperty("Visible", "False");
		mBckgr->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mBckgr->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mBckgr->subscribeEvent(CEGUI::Window::EventMouseMove,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		if (mBckgr) mBckgr->setMousePassThroughEnabled(true);

		/*
		 * The panel at the right with robot (actuator and sensor) information. It starts
		 * at 3/4th of the window (0.75f). If you subscribe an EventMouseEnters event,
		 * dragging the mouse will change viewpoint, but as soon as you drag over this
		 * panel it holds.
		 */
		mRPanel = wm->createWindow("WindowsLook/StaticImage", "Panel right");
		mPanel->addChildWindow(mRPanel);
		mRPanel->setPosition(CEGUI::UVector2(cegui_reldim(0.75f), cegui_reldim(0.0f)));
		mRPanel->setSize(CEGUI::UVector2(cegui_reldim(0.25f), cegui_reldim(1.f)));
		mRPanel->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mRPanel->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mRPanel->setInheritsAlpha(false);
		//		mRPanel->setAlpha(0.3f);

		mSensorValueText = wm->createWindow("WindowsLook/StaticText","Sensor Return Value");
		mRPanel->addChildWindow(mSensorValueText);
		mSensorValueText->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.35f)));
		mSensorValueText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.025f)));
		mSensorValueText->setProperty("Visible", "False");
		mSensorValueText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mSensorValueText->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		/*
		 * Halfway (vertically) on the right side of the window, there is a box that tell
		 * which robot is selected. It is big enough for one line of text (0.025f).
		 */
		mRobotText = wm->createWindow("WindowsLook/StaticText","Robot");
		mRPanel->addChildWindow(mRobotText);
		mRobotText->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.5f)));
		mRobotText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.025f)));
		mRobotText->setProperty("Visible", "False");
		mRobotText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mRobotText->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		/*
		 * A little over halfway (vertically) at the right, there is a box that tells the
		 * position and the orientation. It is big enough for 2 lines (0.05f).
		 */
		mRobotPosRotText = wm->createWindow("WindowsLook/StaticText","Robot Position");
		mRPanel->addChildWindow(mRobotPosRotText);
		mRobotPosRotText->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.525f)));
#ifdef PARTNER_UNIGRAZ
		mRobotPosRotText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.2f)));
#else
		mRobotPosRotText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.05f)));
#endif
		mRobotPosRotText->setProperty("Visible", "False");
		mRobotPosRotText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mRobotPosRotText->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mInfoText = wm->createWindow("WindowsLook/StaticText","Sensor Info");
		mRPanel->addChildWindow(mInfoText);
		mInfoText->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.575f)));
		mInfoText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.25f)));
		mInfoText->setProperty("Visible", "False");
		mInfoText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mInfoText->subscribeEvent(CEGUI::Window::EventMouseButtonDown,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		//        mActuatorInfoText = wm->createWindow("WindowsLook/StaticText","Actuator Info");
		//        mRPanel->addChildWindow(mActuatorInfoText);
		//        mActuatorInfoText->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.7f)));
		//        mActuatorInfoText->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(0.25f)));
		//        mActuatorInfoText->setProperty("Visible", "False");

		mButtonSensorActivity = wm->createWindow("WindowsLook/Button","Sensor Activity");
		mRPanel->addChildWindow(mButtonSensorActivity);
		mButtonSensorActivity->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.825f)));
		mButtonSensorActivity->setSize(CEGUI::UVector2(cegui_reldim(0.25f), cegui_reldim(0.05f)));
		mButtonSensorActivity->setProperty("Text", "Activate");
		mButtonSensorActivity->setProperty("Visible", "False");
		mButtonSensorActivity->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::changeSensorActivity,this));
		mButtonSensorActivity->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonSensorActivity->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonSensorVisibility = wm->createWindow("WindowsLook/Button","Sensor Visibility");
		mRPanel->addChildWindow(mButtonSensorVisibility);
		mButtonSensorVisibility->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.875f)));
		mButtonSensorVisibility->setSize(CEGUI::UVector2(cegui_reldim(0.25f), cegui_reldim(0.05f)));
		mButtonSensorVisibility->setProperty("Text", "Show sensor");
		mButtonSensorVisibility->setProperty("Visible", "False");
		mButtonSensorVisibility->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::changeSensorVisibility,this));
		mButtonSensorVisibility->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonSensorVisibility->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonActuatorFwd = wm->createWindow("WindowsLook/Button","Actuator Move Fwd");
		mRPanel->addChildWindow(mButtonActuatorFwd);
		mButtonActuatorFwd->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.825f)));
		mButtonActuatorFwd->setSize(CEGUI::UVector2(cegui_reldim(0.08333f), cegui_reldim(0.05f)));
		mButtonActuatorFwd->setProperty("Text", "Move Fwd");
		mButtonActuatorFwd->setProperty("Visible", "False");
		mButtonActuatorFwd->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::actuatorFwdActivity,this));
		mButtonActuatorFwd->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonActuatorFwd->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonActuatorBwd = wm->createWindow("WindowsLook/Button","Actuator Move Bwd");
		mRPanel->addChildWindow(mButtonActuatorBwd);
		mButtonActuatorBwd->setPosition(CEGUI::UVector2(cegui_reldim(0.08333f), cegui_reldim(0.825f)));
		mButtonActuatorBwd->setSize(CEGUI::UVector2(cegui_reldim(0.08333f), cegui_reldim(0.05f)));
		mButtonActuatorBwd->setProperty("Text", "Move Bwd");
		mButtonActuatorBwd->setProperty("Visible", "False");
		mButtonActuatorBwd->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::actuatorBwdActivity,this));
		mButtonActuatorBwd->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonActuatorBwd->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonActuatorStop = wm->createWindow("WindowsLook/Button","Actuator Stop");
		mRPanel->addChildWindow(mButtonActuatorStop);
		mButtonActuatorStop->setPosition(CEGUI::UVector2(cegui_reldim(0.16666f), cegui_reldim(0.825f)));
		mButtonActuatorStop->setSize(CEGUI::UVector2(cegui_reldim(0.08334f), cegui_reldim(0.05f)));
		mButtonActuatorStop->setProperty("Text", "Stop");
		mButtonActuatorStop->setProperty("Visible", "False");
		mButtonActuatorStop->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::actuatorStopActivity,this));
		mButtonActuatorStop->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonActuatorStop->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		// menubar at the moment helper to select robots
		mBar = wm->createWindow("WindowsLook/Menubar", "Menubar");
		mPanel->addChildWindow(mBar);
		mBar->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mBar->setSize(CEGUI::UVector2(cegui_reldim(0.75f), cegui_reldim(0.05f)));
		mBar->setProperty("ItemSpacing", "10");
		mBar->setProperty("UnifiedMaxSize", "{{1,0},{1,0}}");
		mBar->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mBar->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mBar->setInheritsAlpha(false);

		// robot selector item
		mRobotSelectorIt = wm->createWindow("WindowsLook/MenuItem", "Robot Selector Item");
		mBar->addChildWindow(mRobotSelectorIt);
		mRobotSelectorIt->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mRobotSelectorIt->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(1.0f)));
		mRobotSelectorIt->setProperty("Text", "Robot");
		mRobotSelectorIt->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mRobotSelectorIt->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::robotMenuHandler,this));
		mRobotSelectorIt->setAlwaysOnTop(true);

		// robot selector popupmenu
		mRobotSelectorPM = wm->createWindow("WindowsLook/PopupMenu", "Robot Selector PopupMenu");
		mRobotSelectorIt->addChildWindow(mRobotSelectorPM);
		mRobotSelectorPM->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mRobotSelectorPM->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(1.0f)));
		mRobotSelectorPM->setProperty("Visible", "False");
		mRobotSelectorPM->setProperty("FadeInTime", "0");
		mRobotSelectorPM->setProperty("FadeOutTime", "0");
		mRobotSelectorPM->setProperty("ItemSpacing", "2");
		mRobotSelectorPM->setProperty("ClippedByParent", "False");
		mRobotSelectorPM->setProperty("AutoResizeEnabled", "True");
		mRobotSelectorPM->setAlwaysOnTop(true);

		mSensorSelectorIt = wm->createWindow("WindowsLook/MenuItem","Sensor Selector Item");
		mBar->addChildWindow(mSensorSelectorIt);
		mSensorSelectorIt->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.5f)));
		mSensorSelectorIt->setSize(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.05f)));
		mSensorSelectorIt->setText("Sensor");
		mSensorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mSensorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::sensorMenuHandler,this));

		mSensorSelectorPM = wm->createWindow("WindowsLook/PopupMenu", "Sensor Selector PopupMenu");
		mSensorSelectorIt->addChildWindow(mSensorSelectorPM);
		mSensorSelectorPM->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mSensorSelectorPM->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(1.0f)));
		mSensorSelectorPM->setProperty("Visible", "False");
		mSensorSelectorPM->setProperty("FadeInTime", "0");
		mSensorSelectorPM->setProperty("FadeOutTime", "0");
		mSensorSelectorPM->setProperty("ItemSpacing", "2");
		mSensorSelectorPM->setProperty("ClippedByParent", "False");
		mSensorSelectorPM->setProperty("AutoResizeEnabled", "True");

		mActuatorSelectorIt = wm->createWindow("WindowsLook/MenuItem","Actuator Selector Item");
		mBar->addChildWindow(mActuatorSelectorIt);
		mActuatorSelectorIt->setPosition(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.5f)));
		mActuatorSelectorIt->setSize(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.05f)));
		mActuatorSelectorIt->setText("Actuator");
		mActuatorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mActuatorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::actuatorMenuHandler,this));

		mActuatorSelectorPM = wm->createWindow("WindowsLook/PopupMenu", "Actuator Selector PopupMenu");
		mActuatorSelectorIt->addChildWindow(mActuatorSelectorPM);
		mActuatorSelectorPM->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mActuatorSelectorPM->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(1.0f)));
		mActuatorSelectorPM->setProperty("Visible", "False");
		mActuatorSelectorPM->setProperty("FadeInTime", "0");
		mActuatorSelectorPM->setProperty("FadeOutTime", "0");
		mActuatorSelectorPM->setProperty("ItemSpacing", "2");
		mActuatorSelectorPM->setProperty("ClippedByParent", "False");
		mActuatorSelectorPM->setProperty("AutoResizeEnabled", "True");

		mConnectorSelectorIt = wm->createWindow("WindowsLook/MenuItem","Connector Selector Item");
		mBar->addChildWindow(mConnectorSelectorIt);
		mConnectorSelectorIt->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.5f)));
		mConnectorSelectorIt->setSize(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.05f)));
		mConnectorSelectorIt->setText("Connector");
		mConnectorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mConnectorSelectorIt->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::connectorMenuHandler,this));

		mConnectorSelectorPM = wm->createWindow("WindowsLook/PopupMenu", "Connector Selector PopupMenu");
		mConnectorSelectorIt->addChildWindow(mConnectorSelectorPM);
		mConnectorSelectorPM->setPosition(CEGUI::UVector2(cegui_reldim(0.0f), cegui_reldim(0.0f)));
		mConnectorSelectorPM->setSize(CEGUI::UVector2(cegui_reldim(1.0f), cegui_reldim(1.0f)));
		mConnectorSelectorPM->setProperty("Visible", "False");
		mConnectorSelectorPM->setProperty("FadeInTime", "0");
		mConnectorSelectorPM->setProperty("FadeOutTime", "0");
		mConnectorSelectorPM->setProperty("ItemSpacing", "2");
		mConnectorSelectorPM->setProperty("ClippedByParent", "False");
		mConnectorSelectorPM->setProperty("AutoResizeEnabled", "True");

		// time scale display
		mTimeScaleText = wm->createWindow("WindowsLook/StaticText","Time scale");
		mBar->addChildWindow(mTimeScaleText);
		mTimeScaleText->setPosition(CEGUI::UVector2(cegui_reldim(0.4f), cegui_reldim(0.0f)));
		mTimeScaleText->setSize(CEGUI::UVector2(cegui_reldim(0.2f), cegui_reldim(1.0f)));
		std::ostringstream t;
		t << (float)((int)(GetGameManager()->GetTimeScale()*100))/100;
		tSValue = t.str();
		mTimeScaleText->setText("Time Scale: " + tSValue);
		mTimeScaleText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mTimeScaleText->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonTimeScaleUp = wm->createWindow("WindowsLook/Button","Time scale Up");
		mTimeScaleText->addChildWindow(mButtonTimeScaleUp);
		mButtonTimeScaleUp->setPosition(CEGUI::UVector2(cegui_reldim(0.9f), cegui_reldim(0.0f)));
		mButtonTimeScaleUp->setSize(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.5f)));
		mButtonTimeScaleUp->setText("+");
		mButtonTimeScaleUp->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::changeTimeScale,this));
		mButtonTimeScaleUp->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonTimeScaleUp->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonTimeScaleDown = wm->createWindow("WindowsLook/Button","Time scale Down");
		mTimeScaleText->addChildWindow(mButtonTimeScaleDown);
		mButtonTimeScaleDown->setPosition(CEGUI::UVector2(cegui_reldim(0.9f), cegui_reldim(0.5f)));
		mButtonTimeScaleDown->setSize(CEGUI::UVector2(cegui_reldim(0.1f), cegui_reldim(0.5f)));
		mButtonTimeScaleDown->setText("-");
		mButtonTimeScaleDown->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::changeTimeScale,this));
		mButtonTimeScaleDown->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mButtonTimeScaleDown->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mFrameRateText = wm->createWindow("WindowsLook/StaticText","Frame Rate");
		mBar->addChildWindow(mFrameRateText);
		mFrameRateText->setPosition(CEGUI::UVector2(cegui_reldim(0.6f), cegui_reldim(0.0f)));
		mFrameRateText->setSize(CEGUI::UVector2(cegui_reldim(0.2f), cegui_reldim(1.0f)));
		mFrameRateText->setText("Frame Rate: ");
		mFrameRateText->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mFrameRateText->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));

		mButtonObservationMode = wm->createWindow("WindowsLook/Button","Observation Mode");
		mBar->addChildWindow(mButtonObservationMode);
		mButtonObservationMode->setPosition(CEGUI::UVector2(cegui_reldim(0.8f), cegui_reldim(0.0f)));
		mButtonObservationMode->setSize(CEGUI::UVector2(cegui_reldim(0.2f), cegui_reldim(1.0f)));
		if(freeCamState)		mButtonObservationMode->setText("To Follow Mode");
		else					mButtonObservationMode->setText("To Free Mode");
		mButtonObservationMode->subscribeEvent(CEGUI::Window::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::changeObservationMode,this));
		mButtonObservationMode->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mButtonObservationMode->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
	}
	// catch to prevent exit (errors will be logged).
	catch(CEGUI::Exception &e)
	{
		std::ostringstream oss;
		oss << "CEGUI::" << e.getMessage().c_str();
		LOG_ERROR(oss.str());
	}

}

/**
 * Each tick the GUI is updated. From SimulationArenaComponent the selected robot is
 * retrieved. If there is no robot selected, this routine sets all information that
 * relates to the previously selected robot to invisible. If there is a robot selected
 * indeed, buildRobotPropertyMap is called. If the menu at the top is clicked,
 * information about sensors, actuators or connectors is updated.
 */
void SimulationGUIComponent::TickGUI()
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	if (index < 0) { LOG_WARNING("Index is smaller than 0"); return; }
	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);

	if (index > (int)allGameActors.size()) {
		LOG_INFO("Not so many game actors"); return;
	}

	mSensorValueText->setProperty("Visible", "False");
	mButtonSensorActivity->setProperty("Visible", "False");
	mButtonSensorVisibility->setProperty("Visible", "False");
	mButtonActuatorFwd->setProperty("Visible", "False");
	mButtonActuatorBwd->setProperty("Visible", "False");
	mButtonActuatorStop->setProperty("Visible", "False");

	hideRenderTexture = true;

	if(index!=-1)
	{
		buildRobotPropertyMap();

		switch (selectedMenuItem) {

		case SENSOR:
			buildSensorPropertyMap();
			break;
		case ACTUATOR:
			buildActuatorPropertyMap();
			break;
		case CONNECTOR:
			buildConnectorPropertyMap();
			break;
		case NONE:
			hideRenderTexture = true;
			mSensorValueText->setProperty("Visible", "False");
			mInfoText->setProperty("Visible", "False");
			break;
		}
	}
	else
	{
		mRobotText->setProperty("Visible", "False");
		mRobotPosRotText->setProperty("Visible", "False");

		mInfoText->setProperty("Visible", "False");
		mButtonSensorActivity->setProperty("Visible", "False");
		mButtonSensorVisibility->setProperty("Visible", "False");
	}
}


void SimulationGUIComponent::sendMessageToAll(dtDAL::GameEvent &event)
{
	dtCore::RefPtr<dtGame::GameEventMessage> eventMsg;
	GetGameManager()->GetMessageFactory().CreateMessage(
			dtGame::MessageType::INFO_GAME_EVENT, eventMsg);

	eventMsg->SetGameEvent(event);
	GetGameManager()->SendMessage(*eventMsg);
}


bool SimulationGUIComponent::robotMenuHandler(const CEGUI::EventArgs& eventArgs)
{
	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);

	for(int i=0; i<(int)mRobot.size(); ++i)
		wm->destroyWindow(mRobot.at(i));
	wm->cleanDeadPool();
	mRobot.erase(mRobot.begin(),mRobot.end());

	for(int i=0; i<(int)GetGameManager()->GetNumGameActors(); i++)
	{
		RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[i]->GetGameActor());
		std::ostringstream t;
		t << actor.getID();
		CEGUI::Window* robot = wm->createWindow("WindowsLook/MenuItem", "Robot" + t.str());
		mRobot.push_back(robot);
		mRobotSelectorPM->addChildWindow(mRobot.back());
		mRobot.back()->setProperty("Text", "Robot" + t.str());
		mRobot.back()->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mRobot.back()->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mRobot.back()->subscribeEvent(CEGUI::MenuItem::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::robotSelectionHandler,this));
	}
	return true;
}


bool SimulationGUIComponent::sensorMenuHandler(const CEGUI::EventArgs& eventArgs)
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	for(int i=0; i<(int)mSensor.size(); ++i)
		wm->destroyWindow(mSensor.at(i));
	wm->cleanDeadPool();
	mSensor.erase(mSensor.begin(),mSensor.end());

	std::vector<dtCore::RefPtr<SensorBase> > sensors = actor.getSensors();

	if (!sensors.empty())
		for(int j=0; j<(int)sensors.size(); j++)
		{

			std::ostringstream t;
			t << j;
			CEGUI::Window* sensor = wm->createWindow("WindowsLook/MenuItem", "Sensor" + t.str());
			mSensor.push_back(sensor);
			mSensorSelectorPM->addChildWindow(mSensor.back());
			std::string test = sensors[j]->getName() + ", Type " + sensors[j]->getType();
			mSensor.back()->setProperty("Text", test);
			mSensor.back()->subscribeEvent(CEGUI::Window::EventMouseEnters,
					CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
			mSensor.back()->subscribeEvent(CEGUI::Window::EventMouseLeaves,
					CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
			mSensor.back()->subscribeEvent(CEGUI::MenuItem::EventMouseClick,
					CEGUI::Event::Subscriber(&SimulationGUIComponent::sensorSelectionHandler,this));
		}

	return true;
}


bool SimulationGUIComponent::actuatorMenuHandler(const CEGUI::EventArgs& eventArgs)
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();
	if (index < 0) {
		LOG_WARNING("No robot selected");
		return false;
	}

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	for(int i=0; i<(int)mActuator.size(); ++i)
		wm->destroyWindow(mActuator.at(i));
	wm->cleanDeadPool();
	mActuator.erase(mActuator.begin(),mActuator.end());

	std::vector<dtCore::RefPtr<ActuatorBase> > actuators = actor.getActuators();
	for(int j=0; j<(int)actuators.size(); j++)
	{
		std::ostringstream t;
		t << j;
		CEGUI::Window* actuator = wm->createWindow("WindowsLook/MenuItem", "Actuator" + t.str());
		mActuator.push_back(actuator);
		mActuatorSelectorPM->addChildWindow(mActuator.back());
		std::string test = actuators[j]->getName() + ", Type " + actuators[j]->getType();
		mActuator.back()->setProperty("Text", test);
		mActuator.back()->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mActuator.back()->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mActuator.back()->subscribeEvent(CEGUI::MenuItem::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::actuatorSelectionHandler,this));
	}

	return true;
}


bool SimulationGUIComponent::connectorMenuHandler(const CEGUI::EventArgs& eventArgs)
{
	//	SimulationRobotSelectionComponent *comp = static_cast<SimulationRobotSelectionComponent*>(GetGameManager()->GetComponentByName("SimulationRobotSelectionComponent"));
	//	int index = comp->getRobFocIndex();

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	for(int i=0; i<(int)mConnector.size(); ++i)
		wm->destroyWindow(mConnector.at(i));
	wm->cleanDeadPool();
	mConnector.erase(mConnector.begin(),mConnector.end());

	std::vector<dtCore::RefPtr<ConnectorBase> > connectors = actor.getConnectors();
	for(int j=0; j<(int)connectors.size(); j++)
	{
		std::ostringstream t;
		t << j;
		CEGUI::Window* connector = wm->createWindow("WindowsLook/MenuItem", "Connector" + t.str());
		mConnector.push_back(connector);
		mConnectorSelectorPM->addChildWindow(mConnector.back());
		std::string test = connectors[j]->getName() + ", Type " + connectors[j]->getType();
		mConnector.back()->setProperty("Text", test);
		mConnector.back()->subscribeEvent(CEGUI::Window::EventMouseEnters,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInGUI,this));
		mConnector.back()->subscribeEvent(CEGUI::Window::EventMouseLeaves,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::mouseInSim,this));
		mConnector.back()->subscribeEvent(CEGUI::MenuItem::EventMouseClick,
				CEGUI::Event::Subscriber(&SimulationGUIComponent::connectorSelectionHandler,this));
	}

	return true;
}

bool SimulationGUIComponent::robotSelectionHandler(const CEGUI::EventArgs& eventArgs)
{
	selectedMenuItem = NONE;
	CEGUI::MenuItem* mI = (CEGUI::MenuItem*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String number = mI->getName();

	number.erase(0,5);

	if (atoi(number.c_str()) < (int)GetGameManager()->GetNumGameActors())
	{
		mSelectRobot.get()->SetDescription(static_cast<std::string>(number.c_str()));
		sendMessageToAll(*mSelectRobot);
	}
	else
		LOG_ERROR("Fehler in SimulationGUIComponent::robotSelectionHandler");
	return true;
}

bool SimulationGUIComponent::sensorSelectionHandler(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::MenuItem* mI = (CEGUI::MenuItem*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String number = mI->getName();

	//	SimulationRobotSelectionComponent *comp = static_cast<SimulationRobotSelectionComponent*>(GetGameManager()->GetComponentByName("SimulationRobotSelectionComponent"));
	//	int index = comp->getRobFocIndex();
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());


	unsigned int oldSensorID = sensorID;

	if (sensorID < actor.getSensors().size()) {
		dtCore::RefPtr<srCore::SensorBase> mySensor =  actor.getSensors()[sensorID];
		//hide robots old camera view
		if (dynamic_cast<srCore::CameraSensorBase*>(mySensor.get())) {
			CameraSensorBase* myCameraSensor = dynamic_cast<srCore::CameraSensorBase*>(mySensor.get());
			myCameraSensor->setRenderOrder(0);
		}

		number.erase(0,6);
		sensorID = atoi(number.c_str());

		//show robots new camera view
		if (dynamic_cast<srCore::CameraSensorBase*>(mySensor.get())) {
			CameraSensorBase* myCameraSensor = dynamic_cast<srCore::CameraSensorBase*>(mySensor.get());
			myCameraSensor->setRenderOrder(1000);
		}
	}


	if (!actor.getSensors().empty())
		if (sensorID < actor.getSensors().size())
		{
			SimulationGUIComponent *guiComp = static_cast<SimulationGUIComponent*>(GetGameManager()->GetComponentByName("SimulationGUIComponent"));
			guiComp->buildSensorPropertyMap();
		}
	if (oldSensorID < actor.getSensors().size()) {
		actor.getSensors()[oldSensorID]->makeVisible(false);

		//		if (dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[oldSensorID])) {
		//			CameraSensorBase* mySensor = dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[oldSensorID]);
		//			mySensor->setRenderOrder(0);
		//		}
	}

	mButtonSensorVisibility->setText("Show sensor");

	return true;
}

bool SimulationGUIComponent::connectorSelectionHandler(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::MenuItem* mI = (CEGUI::MenuItem*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String number = mI->getName();

	//	SimulationRobotSelectionComponent *comp = static_cast<SimulationRobotSelectionComponent*>(GetGameManager()->GetComponentByName("SimulationRobotSelectionComponent"));
	//	int index = comp->getRobFocIndex();
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	//hide robot camera view
	if (sensorID < actor.getSensors().size())
		if (dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[sensorID].get())) {
			CameraSensorBase* mySensor = dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[sensorID].get());
			mySensor->setRenderOrder(0);
		}

	number.erase(0,9);
	connectorID = atoi(number.c_str());

	if (connectorID < actor.getConnectors().size())
	{
		SimulationGUIComponent *guiComp = static_cast<SimulationGUIComponent*>(GetGameManager()->GetComponentByName("SimulationGUIComponent"));
		guiComp->buildConnectorPropertyMap();
	}

	return true;
}

/*
 * Function: actuatorSelectionHandler
 *
 * When someone clicks on an actuator in the Actuator menu, the name of the menu-item is interpreted.
 * It is expected to have the form "ssssssss23", with "ssssssss" as long as the string "actuator". The
 * prefix is interpreted as a number, the actuatorID. If this is a valid number, the function
 * buildActuatorPropertyMap is called.
 *
 * There is no check if there is actually a robot selected. If there is a sensor, it is removed from the
 * scene by setting its rendering order. If there is no sensor, the dynamic cast to a CameraSensorBase
 * returns null.
 *
 * ObjectParameter:
 *   eventArgs	- A CEGUI event (menu-item click)
 *
 * Returns:
 * A boolean denoting success.
 */
bool SimulationGUIComponent::actuatorSelectionHandler(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::MenuItem* mI = (CEGUI::MenuItem*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String number = mI->getName();

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	//hide robot camera view
	if (sensorID < actor.getSensors().size())
		if (dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[sensorID].get())) {
			CameraSensorBase* mySensor = dynamic_cast<srCore::CameraSensorBase*>(actor.getSensors()[sensorID].get());
			mySensor->setRenderOrder(0);
		}

	number.erase(0,8);
	actuatorID = atoi(number.c_str());

	if (actuatorID < actor.getActuators().size())
	{
		SimulationGUIComponent *guiComp = static_cast<SimulationGUIComponent*>(GetGameManager()->GetComponentByName("SimulationGUIComponent"));
		guiComp->buildActuatorPropertyMap();
	}

	return true;
}


bool SimulationGUIComponent::mouseInGUI(const CEGUI::EventArgs& eventArgs)
{
	sendMessageToAll(*mMouseInGUI);
	return true;
}

bool SimulationGUIComponent::mouseInSim(const CEGUI::EventArgs& eventArgs)
{
	if (destroying) return true;
	sendMessageToAll(*mMouseInSim);
	return true;
}

/**
 * Sets the camera status to free roaming when true, and fixed-to-robot when false.
 * This does not send a game message. Use changeObservationMode for that. The button
 * text tells TOWARDS which state the user can switch.
 * @param camstate true when camera has to go in free state
 */
void SimulationGUIComponent::setCameraMode(bool camstate) {
	freeCamState = camstate;
	if (camstate) {
		mButtonObservationMode->setText("To Follow Mode");
	} else {
		mButtonObservationMode->setText("To Free Mode");
	}
}

/**
 * Change the observation mode to the given freeCamState. The actual change is done
 * in simulationCameraController.
 * @param eventArgs
 * @return
 */
bool SimulationGUIComponent::changeObservationMode(const CEGUI::EventArgs& eventArgs) {
	if (freeCamState)
		freeCamState = false;
	else
		freeCamState = true;
	setCameraMode(freeCamState);
	sendMessageToAll(*mChangeCameraMode);
	return true;
}

bool SimulationGUIComponent::changeTimeScale(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::ButtonBase* button = (CEGUI::ButtonBase*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String buttonText = button->getText();
	std::ostringstream t;

	if(buttonText == "+")
	{
		GetGameManager()->ChangeTimeSettings(GetGameManager()->GetSimulationTime(),
				GetGameManager()->GetTimeScale() * 1.1, GetGameManager()->GetSimulationClockTime());
		t << (float)((int)(GetGameManager()->GetTimeScale()*100))/100;
		tSValue = t.str();
		mTimeScaleText->setText("Time Scale: " + tSValue);
	}
	else if (buttonText == "-")
	{
		GetGameManager()->ChangeTimeSettings(GetGameManager()->GetSimulationTime(),
				GetGameManager()->GetTimeScale() * 0.9, GetGameManager()->GetSimulationClockTime());
		t << (float)((int)(GetGameManager()->GetTimeScale()*100))/100;
		tSValue = t.str();
		mTimeScaleText->setText("Time Scale: " + tSValue);
	}
	return true;
}


bool SimulationGUIComponent::actuatorStopActivity(const CEGUI::EventArgs& eventArgs)
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	float value = 0;
	actor.getActuators()[actuatorID]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

	return true;
}

bool SimulationGUIComponent::actuatorFwdActivity(const CEGUI::EventArgs& eventArgs)
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	float value;
	actor.getActuators()[actuatorID]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
	value++;
	actor.getActuators()[actuatorID]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

	return true;
}

bool SimulationGUIComponent::actuatorBwdActivity(const CEGUI::EventArgs& eventArgs)
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	float value;
	actor.getActuators()[actuatorID]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
	value--;
	actor.getActuators()[actuatorID]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

	return true;
}

bool SimulationGUIComponent::changeSensorActivity(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::ButtonBase* button = (CEGUI::ButtonBase*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String buttonText = button->getText();

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	if ((buttonText == "Deactivate") || (buttonText == "Activate"))
	{
		if (sensorID < actor.getSensors().size())
		{
			actor.getSensors()[sensorID]->setActive( actor.getSensors()[sensorID]->isActive()^1 );
			if(actor.getSensors()[sensorID]->isActive()) 	mButtonSensorActivity->setText("Deactivate");
			else 											mButtonSensorActivity->setText("Activate");
		}
	}

	if ((buttonText == "Connect") || (buttonText == "Disconnect"))
	{
		if (buttonText == "Connect") actor.getConnectors()[connectorID]->connect();
		if (buttonText == "Disconnect") actor.getConnectors()[connectorID]->disconnect();

		if(actor.getConnectors()[connectorID]->getState() != ConnectorBase::DISCONNECTED) 	mButtonSensorActivity->setText("Disconnect");
		else 													mButtonSensorActivity->setText("Connect");
	}

	return true;
}

bool SimulationGUIComponent::changeSensorVisibility(const CEGUI::EventArgs& eventArgs)
{
	CEGUI::ButtonBase* button = (CEGUI::ButtonBase*)((const CEGUI::WindowEventArgs&)eventArgs).window;
	CEGUI::String buttonText = button->getText();

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	if (buttonText == "Show sensor")
	{
		if (sensorID < actor.getSensors().size())
		{
			mButtonSensorVisibility->setText("Hide sensor");
			actor.getSensors()[sensorID]->makeVisible(true);
		}
	}
	else
	{
		if (sensorID < actor.getSensors().size())
		{
			mButtonSensorVisibility->setText("Show sensor");
			actor.getSensors()[sensorID]->makeVisible(false);
		}
	}

	return true;
}


void SimulationGUIComponent::buildRobotPropertyMap()
{
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	if (index < 0) return;

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	if (index >= (int)allGameActors.size()) return;
	RobotActorBase &actor = static_cast<RobotActorBase&>(
			allGameActors[index]->GetGameActor());

	std::ostringstream t;
	osg::Vec3f pos, rot;

	rot = actor.getRotation();
	pos = actor.getTranslation();

	t << actor.getID();
	mRobotText->setProperty("Visible", "True");
	mRobotText->setText("Robot " + t.str());
	t.str("");

	t << "Pos: " << (float)((int)(pos.x()*10))/10 << "; " << (float)((int)(pos.y()*10))/10
			<< "; " << (float)((int)(pos.z()*10))/10 << ";\n";
	t << "Rot: " << (float)((int)(rot.x()*10))/10 << "; " << (float)((int)(rot.y()*10))/10
			<< "; " << (float)((int)(rot.z()*10))/10 << ";";

	mRobotPosRotText->setProperty("Visible", "True");
	mRobotPosRotText->setText(t.str());
}


/**
 * There seem only to be screenshots if there is one sensor active. This routine suggests that
 * ONE actor is active, for example selected by the user. So, it is not possible to show the
 * sensors of multiple robots using this routine. Moreover, it assumes that only ONE sensor is
 * currently active, which is retrieved by the class variable SENSORID. This sensor is taken
 * out of the array of sensors attached to the actor.
 *
 * For a camera the image is shown via sensor->getRawData.
 */
void SimulationGUIComponent::buildSensorPropertyMap()
{
	selectedMenuItem = SENSOR;

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	std::vector<dtCore::RefPtr<SensorBase> > sensors = actor.getSensors();

	if (!sensors.empty())
	{
		if(!(sensorID<sensors.size()))
		{
			mInfoText->setProperty("Visible", "False");
			mButtonSensorActivity->setProperty("Visible", "False");
			mButtonSensorVisibility->setProperty("Visible", "False");

		}

		else {

			if (sensors[sensorID]->getReturnValueType() == typeid(osg::Image**).name())
			{
				hideRenderTexture = false;
				mSensorValueText->setProperty("Visible", "False");

				SensorBase* mySensor = sensors[sensorID].get();

				osg::ref_ptr<osg::Image> tmpImage;
				if (mySensor->getType() != "LaserScanner")
					tmpImage = dynamic_cast<CameraSensorBase*>(mySensor)->getRawData().get();
				else {
					ParameterReturnValue prv = MAKE_PARAMETER_VALUE(&tmpImage);
					sensors[sensorID]->evaluate(prv);
				}

				int w,h;
				w = tmpImage->s();
				h = tmpImage->t();

				if (image.get()) image.release();
			}

			else if (sensors[sensorID]->getReturnValueType() == typeid(float*).name())
			{
				hideRenderTexture = true;
				mSensorValueText->setProperty("Visible", "True");

				//					SensorBase* mySensor = sensors[sensorID];
				float sensorValue;
				sensors[sensorID]->evaluate(MAKE_PARAMETER_VALUE(&sensorValue));

				std::ostringstream t;
				t << sensorValue;
				std::string text = "Sensor Value = " + t.str();
				mSensorValueText->setText(text);
			}

			else if (sensors[sensorID]->getReturnValueType() == typeid(unsigned long*).name())
			{
				hideRenderTexture = true;
				mSensorValueText->setProperty("Visible", "True");

				//					SensorBase* mySensor = sensors[sensorID];
				unsigned long sensorValue;
				sensors[sensorID]->evaluate(MAKE_PARAMETER_VALUE(&sensorValue));

				std::ostringstream t;
				t << sensorValue;
				std::string text = "Sensor Value = " + t.str();
				mSensorValueText->setText(text);
			}


			std::vector<ObjectParameter> parameters = sensors[sensorID]->getParameterList();

			std::string text = sensors[sensorID]->getName() + " of Type:\n	  " + sensors[sensorID]->getType() + "\n";

			for(int j=0; j<(int)parameters.size(); j++)
			{
				std::ostringstream t;
				std::string type = parameters[j].valueType;
				if(type == typeid(float*).name())
				{
					float value;
					sensors[sensorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
					t << value;
					text = text + "\n" + parameters[j].valueName + " = " + t.str();
				}
				else if (type == typeid(osg::Vec3f*).name())
				{
					osg::Vec3f value;
					sensors[sensorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
					t << (float)((int)(value.x()*10))/10; t << ", ";
					t << (float)((int)(value.y()*10))/10; t << ", ";
					t << (float)((int)(value.z()*10))/10; t << ";";
					text = text + "\n" + parameters[j].valueName + " = " + t.str();
				}
				else if (type == typeid(osg::Vec4f*).name())
				{
					osg::Vec4f value;
					sensors[sensorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
					t << (float)((int)(value.x()*1000))/1000; t << ", ";
					t << (float)((int)(value.y()*1000))/1000; t << ", ";
					t << (float)((int)(value.z()*1000))/1000; t << ", ";
					t << (float)((int)(value.w()*1000))/1000; t << ";";
					text = text + "\n" + parameters[j].valueName + " =\n" + t.str();
				}
				else if (type == typeid(osg::Vec2*).name())
				{
					osg::Vec2 value;
					sensors[sensorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
					t << (float)((int)(value.x()*100))/100; t << ", ";
					t << (float)((int)(value.y()*100))/100; t << ";";
					text = text + "\n" + parameters[j].valueName + " = " + t.str();
				}
			}
			mInfoText->setText(text);
			mInfoText->setProperty("Visible", "True");

			if(actor.getSensors()[sensorID]->isActive()) 	mButtonSensorActivity->setText("Deactivate");
			else 											mButtonSensorActivity->setText("Activate");

			mButtonSensorActivity->setProperty("Visible", "True");
			mButtonSensorVisibility->setProperty("Visible", "True");

		}

	}
}


void SimulationGUIComponent::buildConnectorPropertyMap()
{
	selectedMenuItem = CONNECTOR;

	//	SimulationRobotSelectionComponent *comp = static_cast<SimulationRobotSelectionComponent*>(GetGameManager()->GetComponentByName("SimulationRobotSelectionComponent"));
	//	int index = comp->getRobFocIndex();
	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());


	std::vector<dtCore::RefPtr<ConnectorBase> > connectors = actor.getConnectors();
	if(!(connectorID<connectors.size())) { mInfoText->setProperty("Visible", "False"); }
	std::vector<ObjectParameter> parameters = connectors[connectorID]->getParameterList();
	std::string text = connectors[connectorID]->getName() + " of Type:\n	  " + connectors[connectorID]->getType() + "\n";

	if(actor.getConnectors()[connectorID]->getState() != ConnectorBase::DISCONNECTED) 	mButtonSensorActivity->setText("Disconnect");
	else 													mButtonSensorActivity->setText("Connect");
	mButtonSensorActivity->setProperty("Visible", "True");
}

/*
 * Function: buildActorPropertyMap
 *
 * Sets the "selected" menu to the Actuator menu. It is called from actuatorSelectionHandler that
 * receives the direct mouse click from the user. This routine again retrieves from the
 * SimulationArenaComponent which robot is selected. It subsequently uses "actuatorID" that is set
 * before by retrieving the proper actuator and write its parameter figures into the information
 * text box, the mInfoText CEGUI window.
 *
 * Assumes that there is a robot selected indeed. It assumes that there are four types of parameters
 * possible for that actuator: float, Vec2f, Vec3f, and Vec4f.
 */
void SimulationGUIComponent::buildActuatorPropertyMap()
{
	selectedMenuItem = ACTUATOR;

	SimulationArenaComponent *comp = dynamic_cast<SimulationArenaComponent*>(
			GetGameManager()->GetComponentByName(SimulationArenaComponent::DEFAULT_NAME));
	int index = comp->getRobFocIndex();

	std::vector< dtGame::GameActorProxy* > allGameActors;
	GetGameManager()->GetAllGameActors(allGameActors);
	RobotActorBase &actor = static_cast<RobotActorBase&>(allGameActors[index]->GetGameActor());

	mButtonActuatorFwd->setProperty("Visible", "True");
	mButtonActuatorBwd->setProperty("Visible", "True");
	mButtonActuatorStop->setProperty("Visible", "True");

	hideRenderTexture = true;

	std::vector<dtCore::RefPtr<ActuatorBase> > actuators = actor.getActuators();
	if(!(actuatorID < actuators.size())) {
		mInfoText->setProperty("Visible", "False");
	}
	std::vector<ObjectParameter> parameters = actuators[actuatorID]->getParameterList();
	std::string text = actuators[actuatorID]->getName() + " of Type:\n	  " + actuators[actuatorID]->getType() + "\n";
	for(unsigned int j=0; j < parameters.size(); j++)
	{
		std::ostringstream t;
		std::string type = parameters[j].valueType;
		if(type == typeid(float*).name())
		{
			float value;
			actuators[actuatorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
			t << value;
			text = text + "\n" + parameters[j].valueName + " = " + t.str();
		}
		else if (type == typeid(osg::Vec3f*).name())
		{
			osg::Vec3f value;
			actuators[actuatorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
			t << (float)((int)(value.x()*10))/10; t << ", ";
			t << (float)((int)(value.y()*10))/10; t << ", ";
			t << (float)((int)(value.z()*10))/10; t << ";";
			text = text + "\n" + parameters[j].valueName + " = " + t.str();
		}
		else if (type == typeid(osg::Vec4f*).name())
		{
			osg::Vec4f value;
			actuators[actuatorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
			t << (float)((int)(value.x()*1000))/1000; t << ", ";
			t << (float)((int)(value.y()*1000))/1000; t << ", ";
			t << (float)((int)(value.z()*1000))/1000; t << ", ";
			t << (float)((int)(value.w()*1000))/1000; t << ";";
			text = text + "\n" + parameters[j].valueName + " =\n" + t.str();
		}
		else if (type == typeid(osg::Vec2f*).name())
		{
			osg::Vec2 value;
			actuators[actuatorID]->getParameter(parameters[j].valueName, MAKE_PARAMETER_VALUE(&value));
			t << (float)((int)(value.x()*100))/100; t << ", ";
			t << (float)((int)(value.y()*100))/100; t << ";";
			text = text + "\n" + parameters[j].valueName + " = " + t.str();
		}
	}
	mInfoText->setText(text);
	mInfoText->setProperty("Visible", "True");
}
