/**
 * @file loadComponents.cpp
 *
 * This file loadComponents.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author 	Anne C. van Rossum
 * @date	Aug 4, 2010
 * @project	Replicator FP7
 * @company	Almende B.V.
 * @case	Plugin
 */


// General files
#include <dlfcn.h>
#include <sstream>

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/loadComponents.h>
#include <srCore/pluginEntryPointBase.h>

namespace srCore {

/* **************************************************************************************
 * Implementation of LoadComponents
 * **************************************************************************************/

/**
 * Default constructor clears the path to load plugins from, and removes all file handles.
 */
LoadComponents::LoadComponents():
		path(), library(), initialised(false), error(false), handles() {
	path.clear();
	handles.clear();
}

/**
 * Closes all file handlers. The symbols become available, so if you have actors defined
 * in your plugin they will not be available anymore to Robot3D.
 */
LoadComponents::~LoadComponents() {
	LOG_INFO("Close all dynamically linked libraries/components (symbols become unavailable)");
	std::vector<void *>::iterator i;
	for (i = handles.begin(); i != handles.end(); i++) {
		dlclose(*i);
	}
}

/**
 * Sets the path from which to load the shared libraries (plugins).
 */
void LoadComponents::SetLibPath(const std::string &libpath) {
	path = libpath;
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Set library path for additional plugins to " << libpath;
	LOG_INFO(msg.str());
	if (!boost::filesystem::is_directory(path)) {
		LOG_ERROR("The path probably does not exist. Stop trying to load libraries.");
		initialised = false;
		error = true;
	}
}

/**
 * Get all the shared libraries (.so files) in the previously assigned path.
 * Do not call this function before you have set SetLibPath.
 */
void LoadComponents::GetLibraries() {
	if (error) return;
	assert (!path.empty());
	LOG_INFO("Try to find libraries in given library path");
	std::ostringstream msg;
	std::string file, extension;
	boost::filesystem::directory_iterator i(path), dir_end;
	for(; i != dir_end; ++i) {
		file = (*i).leaf();
		extension = boost::filesystem::extension(*i);
		if (std::string(extension).find(".so") != std::string::npos) {
			msg.clear(); msg.str("");
			msg << "Found file " << file;
			LOG_DEBUG(msg.str());
			file = path.string() + '/' + file;
			boost::filesystem::complete(file);
			library.push_back(file);
		}
	}
	initialised = true;
}

/**
 * Load additional components or more general, libraries, by searching through a given path that
 * comes from the \<scenario.xml\> file.
 * The plugins are loaded by dlopen and subsequently dlsym. The latter searches for a
 * CreatePluginEntryPoint. Check if that field indeed exist in unmangeled state in the library
 * to be loaded (by for example nm).
 *
 * @param application
 */
void LoadComponents::LoadLibs(dtGame::GameApplication &application) {
	if (error) return;
	assert (initialised);
	LOG_INFO("Load plugins");
	std::ostringstream msg;

	std::vector<std::string>::iterator i;
	for (i = library.begin(); i < library.end(); i++) {
		std::string lib = *i;
		//open the plugin library
		LOG_DEBUG("Open the library by \"dlopen\"");
		void *handle = dlopen(lib.c_str(), RTLD_NOW);// RTLD_LAZY);
		handles.push_back(handle);
		if (!handle) {
			msg.clear(); msg.str("");
			msg << "Library " << dlerror();
			LOG_ERROR(msg.str());
			continue;
		}

		//load "CreatePluginEntryPoint" in the plugin library
		LOG_DEBUG("Get entry point \"CreatePluginEntryPoint\"");
		PluginEntryPointBase* (*entrypoint)();
		entrypoint = (PluginEntryPointBase* (*)())dlsym(handle, "CreatePluginEntryPoint");

		//tell that everything went well
		char *error;
		if ((error = dlerror()) != NULL) {
			msg.clear(); msg.str("");
			msg << "No CreatePluginEntryPoint method found! (" << error << ")";
			LOG_ERROR(msg.str());
			continue;
		}

		//execute PluginEntryPointBase->StartPlugin
		msg.clear(); msg.str("");
		msg << "Run plugin " << lib;
		LOG_INFO(msg.str());
		(*entrypoint)()->StartPlugin(application);
	}

	PrintAllComponents();
}

/**
 * Prints all components (not the controllers).
 */
void LoadComponents::PrintAllComponents() {
	std::ostringstream msg;
	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);

	std::vector<dtGame::GMComponent*> components;
	GA->GetGameManager()->GetAllComponents(components);
	for(unsigned int i=0; i < components.size(); i++) {
		msg.clear(); msg.str("");
		msg << "Component \"";
		msg << components[i]->GetName();
		msg << "\" loaded";
		LOG_INFO(msg.str());
	}
}


} // end of namespace srAlmende
