/*
 * @file objectBase.cpp
 *
 * The Robot3D simulator is created within the European Replicator and Symbrion projects. The
 * Replicator project is funded by the European Commission within the work programme "Cognitive
 * Systems, Interaction, Robotics" under the grant agreement no. 216240. The Symbrion project
 * is funded within the work programme "Future and Emergent Technologies Proactive" under the
 * grant agreement no. 216342. Replicator and Symbrion concerns the design of modular robots
 * in the real world and in 3D simulation.
 *
 * This file is part of the 3D simulator Robot3D which is built on top of the open source game
 * engine Delta3D (copyright (C) 2005, BMH Associates, Inc). This library is free software; you
 * can redistribute it and/or modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version. This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * @author    Lutz Winkler
 * @date      May 6, 2009
 */

// General
#include <sstream>

// Delta3D
#include <dtUtil/log.h>

// Robot3D
#include <srCore/robot/objectBase.h>

namespace srCore {

/************************************************************************************************
 *                      ObjectBase implementation
 ***********************************************************************************************/

/**
 * Constructs an object. This is the subclass for entities that are no bodies, no robots, no
 * sensors, no actuators, but have nevertheless a physical appearance. They can be coupled to a
 * physical body, and have a position and orientation.
 */
ObjectBase::ObjectBase( dtCore::RefPtr<BodyBase> attachedBody, osg::Vec3f translation,
		osg::Vec3f orientation,	std::string name, std::string type, bool active):
		myParameters(),
		name (name),
		type (type),
		relTranslation (translation),
		relOrientation (orientation),
		absTranslation (translation),
		absOrientation (orientation),
		myBody (attachedBody),
		active (active) {

	// Tell that we created another object
	std::ostringstream printmsg;
	printmsg.clear(); printmsg.str("");
	printmsg << "Constructor object \"" << name << "\" of type \"" << type << "\"";
	LOG_DEBUG(printmsg.str());

	// Make it the OpenSceneGraph child of the given body (e.g. a led to the body)
	if (attachedBody != NULL) {
		attachedBody->AddChild(this);
	}

	registerParameters();
}

/**
 * Destructor removes position/orientation parameters and sets physical body to NULL.
 */
ObjectBase::~ObjectBase() {
	myParameters.clear();
	myBody = NULL;
}

/**
 * Set all parameters for the object, the getParameter and setParameter function expect one
 * of the types set here.
 */
void ObjectBase::registerParameters() {

	// Register parameters for relative position and orientation
	ObjectParameter sp;

	sp.valueName = "RelPosition";
	sp.valueType = typeid(osg::Vec3f*).name();
	myParameters.push_back(sp);

	sp.valueName = "RelOrientation";
	sp.valueType = typeid(osg::Vec3f*).name();
	myParameters.push_back(sp);

	sp.valueName = "AbsPosition";
	sp.valueType = typeid(osg::Vec3f*).name();
	myParameters.push_back(sp);

	sp.valueName = "AbsOrientation";
	sp.valueType = typeid(osg::Vec3f*).name();
	myParameters.push_back(sp);
}

/**
 * Print all possible parameters
 */
void ObjectBase::printParameters(dtUtil::Log::LogMessageType type) {
	std::vector<ObjectParameter>::iterator iter;
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Class " << getName() << " accepts parameters: ";
	for (iter = myParameters.begin(); iter != myParameters.end(); iter++) {
		msg << std::endl << (*iter).valueName << " of type " << (*iter).valueType;
	}
	dtUtil::Log::GetInstance().LogMessage(type, __func__, __LINE__,	msg.str().c_str());
}

/**
 * Returns translation of the object.
 */
osg::Vec3f ObjectBase::getTranslation(dtCore::Transformable::CoordSysEnum cs)
{
	dtCore::Transform xform;
	this->GetTransform(xform, cs);
	return xform.GetTranslation();
}

/**
 * Returns orientation of the object. Currently only local orientation.
 */
osg::Vec3f ObjectBase::getOrientation(dtCore::Transformable::CoordSysEnum cs)
{
	dtCore::Transform xform;
	this->GetTransform(xform, cs);
	osg::Vec3 hpr;
	xform.GetRotation(hpr);
	return hpr;
}


/**
 * \brief  Writes the parameter value in parameter
 *
 *		The available parameters can be retrieved by calling getParameterList()
 *
 * \param     		name	   					name of the parameter
 * \param			parameter					pointer where parameter value will be written to
 *
 * Example:
 * parameter is type of float, myHinge is instance of HingeActuator. The following program snippet writes the
 * parameter value of "RelPosition" to parameter. getParameter() checks hereby if the type of parameter is valid.
 *
 * myHinge->getParameter("RelPosition", MAKE_PARAMETER_VALUE(&parameter));
 */
void ObjectBase::getParameter(const std::string &name, ParameterReturnValue parameter) {
	if (name == "RelPosition") {
		// parameter type check
		if (parameter.type != typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}

		// cast void* to Vec3f*
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);

		// return parameter value
		*data = relTranslation;
		parameter.pValue = data;
	}
	else if (name == "RelOrientation") {
		if (parameter.type != typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		*data = relOrientation;
		parameter.pValue = data;
	}
	else if (name == "AbsPosition") {

		// parameter type check
		if (parameter.type != typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}

		// cast void* to Vec3f*
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);

		// return parameter value
		*data = getTranslation(dtCore::Transformable::ABS_CS);
		absTranslation = getTranslation(dtCore::Transformable::ABS_CS);
		*data = absTranslation;
		parameter.pValue = data;
	}

	else if (name == "AbsOrientation") {

		if (parameter.type != typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		*data = getOrientation(dtCore::Transformable::ABS_CS);
		absOrientation = getOrientation(dtCore::Transformable::ABS_CS);
		*data = absOrientation;
		parameter.pValue = data;
	}
	else {
		std::ostringstream msg; msg.clear();
		msg << "Parameter " << name.data() << " is unknown";
		LOG_WARNING(msg.str());
	}
}

/**
 * Return new ParameterReturnValue. In the absolute case the translation and the
 * orientation is first acquired by a preceding call. In the relative case, the
 * stored fields are used as they are.
 */
ParameterReturnValue * ObjectBase::getParameter(const std::string &name) {
	if (name.compare("RelPosition") == 0) {
		osg::Vec3f* data = new osg::Vec3f(relTranslation);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("RelOrientation") == 0) {
		osg::Vec3f* data = new osg::Vec3f(relOrientation);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("AbsPosition") == 0) {
		absTranslation = getTranslation(dtCore::Transformable::ABS_CS);
		osg::Vec3f* data = new osg::Vec3f(absTranslation);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("AbsOrientation") == 0) {
		absOrientation = getOrientation(dtCore::Transformable::ABS_CS);
		osg::Vec3f* data = new osg::Vec3f(absOrientation);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else {
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Parameter " << name.data() << " is unknown";
		LOG_WARNING(msg.str());
		printParameters(dtUtil::Log::LOG_WARNING);
	}
	return NULL;
}

/**
 * \brief  Sets a parameter
 *
 *		See also getParameter()
 *
 */
void ObjectBase::setParameter(const std::string &name, ParameterReturnValue  parameter) {

	if (name == "RelPosition") {

		// parameter type check
		if (parameter.type != typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}

		// cast void* to Vec3f*
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);

		// set parameter
		relTranslation = *data;
		dtCore::Transform xform;
		GetTransform(xform);

		dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(),
				relOrientation.x(), relOrientation.y(), relOrientation.z());

		SetTransform( offset, dtCore::Transformable::REL_CS );
	}
	else if (name == "RelOrientation") {

		if (parameter.type!=typeid(osg::Vec3f*).name()) {
			LOG_ERROR("Parameter type needs to be osg::Vec3f*. Casting could lead to errors.");
		}
		osg::Vec3f* data = reinterpret_cast<osg::Vec3f*> (parameter.pValue);
		relOrientation = *data;

		dtCore::Transform xform;
		GetTransform(xform);

		dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(), relOrientation.x(),
				relOrientation.y(), relOrientation.z());

		SetTransform( offset, dtCore::Transformable::REL_CS );
	}
	else {
		std::ostringstream msg; msg.clear();
		msg << "Parameter " << name.data() << " is unknown";
		LOG_WARNING(msg.str());
	}

}


}
