/*
 * Robot3D Physical realistic 3D simulator for robots
 *
 * @license GNU Lesser General Public License
 *
 * @author Lutz Winkler, Anne C. van Rossum
 */

// General files
#include <stdio.h>

// Open scene graph
#include <osgDB/WriteFile>

// Delta3D library
#include <dtDAL/enginepropertytypes.h>
#include <dtDAL/actorproxyicon.h>
#include <dtDAL/groupactorproperty.h>
#include <dtCore/loadable.h>
#include <dtCore/isector.h>
#include <dtCore/scene.h>
#include <dtGame/gamemanager.h>
#include <dtGame/actorupdatemessage.h>
#include <dtGame/basemessages.h>
#include <dtABC/application.h>
#include <dtCore/camera.h>
#include <dtCore/keyboard.h>
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/body/bodyBase.h>
#include <srCore/robot/robotActorBase.h>

/**
 * It is very important not to have this class depending on srFactory, srSensor, or srActuator. The srFactory class needs
 * to be able to link to srCore, so including a reverse dependency would lead to a cycle.
 *
 * TODO:  This class contains more and more methods about "what can be done to the thing". Such methods are better
 * implemented by the Visitor pattern. Visitors can very naturally store cross-element states, and are perfect to implement
 * for example connections between robots. See for a good document:
 *   http://www.drdobbs.com/184410252;jsessionid=I4DPLASDVO2RFQE1GHRSKH4ATMY32JVN
 */

//TODO: XML-file that describes the robot and using a parser the robot will be generated during runtime
//TODO: An in size scalable robot, so that length unit can be variable  -->  this way different sizes can be tested according to stability of the simulator
//TODO: mechanism to move a whole organism manually. Currently only translation of single robots is possible
//TODO: simulating energy consumption. Each srCore::Object needs therefor a model for its energy consumption and a new object needs to be implemented that simulates the accumulator
//TODO: different message buffers the controller have access to + sending messages needs to be transfered from the controller back to the robot again (underlying mechanism should not be visible to the controller)

namespace srCore {

RobotActorBase::RobotActorBase(dtGame::GameActorProxy &proxy): dtGame::GameActor(proxy),
		myPos(osg::Vec3(0,0,0)),
		myRot(osg::Vec3(0,0,0)),
		globalCommID(),
		globalMsg(std::string()),
		selectionState(false),
		myID(-1),
		connectorStates(0),
		osgGroup(),
		mySensors(),
		myActuators(),
		myBodies(),
		myConnectors(),
		myObjects(),
		sensorData(),
		jointAngles(),
		initPos(osg::Vec3(0,0,0)),
		initRot(osg::Vec3(0,0,0))
{
	globalCommID = const_cast<dtCore::UniqueId* >(&(this->GetUniqueId()));

	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Construct robot";
	LOG_INFO(msg.str());
}

/**
 * Removes a robot
 * @todo Find another way to deallocate the screw drive. This class shouldn't know about specific
 * actuator implementations.
 */
RobotActorBase::~RobotActorBase() {

	setSelectionState(false);
	renderCollisionGeometry(false);

	myActuators.clear();

	for(unsigned int i=0; i<myBodies.size(); i++)
	{
		while(myBodies[i]->GetNumChildren()>0) {

			//dtABC::Application::GetInstance(0)->GetScene()->RemoveDrawable(myBodies[i]->GetChild(0));
			myBodies[i]->RemoveChild(myBodies[i]->GetChild(0));
		}
		dtABC::Application::GetInstance(0)->GetScene()->RemoveDrawable(myBodies[i]);
	}

	// Clear all vectors
	myBodies.clear();
	myConnectors.clear();
	myObjects.clear();

	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Removed robot " << getID();
	LOG_INFO(msg.str());
}

/**
 * Every body has its own collisionCategory bit. No body is allowed to collide with another body.
 * @todo Set collision bits of wheels somewhere else...
 */
void RobotActorBase::initialize() {
	// Debug information if log level high enough
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Initializing robot with " << sensorData.size() << " new sensors.";
	LOG_DEBUG(msg.str()); msg.clear(); msg.str("");

	unsigned int bitCheck = 1;
	for (unsigned int i=0; i<myConnectors.size(); i++) {
		if (connectorStates & bitCheck) {
			myConnectors[i]->setState(ConnectorBase::CHECK_CONNECTION);
		}
		bitCheck <<= 1;
	}

	//set body collision bit
	for (unsigned int i=0; i<myBodies.size(); i++) {

		BodyBase* body = myBodies[i];

		unsigned long int bit = ((myID+1) << 10) | i;

		if ((body->GetCollisionCategoryBits() != 4) || body->GetCollisionCategoryBits() != 2) {
			body->setCollisionBits(bit);
		}
	}
}

void RobotActorBase::setID(long int id)
{
	myID = id;
}

void RobotActorBase::setSelectionState(bool selState)
{
	selectionState = selState;
}

void RobotActorBase::setConnectorStates(int state)
{
	connectorStates = state;
}

void RobotActorBase::addBody(BodyBase* body) {

	myBodies.push_back(body);


}

void RobotActorBase::addActuator(ActuatorBase* actuator) {

	myActuators.push_back(actuator);
	actuator->registerActuator();

}

void RobotActorBase::addConnector(ConnectorBase* connector) {

	myConnectors.push_back(connector);

}

void RobotActorBase::addSensor(SensorBase* sensor){

	mySensors.push_back(sensor);

}

/// Equality to another robot is defined by id
bool RobotActorBase::operator==(const RobotActorBase& other) const {
	return (getID() == other.getID());
}

/**
 * Prints the indices of the robots to which connectors that are connectors are connected.
 */
void RobotActorBase::PrintConnectors() {
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Connectors on robot: " << getID() << " ";
	for (unsigned char i=0; i < getConnectors().size(); i++) {
		RobotActorBase *r = getConnectors()[i]->getConnectedRobot();
		if (r == NULL) {
			msg << "[X]";
		} else {
			msg << "[" << r->getID() << "]";
		}
	}
	LOG_INFO(msg.str());
}

/**
 * Prints the states of the connectors,
 */
void RobotActorBase::PrintConnectorStates() {
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Connectors on robot " << getID() << " connected to: ";
	for (unsigned char i=0; i < getConnectors().size(); i++) {
		RobotActorBase *r = getConnectors()[i]->getConnectedRobot();
		if (r == NULL) {
			msg << "[-]";
		} else {
			msg << "[" << r->getID() << "]";
		}
	}
	msg << " and in state: ";
	for (unsigned char i=0; i < getConnectors().size(); i++) {
		ConnectorBase::ConnectorState state = getConnectors()[i]->getState();
		switch(state) {
		case ConnectorBase::CHECK_CONNECTION: msg << "[?]"; break;
		case ConnectorBase::CONNECTED: msg << "[=]"; break;
		case ConnectorBase::WANT_TO_CONNECT: msg << "[w]"; break;
		case ConnectorBase::DISCONNECTED: msg << "[X]"; break;
		case ConnectorBase::CONNECTING: msg << "[~]"; break;
		}
	}
	LOG_INFO(msg.str());
}
/**
 * Only if the setTranslation and setOrientation routines work properly it is possible to get
 * rid of this routine. For now, it is impossible to have a GMComponent placing and especially
 * rotating a robot somewhere without having the physics engine in its way.
 */
void RobotActorBase::EnableDynamics(bool enable = true) {
	std::vector<dtCore::RefPtr<BodyBase> >::iterator i;
	if (!myBodies.empty()) {
		for (i = myBodies.begin(); i != myBodies.end(); i++) {
			(*i)->EnableDynamics(enable);
		}
	}
	dtCore::Physical::EnableDynamics(enable);
}

//! only use this function when robot is not connected to another one.
void RobotActorBase::setTranslation(const osg::Vec3 &pos)
{
	osg::Vec3 delta = pos - myPos;								//delta distance robot needs to move to
	myPos = pos;

	// for all bodies in the robot, move along delta.
	if (!myBodies.empty())
		for (unsigned int i=0; i<myBodies.size(); i++)
			myBodies[i]->setTranslation(myBodies[i]->getTranslation() + delta);

}

/**
 * Set the robot to an absolute, global position. When starting the simulator, it is possible
 * to use the relative RobotActorBase::setTranslation routine. However, after running a robot for
 * a while it will be quite artificial to use a relative translation operation to set it to an
 * absolute position. It is also important to know that this command is not directly executed,
 * only after the robot gets a tick it will be executed.
 *
 * Todo: We need to use Position for absolute positioning, and Translation for relative positioning
 * consistently.
 */
void RobotActorBase::setPosition(const osg::Vec3 &pos) {
	myPos = pos;
	if (!myBodies.empty())
		for (unsigned int i=0; i<myBodies.size(); i++)
			myBodies[i]->setTranslation(pos);
}

/**
 * Returns the current "rotation" of a robot.
 */
osg::Vec3f RobotActorBase::getRotation() const {

	dtCore::Transform xform;
	osg::Vec3 hpr = osg::Vec3(0,0,0);

	dtCore::Transform initTransform;
	initTransform.MakeIdentity();
	initTransform.SetRotation(initRot);

	osg::Matrix currentMat, initMat;

	initTransform.Get(initMat);

	if (!myBodies.empty()) {

		myBodies[0]->GetTransform(xform);
		xform.Get(currentMat);

		osg::Matrix currentMatInverse, actualMat, initMatInverse;
		currentMatInverse.invert(currentMat);
		initMatInverse.invert(initMat);

		actualMat = initMatInverse * currentMat;
		//actualMat = currentMatInverse * initMat;

		xform.Set(actualMat);
		xform.GetRotation(hpr);

	}

	return hpr;

}

/**
 * This function rotates a robot and all the bodies that are part of the robot, like
 * the body itself, the wheels, etc.
 * Important! Use this function only when a robot is not connected to another one.
 */
void RobotActorBase::setRotation(const osg::Vec3 &rot) {

	//	osg::Vec3 dRot = rot - myRot;
	myRot = rot;

	//	std::cout << "blublub: " << rot << std::endl;
	dtCore::Transform newTransform;
	newTransform.MakeIdentity();
	newTransform.SetRotation(rot);

	dtCore::Transform oldTransformT;
	oldTransformT.MakeIdentity();
	oldTransformT.SetRotation(getRotation());//myRot);

	dtCore::Transform initTransform;
	initTransform.MakeIdentity();
	initTransform.SetRotation(initRot);//myRot);

	osg::Matrix newMat, currentRot, rotBackMat, initMat;
	oldTransformT.Get(currentRot); newTransform.Get(newMat); initTransform.Get(initMat);

	rotBackMat.invert(currentRot);

	osg::Matrix rotate;
	rotate = newMat * rotBackMat;

	// rotate all bodies in the robot
	if (!myBodies.empty()) {

		dtCore::Transform xform;
		myBodies[0]->GetTransform(xform, dtCore::Transformable::ABS_CS);
		osg::Vec3 robotPos = xform.GetTranslation();

		for (unsigned int i=0; i<myBodies.size(); i++) {

			dtCore::Transform xform;
			myBodies[i]->GetTransform(xform, dtCore::Transformable::ABS_CS);
			osg::Vec3 pos = xform.GetTranslation();
			xform.SetTranslation(pos - robotPos);

			osg::Matrix bodyMatrix;
			xform.Get(bodyMatrix);

			bodyMatrix = bodyMatrix * rotate;
			xform.Set(bodyMatrix);
			xform.SetTranslation(xform.GetTranslation() + robotPos);

			myBodies[i]->SetTransform(xform, dtCore::Transformable::ABS_CS);

		}
	}

	myRot = rot;
}

/**
 * Iterates over all connectors and calls getConnectedRobots for the connected robots too. It is hence
 * a recursive function. The recursion is broken by the for loop over the given connectedRobots vector.
 * A robot is considered connected if the state is unequal to DISCONNECTED.
 */
void RobotActorBase::getConnectedRobots (std::vector<RobotActorBase* > &connectedRobots) {

	for (unsigned int i=0;i<connectedRobots.size();i++)
		if (connectedRobots[i]->GetUniqueId() == this->GetUniqueId()) return;	//already visited this node

	connectedRobots.push_back(this);

	for (unsigned int i=0; i<myConnectors.size(); i++) {

		if (myConnectors[i]->getState() != ConnectorBase::DISCONNECTED) {
			RobotActorBase* theOtherRobot = myConnectors[i]->getConnectedRobot();
			if (theOtherRobot)
				theOtherRobot->getConnectedRobots(connectedRobots);
		}
	}
	return;
}

/**
 * Robots that are connected to the given robot. The robot itself will not be added to its neighbors.
 */
void RobotActorBase::getNeighborRobots (std::vector<RobotActorBase* > &neighborRobots) {
	neighborRobots.clear();
	for (unsigned int i=0; i<myConnectors.size(); i++) {
		if (myConnectors[i]->getState() != ConnectorBase::DISCONNECTED) {
			RobotActorBase* theOtherRobot = myConnectors[i]->getConnectedRobot();
			if (theOtherRobot)
				neighborRobots.push_back(theOtherRobot);
		}
	}
}

/**
 * Fills given vector with all bodies that together make up the robot.
 */
void RobotActorBase::getConnectedBodies (std::vector<BodyBase* > &connectedBodies, BodyBase* bodyNode) {

	for (unsigned int i=0;i<connectedBodies.size();i++)
		if (connectedBodies[i]->GetUniqueId() == bodyNode->GetUniqueId()) return;	//already visited this node

	connectedBodies.push_back(bodyNode);
	std::vector<ConnectorBase*> connectors;

	for (unsigned int i=0; i< bodyNode->GetNumChildren(); i++)
		if (dynamic_cast<ConnectorBase*>(bodyNode->GetChild(i)))
			connectors.push_back(dynamic_cast<ConnectorBase*>(bodyNode->GetChild(i)));

	for (unsigned int i=0; i<connectors.size(); i++) {

		if (connectors[i]->getState() != ConnectorBase::DISCONNECTED) {

			RobotActorBase* theOtherRobot = connectors[i]->getConnectedRobot();
			if (theOtherRobot)
				theOtherRobot->getConnectedBodies(
						connectedBodies, connectors[i]->getOtherConnector()->getSRBody());
		}
	}
	return;
}

// register this robot and all other robots connected to this robot to the same message type
void RobotActorBase::registerToCommBus(dtCore::UniqueId* id) {

	globalCommID = id;

}

// register this robot and all other robots connected to this robot to the same message type
void RobotActorBase::unregisterFromCommBus(){
	globalCommID = new dtCore::UniqueId();
}



// updating behaviour of sensors, connectors, actuators
// TODO: Updating behaviour of sensors, connectors and actuators every tick is not necessary.
void RobotActorBase::TickLocal(const dtGame::Message &tickMessage)
{
	//#pragma omp parallel <-- does not work

	const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(tickMessage);
	float deltaSimTime = tick.GetDeltaSimTime();

	if (!mySensors.empty())
		for (unsigned int i=0; i<mySensors.size(); i++)
			mySensors[i]->update(deltaSimTime);

	if (!myConnectors.empty())
		for (unsigned int i=0; i<myConnectors.size(); i++)
			myConnectors[i]->update(deltaSimTime);

	if (!myActuators.empty())
		for (unsigned int i=0; i<myActuators.size(); i++)
			myActuators[i]->update(deltaSimTime);

	if (!myObjects.empty())
		for (unsigned int i=0; i<myObjects.size(); i++)
			myObjects[i]->update(deltaSimTime);

}

void RobotActorBase::TickRemote(const dtGame::Message &tickMessage)
{
	//const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(tickMessage);
	//float deltaSimTime = tick.GetDeltaSimTime();

	// do NOT recompute velocity and turn rate since we don't own this tank!

	//MoveTheRobot(deltaSimTime);
}

void RobotActorBase::ProcessMessage(const dtGame::Message &message)
{

	if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT) {
		//
	}

	//	if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT)
	//	{
	//		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);
	//		if (eventMsg.GetGameEvent()->GetName() == DESTROY_ACTOR)
	//		{
	//			for(int i=0; i<mySensors.size(); i++)
	//				delete mySensors[i];
	//		}
	//	}
}

void RobotActorBase::sendBusMessage(const std::string& message){

	//	dtCore::RefPtr<CommMessage> commMsg = new CommMessage();
	//	GetGameActorProxy().GetGameManager()->GetMessageFactory().CreateMessage(SimulationMessageType::COMM_BUS, commMsg);
	//	commMsg->setBusID(globalCommID);
	//	commMsg->setMessage(message);
	//	GetGameActorProxy().GetGameManager()->SendMessage(*commMsg);

}

/**
 * Initialising basic robot properties. Its position and rotation. Moreover, for each sensor
 * the setup() function is called. After all that, the overloaded Delta3D function
 * GameActor::OnEnteredWorld() is evoked.
 */
void RobotActorBase::OnEnteredWorld()
{
	dtCore::Transform xform;
	osg::Vec3 hpr = osg::Vec3(0,0,0);
	osg::Vec3 pos = osg::Vec3(0,0,0);

	if (!myBodies.empty()) {
		myBodies[0]->GetTransform(xform);
		xform.GetRotation(hpr);
		xform.GetTranslation(pos);
	}
	else LOG_WARNING("Initializing robot with no bodies !!!");

	initRot = hpr;
	initPos = pos;

	//setting position and orientation
	float x		= GetGameActorProxy().GetTranslation()[0];
	float y 		= GetGameActorProxy().GetTranslation()[1];
	float z 		= GetGameActorProxy().GetTranslation()[2];
	float heading 	= GetGameActorProxy().GetRotation()[2];
	float roll		= GetGameActorProxy().GetRotation()[0];
	float pitch 	= GetGameActorProxy().GetRotation()[1];


	setTranslation(osg::Vec3(x,y,z));
	setRotation(osg::Vec3(heading, roll, pitch));

	RobotActorBase::initialize();

	std::vector<dtCore::RefPtr<SensorBase> > allSensors = this->getSensors();
	if (!allSensors.empty())
		for (unsigned int j=0; j<allSensors.size(); j++) {
			if (allSensors[j] == NULL) {
				std::ostringstream msg;	msg.clear(); msg.str("");
				msg << "Sensor " << j << " does not exist! Does the sensor factory know this one?";
				LOG_ERROR(msg.str());
				continue;
			}
			allSensors[j]->setup();
		}

	GameActor::OnEnteredWorld();
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

//todo: program RobotActorProxy so that it can be used within STAGE
RobotActorBaseProxy::RobotActorBaseProxy()
{
	SetClassName("Robot");
}

///////////////////////////////////////////////////////////////////////////////
void RobotActorBaseProxy::BuildPropertyMap()
{
	const std::string GROUP = "Robot";

	dtGame::GameActorProxy::BuildPropertyMap();
	RobotActorBase *actor = static_cast<RobotActorBase*>(&GetGameActor());

	// ID property
	AddProperty(new dtDAL::LongActorProperty("ID","ID",
			dtDAL::LongActorProperty::SetFuncType(actor, &RobotActorBase::setID),
			dtDAL::LongActorProperty::GetFuncType(actor, &RobotActorBase::getID),
			"Sets/gets the robot's ID.",GROUP));

	// position property
	AddProperty(new dtDAL::Vec3ActorProperty("Pos", "Pos",
			dtDAL::Vec3ActorProperty::SetFuncType(actor, &RobotActorBase::setTranslation),
			dtDAL::Vec3ActorProperty::GetFuncType(actor, &RobotActorBase::getTranslation),
			"gets the robot's position, setPos is an empty function.",GROUP));

	// rotation property
	AddProperty(new dtDAL::Vec3ActorProperty("Rot", "Rot",
			dtDAL::Vec3ActorProperty::SetFuncType(actor, &RobotActorBase::setRotation),
			dtDAL::Vec3ActorProperty::GetFuncType(actor, &RobotActorBase::getRotation),
			"gets the robot's rotation, setRot is an empty function.",GROUP));


	//	dtDAL::GroupActorProperty* gap = 		new dtDAL::GroupActorProperty("connectors","connectors",
	//			dtDAL::MakeFunctor(actor,&RobotActor::setConnectors),
	//			dtDAL::MakeFunctorRet(actor,&RobotActor::getConnectors),
	//			"gets rotation and position of the robots connectors.",GROUP);
	//
	//
	//	AddProperty( gap );


}

///////////////////////////////////////////////////////////////////////////////
RobotActorBaseProxy::~RobotActorBaseProxy()
{

}

///////////////////////////////////////////////////////////////////////////////
void RobotActorBaseProxy::CreateActor()
{
	SetActor(*new RobotActorBase(*this));
}

///////////////////////////////////////////////////////////////////////////////
void RobotActorBaseProxy::OnEnteredWorld()
{
	RegisterForMessages(dtGame::MessageType::INFO_GAME_EVENT);
	RegisterForMessages(SimulationMessageType::COMM_BUS);

	// Register an invokable for tick messages. Local or Remote only, not both!
	if (IsRemote())
		RegisterForMessages(dtGame::MessageType::TICK_REMOTE, dtGame::GameActorProxy::TICK_REMOTE_INVOKABLE);
	else
		RegisterForMessages(dtGame::MessageType::TICK_LOCAL, dtGame::GameActorProxy::TICK_LOCAL_INVOKABLE);

	dtGame::GameActorProxy::OnEnteredWorld();
}


void RobotActorBaseProxy::setID(long int id)
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setID(id);
}

long int RobotActorBaseProxy::getID()
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	return actor.getID();
}


void RobotActorBaseProxy::setConnectorStates(int state)
{
	//	printf("\n\n set connector state to %d \n\n", state);
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setConnectorStates(state);
}

int RobotActorBaseProxy::getConnectorStates()
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	return actor.getConnectorStates();
}


void RobotActorBaseProxy::setRotation(osg::Vec3 rot)
{

	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setRotation(rot);
}

osg::Vec3 RobotActorBaseProxy::getRotation()
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	return actor.getRotation();
}

void RobotActorBaseProxy::setTranslation(osg::Vec3 pos)
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setTranslation(pos);
}

osg::Vec3 RobotActorBaseProxy::getTranslation()
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	return actor.getTranslation();
}
void RobotActorBaseProxy::setInitialJointAngles(std::vector<float> jointAngles)
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setInitialJointAngles(jointAngles);
}

void RobotActorBaseProxy::setSensors(std::vector<SensorInitData> sensors)
{
	RobotActorBase &actor = static_cast<RobotActorBase&>(GetGameActor());
	actor.setSensorData(sensors);
}

}

