/**
 * @file RobotCreator.cpp
 * @brief 
 *
 * This file RobotCreator.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 * but the software is not allowed to be used for military use, within the bio-industry
 * or for animal experimentation.
 *
 * @author  Anne C. van Rossum
 * @date    Jul 13, 2010
 * @project Replicator FP7
 * @company Almende B.V.
 * @case	
 */


// General files
#include <sstream>

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/robotCreator.h>
#include <srCore/robot/robotActorBase.h>

namespace srCore {

/* **************************************************************************************
 * Implementation of RobotCreator
 * **************************************************************************************/

RobotCreator::RobotCreator() {

}

RobotCreator::~RobotCreator() {

}

/*
 * Adds a robot to the game manager. The robot type needs to be registered before. If
 * something fails we return false;
 */
bool RobotCreator::addRobotToGM(int id, const RobotInitData &parameters, float height) {

	std::ostringstream msg;	msg.clear(); msg.str("");
	msg << "Add robot \"" << id << "\" named \"" << parameters.name;
	msg << "\" of type \"" << parameters.type << "\" to the game manager";
	LOG_INFO(msg.str());

	dtCore::RefPtr<RobotActorBaseProxy> mRobotActorProxy;
	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);

	std::string robotType = std::string(parameters.type);
	// Create the robot
	if (std::string(parameters.type).find("RobotKIT2") != std::string::npos) {
		GA->GetGameManager()->CreateActor("RobotKIT2", "RobotKIT2", mRobotActorProxy);
//		GA->GetGameManager()->CreateActor(*RobotKITType.get(), mRobotActorProxy);
	} else if (robotType.compare("RobotKIT") == 0) {
		GA->GetGameManager()->CreateActor("RobotKIT", "RobotKIT", mRobotActorProxy);
//		GA->GetGameManager()->CreateActor(*RobotKITType.get(), mRobotActorProxy);
	} else {
		LOG_ERROR("That robot type we do not know!");
		return false;
	}
	assert (mRobotActorProxy != NULL);

	// Set identification
	mRobotActorProxy->setID(id);
	mRobotActorProxy->setConnectorStates(parameters.connectorStates);
	mRobotActorProxy->SetName(parameters.name);

	// Set position and rotation
	osg::Vec3f pos = parameters.pos + osg::Vec3f(0.0, 0.0, height);
	msg.clear(); msg.str("");
	msg << "Position at: [" << pos.x() << "," << pos.y() << "]";
	LOG_INFO(msg.str());

	mRobotActorProxy->SetTranslation(pos);
	osg::Matrix matrix = osg::Matrix(parameters.quat);
	mRobotActorProxy->SetRotationFromMatrix(matrix);

	// Initialise joint angles and sensor values
	mRobotActorProxy->setInitialJointAngles(parameters.jointAngles);
	mRobotActorProxy->setSensors(parameters.sensors);

	// Debug information
	msg.clear(); msg.str("");
	msg << "Joint angles are: " << parameters.jointAngles[0];
	LOG_DEBUG(msg.str());

	// Finally add actor to the game manager
	GA->GetGameManager()->AddActor(*mRobotActorProxy, false, true);

	// Success
	return true;
}

} // end of namespace srCore
