/*
 * cameraSensorBase.cpp
 *
 *  Created on: May 5, 2009
 *      Author: winkler
 */

// Delta3D libraries
#include <dtUtil/log.h>

// Robot3D libraries
#include <srCore/sensor/cameraSensorBase.h>
#include <srCore/export.h>

using namespace srCore;

/**
 * The default constructor. Additional fields need to be set in the setup() routine,
 * namely the window, the view, the sensor and the OSG camera.
 */
CameraSensorBase::CameraSensorBase(dtCore::RefPtr<BodyBase> attachedTo, osg::Vec3f pos,
		osg::Vec3f orient, std::string name, std::string type, glType cameraType, bool active): SensorBase(
				attachedTo, pos, orient, name, type, cameraType, active),
				minDistance		(1),
				maxDistance		(1000),
				frustumLeft		(-1),
				frustumRight	(1),
				frustumUp		(-1),
				frustumDown		(1),
				flareAngle		(0.5),
				clearColor		(osg::Vec4(0,0,0,0)),
				cameraType		(cameraType),
				sensor			(NULL),
				resolution		(osg::Vec2(16,16)),
				win				(NULL),
				view			(NULL),
				mOsgCamera		(NULL),
				sensorImage 	(new osg::Image),
				mScreenShotTaker(new CameraScreenShotCallback(cameraType)){
	mScreenShotTaker->requestData();
}

/**
 * Deallocates this camera and removes it also from the body it is attached to. Each camera
 * does have its own view defined, which is also removed from the application.
 * @return
 */
CameraSensorBase::~CameraSensorBase() {
	std::ostringstream msg;	msg.clear(); msg.str("");
	msg << "Remove camera from \"" << myBody->GetName() << "\"";
	LOG_INFO(msg.str());

	this->myBody->RemoveChild(sensor.get());

	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
	GA->RemoveView(*view);

	view = NULL;
	win  = NULL;
};


osg::ref_ptr<osg::Image> CameraSensorBase::getRawData() {

	sensorImage = mScreenShotTaker->getOSGImage();
	return sensorImage;

}

void CameraSensorBase::requestData() {

	mScreenShotTaker->requestData();

}

void CameraSensorBase::update(float deltaSimTme) {

	if (active) {
		requestData();
		if (isActiveSensor) getNewSensorValue();
	}

}

/**
 * The camera sensor base class accepts a set of parameters, under which:
 * <ul>
 * <li>MinDistance
 * <li>FlareAngle
 * <li>MaxDistance
 * <li>Frustum
 * <li>Resolution
 * <li>BackgroundColor
 * </ul>
 */
void CameraSensorBase::setParameter(const std::string &name, ParameterReturnValue parameter) {
	if (name == "MinDistance") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(float*).name())) return;

		// cast void* to float*
		float* data = reinterpret_cast<float*> (parameter.pValue);

		// set parameter ...
		minDistance = *data;

		// ... and adopt other parameter, as they might have changed
		float frustum = tan(this->flareAngle/2) * this->minDistance;

		if (sensor.get() && (frustum > 0.02 / this->minDistance)) {
			this->frustumDown = frustum;
			this->frustumUp = frustum;
			this->frustumRight = frustum;
			this->frustumLeft = frustum;

			dynamic_cast<dtCore::Camera*>(sensor.get())->SetFrustum (
					-frustumLeft, frustumRight, -frustumUp, frustumDown, minDistance, maxDistance);
		}

	}
	else if (name == "FlareAngle") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(float*).name())) return;

		float* data = reinterpret_cast<float*> (parameter.pValue);
		flareAngle = *data;

		float frustum = tan(this->flareAngle/2) * this->minDistance;
		if (sensor.get()) {

			this->frustumDown = frustum;
			this->frustumUp = frustum;
			this->frustumRight = frustum;
			this->frustumLeft = frustum;

			dynamic_cast<dtCore::Camera*>(sensor.get())->SetFrustum (
					-frustumLeft, frustumRight, -frustumUp, frustumDown, minDistance, maxDistance);
		}

	}
	else if (name == "MaxDistance") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(float*).name())) return;
		float* data = reinterpret_cast<float*> (parameter.pValue);
		maxDistance = *data;
		if (sensor.get()) dynamic_cast<dtCore::Camera*>(sensor.get())->SetFrustum (
				-frustumLeft, frustumRight, -frustumUp, frustumDown, minDistance, maxDistance);

	}
	else if (name == "Frustum") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(osg::Vec4f*).name())) return;
		osg::Vec4f* data = reinterpret_cast<osg::Vec4f*> (parameter.pValue);
		frustumLeft 	= data->x();
		frustumRight 	= data->y();
		frustumUp 		= data->z();
		frustumDown		= data->w();

		if (sensor.get()) {
			dynamic_cast<dtCore::Camera*>(sensor.get())->SetFrustum (-frustumLeft, frustumRight, -frustumUp, frustumDown, minDistance, maxDistance);
			view->SetCamera(dynamic_cast<dtCore::Camera*>(sensor.get()));
		}
	}
	else if (name == "Resolution") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(osg::Vec2d*).name())) return;
		osg::Vec2d* data = reinterpret_cast<osg::Vec2d*> (parameter.pValue);
		resolution = *data;
		if (sensor.get()) {

			dtCore::DeltaWin::PositionSize posSize = win->GetPosition();
			posSize.mHeight = static_cast<int>(resolution.y());
			posSize.mWidth = static_cast<int>(resolution.x());
			win->SetPosition (posSize);
		}

	}
	else if (name == "BackgroundColor") {
		if (!PRVUtils::Instance()->isCorrectParameterType(parameter, typeid(osg::Vec4*).name())) return;
		osg::Vec4* data = reinterpret_cast<osg::Vec4*> (parameter.pValue);
		clearColor = *data;
		if (sensor.get()) dynamic_cast<dtCore::Camera*>(sensor.get())->SetClearColor(clearColor);

		view->SetCamera(dynamic_cast<dtCore::Camera*>(sensor.get()));
	}
	else SensorBase::setParameter(name, parameter);

	dtCore::Transform offset;
	GetTransform(offset);
	sensor->SetTransform( offset, dtCore::Transformable::REL_CS );

}


/**
 * \brief  Writes the parameter value in parameter
 *
 *		The available parameters can be retrieved by calling getParameterList()
 *		See also: ObjectBase::getParameter()
 *
 *
 */
void CameraSensorBase::getParameter(const std::string &name, ParameterReturnValue parameter) {
	if (name == "MinDistance") {
		// check parameter type
		if (parameter.type != typeid(float*).name()) printf("Parameter type needs to be float*. Casting could lead to errors.  \n");
		parameter.pValue = &minDistance;
	}
	else if (name == "FlareAngle") {
		if (parameter.type != typeid(float*).name()) printf("Parameter type needs to be float*. Casting could lead to errors.  \n");
		parameter.pValue = &flareAngle;
	}
	else if (name == "MaxDistance") {
		if (parameter.type!=typeid(float*).name()) printf("Parameter type needs to be float*. Casting could lead to errors.  \n");
		parameter.pValue = &maxDistance;
	}
	else if (name == "Frustum") {
		if (parameter.type!=typeid(osg::Vec4f*).name()) printf("Parameter type needs to be osg::Vec4f*. Casting could lead to errors.  \n");
		osg::Vec4f* data = reinterpret_cast<osg::Vec4f*> (parameter.pValue);
		*data = osg::Vec4f(frustumLeft, frustumRight, frustumUp, frustumDown);
		parameter.pValue = data;
	}
	else if (name == "Resolution") {
		if (parameter.type!=typeid(osg::Vec2*).name()) printf("Parameter type needs to be osg::Vec2*. Casting could lead to errors.  \n");
		osg::Vec2* data = reinterpret_cast<osg::Vec2*> (parameter.pValue);
		*data = resolution;
		parameter.pValue = data;
	}
	else if (name == "BackgroundColor") {
		if (parameter.type!=typeid(osg::Vec4*).name()) printf("Parameter type needs to be osg::Vec4*. Casting could lead to errors.  \n");
		osg::Vec4* data = reinterpret_cast<osg::Vec4*> (parameter.pValue);
		*data = clearColor;
		parameter.pValue = data;
	}
	else SensorBase::getParameter(name, parameter);
}

ParameterReturnValue* CameraSensorBase::getParameter(const std::string &name) {
	//! First retrieve from sensor base, if no result check one of the parameters here.
	ParameterReturnValue* result = SensorBase::getParameter(name);
	if (result != NULL) return result;

	if (name.compare("MinDistance") == 0) {
		float* data = new float(minDistance);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("MaxDistance") == 0) {
		float* data = new float(maxDistance);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("FlareAngle") == 0) {
		float* data = new float(flareAngle);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("Frustum") == 0) {
		osg::Vec4f* data = new osg::Vec4f(frustumLeft, frustumRight, frustumUp, frustumDown);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("Resolution") == 0) {
		osg::Vec2* data = new osg::Vec2(resolution);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else if (name.compare("BackgroundColor") == 0) {
		osg::Vec4* data = new osg::Vec4(clearColor);
		ParameterReturnValue *prv = MAKE_PRV(data);
		return prv;
	} else {
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Parameter " << name.data() << " is unknown";
		LOG_WARNING(msg.str());
		printParameters(dtUtil::Log::LOG_WARNING);
	}
	return NULL;

}

/**
 * Sets up the sensor that is built on top of the dtCore::Camera class. A new view is created
 * and a new scene too. The latter is only done outside USE_SEPARATE_GUI mode. When the simulator
 * is in this mode, the cameras can just use the main scene (there will be no GUI elements in
 * them). A new view is by the way necessary for each new camera.
 */
void CameraSensorBase::setup() {
	LOG_INFO("Setup camera by creating a new dtCore::Camera and corresponding view");
	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
	view = new dtCore::View(name);

#ifdef USE_SEPARATE_GUI
	//populate the view with the robots
	view->SetScene(GA->GetScene());
#else
	dtCore::Scene* scene = new dtCore::Scene;
	view->SetScene(scene);
#endif

	//add the view to the application, so it is actually rendered
	GA->AddView(*view);

	//setup the camera
	osg::ref_ptr<dtCore::Camera> mCamera = new dtCore::Camera(name);

	mCamera->GetOSGCamera()->setCullMask(0xFFFFFFFE & ~(SimulationUtils::Instance()->getCollisionNodeMask()));
	mCamera->SetWindow(GA->GetWindow());
	mCamera->GetOSGCamera()->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
	mCamera->GetOSGCamera()->setClearMask(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	mCamera->GetOSGCamera()->setViewport(0, 0, (int) resolution.x(), (int) resolution.y());
	mCamera->GetOSGCamera()->setClearDepth (10);
	mCamera->SetNearFarCullingMode (dtCore::Camera::NO_AUTO_NEAR_FAR);
	mCamera->SetFrustum(-frustumLeft, frustumRight, -frustumUp, frustumDown,
			minDistance, maxDistance);

	//assign the camera to the class member "sensor"
	sensor = mCamera;
	view->SetCamera(mCamera);

	///calculate viewport in the right corner
	int screenWidth, screenHeight, winX, winY;
	GA->GetWindow()->GetPosition(winX, winY, screenWidth, screenHeight);
	float margin = 0.03;
	float width  = (float) (0.25 * screenWidth - margin * 2);
	float height = width; //we want it square
	float xPos   = (float) (0.75 * screenWidth + margin);
	float yPos   = (float) (screenHeight - height - margin);

	///@todo viewport is set twice, check if that is needed
	std::ostringstream msg;	msg.clear(); msg.str("");
	msg.clear(); msg.str("");
	msg << "Window [x,y,width,height] = [" << xPos << "," << yPos << "," << width << "," <<
			height << "]";
	LOG_INFO(msg.str());
	mCamera->GetOSGCamera()->setViewport(new osg::Viewport(xPos, yPos, width, height));

	///attach camera to the body
	this->myBody->AddChild(sensor.get());

	///and set the local transform that positions it somewhere at the periphery of the actual
	///body carrying this camera around
	///@todo Why is the orientation corrected?
	dtCore::Transform offset(
			relTranslation.x(), relTranslation.y(), relTranslation.z(),
			relOrientation.x()+180, relOrientation.y(), relOrientation.z());
	sensor->SetTransform( offset, dtCore::Transformable::REL_CS );

	///initialise the screen shot taker.
	this->mScreenShotTaker = new CameraScreenShotCallback(cameraType);
	dynamic_cast<dtCore::Camera*>(sensor.get())->GetOSGCamera()->setPostDrawCallback(
			mScreenShotTaker.get());

	///when there are shadows, we have to render the scene differently
	if (SimulationUtils::Instance()->getWithShadows()) {
		view->GetOsgViewerView()->setSceneData(
				srCore::SimulationUtils::Instance()->getShadowScene().get());
		view->GetOsgViewerView()->init();
	}

	///set the order of rendering, at least later than the main window (0) and the GUI (0 or 1)
	view->SetRenderOrder(GA->GetView()->GetRenderOrder()+3);
}

