/*
 * raySensorBase.cpp
 *
 *  Created on: May 7, 2009
 *      Author: winkler
 */

#include "raySensorBase.h"
#include <osg/LineWidth>

#define RTOD(x) 180 * x /M_PI
#define SHOW_RAYS

namespace srCore {

/**
 * \brief  Initializes the sensor
 *
 * \param			gm							GameManager, which handles the corresponding robot
 * \param			attachedTo					Body, where the sensor is attached to
 * \param     		pos			   				Local position of the sensor
 * \param			orient 						Local orientation of the sensor
 * \param			flareAngle					flare angle of the sensor
 * \param			resolution					resolution of the sensor
 * \param			name						name of the sensor
 *
 *
 */
RaySensorBase::RaySensorBase(dtCore::RefPtr<BodyBase> attachedTo, osg::Vec3f pos,osg::Vec3f orient, const std::string &name, std::string type,bool active) :
	SensorBase(attachedTo, pos, orient, name, type, RGB_CAMERA, active),
	minDistance(1),
	maxDistance(1000),
	frustumLeft(-1),
	frustumRight(1),
	frustumUp(-1),
	frustumDown(1),
	flareAngle(0.5),
	sensorImage(new osg::Image),
	resolution(osg::Vec2(1, 1)),
	x(0),
	y(0),
	displayRays(true),
	renderedRays(false)
{

	// Register parameters
	ObjectParameter sp;

	sp.valueName = "MinDistance";
	sp.valueType = typeid(float*).name();
	myParameters.push_back(sp);

	sp.valueName = "FlareAngle";
	sp.valueType = typeid(float*).name();
	myParameters.push_back(sp);

	sp.valueName = "MaxDistance";
	sp.valueType = typeid(float*).name();
	myParameters.push_back(sp);

	sp.valueName = "Frustum";
	sp.valueType = typeid(osg::Vec4f*).name();
	myParameters.push_back(sp);

	sp.valueName = "Resolution";
	sp.valueType = typeid(osg::Vec2d*).name();
	myParameters.push_back(sp);

	// allocate memory for the sensor image
	sensorImage->allocateImage(static_cast<int> (resolution.x()),
			static_cast<int> (resolution.y()), 1, GL_RGB, GL_UNSIGNED_BYTE);

	// not tested yet, but might be useful to introduce noise to the sensor
	mNoise = new dtUtil::Noise2f();
	mNoise->Reseed(341341);



}
;

RaySensorBase::~RaySensorBase() {

	sensorImage->unref();
	delete mNoise;

}
;

//! calculates sensor values
void RaySensorBase::requestData() {

	// 1. creating initial matrix along the Y-Z-plane
	// rotating the plane according to the orientation of the sensor
	// retrieving intersection points
	int count = static_cast<int>(resolution.x()) * static_cast<int>(resolution.y());
	while(count--)
	{
		//TODO: optimise drawing rays
	float alpha, beta;

	if(static_cast<int>(resolution.x())==1)
		alpha=0;
	else
		alpha = x * flareAngle / (resolution.x() -1) - flareAngle/2.0;

	if(static_cast<int>(resolution.y())==1)
		beta=0;
	else
		beta = y * flareAngle / (resolution.y()-1) - flareAngle/2.0;

	//calculate position of the ray
	osg::Vec3f startVec = osg::Vec3f(minDistance*sin(alpha), -minDistance*cos(alpha), minDistance*sin(beta));
	osg::Vec3f stopVec = osg::Vec3f(maxDistance*sin(alpha), -maxDistance*cos(alpha), maxDistance*sin(beta));



	// get global rotation matrix of the sensor
	dtCore::Transform xform;
	osg::Matrix rotMatrix;
	//GetTransform(xform, dtCore::Transformable::ABS_CS);
	//xform.GetRotation(rotMatrix);

	//calculating from hpr to rotation matrix
	xform.SetRotation(this->getOrientation(dtCore::Transformable::ABS_CS));
	xform.GetRotation(rotMatrix);

	// get global position of the sensor
	osg::Vec3d translation = getTranslation(dtCore::Transformable::ABS_CS);

	// calculate start and stop position of the ray
	osg::Vec3d startPosition = startVec * rotMatrix  + translation;
	osg::Vec3d stopPosition = stopVec* rotMatrix  + translation;


	float z=0;
	// retrieve sensor value
	 osg::Vec3f intersectPos = getNextIntersection(startPosition, stopPosition, &z);


	//this->displayRays = true;
	if(this->displayRays){
			lineGeometry = new osg::Geometry();
			vArray = new osg::Vec3Array(2);
			colors = new osg::Vec4Array(1);

			(*vArray)[0] = startPosition;
			(*vArray)[1] = intersectPos;
			(*colors)[0] = osg::Vec4(1, 0, 0, 1);

			lineGeometry->setVertexArray(vArray);
			lineGeometry->setColorArray(colors);
			lineGeometry->setColorBinding(osg::Geometry::BIND_OVERALL);
			osg::DrawArrays* drawable = new osg::DrawArrays(
					osg::PrimitiveSet::LINES, 0, 2);
			lineGeometry->addPrimitiveSet(drawable);
			mGeode->addDrawable(lineGeometry);
			mGeode->setNodeMask(SimulationUtils::laserRayNodeMask);

			//flag for remove rays later
			this->renderedRays = true;
		}




	// normalize, so that values are between 0 and 255.

	if ((stopVec - startVec).length() != 0)
	z *= 255 / (stopVec - startVec).length();

	if (z < 0)
		z = 0;
	if (z > 255)
		z = 255;

	currentSensorValue = z;

	//draw image
	if ((x >= 0) && (x < resolution.x()) && (y >= 0)
			&& (y < resolution.y())) {

		pixelPos = ((3 * x) + sensorImage->getRowSizeInBytes() * y);

		sensorImage->data()[pixelPos] = static_cast<unsigned char> (z);
		sensorImage->data()[pixelPos + 1] = static_cast<unsigned char> (z);
		sensorImage->data()[pixelPos + 2] = static_cast<unsigned char> (z);

	}

	// move ray to the next position
	x++;
	if (x >= resolution.x()) {
		x = 0;
		y++;
		if (y >= resolution.y()) {
			y = 0;

		}
	}
}
	//	osgDB::writeImageFile( *sensorImage, "blub.jpg" );


}

/**
 * \brief  Sets a parameter
 *
 *		See also getParameter()
 *
 */
void RaySensorBase::setParameter(const std::string &name,
		ParameterReturnValue parameter) {

	if (name == "MinDistance") {

		// check parameter type
		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");

		// cast void* to float*
		float* data = reinterpret_cast<float*> (parameter.pValue);

		// set parameter ...
		minDistance = *data;

		// .. and adjust other parameters as well, as they might have changed
		float frustum = tan(this->flareAngle / 2) * this->minDistance;

		//if (sensor.get() && (frustum > 0.02 / this->minDistance)) {
		if (frustum > 0) {

			this->frustumDown = frustum;
			this->frustumUp = frustum;
			this->frustumRight = frustum;
			this->frustumLeft = frustum;

			printf("frustum %f, flareAngle %f", frustum, flareAngle);

		}

	}

	else if (name == "FlareAngle") {

		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");
		float* data = reinterpret_cast<float*> (parameter.pValue);
		flareAngle = *data;

		float frustum = tan(this->flareAngle / 2) * this->minDistance;
		if (frustum > 0) {

			this->frustumDown = frustum;
			this->frustumUp = frustum;
			this->frustumRight = frustum;
			this->frustumLeft = frustum;

		}

	}

	else if (name == "MaxDistance") {

		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");
		float* data = reinterpret_cast<float*> (parameter.pValue);
		maxDistance = *data;

	}

	else if (name == "Frustum") {

		if (parameter.type != typeid(osg::Vec4f*).name())
			printf(
					"Parameter type needs to be osg::Vec4f*. Frustum is not set.  \n");
		else {

			osg::Vec4f* data = reinterpret_cast<osg::Vec4f*> (parameter.pValue);
			frustumLeft = data->x();
			frustumRight = data->y();
			frustumUp = data->z();
			frustumDown = data->w();

		}

	}

	else if (name == "Resolution") {

		if (parameter.type != typeid(osg::Vec2d*).name())
			printf(
					"Parameter type needs to be osg::Vec2d*. Casting could lead to errors.  \n");
		osg::Vec2d* data = reinterpret_cast<osg::Vec2d*> (parameter.pValue);
		resolution = *data;

		sensorImage->allocateImage(static_cast<int> (resolution.x()),
				static_cast<int> (resolution.y()), 1, GL_RGB, GL_UNSIGNED_BYTE);

	}

	else
		SensorBase::setParameter(name, parameter);

}

/**
 * \brief  Writes the parameter value in parameter
 *
 *		The available parameters can be retrieved by calling getParameterList()
 *		See also: ObjectBase::getParameter()
 *
 *
 */
void RaySensorBase::getParameter(const std::string &name,
		ParameterReturnValue parameter) {

	if (name == "MinDistance") {

		//check parameter type
		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");

		// cast void* to float*
		float* data = reinterpret_cast<float*> (parameter.pValue);

		// get parameter
		*data = minDistance;
		parameter.pValue = data;

	}

	else if (name == "FlareAngle") {

		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");
		float* data = reinterpret_cast<float*> (parameter.pValue);
		*data = flareAngle;
		parameter.pValue = data;

	}

	else if (name == "MaxDistance") {

		if (parameter.type != typeid(float*).name())
			printf(
					"Parameter type needs to be float*. Casting could lead to errors.  \n");
		float* data = reinterpret_cast<float*> (parameter.pValue);
		*data = maxDistance;
		parameter.pValue = data;

	}

	else if (name == "Frustum") {

		if (parameter.type != typeid(osg::Vec4f*).name())
			printf(
					"Parameter type needs to be osg::Vec4f*. Casting could lead to errors.  \n");
		osg::Vec4f* data = reinterpret_cast<osg::Vec4f*> (parameter.pValue);
		*data = osg::Vec4f(frustumLeft, frustumRight, frustumUp, frustumDown);
		parameter.pValue = data;

	}

	else if (name == "Resolution") {

		if (parameter.type != typeid(osg::Vec2*).name())
			printf(
					"Parameter type needs to be osg::Vec2*. Casting could lead to errors.  \n");
		osg::Vec2* data = reinterpret_cast<osg::Vec2*> (parameter.pValue);
		*data = resolution;
		parameter.pValue = data;

	}

	else
		SensorBase::getParameter(name, parameter);

}

//sets up the sensor
void RaySensorBase::setup() {

	//		//sensor = new dtCore::Transformable();
	//		gm->GetApplication().GetScene()->AddDrawable(sensor.get());
	//
	//		//attach sensor to the deltaDrawable
	//		//this->attachedTo->AddChild(sensor.get());
	//		this->myBody->getObject()->AddChild(sensor.get());
	//
	//		dtCore::Transform offset(relTranslation.x(), relTranslation.y(), relTranslation.z(), relOrientation.x(), relOrientation.y(), relOrientation.z());
	//		sensor->SetTransform( offset, dtCore::Transformable::REL_CS );
	//for drawing

	//for drawing
	this->mGeode = new osg::Geode();
	mGeode->setNodeMask(SimulationUtils::laserRayNodeMask);
	osg::StateSet *ss = this->mGeode->getOrCreateStateSet();
	osg::LineWidth* linewidth = new osg::LineWidth;
	linewidth->setWidth(1);
	ss->setAttributeAndModes(linewidth, osg::StateAttribute::ON);
	ss->setMode(GL_LIGHTING, osg::StateAttribute::OFF);
	dtABC::Application::GetInstance(0)->GetScene()->GetSceneNode()->addChild(mGeode);



}


//! Will be called in RobotActorBase::TickLocal()
void RaySensorBase::update(float deltaSimTme) {

	// get n ray trace values per frame


	//remove the rays
	if (this->displayRays && this->renderedRays) {
		mGeode->removeDrawables(0, mGeode->getNumDrawables());
		this->renderedRays = false;
	}

//	if (active)
	requestData();

//	if (isActiveSensor)
    getNewSensorValue();


}

/**
 *  Generates a ray, given by start and end position and calculates a value, depending on the distance
 *  of the object that reflected the ray and its orientation towards the ray.
 */
osg::Vec3f RaySensorBase::getNextIntersection(osg::Vec3f startPos, osg::Vec3f endPos, float *z) {

	osg::Vec3f p;

	// create Isector, which calculates intersection points of the ray with the environment
	dtCore::RefPtr<dtCore::Isector> query = new dtCore::Isector(
			dtABC::Application::GetInstance(0)->GetScene());//GetSceneParent());
	query->Reset();
	query->SetStartPosition(startPos);
	query->SetEndPosition(endPos);
	query->SetDirection(endPos - startPos);
	//query->SetTraversalMask(~0);
//	query->SetTraversalMask(~SimulationUtils::geometryDetailsNodeMask);

	if (query.get()->Update()) {

		osgUtil::IntersectVisitor &iv = query.get()->GetIntersectVisitor();

		//get intersection point
		p  = iv.getHitList(query.get()->GetLineSegment())[0].getWorldIntersectPoint();

		//get distance between sensor and intersect point
		*z = (p-startPos).length();

		return p;

	}

	else {
		*z=(endPos - startPos).length();
		return endPos;
	}

}

}
