/*
 * Robot3D Physical realistic 3D simulator for robots
 *
 * @license GNU Lesser General Public License
 *
 * @author Lutz Winkler
 * @author Anne C. van Rossum
 */

// General libraries
#include <dlfcn.h>
#include <sys/stat.h>

// Delta3D libraries
#include <dtUtil/log.h>
#include <dtDAL/librarymanager.h>

// Robot3D files
#include <srCore/export.h>
#include <srCore/simulationEntryPoint.h>
#include <srCore/robotCreator.h>
#include <srCore/robot/robotActorBase.h>
#include <srCore/loadComponents.h>
#include <srCore/robot3dComponent.h>

/************************************************************************************************
 *                      Global properties
 ***********************************************************************************************/

#define TIME_SCALE 							0.1				//1
#define GRAVITY								200				//n*(5cm)/s²

/************************************************************************************************
 *                      Default strings / names / paths
 ***********************************************************************************************/

std::string XML_TAG_SCENE_FILE			= "ScenarioFile";
std::string XML_TAG_CEGUI_SCHEME_FILE	= "CEGUISchemeFile";
std::string XML_TAG_CEGUI_PATH			= "CEGUIPath";
std::string XML_TAG_COMPONENT_PATH		= "ComponentPath";
std::string XML_TAG_STAGEPROJECT_PATH	= "StageProject";

std::string DEFAULT_COMPONENT_PATH 		= "/usr/lib/robot3d/components";

std::string DEFAULT_STAGEPROJECT_PATH	= "/var/lib/robot3d/StageProject";
std::string DEFAULT_CEGUI_PATH	        = "/var/lib/robot3d";
std::string DEFAULT_CEGUI_SCHEME_FILE   = "/var/lib/robot3d/gui/schemes/WindowsLook.scheme";
std::string DEFAULT_SCENE_FILE			= "/var/lib/robot3d/scenarios/scene.xml";

// Will be prepended by $HOME/
std::string DEFAULT_SCENE_FILE_HOME		= ".robot3d/scene.xml";
std::string DEFAULT_SCREENSHOT_PATH		= ".robot3d/screenshots/";

/************************************************************************************************
 *                      Shortcuts
 ***********************************************************************************************/

using dtCore::RefPtr;
using namespace srCore;

/************************************************************************************************
 *                      C Hooks for Delta3d
 ***********************************************************************************************/

/**
 * This function is obligated by the Delta3D library. It makes it possible to dynamically load the
 * Robot3D library.
 */
extern "C" ROBOT_EXPORT dtGame::GameEntryPoint* CreateGameEntryPoint() {
	return new SimulationEntryPoint();
}

/**
 * This function is obligated by the Delta3D library. It makes it possible to dynamically unload the
 * Robot3D library.
 */
extern "C" ROBOT_EXPORT void DestroyGameEntryPoint(dtGame::GameEntryPoint* entryPoint) {
	delete entryPoint;
}

/************************************************************************************************
 *                      Implementation of SimulationEntryPoint
 ***********************************************************************************************/

/**
 * The constructor of the Robot3D simulator. It sets the default log level to LOG_INFO and
 * displays a welcome message.
 */
SimulationEntryPoint::SimulationEntryPoint():
		guiComp(), loadComp() {
	dtUtil::Log::GetInstance().SetLogLevel(dtUtil::Log::LOG_INFO);
	LOG_ALWAYS("Welcome to the Robot3D simulator!");
	LOG_ALWAYS("Version 0.0.1.11.22");
	osg::DisplaySettings::instance()->setNumMultiSamples(4);
}

/**
 * Destroy the entire Robot3D simulator.
 * @return the void
 * @todo Check what can be deallocated...
 */
SimulationEntryPoint::~SimulationEntryPoint() {
	LOG_INFO("Unload everything...");
}


/**
 * The start up routine. It loads a predefined XML configuration file and initiates the simulator
 * with the data found in it. It is not the idea that this function needs to know the different types
 * of robot, sensor, and actuator classes that can be defined. This needs to be registered by the
 * user.
 */
void SimulationEntryPoint::OnStartup(dtGame::GameApplication &app)
{
	//	app.GetGameManager()->DebugStatisticsTurnOn(true,true,1,true,"gameManagerDebugInfo.txt");
	bool withRobotControlEnabled = true;
	bool withGUI = false;
	bool freeCamState = true;
	int  robotFocus = 0;
	osg::Vec3f camPos;

	std::ostringstream msg;	msg.clear(); msg.str("");
	dtUtil::Log::GetInstance().SetLogLevel(dtUtil::Log::LOG_INFO);
	LOG_INFO("Beginning of Robot3D simulator startup");

	app.GetWindow()->SetWindowTitle("Robot3D Simulator");
	app.GetWindow()->ShowCursor(false);

	// Set stage project path
	std::string stageprojectPath=app.GetConfigPropertyValue(XML_TAG_STAGEPROJECT_PATH, DEFAULT_STAGEPROJECT_PATH);
	SimulationUtils::Instance()->setStageProjectPath(stageprojectPath);
	msg.clear(); msg.str("");
	msg << "Set StageProject path to: " << SimulationUtils::Instance()->getStageProjectPath();
	LOG_INFO(msg.str());

	// Register new message types
	SimulationMessageType::RegisterMessageTypes( app.GetGameManager()->GetMessageFactory() );

	LOG_WARNING("Load actors explicitly from libRobot3D.so library");
	dtDAL::LibraryManager::GetInstance().LoadActorRegistry("Robot3D");

	dtCore::RefPtr<Robot3DComponent> robot3DComp = new Robot3DComponent("Robot3DComponent");
	app.GetGameManager()->AddComponent(*robot3DComp, dtGame::GameManager::ComponentPriority::LOWER);

	// audio
#ifdef AUDIO_ON
	dtAudio::AudioManager::Instantiate();
	dtAudio::AudioManager::GetInstance().Config();
#endif

	std::string file = srCore::SimulationUtils::Instance()->getSceneFile();

	// First try to find the config file in the $HOME directory
	std::ostringstream f; f.clear();
	f << boost::filesystem::path(getenv("HOME")).string() << '/';
	f << DEFAULT_SCENE_FILE_HOME;
	if (boost::filesystem::exists(file)) {
		// then don't overwrite it....
	} else if (boost::filesystem::exists(f.str())) {
		file = app.GetConfigPropertyValue(XML_TAG_SCENE_FILE, f.str());
	} else {
		// Now it's trying to find it in /var
		LOG_ALWAYS("Scene file does not exist in home directory, get from default installation location");
		file = app.GetConfigPropertyValue(XML_TAG_SCENE_FILE, DEFAULT_SCENE_FILE);
	}
	srCore::SimulationUtils::Instance()->setSceneFile(file);

	try {
		msg.clear(); msg.str("");
		msg << "Reading \"scene.xml\" file: " << srCore::SimulationUtils::Instance()->getSceneFile();
		LOG_INFO(msg.str());
	} catch(std::runtime_error error) {
		LOG_ERROR("Configuration file does not exist!");
		return;
	}

	// Read configuration file
	robot3DComp->GetConfigObject().readConfigFile(srCore::SimulationUtils::Instance()->getSceneFile());

	msg.clear(); msg.str("");
	msg << "Map: " << robot3DComp->GetConfigObject().getMap();
	LOG_INFO(msg.str());
	msg.clear(); msg.str("");
	msg << "With Shadows: " << robot3DComp->GetConfigObject().getWithShadows();
	LOG_INFO(msg.str());
	msg.clear(); msg.str("");
	msg << "With Detailed Models: " << robot3DComp->GetConfigObject().getWithDetailedModels();
	LOG_INFO(msg.str());

	msg.clear(); msg.str("");
	msg << "Load " << robot3DComp->GetConfigObject().getRobots().size() << " robots";
	LOG_INFO(msg.str());

	for (unsigned int i=0; i< robot3DComp->GetConfigObject().getRobots().size(); i++) {
		msg.clear(); msg.str("");
		msg << "Named " << robot3DComp->GetConfigObject().getRobots()[i].name << " with type " << robot3DComp->GetConfigObject().getRobots()[i].type;
		msg << " at position: [" << robot3DComp->GetConfigObject().getRobots()[i].pos.x() << "," <<
				robot3DComp->GetConfigObject().getRobots()[i].pos.y() << "," << robot3DComp->GetConfigObject().getRobots()[i].pos.z() << "]";
		msg << " oriented as: [" << robot3DComp->GetConfigObject().getRobots()[i].quat << "] ";
		if (!robot3DComp->GetConfigObject().getRobots()[i].jointAngles.empty()) {
			msg << " with joint angle: " << robot3DComp->GetConfigObject().getRobots()[i].jointAngles[0];
		} else {
			msg << " without joint angle";
		}
		LOG_DEBUG(msg.str());
	}

	msg.clear(); msg.str("");
	msg << "Display controllers that can be loaded (potentially)...";
	LOG_INFO(msg.str());

	for (unsigned int i=0; i< robot3DComp->GetConfigObject().getControllers().size(); i++)
	{
		if (robot3DComp->GetConfigObject().getControllers().at(i).name == NULL) {
			LOG_WARNING("There seems to be something wrong with the vector with ControllerInitData");
			continue;
		}
		msg.clear(); msg.str("");
		msg << i << ": ";
		msg << "A " << robot3DComp->GetConfigObject().getControllers()[i].type << " controller";
		if (strcmp(robot3DComp->GetConfigObject().getControllers()[i].name, "")) {
			msg << " named \"" << robot3DComp->GetConfigObject().getControllers()[i].name << "\"";
		}
		if (strncmp(robot3DComp->GetConfigObject().getControllers()[i].type, "Input", strlen("Input")) == 0)
		{
			withRobotControlEnabled = atof(robot3DComp->GetConfigObject().getControllers()[i].parameter);
			msg <<" withRobotControlEnabled="<< withRobotControlEnabled;

		}
		if (strncmp(robot3DComp->GetConfigObject().getControllers()[i].type, "GUI", strlen("GUI")) == 0)
			withGUI = true;

		if (strncmp(robot3DComp->GetConfigObject().getControllers()[i].type, "FreeCam", strlen("FreeCam")) == 0)
		{
			robotFocus = atoi(robot3DComp->GetConfigObject().getControllers()[i].robot);
			camPos = robot3DComp->GetConfigObject().getControllers()[i].pos;

			msg << " at position: [" << camPos.x() << "," << camPos.y() << "," << camPos.z() << "]";

			freeCamState = true;
		}
		else if (strncmp(robot3DComp->GetConfigObject().getControllers()[i].type, "FollowCam", strlen("FollowCam")) == 0)
		{
			robotFocus = atoi(robot3DComp->GetConfigObject().getControllers()[i].robot);
			camPos = robot3DComp->GetConfigObject().getControllers()[i].pos;

			msg << " at position: [" << camPos.x() << "," << camPos.y() << "," << camPos.z() << "]";

			freeCamState = false;
		}

		if (strcmp(robot3DComp->GetConfigObject().getControllers()[i].robot, ""))
			msg << " controlling robot \"" << robot3DComp->GetConfigObject().getControllers()[i].robot << "\" ";
		LOG_INFO(msg.str());

	}

	SimulationUtils::Instance()->setDetailedGraphics(robot3DComp->GetConfigObject().getWithDetailedModels());
	SimulationUtils::Instance()->setWithShadows(robot3DComp->GetConfigObject().getWithShadows());
	SimulationUtils::Instance()->setCollisionGeometryDetails(robot3DComp->GetConfigObject().getCollisionGeometryDetails());
	SimulationUtils::Instance()->setConnectorsAsJoints(robot3DComp->GetConfigObject().getConnectorsAsJoints());
	SimulationUtils::Instance()->setWithWheels(robot3DComp->GetConfigObject().getRobotsWithWheels());

	//Load Map and apply Light (set log level temporary to WARNING to get less log messages)
	LOG_INFO("Set context, load all information from StageProject directory");
	dtUtil::Log::LogMessageType lmt = dtUtil::Log::GetInstance().GetLogLevel();
	if (lmt < dtUtil::Log::LOG_WARNING)
		dtUtil::Log::GetInstance().SetLogLevel(dtUtil::Log::LOG_WARNING);

	dtDAL::Project::GetInstance().SetContext(SimulationUtils::Instance()->getStageProjectPath());
	dtDAL::Map &map = dtDAL::Project::GetInstance().GetMap(robot3DComp->GetConfigObject().getMap());
	dtDAL::Project::GetInstance().LoadMapIntoScene(map, app.GetGameManager()->GetScene(), false);
	dtUtil::Log::GetInstance().SetLogLevel(lmt);
	LOG_INFO("Map loaded");

	// SetOSGParameters
	app.GetView(0)->GetDatabasePager()->SetTargetFrameRate(60);		//20

	SetPhysicalParameters(app);

	//----------------------Components----------------------------------------------------------------------------------------

	LOG_INFO("Creating components");

	//Add DefaultMessageProcessor
	dtGame::DefaultMessageProcessor *dmp = new dtGame::DefaultMessageProcessor("DefaultMessageProcessor");
	app.GetGameManager()->AddComponent(*dmp,dtGame::GameManager::ComponentPriority::HIGHEST);

	//Add Arena Component - doing all the important stuff
	dtCore::RefPtr<SimulationArenaComponent> arenaComp = new SimulationArenaComponent(SimulationArenaComponent::DEFAULT_NAME);
	app.GetGameManager()->AddComponent(*arenaComp, dtGame::GameManager::ComponentPriority::LOWEST);

	arenaComp->createRobotsFromSceneFile(robot3DComp->GetConfigObject());

	//Add all GUI elements
	guiComp = NULL;
	if (withGUI) {
		InitGUI(app);
		LoadGUIComponents(app);
	}

	//Add Camera Component
	dtCore::RefPtr<SimulationCameraController> cameraContr = new SimulationCameraController(
			"SimulationCameraController", camPos, freeCamState);
	app.GetGameManager()->AddComponent(*cameraContr, dtGame::GameManager::ComponentPriority::LOWER);
	if (withGUI) {
		cameraContr->init(guiComp->GetMouse(), guiComp->GetKeyboard());
	} else {
		cameraContr->init(app.GetMouse(), app.GetKeyboard());
	}

#ifdef USE_SEPARATE_GUI
	cameraContr->setViewport(false);
#else
	cameraContr->setViewport(true);
#endif
	f.clear(); f.str("");
	f << boost::filesystem::path(getenv("HOME")).string() << '/';
	f << DEFAULT_SCREENSHOT_PATH;
	std::string path = f.str();
	if (boost::filesystem::exists(path)) {
		cameraContr->setScreenshotPath(path);
	} else {
		std::ostringstream msg; msg.clear();
		msg << "Screenshot path \"" << path << "\" does not exist, it will be created";
		LOG_INFO(msg.str());
		boost::filesystem::create_directories(path);
		cameraContr->setScreenshotPath(path);
	}

	//Add Input Component
	dtCore::RefPtr<SimulationInputComponent> inputComp = new SimulationInputComponent("SimulationInputComponent", withRobotControlEnabled);
	app.GetGameManager()->AddComponent(*inputComp, dtGame::GameManager::ComponentPriority::LOWER);


	//-----------------------------------------------------------------------------------------------------------------------
	// Shadows

	if (SimulationUtils::Instance()->getWithShadows()) {
		app.GetView(0)->GetOsgViewerView()->setSceneData(srCore::SimulationUtils::Instance()->getShadowScene().get());
		app.GetView(0)->GetOsgViewerView()->init();
	}

	std::vector<dtGame::GMComponent*> allComp;
	app.GetGameManager()->GetAllComponents(allComp);
	for (unsigned int i=0; i< allComp.size(); i++) {
		if (dynamic_cast<dtCore::Light*>(allComp[i])) {
			LOG_INFO("Found light (as component)");
		}
	}

	LOG_INFO("Load all plugins from ComponentPath");
	loadComp = new LoadComponents();
	std::string comp_path = app.GetConfigPropertyValue(XML_TAG_COMPONENT_PATH, DEFAULT_COMPONENT_PATH);
	loadComp->SetLibPath(comp_path);
	loadComp->GetLibraries();
	loadComp->LoadLibs(app);

	LOG_INFO("End of Robot3D simulator startup");

}

/**
 * Todo: For every component create a separate function and bundle the components in a map,
 * so they can be easily swapped (and retrieved by plugins), and perhaps even reconfigured on the
 * fly.
 */
void SimulationEntryPoint::AddCamera() {

}

/**
 * Sets a bunch of parameters for the physics engine.
 *
 * @todo Move Physics related stuff to a separate component. The component may be build on
 * dtPhysics, so it enables us to switch physics engines.
 */
void SimulationEntryPoint::SetPhysicalParameters(dtGame::GameApplication &app) {

	//Setting Collision NearCallback function 
//	void* data = NULL;
//	app.GetScene()->SetUserCollisionCallback (SimulationUtils::Instance()->NearCallback, data);

	//Setting PhysicsStepSize and TargetFrameRate for stabilizing the simulation environment:
	app.GetGameManager()->GetScene().SetPhysicsStepSize(0.05);		//0.05

	// The Constraint Force Mixing is invented by the ODE guys. It allows for relaxing of the
	// constraint equation for joints. The original hard constraint is 0.
	// @see <a>http://opende.sourceforge.net/wiki/index.php/Manual_%28Concepts%29#Constraint_Force_Mixing_.28CFM.29</a>
	dWorldSetCFM(app.GetGameManager()->GetScene().GetWorldID(), 0.005);			//0.05

	// The Error Reduction Parameter is also invented by the ODE guys. It is for example
	// possible to set the position of the multiple bodies that make up a robot in slightly
	// wrong relative locations. Each joint can apply a force to bring its bodies back into
	// the right positions.
	// <a>@see http://opende.sourceforge.net/wiki/index.php/Manual_%28Concepts%29#Joint_error_and_the_error_reduction_parameter_.28ERP.29</a>
	dWorldSetERP(app.GetGameManager()->GetScene().GetWorldID(), 0.9);			//0.9

	//Gravity
	app.GetGameManager()->GetScene().SetGravity(0.f, 0.f, -GRAVITY);

	// If you see a robot slowly sinking into the surface and then suddenly popping out in an
	// incredibly high speed, that speed is set by this parameter. Of course, we normally don't
	// want that to happen at all...
	dWorldSetContactMaxCorrectingVel(app.GetGameManager()->GetScene().GetWorldID(), 250);			//200

	// This parameter tells how much the robots can sink into the ground, as described
	// by the ODE lib, increasing this to some small value (e.g. 0.001) can help prevent
	// jittering problems due to contacts being repeatedly made and broken.
	dWorldSetContactSurfaceLayer (app.GetGameManager()->GetScene().GetWorldID(), 0.005);			//0.001 (0.003)

	app.GetGameManager()->ChangeTimeSettings(app.GetGameManager()->GetSimulationTime(), TIME_SCALE,
			app.GetGameManager()->GetSimulationClockTime());

}

/**
 * Set some parameters
 */
void SimulationEntryPoint::InitGUI(dtGame::GameApplication &app) {
	std::string ceguiPath = app.GetConfigPropertyValue(XML_TAG_CEGUI_PATH, DEFAULT_CEGUI_PATH);
	SimulationUtils::Instance()->setCEGUIPath(ceguiPath);
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Set CEGUI path to: " << SimulationUtils::Instance()->getCEGUIPath();
	LOG_INFO(msg.str());

	std::string ceguiSchemeFile = app.GetConfigPropertyValue(XML_TAG_CEGUI_SCHEME_FILE, DEFAULT_CEGUI_SCHEME_FILE);
	SimulationUtils::Instance()->setCEGUISchemeFile(ceguiSchemeFile);
	msg.clear(); msg.str("");
	msg << "Set CEGUI SchemeFile to: " << SimulationUtils::Instance()->getCEGUISchemeFile();
	LOG_INFO(msg.str());
}

/**
 * Loads GUI components in the same dtCore::DeltaWin window, but using a different camera, a
 * different view and a different scene.
 * The GUI adds the information panel at the right. Where you will see information about the
 * selected robot, the output of its sensors and the input to its actuators. It also adds a
 * menu at the top to select certain components in the scene.
 * @todo Freecam state or not is not the business of the simulationguicomponent
 * @param app
 */
void SimulationEntryPoint::LoadGUIComponents(dtGame::GameApplication &app) {
	// The default application view contains everything
	dtCore::RefPtr<dtCore::View> appView = app.GetView();
	dtCore::RefPtr<dtCore::Camera> appCamera = app.GetCamera();
	dtCore::RefPtr<dtCore::Scene> appScene = app.GetScene();
	dtCore::RefPtr<dtCore::DeltaWin> appWindow = app.GetWindow();
#ifdef USE_SEPARATE_GUI
	// In this case not everything is added to the default application view
	// The gui components are maintained in a separate guiView and guiScene and can be seen through a guiCamera
	// This makes it possible for robots NOT to notice them
	dtCore::RefPtr<dtCore::View> guiView;
	dtCore::RefPtr<dtCore::Camera> guiCamera;
	dtCore::RefPtr<dtCore::Scene> guiScene;

	//Create also a new view and a new scene
	LOG_INFO("Create a new view and a new scene for the GUI");
	guiView = new dtCore::View("GUI View");
	guiCamera = new dtCore::Camera("GUI Camera");
	guiScene = new dtCore::Scene("GUI Scene");

	// Set some default parameters
	dtCore::Transform transform(20.0f, -30.0f, 15.0f);
	guiCamera->SetTransform(transform);
	guiCamera->SetClearColor(0.f, 0.f, 0.f, 0.f);
	guiView->SetCamera(guiCamera);

	// Add the view to the application
	app.AddView(*guiView);
	guiCamera->SetWindow(appWindow);
	guiView->SetScene(guiScene);
	guiCamera->GetOSGCamera()->setClearMask( GL_DEPTH_BUFFER_BIT );

	//Load the CEGUI window layout
	LOG_INFO("Load CEGUI window layout");
	guiComp = new SimulationGUIComponent("SimulationGUIComponent", *appWindow,
			*guiView->GetKeyboard(), *guiView->GetMouse());

	// Create a new camera that sees the GUI
	LOG_INFO("Create a new camera that sees the GUI (the default one will only see the scene with robots)");
	guiScene->AddDrawable(guiComp->GetGUIDrawable());

	// Adjust render order, so the GUI is rendered after the overall scene

	app.GetView()->SetRenderOrder(2);
	guiView->SetRenderOrder(app.GetView()->GetRenderOrder() + 1);
	CEGUI::System::getSingleton().getGUISheet()->setAlpha(0.2f);

	// Add component to game manager for later retrieval
	app.GetGameManager()->AddComponent(*guiComp, dtGame::GameManager::ComponentPriority::LOWER);
#else
	//Load the CEGUI window layout
	LOG_INFO("Load CEGUI window layout");
	guiComp = new SimulationGUIComponent("SimulationGUIComponent", *appWindow,
			*appView->GetKeyboard(), *appView->GetMouse());

	// Add to scene
	appScene->AddDrawable(guiComp->GetGUIDrawable());
	guiComp->GetGUIDrawable()->GetOSGNode()->setNodeMask(0x1);

	app.GetGameManager()->AddComponent(*guiComp, dtGame::GameManager::ComponentPriority::LOWER);
#endif
}


