/*
 * simulationUtils.cpp
 *
 *  Created on: Nov 21, 2008
 *      Author: winkler
 */

#include <srCore/simulationUtils.h>

#include <dtCore/odespacewrap.h>

#include <sys/stat.h>

#include <srCore/body/bodyBase.h>

using namespace dtCore;

namespace srCore {

IMPLEMENT_ENUM(SimulationMessageType);
const SimulationMessageType SimulationMessageType::COMM_BUS("Communication Bus", "comm",
		"messages robots can send", USER_DEFINED_MESSAGE_TYPE + 1);


/* **************************************************************************************
 * Implementation of dedicated Robot3D simulation messages
 * **************************************************************************************/

void SimulationMessageType::RegisterMessageTypes(dtGame::MessageFactory& factory)
{
	factory.RegisterMessageType<CommMessage>(COMM_BUS);
}

CommMessage::CommMessage()
{
	AddParameter( new dtGame::ActorMessageParameter( "BusID" ) );
	AddParameter( new dtGame::ActorMessageParameter( "BusMessage" ) );
	AddParameter( new dtGame::ActorMessageParameter( "CommType" ) );
}

CommMessage::CommMessage(SimulationMessageType::commType type, dtCore::UniqueId* id, const std::string& message)
{
	AddParameter( new dtGame::ActorMessageParameter( "BusID" ) );
	AddParameter( new dtGame::ActorMessageParameter( "BusMessage" ) );
	AddParameter( new dtGame::ActorMessageParameter( "CommType" ) );

	busID = id;
	this->message = message;
	this->type = type;
}


CommMessage::~CommMessage()
{
}

void CommMessage::setBusID( dtCore::UniqueId* uniqueId )
{
	busID = uniqueId;
}

dtCore::UniqueId* CommMessage::getBusID() const
{
	return busID;
}

void CommMessage::setMessage( const std::string& message )
{
	this->message = message;
}

const std::string& CommMessage::getMessage() const
{
	return this->message;
}

void CommMessage::setCommType( SimulationMessageType::commType type )
{
	this->type = type;
}

const SimulationMessageType::commType& CommMessage::getCommType() const
{
	return this->type;
}

/* **************************************************************************************
 * Implementation of SimulationUtils_
 * **************************************************************************************/

SimulationUtils_::SimulationUtils_():
					geometryDetailsNodeMask(0x2),
					collisionNodeMask(0x4),
					laserRayNodeMask(0x8),

					detailedGraphics(true),
					withShadows(true),
					shadowedScene(new osgShadow::ShadowedScene()),
					collisionGeometryDetails(1),

					connectorsAsJoints(false),
					withWheels(true),
					mContactJointGroupID(dJointGroupCreate(0)),

					sceneFile("./scenarios/scene.xml"),
					stageProjectPath("./StageProject"),

					startup (true) {
}

/**
 * Default destructor
 */
SimulationUtils_::~SimulationUtils_() {
	// TODO Auto-generated destructor stub
}

bool SimulationUtils_::fileExists(std::string fileName) {
	struct stat stFileInfo;
	return (!(stat(fileName.c_str(),&stFileInfo)));
}

void SimulationUtils_::NearCallback(void* data, dGeomID o1, dGeomID o2)
{

	dtGame::GameApplication *app = dtGame::GameApplication::GetInstance(0);

	if (data == 0 || o1 == 0 || o2 == 0)
	{
		return;
	}

	//   ODESpaceWrap* spaceWrap = static_cast<ODESpaceWrap*>(data);

	Transformable* c1 = static_cast<Transformable*>(dGeomGetData(o1));
	Transformable* c2 = static_cast<Transformable*>(dGeomGetData(o2));

	dContactGeom contactGeoms[8];

	int numContacts = dCollide(o1, o2, 8, contactGeoms, sizeof(dContactGeom));

	if (numContacts > 0 && c1 != 0 && c2 != 0)
	{
		ODESpaceWrap::CollisionData cd;

		cd.mBodies[0] = c1;
		cd.mBodies[1] = c2;

		if (dynamic_cast<BodyBase*>(c1)) dynamic_cast<BodyBase*>(c1)->setContact(true);
		if (dynamic_cast<BodyBase*>(c2)) dynamic_cast<BodyBase*>(c2)->setContact(true);

		cd.mLocation.set(
				contactGeoms[0].pos[0], contactGeoms[0].pos[1], contactGeoms[0].pos[2]
		);

		cd.mNormal.set(
				contactGeoms[0].normal[0], contactGeoms[0].normal[1], contactGeoms[0].normal[2]
		);

		cd.mDepth = contactGeoms[0].depth;

		//       if (spaceWrap->mCollisionCBFunc.valid())
		//       {
		//          spaceWrap->mCollisionCBFunc(cd);
		//       }

		if (c1 != 0 || c2 != 0)
		{
			dContact contact;

			for (int i = 0; i < numContacts; ++i)
			{
				//contact parameters, the ones in comment are, how they are set by standard in Delta3D
				contact.surface.mode       = dContactBounce | dContactSlip1 | dContactSlip2 | dContactApprox1 | dContactSoftCFM;		//dContactBounce;
				contact.surface.mu         = (dReal)2.0;											//1000
				contact.surface.bounce     = (dReal)0.0;											//0.75
				contact.surface.bounce_vel = (dReal)0;											//0.001
				contact.surface.slip1	= (dReal) 0.1;
				contact.surface.slip2	= (dReal) 0.1;
				contact.surface.soft_cfm	= (dReal) 0.1;



				contact.geom = contactGeoms[i];

				// Make sure to call these both, because in the case of
				// Trigger, meaningful stuff happens even if the return
				// is false.
				bool contactResult1 = c1->FilterContact(&contact, c2);
				bool contactResult2 = c2->FilterContact(&contact, c1);

				if (contactResult1 && contactResult2)
				{

					// All this also should be in a virtual function.
					Physical* p1 = dynamic_cast<Physical*>(c1);
					Physical* p2 = dynamic_cast<Physical*>(c2);

					if (p1 != 0 || p2 != 0)
					{


						dJointID joint = dJointCreateContact( app->GetGameManager()->GetScene().GetWorldID(),
								mContactJointGroupID,
								&contact);

						dJointAttach(joint,
								p1 != 0 && p1->DynamicsEnabled() ? p1->GetBodyID() : 0,
										p2 != 0 && p2->DynamicsEnabled() ? p2->GetBodyID() : 0);
					}
				}
			}
		}
	}
}

/**
 * This routine is only useful in the case shadows are turned on.
 */
osg::ref_ptr<osgShadow::ShadowedScene> SimulationUtils_::getShadowScene() {

	if (!startup) return shadowedScene;

	startup = false;

	dtGame::GameApplication *app = dtGame::GameApplication::GetInstance(0);
	osg::ref_ptr<osg::Group> rootNode = app->GetGameManager()->GetScene().GetSceneNode();

	dtCore::Light* sceneLight;

	sceneLight = app->GetGameManager()->GetScene().GetLight(0);
	rootNode->addChild(sceneLight->GetOSGNode());

	int ReceivesShadowTraversalMask  = ~(laserRayNodeMask);
	int CastsShadowTraversalMask     = ~(laserRayNodeMask);

	shadowedScene->setReceivesShadowTraversalMask(ReceivesShadowTraversalMask);
	shadowedScene->setCastsShadowTraversalMask(CastsShadowTraversalMask);

	osg::ref_ptr<osgShadow::SoftShadowMap> sm = new osgShadow::SoftShadowMap();

	sm->setTextureUnit(1);
	sm->setAmbientBias(osg::Vec2(0.5,0.5));			//(0,0,8)
	sm->setBias(0);//0.002);
	//	sm->setTextureSize(256,256);
	sm->setSoftnessWidth(0.0002);					// 0.005 );
	sm->setJitteringScale(32);
	int mapres = 256;// 4068;
	sm->setTextureSize(osg::Vec2s(mapres,mapres));
	sm->init();

	shadowedScene->setShadowTechnique(sm.get());

	shadowedScene->addChild(app->GetView()->GetOsgViewerView()->getSceneData());//rootNode.get());

	return shadowedScene;
}

/**
 * This function is called from takeScreenShot(filename).
 */
void ScreenShotCallback::setRequestDataToTrue() {
	mTakeScreenShotNextFrame = true;
}

/**
 * Every time this function is called a call back will actually have consequences.
 * It is called from SimulationCameraController::TakeScreenShot().
 */
void ScreenShotCallback::takeScreenShot(const std::string filename) {
	this->filename = filename;
	setRequestDataToTrue();
}

/**
 * Only one time is the content of this function executed. An actual image will
 * be written again when setRequestDataToTrue() is called in between.
 */
void ScreenShotCallback::operator()(const osg::Camera &camera) const {

	// If screen shot is taken than back off directly
	if(!mTakeScreenShotNextFrame) return;

	mTakeScreenShotNextFrame = false;

	int x = static_cast<int>(camera.getViewport()->x());
	int y = static_cast<int>(camera.getViewport()->y());
	unsigned int width = static_cast<unsigned int>(camera.getViewport()->width());
	unsigned int height = static_cast<unsigned int>(camera.getViewport()->height());

	switch (type) {

	case DEPTH_CAMERA:
		image->allocateImage(width, height, 1, GL_DEPTH_COMPONENT, GL_FLOAT);		//GL_RGB, GL_UNSIGNED_BYTE
		image->readPixels(x, y, width, height, GL_DEPTH_COMPONENT, GL_FLOAT);		//GL_DEPTH_COMPONENT
		break;

	case RGB_CAMERA:
		image->allocateImage(x + width, y + height, 1, GL_RGB, GL_UNSIGNED_BYTE);		//GL_RGB, GL_UNSIGNED_BYTE
		image->readPixels(0, 0, x + width, y + height, GL_RGB, GL_UNSIGNED_BYTE);		//GL_DEPTH_COMPONENT

		if ((filename != "") && (image->getImageSizeInBytes() > 0)) {
			osgDB::writeImageFile(*image, filename);
		}

		break;

	default:
		break;

	}

}


} // end of srCore namespace
