/**
 * An XML parser that parses the configuration defined in the <scene.xml> file that is
 * referenced by the <config.xml> in the root directory. It can contain controllers,
 * robots, etc.
 *
 * @date Nov 25, 2008
 * @author Lutz Winkler, Anne van Rossum
 */

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <list>
#include <cassert>

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "srCore/xmlConfig.h"
#include "srCore/simulationUtils.h"

#include <dtUtil/log.h>
#include <dtUtil/matrixutil.h>

#ifdef WITH_ORGANISM_STRING_CHECK
#include <boost/spirit.hpp>
#include <boost/spirit/actor.hpp>
#endif

#include <osg/Matrix>
#include <dtCore/transform.h>

using namespace xercesc;
using namespace std;

/**
 * Copy by copy constructor
 * See: http://www.fredosaurus.com/notes-cpp/oop-condestructors/copyconstructors.html
 */
RobotInitData::RobotInitData(const RobotInitData& data): name(data.name), type(data.type), pos(data.pos), quat(data.quat),
		connectorStates(data.connectorStates), stateTestFlags(data.stateTestFlags),
		jointAngles(),
		sensors() {
	jointAngles.clear();

	//	LOG_WARNING("Default joint angle is pushed on stack, because of bad range checking later on");
	jointAngles.push_back(0.0);

	sensors.clear();
}

//! Copy by assignment
RobotInitData& RobotInitData::operator=(const RobotInitData& data) {
	LOG_ERROR("Needs to be implemented");
	assert(false);
}


/**
* A singleton pattern for GetConfig.
*/
//GetConfig* GetConfig::m_pInstance = NULL;

/**
*  Constructor initializes xerces-C libraries.
*  The XML tags and attributes which we seek are defined.
*  The xerces-C DOM parser infrastructure is initialized.
*/
GetConfig::GetConfig()
{
	try
	{
		XMLPlatformUtils::Initialize();  // Initialize Xerces infrastructure
	}
	catch( XMLException& e )
	{
		std::ostringstream printmsg;
		printmsg.clear(); printmsg.str("");
		char* message = XMLString::transcode( e.getMessage() );
		printmsg << "XML toolkit initialisation: " << message << endl;
		LOG_ERROR(printmsg.str());
		XMLString::release( &message );
		// throw exception here to return ERROR_XERCES_INIT
	}

	// Tags and attributes used in XML file.
	// Can't call transcode till after Xerces Initialize()


	LOG_INFO("Start reading <config.xml>  file.")

	TAG_scene     = XMLString::transcode("Scene");
	ATTR_activateShadows = XMLString::transcode("activateShadows");
	ATTR_loadDetailedModels = XMLString::transcode("loadDetailedModels");
	ATTR_connectorsAsJoints = XMLString::transcode("ConnectorsAsJoints");
	ATTR_withWheels = XMLString::transcode("RobotModelsWithWheels");
	ATTR_collisionDetails = XMLString::transcode("CollisionGeometryDetails");

	ATTR_map = XMLString::transcode("map");

	TAG_organism    	= XMLString::transcode("Organism");
	ATTR_org_structure 	= XMLString::transcode("structure");
	ATTR_org_x			= XMLString::transcode("x");
	ATTR_org_y			= XMLString::transcode("y");
	ATTR_org_z			= XMLString::transcode("z");
	ATTR_org_h			= XMLString::transcode("h");
	ATTR_org_p			= XMLString::transcode("p");
	ATTR_org_r			= XMLString::transcode("r");
	ATTR_org_quat_x     = XMLString::transcode("q_x");
	ATTR_org_quat_y     = XMLString::transcode("q_y");
	ATTR_org_quat_z     = XMLString::transcode("q_z");
	ATTR_org_quat_w     = XMLString::transcode("q_w");

	TAG_terminal	= XMLString::transcode("Terminal");
	ATTR_termId		= XMLString::transcode("id");

	TAG_robot = XMLString::transcode("Robot");
	ATTR_robotName = XMLString::transcode("name");
	ATTR_robotType = XMLString::transcode("type");
	ATTR_x = XMLString::transcode("x");
	ATTR_y = XMLString::transcode("y");
	ATTR_z = XMLString::transcode("z");
	ATTR_h = XMLString::transcode("h");
	ATTR_r = XMLString::transcode("r");
	ATTR_p = XMLString::transcode("p");
	ATTR_connected = XMLString::transcode("connected");
	ATTR_q0 = XMLString::transcode("q0");
	ATTR_q1 = XMLString::transcode("q1");

	TAG_sensor  = XMLString::transcode("Sensor");
	ATTR_sensorName = XMLString::transcode("name");
	ATTR_sensorType = XMLString::transcode("type");
	ATTR_position_x = XMLString::transcode("x");
	ATTR_position_y = XMLString::transcode("y");
	ATTR_position_z = XMLString::transcode("z");
	ATTR_orientation_x = XMLString::transcode("p");
	ATTR_orientation_y = XMLString::transcode("q");
	ATTR_orientation_z = XMLString::transcode("r");
	ATTR_apertureAngle = XMLString::transcode("aperture");
	ATTR_resolution_x = XMLString::transcode("res_x");
	ATTR_resolution_y = XMLString::transcode("res_y");
	ATTR_bodyPart = XMLString::transcode("bodyPart");

	TAG_controller = XMLString::transcode("Controller");
	ATTR_controllerName = XMLString::transcode("name");
	ATTR_controllerType = XMLString::transcode("type");
	ATTR_robotToControl = XMLString::transcode("robotToControl");
	ATTR_controllerParameter = XMLString::transcode("parameter");

	TAG_plugin = XMLString::transcode("Plugin");
	ATTR_pluginName = XMLString::transcode("name");
	ATTR_pluginPath = XMLString::transcode("path");

	TAG_param = XMLString::transcode("Param");
	ATTR_name = XMLString::transcode("name");
	ATTR_value = XMLString::transcode("value");

	m_ConfigFileParser = new XercesDOMParser;

	controllers.clear();
	robots.clear();

//	controllers.clear();
//	controllers.reserve(4);
//	ostringstream msg; msg.clear(); msg.str("");
//	msg << "We start with a capacity for " << controllers.capacity() << " controllers";
//	LOG_ALWAYS(msg.str());
}

/**
*  Class destructor frees memory used to hold the XML tag and
*  attribute definitions. It als terminates use of the xerces-C
*  framework.
*/

GetConfig::~GetConfig()
{
	LOG_ALWAYS("Free structs defined from <config.xml>  file.");
	try
	{
				XMLPlatformUtils::Terminate();  // Terminate Xerces
	}
	catch( xercesc::XMLException& e )
	{
		std::ostringstream printmsg;
		printmsg.clear(); printmsg.str("");
		char* message = xercesc::XMLString::transcode( e.getMessage() );
		printmsg << "XML toolkit tear down: " << message << endl;
		LOG_ERROR(printmsg.str());
		XMLString::release( &message );
	}


	// Free memory

	try
	{

		if(map) XMLString::release( &map );

		XMLString::release(&(organism.structure) );
		for (unsigned int i=0; i<organism.terminals.size(); i++) {
			XMLString::release(&(organism.terminals[i].id) );
			XMLString::release(&(organism.terminals[i].robot.name) );
			XMLString::release(&(organism.terminals[i].robot.type) );
			for (unsigned int j=0; j < organism.terminals[i].robot.sensors.size(); j++) {
				XMLString::release( &(organism.terminals[i].robot.sensors[j].name) );
				XMLString::release( &(organism.terminals[i].robot.sensors[j].type) );
			}
			if(organism.terminals[i].controller.name != NULL)
			{
			//TODO: release macht aerger wenn kein controller im terminal angegeben wird
				XMLString::release(&(organism.terminals[i].controller.name) );
				XMLString::release(&(organism.terminals[i].controller.type) );
				XMLString::release(&(organism.terminals[i].controller.parameter) );
			}
		}


		for (unsigned int i=0; i<robots.size(); i++) {
			XMLString::release(&(robots[i].name) );
			for (unsigned int j=0; j < robots[i].sensors.size(); j++) {
				XMLString::release( &(robots[i].sensors[j].name) );
				XMLString::release( &(robots[i].sensors[j].type) );
			}
		}

		for (unsigned int i=0; i<controllers.size(); i++) {
			XMLString::release(&(controllers[i].name) );
			XMLString::release(&(controllers[i].type) );
			XMLString::release(&(controllers[i].robot) );
			XMLString::release(&(controllers[i].parameter) );
		}

		XMLString::release( &TAG_scene );
		XMLString::release( &TAG_robot );
		XMLString::release( &TAG_controller );
		XMLString::release( &TAG_plugin );
		XMLString::release( &TAG_sensor );
		XMLString::release( &TAG_organism );
		XMLString::release( &TAG_terminal );
		XMLString::release( &ATTR_activateShadows );
		XMLString::release( &ATTR_loadDetailedModels );
		XMLString::release( &ATTR_map );
		XMLString::release( &ATTR_org_structure );
		XMLString::release( &ATTR_org_x );
		XMLString::release( &ATTR_org_y );
		XMLString::release( &ATTR_org_z );
		XMLString::release( &ATTR_org_h );
		XMLString::release( &ATTR_org_p );
		XMLString::release( &ATTR_org_r );
		XMLString::release( &ATTR_org_quat_x );
		XMLString::release( &ATTR_org_quat_y );
		XMLString::release( &ATTR_org_quat_z );
		XMLString::release( &ATTR_org_quat_w );
		XMLString::release( &ATTR_termId );
		XMLString::release( &ATTR_robotName );
		XMLString::release( &ATTR_robotType );
		XMLString::release( &ATTR_x );
		XMLString::release( &ATTR_y );
		XMLString::release( &ATTR_z );
		XMLString::release( &ATTR_h );
		XMLString::release( &ATTR_p );
		XMLString::release( &ATTR_r );
		XMLString::release( &ATTR_connected );
		XMLString::release( &ATTR_q0 );
		XMLString::release( &ATTR_q1 );
		XMLString::release( &ATTR_controllerName );
		XMLString::release( &ATTR_controllerType );
		XMLString::release( &ATTR_robotToControl );
		XMLString::release( &ATTR_controllerParameter );
		XMLString::release( &ATTR_connectorsAsJoints );
		XMLString::release( &ATTR_withWheels );
		XMLString::release( &ATTR_collisionDetails );
		XMLString::release( &ATTR_pluginName );
		XMLString::release( &ATTR_pluginPath );
		XMLString::release( &ATTR_sensorName );
		XMLString::release( &ATTR_sensorType );
		XMLString::release( &ATTR_position_x );
		XMLString::release( &ATTR_position_y );
		XMLString::release( &ATTR_position_z );
		XMLString::release( &ATTR_orientation_x );
		XMLString::release( &ATTR_orientation_y );
		XMLString::release( &ATTR_orientation_z );
		XMLString::release( &ATTR_apertureAngle );
		XMLString::release( &ATTR_resolution_x );
		XMLString::release( &ATTR_resolution_y );
		XMLString::release( &ATTR_bodyPart );
		XMLString::release( &TAG_param );
		XMLString::release( &ATTR_name );
		XMLString::release( &ATTR_value );
	}
	catch( ... )
	{
		LOG_ERROR("Unknown exception encountered in TagNamesdtor");
	}

	delete m_ConfigFileParser;

}

/**
*  This function:
*  - Tests the access and availability of the XML configuration file.
*  - Configures the xerces-c DOM parser.
*  - Reads and extracts the pertinent information from the XML config file.
*
*  @param in configFile The text string name of the HLA configuration file.
*/

bool atob(char* c) {

	if ((strncmp(c, "1", strlen("1")) == 0)  || (strncmp(c, "true", strlen("true")) == 0))
		return true;
	else return false;

}

/**
 * Reads the configuration file. It also checks if the file exists, etc. No need to do that
 * beforehand.
 * @param configFile
 */
void GetConfig::readConfigFile(string configFile)
throw( std::runtime_error )
{
	// Test to see if the file is ok.
	if (!srCore::SimulationUtils::Instance()->fileExists(configFile)) {
		LOG_ERROR("File does not exist!");
	}

	struct stat fileStatus;

	int iretStat = stat(configFile.c_str(), &fileStatus);
	if( iretStat == ENOENT )
		throw ( std::runtime_error("Path file_name does not exist, or path is an empty string.") );
	else if( iretStat == ENOTDIR )
		throw ( std::runtime_error("A component of the path is not a directory."));
	else if( iretStat == ELOOP )
		throw ( std::runtime_error("Too many symbolic links encountered while traversing the path."));
	else if( iretStat == EACCES )
		throw ( std::runtime_error("Permission denied."));
	else if( iretStat == ENAMETOOLONG )
		throw ( std::runtime_error("File can not be read\n"));

	// Configure DOM parser.

	m_ConfigFileParser->setValidationScheme( XercesDOMParser::Val_Never );
	m_ConfigFileParser->setDoNamespaces( false );
	m_ConfigFileParser->setDoSchema( false );
	m_ConfigFileParser->setLoadExternalDTD( false );

	try
	{
		m_ConfigFileParser->parse( configFile.c_str() );

		// no need to free this pointer - owned by the parent parser object
		DOMDocument* xmlDoc = m_ConfigFileParser->getDocument();

		// Get the top-level element: NAme is "root". No attributes for "root"
		assert( xmlDoc != NULL);
		DOMElement* elementRoot = xmlDoc->getDocumentElement();
		if( !elementRoot ) throw(std::runtime_error( "empty XML document" ));

		// Parse XML file for tags of interest: "ApplicationSettings"
		// Look one level nested within "root". (child of root)


		// loading global simulation parameters
		map = XMLString::transcode(elementRoot->getAttribute(ATTR_map));
		//      char* tmpStr = ;
		withShadows = atob(XMLString::transcode(elementRoot->getAttribute(ATTR_activateShadows)));//tmpStr);

		//      tmpStr =
		withDetailedModels = atob(XMLString::transcode(elementRoot->getAttribute(ATTR_loadDetailedModels)));//);
		//      XMLString::release(&tmpStr);

		connectorsAsJoints = atob( XMLString::transcode(elementRoot->getAttribute(ATTR_connectorsAsJoints)));
		withWheels = atob ( XMLString::transcode(elementRoot->getAttribute(ATTR_withWheels)));

		collisionGeometryDetails = atoi(XMLString::transcode(elementRoot->getAttribute(ATTR_collisionDetails)));


		//loading robots and controllers
		DOMNodeList*      children = elementRoot->getChildNodes();
		const  XMLSize_t nodeCount = children->getLength();

		// For all nodes, children of "scene" in the XML tree.
		for( XMLSize_t xx = 0; xx < nodeCount; ++xx )
		{
			DOMNode* currentNode = children->item(xx);
			if( currentNode->getNodeType() &&  // true is not NULL
					currentNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
			{
				// Found node which is an Element. Re-cast node as element
				DOMElement* currentElement = dynamic_cast< xercesc::DOMElement* >( currentNode );

				if( XMLString::equals(currentElement->getTagName(), TAG_organism))
				{
					organism.structure = XMLString::transcode(currentElement->getAttribute(ATTR_org_structure));
					organism.pos = osg::Vec3f(
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_x))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_y))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_z)))
					);
					if( currentElement->hasAttribute(ATTR_org_h) )
					{
						osg::Vec3f hpr = osg::Vec3f(
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_h))),
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_p))),
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_r)))
						);
						osg::Matrix m;
						dtUtil::MatrixUtil::HprToMatrix(m, hpr);
						organism.quat = m.getRotate();
					}
					if( currentElement->hasAttribute(ATTR_org_quat_x) )
					{
						organism.quat = osg::Quat(
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_quat_x))),
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_quat_y))),
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_quat_z))),
								atof(XMLString::transcode(currentElement->getAttribute(ATTR_org_quat_w)))
						);
					}
					DOMNodeList* orgChildren = currentNode->getChildNodes();
					const  XMLSize_t orgNodeCount = orgChildren->getLength();

					for( XMLSize_t org_i = 0; org_i < orgNodeCount; ++org_i )				//loop terminals
					{
						DOMNode* currentTerminalNode = orgChildren->item(org_i);
						if( currentTerminalNode->getNodeType() &&  // true is not NULL
								currentTerminalNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
						{
							// Found node which is an Element. Re-cast node as element
							DOMElement* currentTerminalElement = dynamic_cast< xercesc::DOMElement* >( currentTerminalNode );

							if( XMLString::equals(currentTerminalElement->getTagName(), TAG_terminal))
							{
								Terminal newTerminal;

								newTerminal.id = XMLString::transcode(currentTerminalElement->getAttribute(ATTR_termId));

								DOMNodeList* termChildren = currentTerminalNode->getChildNodes();
								const  XMLSize_t termNodeCount = termChildren->getLength();
								for( XMLSize_t term_i = 0; term_i < termNodeCount; ++term_i )
								{
									DOMNode* currentTermChildNode = termChildren->item(term_i);

									if( currentTermChildNode->getNodeType() &&  // true is not NULL
											currentTermChildNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
									{
										// Found node which is an Element. Re-cast node as element
										DOMElement* currentTermChildElement = dynamic_cast< xercesc::DOMElement* >( currentTermChildNode );

										if( XMLString::equals(currentTermChildElement->getTagName(), TAG_robot))
										{
											// Already tested node as type element and of name "ApplicationSettings".
											// Read attributes of element "ApplicationSettings".

											newTerminal.robot.name = XMLString::transcode(currentTermChildElement->getAttribute(ATTR_robotName));
											newTerminal.robot.type = XMLString::transcode(currentTermChildElement->getAttribute(ATTR_robotType));
											newTerminal.robot.connectorStates = atoi(XMLString::transcode(currentTermChildElement->getAttribute(ATTR_connected)));

											DOMNodeList* robChildren = currentTermChildNode->getChildNodes();
											const  XMLSize_t robNodeCount = robChildren->getLength();

											for( XMLSize_t rob_i = 0; rob_i < robNodeCount; ++rob_i )
											{
												DOMNode* currentSensorNode = robChildren->item(rob_i);
												if( currentSensorNode->getNodeType() &&  // true is not NULL
														currentSensorNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
												{
													// Found node which is an Element. Re-cast node as element
													DOMElement* currentSensorElement = dynamic_cast< xercesc::DOMElement* >( currentSensorNode );

													if( XMLString::equals(currentSensorElement->getTagName(), TAG_sensor))
													{
														SensorInitData newSensor;
														newSensor.name = XMLString::transcode(currentSensorElement->getAttribute(ATTR_sensorName));
														newSensor.type = XMLString::transcode(currentSensorElement->getAttribute(ATTR_sensorType));
														newSensor.position = osg::Vec3f(
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_x))),
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_y))),
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_z))));
														newSensor.orientation = osg::Vec3f(
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_x))),
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_y))),
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_z))));
														newSensor.apertureAngle = atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_apertureAngle)));
														newSensor.resolution = osg::Vec2f(
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_resolution_x))),
															atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_resolution_y))));
														newSensor.bodyPart = atoi (XMLString::transcode(currentTermChildElement->getAttribute(ATTR_bodyPart)));

														newTerminal.robot.sensors.push_back(newSensor);
													}
												}
											}
										}
										else if( XMLString::equals(currentTermChildElement->getTagName(), TAG_controller))
										{
											newTerminal.controller.name = XMLString::transcode(currentTermChildElement->getAttribute(ATTR_controllerName));
											newTerminal.controller.type = XMLString::transcode(currentTermChildElement->getAttribute(ATTR_controllerType));
											newTerminal.controller.parameter = XMLString::transcode(currentTermChildElement->getAttribute(ATTR_controllerParameter));
											newTerminal.controller.pos = osg::Vec3f(
												atof(XMLString::transcode(currentTermChildElement->getAttribute(ATTR_x))),
												atof(XMLString::transcode(currentTermChildElement->getAttribute(ATTR_y))),
												atof(XMLString::transcode(currentTermChildElement->getAttribute(ATTR_z)))
											);
#ifdef WITH_CONFIG_PARAMS
											DOMNodeList* ctrlChildren = currentTermChildNode->getChildNodes();
											const  XMLSize_t ctrlNodeCount = ctrlChildren->getLength();
											for( XMLSize_t ctrl_i = 0; ctrl_i < ctrlNodeCount; ++ctrl_i )
											{
												DOMNode* currentParamNode = ctrlChildren->item(ctrl_i);
												if( currentParamNode->getNodeType() &&  // true is not NULL
														currentParamNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
												{
													// Found node which is an Element. Re-cast node as element
													DOMElement* currentParamElement = dynamic_cast< xercesc::DOMElement* >( currentParamNode );

													if( XMLString::equals(currentParamElement->getTagName(), TAG_param))
													{
														std::string str_name(XMLString::transcode(currentParamElement->getAttribute(ATTR_name)));
														std::string str_value(XMLString::transcode(currentParamElement->getAttribute(ATTR_value)));
														newTerminal.controller.paramMap.insert(make_pair( str_name, str_value ));
													}
												}
											}
#endif
										}
									}
								}
								organism.terminals.push_back(newTerminal);
							}
						}
					}
				}

				else if( XMLString::equals(currentElement->getTagName(), TAG_robot))
				{

					// Already tested node as type element and of name "ApplicationSettings".
					// Read attributes of element "ApplicationSettings".
					RobotInitData newRobot;

					newRobot.name = XMLString::transcode(currentElement->getAttribute(ATTR_robotName));
					newRobot.type = XMLString::transcode(currentElement->getAttribute(ATTR_robotType));
					newRobot.pos = osg::Vec3f(
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_x))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_y))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_z)))
					);

					std::ostringstream msg; msg.clear(); msg.str("");
					msg << "XML position at " << newRobot.pos.x() << "," << newRobot.pos.y();
					LOG_INFO(msg.str());

					osg::Vec3 hpr = osg::Vec3f(
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_h))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_p))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_r)))
					);

					osg::Matrix m;
					dtUtil::MatrixUtil::HprToMatrix(m, hpr);
					newRobot.quat = m.getRotate();


					newRobot.connectorStates = atoi(XMLString::transcode(currentElement->getAttribute(ATTR_connected)));

					newRobot.jointAngles.push_back(atoi(XMLString::transcode(currentElement->getAttribute(ATTR_q0))));
					newRobot.jointAngles.push_back(atoi(XMLString::transcode(currentElement->getAttribute(ATTR_q1))));

					DOMNodeList* robChildren = currentNode->getChildNodes();
					const  XMLSize_t robNodeCount = robChildren->getLength();

					for( XMLSize_t rob_i = 0; rob_i < robNodeCount; ++rob_i )
					{
						DOMNode* currentSensorNode = robChildren->item(rob_i);
						if( currentSensorNode->getNodeType() &&  // true is not NULL
								currentSensorNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
						{
							// Found node which is an Element. Re-cast node as element
							DOMElement* currentSensorElement = dynamic_cast< xercesc::DOMElement* >( currentSensorNode );

							if( XMLString::equals(currentSensorElement->getTagName(), TAG_sensor))
							{
								SensorInitData newSensor;
								newSensor.name = XMLString::transcode(currentSensorElement->getAttribute(ATTR_sensorName));
								newSensor.type = XMLString::transcode(currentSensorElement->getAttribute(ATTR_sensorType));
								newSensor.position = osg::Vec3f(
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_x))),
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_y))),
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_position_z))));
								newSensor.orientation = osg::Vec3f(
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_x))),
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_y))),
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_orientation_z))));
								newSensor.apertureAngle = atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_apertureAngle)));
								newSensor.resolution = osg::Vec2f(
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_resolution_x))),
										atof(XMLString::transcode(currentSensorElement->getAttribute(ATTR_resolution_y))));
								newSensor.bodyPart = atoi (XMLString::transcode(currentSensorElement->getAttribute(ATTR_bodyPart)));

								newRobot.sensors.push_back(newSensor);
							}
						}
					}


					robots.push_back(newRobot);
				}

				else if( XMLString::equals(currentElement->getTagName(), TAG_controller)) {
					ControllerInitData newController;
					newController.name = XMLString::transcode(currentElement->getAttribute(ATTR_controllerName));
					newController.type = XMLString::transcode(currentElement->getAttribute(ATTR_controllerType));
					newController.robot = XMLString::transcode(currentElement->getAttribute(ATTR_robotToControl));
					newController.parameter = XMLString::transcode(currentElement->getAttribute(ATTR_controllerParameter));
					newController.pos = osg::Vec3f(
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_x))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_y))),
							atof(XMLString::transcode(currentElement->getAttribute(ATTR_z)))
					);

#ifdef WITH_CONFIG_PARAMS
					DOMNodeList* ctrlChildren = currentNode->getChildNodes();
					const  XMLSize_t ctrlNodeCount = ctrlChildren->getLength();
					for( XMLSize_t ctrl_i = 0; ctrl_i < ctrlNodeCount; ++ctrl_i )
					{
						DOMNode* currentParamNode = ctrlChildren->item(ctrl_i);
						if( currentParamNode->getNodeType() &&  // true is not NULL
								currentParamNode->getNodeType() == DOMNode::ELEMENT_NODE ) // is element
						{
							// Found node which is an Element. Re-cast node as element
							DOMElement* currentParamElement = dynamic_cast< xercesc::DOMElement* >( currentParamNode );

							if( XMLString::equals(currentParamElement->getTagName(), TAG_param))
							{
								std::string str_name(XMLString::transcode(currentParamElement->getAttribute(ATTR_name)));
								std::string str_value(XMLString::transcode(currentParamElement->getAttribute(ATTR_value)));
								newController.paramMap.insert(make_pair( str_name, str_value ));
							}
						}
					}
#endif

					int size = controllers.size();
					controllers.push_back(newController);
					size++;
					assert (controllers.size() == (unsigned int)size);
					std::ostringstream msg; msg.clear(); msg.str("");
					msg << "Add controller: " << newController.type;
					if (size == 1)
						msg << " (there is now " << size << " controller)";
					else
						msg << " (there are now " << size << " controllers)";
					LOG_INFO(msg.str());
				}

				else if( XMLString::equals(currentElement->getTagName(), TAG_plugin)) {
					PluginInitData newPlugin;
					newPlugin.name = XMLString::transcode(
							currentElement->getAttribute(ATTR_pluginName));
					newPlugin.path = XMLString::transcode(
							currentElement->getAttribute(ATTR_pluginPath));

					std::ostringstream msg; msg.clear(); msg.str("");
					msg << "There are now " << plugins.size() << " plugins";
					LOG_ALWAYS(msg.str());
					plugins.push_back(newPlugin);
					msg.clear(); msg.str("");
					msg << "There are now " << plugins.size() << " plugins";
					LOG_ALWAYS(msg.str());

				}

			}
		}
#ifdef WITH_ORGANISM_STRING
		if (organism.structure != NULL) processOrganismString();
#endif
	}
	catch( xercesc::XMLException& e )
	{
		char* message = xercesc::XMLString::transcode( e.getMessage() );
		ostringstream errBuf;
		errBuf << "Error parsing file: " << message << flush;
		XMLString::release( &message );
	}
	LOG_INFO("The XML configuration file has been read.");
}

#ifdef WITH_ORGANISM_STRING_CHECK
void GetConfig::structureParser (char* structure)
{
	using namespace boost::spirit;


	rule<> Production;
	rule<> RobotVar   = +(anychar_p -'A'-'B'-'C'-'D'-'['-']'-eol_p-end_p);
	rule<> Number     = chlit<char>('0') | chlit<char>('1') | chlit<char>('2') | chlit<char>('3');
	rule<> Side1       = (chlit<char>('A') | chlit<char>('B') | chlit<char>('C') | chlit<char>('D'));
	rule<> Side2       = (chlit<char>('A') | chlit<char>('B') | chlit<char>('C') | chlit<char>('D')) >> !Number;
	rule<> ProdWBr    = chlit<char>('[') >> Side1 >> Side2 >> RobotVar >> !Production >> chlit<char>(']') >> !Production;
	rule<> ProdWoBr   = Side1 >> Side2 >> RobotVar >> !Production;
	Production 		  = ProdWBr | ProdWoBr;

	rule<> organismStructure = RobotVar >> !Production;

	if(!parse(structure, organismStructure).full)
	{
		cout << "Error: <xmlConfig.cpp>   Organism Structure Parsing Error" << endl;
		exit(0);
	}
}
#endif

#ifdef WITH_ORGANISM_STRING

void GetConfig::processOrganismString()
{
	std::string struc = organism.structure;
	std::string terminal;
	std::vector<int> robotStack;
	std::vector<organismTree::iterator> iterStack;
	int robot = robots.size();

	organismTree orgTree;
	organismTree::iterator iter;

	int si  = 0;
	int rot = 0;
	string type;
	int lastSide = 0;
	short flag   = 0;

#ifdef WITH_ORGANISM_STRING_CHECK
	structureParser(organism.structure);
#endif

	unsigned int ind=1;
	while(ind<struc.length() && struc.at(ind) != '[' && struc.at(ind) != ']' && struc.at(ind) != 'A' && struc.at(ind) != 'B' && struc.at(ind) != 'C' && struc.at(ind) != 'D') ind++;
	terminal = struc.substr(0,ind);
	cout << "Info: <xmlConfig> Parsing...  i: " << 0 << ", Ind: " << ind << ", terminal: " << terminal << "." << endl;


	for(unsigned int i=0; i<organism.terminals.size(); i++)
	{
		if(terminal.compare(organism.terminals[i].id) == 0) {
			RobotInitData newRobot;
			std::ostringstream nameRobot;
			nameRobot << organism.terminals[i].robot.name << robots.size()+1;
			newRobot.name = XMLString::replicate(nameRobot.str().c_str());
			newRobot.type = organism.terminals[i].robot.type;
			newRobot.pos  =	organism.pos;
			newRobot.quat = organism.quat;
			newRobot.jointAngles.push_back(0);
			newRobot.jointAngles.push_back(0);
			newRobot.connectorStates = organism.terminals[i].robot.connectorStates;
			newRobot.stateTestFlags  = 0;
			newRobot.sensors = organism.terminals[i].robot.sensors;

			type = newRobot.type;
			robots.push_back(newRobot);

			if(XMLString::stringLen(organism.terminals[i].controller.type) > 0) {
				ControllerInitData newController;
				std::ostringstream nameController;
				nameController << organism.terminals[i].controller.name << robots.size()+1;
				newController.name  = XMLString::replicate(nameController.str().c_str());
				newController.type  = organism.terminals[i].controller.type;
				newController.robot = robots[robot].name;
				newController.pos   = organism.terminals[i].controller.pos;
				newController.parameter = organism.terminals[i].controller.parameter;
#ifdef WITH_CONFIG_PARAMS
				newController.paramMap  = organism.terminals[i].controller.paramMap;
#endif
				controllers.push_back(newController);

				iter = orgTree.seed(terminal, XMLString::replicate(nameRobot.str().c_str()), XMLString::replicate(nameController.str().c_str()), newRobot.pos, newRobot.quat);
			}
			else
				iter = orgTree.seed(terminal, XMLString::replicate(nameRobot.str().c_str()), "NULL", newRobot.pos, newRobot.quat);

			break;
		}
		else {
			if(i==organism.terminals.size()-1) {
				cout << "Error: <xmlConfig.cpp>   Parsing Error: " << terminal << " is not a terminal" << endl;
				exit(0);
			}
		}
	}


	for(unsigned int i=ind; i<struc.length(); i++)
	{
		terminal = struc.at(i);
		cout<<"Info: <xmlConfig> Parsing...  struc: " << struc << ", i: " << i << ", Terminal: "<< terminal << endl;

		if(terminal.compare("A") == 0) {

			if(lastSide == 0) {
				lastSide = 1;
				if((robots[robot].stateTestFlags & 0x0001) != 0) {
					cout << "Error: <xmlConfig.cpp>   Parsing Error: connector A of robot " << robot << " is already allocated" << endl;
					exit(0);
				}
				else robots[robot].stateTestFlags = robots[robot].stateTestFlags | 0x0001;
			}
			else {
				if(struc.at(i+1)=='0' || struc.at(i+1)=='1'||struc.at(i+1)=='2'||struc.at(i+1)=='3') {
					rot = atoi(&(struc.at(i+1)));
					i++;
				}
				else rot = 0;

				flag   = 0x0001;
				si     = 1;
			}
		}
		else if(terminal.compare("B") == 0) {

			if(lastSide == 0) {
				lastSide = 2;
				if((robots[robot].stateTestFlags & 0x0010) != 0) {
					cout << "Error: <xmlConfig.cpp>   Parsing Error: connector B of robot " << robot << " is already allocated" << endl;
					exit(0);
				}
				else robots[robot].stateTestFlags = robots[robot].stateTestFlags | 0x0010;
			}
			else {
				if(struc.at(i+1)=='0' || struc.at(i+1)=='1'||struc.at(i+1)=='2'||struc.at(i+1)=='3') {
					rot = atoi(&(struc.at(i+1)));
					i++;
				}
				else rot = 0;

				flag   = 0x0010;
				si     = 2;
			}
		}
		else if(terminal.compare("C") == 0) {

			if(lastSide == 0) {
				lastSide = 3;
				if((robots[robot].stateTestFlags & 0x0100) != 0) {
					cout << "Error: <xmlConfig.cpp>   Parsing Error: connector C of robot " << robot << " is already allocated" << endl;
					exit(0);
				}
				else robots[robot].stateTestFlags = robots[robot].stateTestFlags | 0x0100;
			}
			else {
				if(struc.at(i+1)=='0' || struc.at(i+1)=='1'||struc.at(i+1)=='2'||struc.at(i+1)=='3') {
					rot = atoi(&(struc.at(i+1)));
					i++;
				}
				else rot = 0;

				flag   = 0x0100;
				si     = 3;
			}
		}
		else if(terminal.compare("D") == 0) {

			if(lastSide == 0) {
				lastSide = 4;
				if((robots[robot].stateTestFlags & 0x1000) != 0) {
					cout << "Error: <xmlConfig.cpp>   Parsing Error: connector D of robot " << robot << " is already allocated" << endl;
					exit(0);
				}
				else robots[robot].stateTestFlags = robots[robot].stateTestFlags | 0x1000;
			}
			else {
				if(struc.at(i+1)=='0' || struc.at(i+1)=='1'||struc.at(i+1)=='2'||struc.at(i+1)=='3') {
					rot = atoi(&(struc.at(i+1)));
					i++;
				}
				else rot = 0;

				flag   = 0x1000;
				si     = 4;
			}
		}
		else if(terminal.compare("[") == 0) {

			robotStack.push_back(robot);
			iterStack.push_back(iter);
		}
		else if(terminal.compare("]") == 0) {

			robot = robotStack[robotStack.size()-1];
			iter  = iterStack[iterStack.size()-1];
			robotStack.pop_back();
			iterStack.pop_back();
		}
		else {
			unsigned int ind=1;
			while(i+ind<struc.length() && struc.at(i+ind) != '[' && struc.at(i+ind) != ']' && struc.at(i+ind) != 'A' && struc.at(i+ind) != 'B' && struc.at(i+ind) != 'C' && struc.at(i+ind) != 'D') ind++;
			terminal = struc.substr(i,ind);
			cout << "Info: <xmlConfig> Parsing...  i: " << i << ", Ind: " << ind << ", terminal: " << terminal << endl;
			i = i+ind-1;

			for(unsigned int j=0; j<organism.terminals.size(); j++)
			{
				if(terminal.compare(organism.terminals[j].id) == 0)
				{
					RobotInitData newRobot;
					std::ostringstream nameRobot;

					nameRobot << organism.terminals[j].robot.name << (int)robots.size()+1;
					newRobot.name = XMLString::replicate(nameRobot.str().c_str());
					newRobot.type = organism.terminals[j].robot.type;

					computePosQuat(newRobot.type, type, robot, si, lastSide, rot, newRobot.pos, newRobot.quat);
					type = newRobot.type;

					newRobot.jointAngles.push_back(0);
					newRobot.jointAngles.push_back(0);
					newRobot.connectorStates = organism.terminals[j].robot.connectorStates;
					newRobot.stateTestFlags  = flag;
					newRobot.sensors = organism.terminals[j].robot.sensors;

					robots.push_back(newRobot);
					robot = robots.size()-1;

					if(XMLString::stringLen(organism.terminals[j].controller.type) > 0) {
						ControllerInitData newController;
						std::ostringstream nameController;
						nameController << organism.terminals[j].controller.name << (int)controllers.size()+1;
						newController.name  = XMLString::replicate(nameController.str().c_str());
						newController.type  = organism.terminals[j].controller.type;
						newController.robot = robots[robots.size()-1].name;
						newController.pos   = organism.terminals[j].controller.pos;
						newController.parameter = organism.terminals[j].controller.parameter;
#ifdef WITH_CONFIG_PARAMS
						newController.paramMap  = organism.terminals[j].controller.paramMap;
#endif
						controllers.push_back(newController);

						iter = orgTree.append_child(iter, terminal, XMLString::replicate(nameRobot.str().c_str()), XMLString::replicate(nameController.str().c_str()), newRobot.pos, newRobot.quat, si-1, lastSide-1, rot);
					}
					else
						iter = orgTree.append_child(iter, terminal, XMLString::replicate(nameRobot.str().c_str()), "NULL", newRobot.pos, newRobot.quat, si-1, lastSide-1, rot);

					lastSide = 0;

					break;
				}
				else {
					if(j==organism.terminals.size()-1) {
						cout << "Error: <xmlConfig.cpp>   Parsing Error: " << terminal << " is not a terminal" << endl;
						exit(0);
					}
				}
			}
		}
	}

	orgTree.printTree();
//	organismTree::iterator i = orgTree.begin();
//	while (i!=orgTree.end())
//	{
//		cout << "organismTree: robots " << i->robotName << " quat is " << i->quat << endl;
//		++i;
//	}
//	cout << "organismTree: robots " << i->robotName << " quat is " << i->quat << endl;

	orgTree.generUnambigTree();
	string unambigOrgString1 = orgTree.generateString();
	cout << "organismTree: The unambiguous Organism String is: " << unambigOrgString1 << endl;
	cout << "organismTree: new Organism pos:  " << orgTree.getOrgPos() << endl;
	cout << "organismTree: new Organism quat: " << orgTree.getOrgQuat().x() << ", " << orgTree.getOrgQuat().y() << ", " << orgTree.getOrgQuat().z() << ", " << orgTree.getOrgQuat().w() << endl;

//	orgTree.printTree();
//	organismTree::iterator i = orgTree.begin();	i+=1;
//	orgTree.erase(i);
//	orgTree.printTree();
}
#endif

std::vector<osg::Matrix> GetConfig::getSideMatrix(float d)
{
	std::vector<osg::Matrix> side;
	side.push_back(osg::Matrix(1,0,0,0,  0,1,0,0,  0,0,1,0, -d,0,0,1));		//side[0] = A
	side.push_back(osg::Matrix(0,1,0,0,  -1,0,0,0, 0,0,1,0, 0,-d,0,1));		//side[1] = B
	side.push_back(osg::Matrix(-1,0,0,0, 0,-1,0,0, 0,0,1,0, d,0,0,1));		//side[2] = C
	side.push_back(osg::Matrix(0,-1,0,0, 1,0,0,0,  0,0,1,0, 0,d,0,1));		//side[3] = D

	return side;
}

void GetConfig::computePosQuat(string type, string lastType, int robot, int si, int lastSide, int rot, osg::Vec3f &pos, osg::Quat &quat)
{
	float unknown       = 5;
	float robotKIT      = 1.1;
	float robotSSSA_ACD = 1.1;
	float robotSSSA_B   = 1.3;


	osg::Matrix R  = osg::Matrix(-1,0,0,0, 0,-1,0,0, 0,0,1,0,  0,0,0,1);

	std::vector<osg::Matrix> rotation;
	rotation.push_back(osg::Matrix(1,0,0,0, 0,1,0,0,  0,0,1,0,  0,0,0,1));		// rot = 0
	rotation.push_back(osg::Matrix(1,0,0,0, 0,0,1,0,  0,-1,0,0, 0,0,0,1));		// rot = 1
	rotation.push_back(osg::Matrix(1,0,0,0, 0,-1,0,0, 0,0,-1,0, 0,0,0,1));		// rot = 2
	rotation.push_back(osg::Matrix(1,0,0,0, 0,0,-1,0, 0,1,0,0,  0,0,0,1));		// rot = 3

	osg::Matrix parentRobMatrix;
	osg::Matrix childRobMatrix;

	parentRobMatrix.setTrans(robots[robot].pos);
	parentRobMatrix.setRotate(robots[robot].quat);

	osg::Matrix invertedSideMatrix;

	if(lastType.compare("RobotKIT") == 0)
		invertedSideMatrix.invert(getSideMatrix(robotKIT)[lastSide-1]);
	else if (lastType.compare("RobotSSSA") == 0) {
		if(lastSide==2)
			invertedSideMatrix.invert(getSideMatrix(robotSSSA_B)[lastSide-1]);
		else
			invertedSideMatrix.invert(getSideMatrix(robotSSSA_ACD)[lastSide-1]);
	}
	else {
		invertedSideMatrix.invert(getSideMatrix(unknown)[lastSide-1]);
		cout << "Warn: <xmlConfig.cpp>   unknown robot type; can not append robots correctly; " << endl;
	}

	osg::Matrix transformationMatrix;

	if(type.compare("RobotKIT") == 0)
		transformationMatrix = invertedSideMatrix * R * rotation[rot] * getSideMatrix(robotKIT)[si-1];
	else if (type.compare("RobotSSSA") == 0) {
		if(si==2)
			transformationMatrix = invertedSideMatrix * R * rotation[rot] * getSideMatrix(robotSSSA_B)[si-1];
		else
			transformationMatrix = invertedSideMatrix * R * rotation[rot] * getSideMatrix(robotSSSA_ACD)[si-1];
	}
	else {
		transformationMatrix = invertedSideMatrix * R * rotation[rot] * getSideMatrix(unknown)[si-1];
		cout << "Warn: <xmlConfig.cpp>   unknown robot type; can not append robots correctly; " << endl;
	}

	osg::Matrix invertedTransfMatrix;

	invertedTransfMatrix.invert(transformationMatrix);

	childRobMatrix = invertedTransfMatrix * parentRobMatrix;

	pos  = childRobMatrix.getTrans();
	quat = childRobMatrix.getRotate();


//		cout << "parentRobMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(parentRobMatrix);
//		cout << "invertedSideMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(invertedSideMatrix);
//		cout << "rotationMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(rotation[rot]);
//		cout << "SideMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(getSideMatrix(robotKIT)[si-1]);
//		cout << "transformationMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(transformationMatrix);
//		cout << "invertedTransfMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(invertedTransfMatrix);
//		cout << "childRobMatrix: " << endl;
//		dtUtil::MatrixUtil::Print(childRobMatrix);
//		cout << "transformationMatrix: "<< endl;
//		dtUtil::MatrixUtil::Print(transformationMatrix);
}


#ifdef MAIN_TEST
/* This main is provided for unit test of the class. */

int main()
{
	string configFile="scene.xml"; // stat file. Get ambigious segfault otherwise.

	GetConfig appConfig;

	appConfig.readConfigFile(configFile);

	cout << "Map: " << appConfig.getMap() << endl;
	cout << "With Shadows: " << appConfig.getWithShadows() << endl;
	cout << "With Detailed Models: " << appConfig.getWithDetailedModels() << endl;

	cout << "Organism:" << endl;
	cout << "Structure: " << appConfig.getOrganism().structure << endl;
	cout << "Position: " << appConfig.getOrganism().pos.x() << "," << appConfig.getOrganism().pos.y() << "," << appConfig.getOrganism().pos.z() << endl;

	cout << "Terminals: "
	for (int i=0; i< appConfig.getOrganism().terminals.size(); i++) {
		cout << "ID: " << appConfig.getOrganism().terminals[i].id << endl;
		cout << "Robot: " << appConfig.getOrganism().terminals[i].robot.name << " of type " << appConfig.getOrganism().terminals[i].robot.type << endl;
		cout << "Contoller: " << appConfig.getOrganism().terminals[i].controller.name << " of type " << appConfig.getOrganism().terminals[i].controller.type <<
	}

	cout << "Robots:" << endl;
	for (int i=0; i< appConfig.getRobots().size(); i++) {
		cout << appConfig.getRobots()[i].name << " of type " << appConfig.getRobots()[i].type << endl;
		cout << "Position: " << appConfig.getRobots()[i].pos.x() << "," << appConfig.getRobots()[i].pos.y() << "," << appConfig.getRobots()[i].pos.z() << endl;
		cout << "Orientation: " << appConfig.getRobots()[i].hrp.x() << "," << appConfig.getRobots()[i].hrp.y() << "," << appConfig.getRobots()[i].hrp.z() << endl;
	}

	cout << "Controllers:" << endl;
	for (int i=0; i< appConfig.getControllers().size(); i++)
		cout << appConfig.getControllers()[i].name << " of type " << appConfig.getControllers()[i].type <<
		" controlling robot " << appConfig.getControllers()[i].robot << endl;

	cout << "done" <<endl;

	return 0;
}
#endif

