/**
 * @file Agent.cpp
 * @brief There is one Agent running on one robot
 *
 *
 * @author  Lutz Winkler
 * @date
 * @project Replicator FP7
 * @case	YARP
 */

// Delta3D libraries
#include <dtGame/basemessages.h>
#include <dtGame/messagetype.h>
#include <dtDAL/enginepropertytypes.h>
#include <dtABC/application.h>
#include <dtGame/gamemanager.h>
#include <dtUtil/log.h>

// Robot3D libs
#include <srInterface/Agent.h>


namespace srInterface {

////////////////////////////////////////////////////////////////////
/**
 * Function: Agent
 *
 * An agent is a wrapper around a robot that can be used by interface classes to
 * decorate it with certain functionality. For example, if YARP or UART is defined in
 * the configuration files, the corresponding functions become available via the
 * YARPInterface and the UARTInterface classes that both derive from Agent. Each agent
 * is related to one and only one robot. This makes it possible to decorate each robot
 * in the simulator with e.g. a YARP ("/laser") port.
 *
 * It is not necessary to equip this component with (distributed) HLA functionality.
 * Use rather a YARP interface if distributed control is required.
 *
 * ObjectParameter:
 *   name	- Default name for the interface, e.g. YARPInterface
 *   robot	- The robot that this Agent controls
 *
 */
Agent::Agent(const std::string &name, RobotActorBase* robot): dtGame::GMComponent(name),
		paramMap(),
		robot(robot),
		deltaSimTime(0.0),
		lastMessages(),
		manuallySetAngularVel(false) {
}

/*
 * Function: ~Agent
 *
 * Default destructor. Robot is already dereferenced in OnRemovedFromGM.
 */
Agent::~Agent() {

}

/*
 * Function: OnRemovedFromGM
 *
 * Does not (yet) remove the Agent itself. However, its reference to a RobotActorBase
 * entity is already set to NULL.
 */
void Agent::OnRemovedFromGM() {
	clearMsgList();
	robot = NULL;
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Agent " << this->GetName() << " is removed from the game manager";
	LOG_INFO(msg.str());
}

//get robot state:
ParameterReturnValue Agent::getSensorValue(dtCore::RefPtr<SensorBase> sensor)
{
	if (sensor->getReturnValueType() == typeid(unsigned long*).name()) {
		unsigned long value;
		ParameterReturnValue srv = MAKE_PARAMETER_VALUE(&value);
		sensor->evaluate(srv);
		return srv;
	}
	else if (sensor->getReturnValueType() == typeid(osg::Image**).name()) {

		osg::Image* value;
		ParameterReturnValue srv = MAKE_PARAMETER_VALUE(&value);
		sensor->evaluate(srv);
		return srv;
	}
	else {
		int value = 0;
		printf("Warning: sensor return type: %s currently not supported", sensor->getReturnValueType().c_str());
		return MAKE_PARAMETER_VALUE(&value);
	}

}

void Agent::__colorRobot(osg::Vec4 color) {

	robot->colorCollisionGeometry(color);

}

void Agent::move2D(osg::Vec2 velocity) {

	osg::Vec2 ds = velocity * this->getDeltaSimTime();
	float theta = robot->getRotation().x() * 3.1415 / 180;

	osg::Vec3 v = robot->getTranslation() + osg::Vec3(cos(theta) * ds.x() - sin(theta) * ds.y(),
			sin(theta) * ds.x() + sin(theta) * ds.y(), 0);

	robot->setTranslation(robot->getTranslation() + osg::Vec3(cos(theta) * ds.x() - sin(theta) * ds.y(),
			sin(theta) * ds.x() + sin(theta) * ds.y(), 0));
}

void Agent::rotate2D(float angular_v) {
	float dtheta = angular_v * this->getDeltaSimTime();
	robot->setRotation(osg::Vec3(dtheta,0,0));
}

/**
 * The preferred way to set the position of a robot is not by @c setRobotTranslation in
 * the Agent. It should be a function on the robot entity though. Then
 * to ensure that ControllerComponent derivatives actually work on the real robot, they
 * should only direct the actuators. To simulate human intervention it is best to create
 * a new dtGame::GMComponent like SimulationArenaComponent that receives messages from a
 * Agent and responds to it by setting it in a certain position.
 */
void Agent::setRobotTranslation(const osg::Vec3& pos) {
	robot->setTranslation(pos);
}

/**
 * @see setRobotTranslation.
 */
void Agent::setRobotRotation(const osg::Vec3& rot) {
	robot->setRotation(rot);
}

float Agent::getActuatorPosition(dtCore::RefPtr<ActuatorBase> actuator)
{
	return actuator->getPos();
}

float Agent::getActuatorVelocity(dtCore::RefPtr<ActuatorBase> actuator)
{
	return actuator->getVel(deltaSimTime);
}

//float Agent::getActuatorAcceleration(ActuatorBase* actuator)
//{
//
//}

int Agent::getConnectorState(dtCore::RefPtr<ConnectorBase> connector)
{
	return connector->getState();
}

/**
 * Just relays the connect request to the given ConnectorBase.
 */
void Agent::connect(dtCore::RefPtr<ConnectorBase> connector) {
	connector->connect();
}

void Agent::disconnect(dtCore::RefPtr<ConnectorBase> connector)
{
	connector->disconnect();
}

//void Agent::setActuatorPosition(dtCore::RefPtr<ActuatorBase> actuator, float value) {
//	if (dynamic_cast<HingeActuator*>(actuator.get()))
//		dynamic_cast<HingeActuator*>(actuator.get())->gotoAngle(value);
//}

void Agent::setActuatorVelocity(dtCore::RefPtr<ActuatorBase> actuator, float value) {
	actuator->setVel(value, deltaSimTime);
}

//void Agent::setActuatorAcceleration(ActuatorBase* actuator, float value)
//{
//
//}

/**
 * Send a message to everyone, not just to the swarm, also to components in the simulator.
 */
void Agent::sendGlobalMessage(const std::string& msg) {
	sendMessage(msg, SimulationMessageType::GLOBAL, const_cast<dtCore::UniqueId* >(&robot->GetUniqueId()));
}

/**
 * Send a local message to the robot connected to this robot by the connector with id @c connectorID.
 * The range of connectorID is checked and it is checked if there is actually a dual connector (of the other
 * robot) on the other side.
 */
void Agent::sendLocalMessage(const std::string& msg, unsigned int connectorID) {
	if (connectorID < robot->getConnectors().size())
		if (robot->getConnectors()[connectorID]->getOtherConnector().get()) {
			sendMessage(msg, SimulationMessageType::LOCAL,
					robot->getConnectors()[connectorID]->getOtherConnector()->getID());
		}
}

/**
 * We don't know how to resolve connectorName%s yet. Don't use this function.
 * @deprecated
 */
void Agent::sendLocalMessage(const std::string& msg, const std::string& connectorName) {

}

/**
 * To send the organism a message. The identifier will be set to the UniqueID (bus id) of the robot,
 * not of the controller component base itself.
 */
void Agent::sendOrganismMessage(const std::string& msg) {
	sendMessage(msg, SimulationMessageType::ORGANISM, robot->getBusID());
}

/**
 * Helper function for the different types of messages that can be sent.
 */
void Agent::sendMessage(const std::string& msg,
		SimulationMessageType::commType type, dtCore::UniqueId* commID) {
	dtCore::RefPtr<CommMessage> commMsg = new CommMessage();
	GetGameManager()->GetMessageFactory().CreateMessage(SimulationMessageType::COMM_BUS, commMsg);
	commMsg->setBusID(commID);
	commMsg->setMessage(msg);
	commMsg->setCommType(type);
	GetGameManager()->SendMessage(*commMsg);

}

//////////////////////////////////////////////////////////////////////////

/**
 * This is the routine that you can overload in your own robot main controller. Be aware that
 * in the end multiple of those Agent entities need to run in parallel on
 * the robot. So, actually subcontroller entities will be needed and an architecture between
 * them like the Task entities and the hooks in the Robot3DRTOS operating system.
 *
 * Don't forget to call this routine if you overload it. It will do two things. First, it will
 * set the angular velocity of all body parts of the robot to 0. And it will conveniently push
 * messages on a local lastMessages stack. This is important to know, because you will need to
 * empty that stack by clearMsgList().
 */
void Agent::ProcessMessage(const dtGame::Message& message)
{
	GMComponent::ProcessMessage(message);

	if (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL)
	{
		const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(message);
		deltaSimTime = tick.GetDeltaSimTime();

		for (unsigned int i=0; i < this->robot->getBodies().size(); i++) {
			dBodySetAngularVel(robot->getBodies()[i]->GetBodyID(), 0, 0, 0);
			//dBodySetLinearVel(robot->getBodies()[i]->GetBodyID(), 0, 0, 0);
		}
	}


	if (message.GetMessageType() == SimulationMessageType::COMM_BUS)
	{

		dtCore::UniqueId* busID = dynamic_cast<CommMessage*>(
				const_cast<dtGame::Message*>(&message))->getBusID();
		std::string msg = dynamic_cast<CommMessage*>(
				const_cast<dtGame::Message*>(&message))->getMessage();
		SimulationMessageType::commType type = dynamic_cast<CommMessage*>(
				const_cast<dtGame::Message*>(&message))->getCommType();

		SimMessage message;

		switch (type) {

		case SimulationMessageType::ORGANISM:

			//global bus communication
			if (busID == robot->getBusID()) {
				//printf("robot no. %ld received comm message: %s on bus id %s, my bus id: %s \n", this->getID(), msg.c_str(), busID.ToString().c_str(), globalCommID.ToString().c_str());
				message.data = msg;
				message.type = SimulationMessageType::ORGANISM;
				message.port = -1;
				lastMessages.push_back(message);
			}
			break;

		case SimulationMessageType::LOCAL:

			//local communication between connectors
			for (unsigned int i=0; i<robot->getConnectors().size(); i++) {
				if (busID == robot->getConnectors()[i]->getID()) {
					message.data = msg;
					message.type = SimulationMessageType::LOCAL;
					message.port = i;
					lastMessages.push_back(message);
				}
			}
			break;


		case SimulationMessageType::GLOBAL:

			message.data = msg;
			message.type = SimulationMessageType::GLOBAL;
			message.port = -1;
			lastMessages.push_back(message);
			break;

		default:
			break;

		}

	}
}

} // end of namespace


