/**
 * @file YarpRobotNavigation.cpp
 * @brief Module to allow for navigation over YARP of a robot in Robot3D
 *
 * This file YarpRobotNavigation.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Andrei Rusu <andrei@almende.org>
 *
 * @author	Andrei A. Rusu
 * @date	Jun 15, 2010
 * @project	Replicator FP7
 * @company	Almende B.V.
 * @case	YARP
 */

// General files
#include <iostream>
#include <iomanip>
#include <deque>

// YARP
#include <yarp/sig/all.h>
#include <yarp/os/all.h>
#include <yarp/dev/all.h>

// Robot3D files
#include <srInterface/YarpRobotNavigation.h>

using namespace srCore;

namespace srInterface {

/* **************************************************************************************
 * Protocol vocab commands
 * **************************************************************************************/

#define RPC_VOCAB_HELP VOCAB4('h','e','l','p')
#define RPC_VOCAB_SET VOCAB3('s','e','t')
#define RPC_VOCAB_GET VOCAB3('g','e','t')
#define RPC_VOCAB_IS VOCAB2('i','s')
#define RPC_VOCAB_FAILED VOCAB4('f','a','i','l')
#define RPC_VOCAB_OK VOCAB2('o','k')
#define RPC_VOCAB_ADV VOCAB3('a','d','v')
#define RPC_VOCAB_ROT VOCAB3('r','o','t')
#define RPC_VOCAB_UADV VOCAB4('u','a','d','v')
#define RPC_VOCAB_UROT VOCAB4('u','r','o','t')
#define RPC_VOCAB_CLR VOCAB3('c','l','r')
#define RPC_VOCAB_QLMT VOCAB4('q','l','m','t')
#define RPC_VOCAB_NOW VOCAB3('n','o','w')
#define RPC_VOCAB_SYNC VOCAB4('s','y','n','c')
#define RPC_VOCAB_STRF VOCAB4('s','t','r','f')

/* **************************************************************************************
 * Implementation of YarpRobotNavigation
 * **************************************************************************************/

/**
 * Default constructor, the thing can best be stored in RefPtr.
 */
YarpRobotNavigation::YarpRobotNavigation(const std::string &name, RobotActorBase *robotActor) : Agent(name, robotActor), Thread(),
		yarp(),
		rpcPort(),
		rpcMutex(),
		rpcThread(),
		mainCommandQueue(),
		syncCommandQueues(),
		commandQueueMutex(),
		RPC_ADV_UNIT(0.1), RPC_ROT_UNIT(1.0), RPC_QLMT(10000) {
	osg::Vec3 rot = getRobotRotation();
	rot[0] = -1;
	setRobotRotation(rot);

	LOG_INFO("Construct robot navigation component");
}

YarpRobotNavigation::~YarpRobotNavigation() {
	LOG_INFO("Destruct robot navigation component");
}

void YarpRobotNavigation::SetPort(const std::string &portPrefix) {
	std::string portName = portPrefix + "/navigation";
	if (!rpcPort.open(portName.c_str())) {
		std::stringstream msg; msg.clear(); msg.str("");
		msg << "Could not open RpcPort name: " << portName << ". Exiting...";
		LOG_ERROR(msg.str());
		LOG_ALWAYS("The RPC thread will not be started");
		return;
	}

	// start RPC thread
	rpcThread = new YarpRobotNavigationRpcThread(this);
	rpcThread->start();

}

//#define DEBUG_AGGREGATE

bool YarpRobotNavigation::respond(
		const std::list<yarp::os::Bottle> &commandList, yarp::os::Bottle &reply) {

	// clear reply
	reply.clear();

	// quick exit on empty command list
	if (commandList.size() == 0) {

#ifdef DEBUG_AGGREGATE
		std::stringstream ossLog;
		ossLog << "Empty command list.";
		LOG_INFO(ossLog.str().c_str());
#endif //DEBUG_AGGREGATE
		reply.addVocab(RPC_VOCAB_OK);
		return true;
	}

	// get exclusive rights to resource
	rpcMutex.wait();

	bool ok = false;
	// aggregate advancement
	double aggregateAdv = 0.0;
	// aggregate strafe
	double aggregateStrf = 0.0;
	// aggregate rotation
	double aggregateRot = 0.0;

#ifdef DEBUG_AGGREGATE
	std::stringstream ossLog;

	ossLog << "Aggregating " << commandList.size() << " commands." << endl;
#endif //DEBUG_AGGREGATE
	std::list<yarp::os::Bottle>::const_iterator command = commandList.begin();

	// evaluate all commands
	while (command != commandList.end()) {
#ifdef DEBUG_AGGREGATE
		ossLog << command->toString() << endl;
#endif //DEBUG_AGGREGATE
		ok = false;
		switch (command->get(0).asVocab()) {
		case RPC_VOCAB_ADV:
			aggregateAdv += (command->get(1).asDouble());
			ok = true;
			break;
		case RPC_VOCAB_STRF:
			aggregateStrf += (command->get(1).asDouble());
			ok = true;
			break;
		case RPC_VOCAB_ROT:
			aggregateRot += (command->get(1).asDouble());
			ok = true;
			break;
		}
		// if one command is not ok, break operation
		if (!ok) {
			break;
		}
		// advance to the next command
		++command;
	}

	if (!ok) {
		reply.addVocab(RPC_VOCAB_FAILED);
	} else {
#ifdef DEBUG_AGGREGATE
		ossLog << "Executing commands: adv " << aggregateAdv << "; rot: "
				<< aggregateRot << "; strf: "<<aggregateStrf << endl;
		LOG_INFO(ossLog.str().c_str());
#endif //DEBUG_AGGREGATE
		// execute aggregate commands
		if (abs(aggregateAdv) > 0.0)
			advanceX(aggregateAdv);
		if (abs(aggregateStrf) > 0.0)
			strafeX(aggregateStrf);
		if (abs(aggregateRot) > 0.0)
			rotateX(aggregateRot);
		reply.addVocab(RPC_VOCAB_OK);
	}

	// release lock
	rpcMutex.post();

	return ok;
}

bool YarpRobotNavigation::respond(const yarp::os::Bottle &command,
		yarp::os::Bottle &reply) {

	// get exclusive rights to resource
	rpcMutex.wait();

	bool ok = false;
	bool rec = false; // is the command recognized?
	double argDouble;

	switch (command.get(0).asVocab()) {
	case RPC_VOCAB_STRF:
		rec = true;
		argDouble = command.get(1).asDouble();
		strafeX(argDouble);
		reply.addVocab(RPC_VOCAB_STRF);
		reply.addDouble(argDouble);
		ok = true;
		break;
	case RPC_VOCAB_ADV:
		rec = true;
		argDouble = command.get(1).asDouble();
		advanceX(argDouble);
		reply.addVocab(RPC_VOCAB_ADV);
		reply.addDouble(argDouble);
		ok = true;
		break;
	case RPC_VOCAB_ROT:
		rec = true;
		argDouble = command.get(1).asDouble();
		rotateX(argDouble);
		reply.addVocab(RPC_VOCAB_ROT);
		reply.addDouble(argDouble);
		ok = true;
		break;

	}

	if (!ok) {
		reply.clear();
		reply.addVocab(RPC_VOCAB_FAILED);
	} else
		reply.addVocab(RPC_VOCAB_OK);

	// release lock
	rpcMutex.post();

	return ok;
}

//////////////////////////////////////////////////////////////////////////
void YarpRobotNavigation::ProcessMessage(const dtGame::Message& message) {

	Agent::ProcessMessage(message);

	if (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL) {
		resume();
	}

	// Evaluate front left distance sensor
	//	unsigned long val;
	//	DistanceSensor *flDistanceSensor =
	//			dynamic_cast<DistanceSensor*> (getSensor(
	//					"Front Left DistanceSensor").get());
	//	flDistanceSensor->evaluate(MAKE_PARAMETER_VALUE(&(val)));
	//
	//	// Evaluate front right distance sensor
	//	DistanceSensor *frDistanceSensor =
	//			dynamic_cast<DistanceSensor*> (getSensor(
	//					"Front Right DistanceSensor").get());
	//	frDistanceSensor->evaluate(MAKE_PARAMETER_VALUE(&(val)));


	std::list<yarp::os::Bottle> commandList;

	// execute a command from the command queue, if any
	if (commandQueueMutex.check()) {
		//		LOG_ALWAYS("Lock queue to check size.");
		if (mainCommandQueue.size() > 0) {
			//			LOG_ALWAYS("Executing first command in queue.");
			// handle the call using the controller's respond method
			commandList.push_back(mainCommandQueue.front());
			// send replay back
			mainCommandQueue.pop_front();
			//			LOG_ALWAYS("First command in queue executed.");
		}

		// execute commands from remaining sync queues and remove the empty ones
		std::list<std::deque<yarp::os::Bottle> >::iterator qsit =
				syncCommandQueues.begin();

#ifdef DEBUG_QUEUES
		{
			std::stringstream oss;
			oss << "Size of MAIN queue: " << mainCommandQueue.size();
			LOG_ALWAYS(oss.str());
		}

		{
			std::stringstream oss;
			oss << "Sync queue count: " << syncCommandQueues.size();
			LOG_ALWAYS(oss.str());
		}

		std::stringstream ossLog;

		if (qsit != syncCommandQueues.end())
			ossLog << "Size of sync queue: ";
#endif //DEBUG_QUEUES
		// execute a command from all non-empty synchronous queues,
		// remove empty ones
		while (qsit != syncCommandQueues.end()) {

#ifdef DEBUG_QUEUES
			ossLog << qsit->size() << " ";
#endif //DEBUG_QUEUES
			if (qsit->empty()) {
				// erase empty queue, move forward
				qsit = syncCommandQueues.erase(qsit);
			} else {
				// commands still present in this sync queue, execute
				commandList.push_back(qsit->front());
				// send replay back
				qsit->pop_front();
				++qsit;
			}
#ifdef DEBUG_QUEUES
			if (qsit == syncCommandQueues.end())
				LOG_ALWAYS(ossLog.str());
#endif //DEBUG_QUEUES
		}

		if (commandList.size() > 0) {
			/// EXECUTE COMMAND QUEUE
			yarp::os::Bottle reply;
			respond(commandList, reply);
			if (reply.get(0).asVocab() != RPC_VOCAB_OK) {
				{
					std::stringstream oss;
					oss
					<< "Respond() issued a failure reply. Command list not correct: "
					<< reply.toString() << std::endl;
					LOG_ALWAYS(oss.str());
				}
			}
		}

		// release queues
		commandQueueMutex.post();
	}
}

//#define DEBUG

void YarpRobotNavigation::setSpeed(float speedL, float speedR) {

	float dt = getDeltaSimTime();

	osg::Vec3 trans = getRobotPosition();
	osg::Vec3 rot = getRobotRotation();

#ifdef DEBUG
	std::ostringstream oss;
	oss << "Trans: " << trans << "; Rot: " << rot << endl;
#endif

	float d = dt * (speedL + speedR) / 2.0;
	float phi = dt * (speedR - speedL) / 0.001;

#ifdef DEBUG
	oss << "Phi: " << phi << "; Dt: " << dt << endl;
#endif
	trans[0] += (d * cos(phi));
	trans[1] += (d * sin(phi));

	rot[0] = phi;

#ifdef DEBUG
	oss << "NEW Trans: " << trans << "; NEW Rot: " << rot << endl;
	LOG_INFO(oss.str());
#endif

	setRobotRotation(rot);
	setRobotTranslation(trans);
}

void YarpRobotNavigation::advanceX(float X) {
	osg::Vec3 trans = getRobotPosition();
	osg::Vec3 rot = getRobotRotation();

	rot[0] += 90;

	const float angleRad = ((rot[0] > 0 ? rot[0] : (360 + rot[0])) * M_PI
			/ 180.0);

	trans[0] += (X * cos(angleRad));
	trans[1] += (X * sin(angleRad));

	setRobotTranslation(trans);
}

void YarpRobotNavigation::strafeX(float X) {
	osg::Vec3 trans = getRobotPosition();
	osg::Vec3 rot = getRobotRotation();

	rot[0] += 180;

	const float angleRad = ((rot[0] > 0 ? rot[0] : (360 + rot[0])) * M_PI
			/ 180.0);

	trans[0] += (X * cos(angleRad));
	trans[1] += (X * sin(angleRad));

	setRobotTranslation(trans);
}

void YarpRobotNavigation::rotateX(float X) {
	osg::Vec3 rot = getRobotRotation();

#ifdef DEBUG
	std::ostringstream oss;
	oss << "Rot: " << rot << endl;
#endif

	rot[0] += X;

#ifdef DEBUG
	oss << "NEW Rot: " << rot << endl;
	LOG_INFO(oss.str());
#endif
	setRobotRotation(rot);
}

void YarpRobotNavigation::OnAddedToGM() {
	Agent::OnAddedToGM();
	start();
	LOG_INFO("Navigation added to GM");
}

void YarpRobotNavigation::OnRemovedFromGM() {
	Agent::OnRemovedFromGM();
	exit();
	LOG_INFO("Navigation removed from GM");
}

void YarpRobotNavigation::task() {
	while (true) {
		wait();
		suspend();
	}
}


/* **************************************************************************************
 * Implementation of YarpRobotNavigationRpcThread
 * **************************************************************************************/

void YarpRobotNavigationRpcThread::run() {
	yarp::os::Bottle command, replay;

	// repeat while this thread is not asked to stop
	while (isStopping() != true) {
		// do a blocking read from the rpc port
		if (ctrl->rpcPort.read(command, true)) {
			// confirm reception
			yarp::os::Bottle reply;

			// pass the command to the sequencer
			sequencer(command, reply);

			ctrl->rpcPort.reply(reply);
		}
	}

}

void YarpRobotNavigationRpcThread::sequencer(yarp::os::Bottle &command,
		yarp::os::Bottle &reply) {

	// wait for exclusive privileges to the queue
	ctrl->commandQueueMutex.wait();

	//	{
	//		std::stringstream oss;
	//		oss << "Inserting command: " << command.toString()
	//				<< " into queue of size: " << (ctrl->commandQueue.size())
	//				<< endl;
	//		LOG_ALWAYS(oss.str());
	//	}

	bool ok = false;
	bool rec = false; // is the command recognized?
	double argDouble;
	std::list<std::deque<yarp::os::Bottle> >::iterator qsit;

	yarp::os::Bottle subCommand;
	unsigned int repeats = 0;

	switch (command.get(0).asVocab()) {
	case RPC_VOCAB_HELP: {
		rec = true;
		reply.addString("Commands:");
		reply.addString(
				"adv <double> // enqueues advance commands with a total effect of <double> simulator length units at the BACK of the main queue ******");
		reply.addString(
				"strf <double> // enqueues strafe commands with a total effect of <double> simulator length units at the BACK of the main queue ******");
		reply.addString(
				"rot <double> // enqueues rotate commands with a total effect of <double> simulator rotation units at the BACK of the main queue ******");
		reply.addString(
				"now adv <double> // enqueues  rotate commands with a total effect of <double> simulator length units at the FRONT of the main queue ******");
		reply.addString(
				"now strf <double> // enqueues  strafe commands with a total effect of <double> simulator length units at the FRONT of the main queue ******");
		reply.addString(
				"now rot <double> // enqueues  rotate commands with a total effect of <double> simulator rotation units at the FRONT of the main queue ******");
		reply.addString(
				"sync adv <double> // enqueues  rotate commands with a total effect of <double> simulator length units in a new queue, which is executed in parallel with the main queue ******");
		reply.addString(
				"sync strf <double> // enqueues  strafe commands with a total effect of <double> simulator length units in a new queue, which is executed in parallel with the main queue ******");
		reply.addString(
				"sync rot <double> // enqueues  rotate commands with a total effect of <double> simulator rotation units in a new queue, which is executed in parallel with the main queue ******");
		reply.addString(
				"set uadv <double> // sets a new value for advancement unit, executed in a ProcessMessage call; this affects the speed of advancement ******");
		reply.addString(
				"set urot <double> // sets a new value for rotation unit, executed in a ProcessMessage call; this affects the speed of rotation ******");
		reply.addString(
				"set qlmt <double> // modifies the maximum length of all execution queues; queues above this size are trimmed from the front end ******");
		reply.addString(
				"get uadv // gets the current value for advancement unit ******");
		reply.addString(
				"get urot // gets current value for rotation unit ******");
		reply.addString(
				"get qlmt // gets current value for maximum queue length ******");
		reply.addString("clr // clears all execution queues ******");

		ok = true;
	}
	break;
	case RPC_VOCAB_CLR: {
		rec = true;
		reply.addVocab(RPC_VOCAB_CLR);
		// clear main queue
		ctrl->mainCommandQueue.clear();
		qsit = ctrl->syncCommandQueues.begin();
		// clear all synchronous queues
		while (qsit != ctrl->syncCommandQueues.end()) {
			qsit->clear();
			++qsit;
		}

		ok = true;
	}
	break;
	case RPC_VOCAB_SYNC: {
		rec = true;
		reply.addVocab(RPC_VOCAB_SYNC);
		switch (command.get(1).asVocab()) {
		case RPC_VOCAB_STRF:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

			// create a new synchronous queue
			ctrl->syncCommandQueues.push_back(std::deque<yarp::os::Bottle>());

			// insert sub-commands into the new sync queue
			while (repeats-- > 0)
				ctrl->syncCommandQueues.back().push_back(subCommand);

			reply.addVocab(RPC_VOCAB_STRF);
			reply.addDouble(argDouble);
			ok = true;
			break;
		case RPC_VOCAB_ADV:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

			// create a new synchronous queue
			ctrl->syncCommandQueues.push_back(std::deque<yarp::os::Bottle>());

			// insert sub-commands into the new sync queue
			while (repeats-- > 0)
				ctrl->syncCommandQueues.back().push_back(subCommand);

			reply.addVocab(RPC_VOCAB_ADV);
			reply.addDouble(argDouble);
			ok = true;
			break;
		case RPC_VOCAB_ROT:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ROT_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ROT_UNIT);

			// create a new synchronous queue
			ctrl->syncCommandQueues.push_back(std::deque<yarp::os::Bottle>());

			// insert sub-commands into the new sync queue
			while (repeats-- > 0)
				ctrl->syncCommandQueues.back().push_back(subCommand);

			reply.addVocab(RPC_VOCAB_ROT);
			reply.addDouble(argDouble);
			ok = true;
			break;
		}
	}
	break;
	case RPC_VOCAB_SET: {
		rec = true;
		reply.addVocab(RPC_VOCAB_SET);
		switch (command.get(1).asVocab()) {
		case RPC_VOCAB_UADV: {
			rec = true;
			reply.addVocab(RPC_VOCAB_UADV);
			ctrl->RPC_ADV_UNIT = command.get(2).asDouble();
			ok = true;
		}
		break;
		case RPC_VOCAB_UROT: {
			rec = true;
			reply.addVocab(RPC_VOCAB_UROT);
			ctrl->RPC_ROT_UNIT = command.get(2).asDouble();
			ok = true;
		}
		break;
		case RPC_VOCAB_QLMT: {
			rec = true;
			reply.addVocab(RPC_VOCAB_QLMT);
			ctrl->RPC_QLMT = (unsigned int) command.get(2).asInt();
			ok = true;
		}
		break;
		}
	}
	break;
	case RPC_VOCAB_GET: {
		rec = true;
		switch (command.get(1).asVocab()) {
		case RPC_VOCAB_UADV: {
			rec = true;
			reply.addVocab(RPC_VOCAB_UADV);
			reply.addDouble(ctrl->RPC_ADV_UNIT);
			ok = true;
		}
		break;
		case RPC_VOCAB_UROT: {
			rec = true;
			reply.addVocab(RPC_VOCAB_UROT);
			reply.addDouble(ctrl->RPC_ROT_UNIT);
			ok = true;
		}
		break;
		case RPC_VOCAB_QLMT: {
			rec = true;
			reply.addVocab(RPC_VOCAB_QLMT);
			reply.addInt(ctrl->RPC_QLMT);
			ok = true;
		}
		break;
		}
	}
	break;

	case RPC_VOCAB_NOW: {
		rec = true;
		reply.addVocab(RPC_VOCAB_NOW);
		switch (command.get(1).asVocab()) {
		case RPC_VOCAB_STRF:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

			// insert sub-commands into the queue
			while (repeats-- > 0)
				ctrl->mainCommandQueue.push_front(subCommand);

			reply.addVocab(RPC_VOCAB_STRF);
			reply.addDouble(argDouble);
			ok = true;
			break;
		case RPC_VOCAB_ADV:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

			// insert sub-commands into the queue
			while (repeats-- > 0)
				ctrl->mainCommandQueue.push_front(subCommand);

			reply.addVocab(RPC_VOCAB_ADV);
			reply.addDouble(argDouble);
			ok = true;
			break;
		case RPC_VOCAB_ROT:
			argDouble = command.get(2).asDouble();

			// break down the command in sub-commands
			subCommand.clear();
			subCommand.addVocab(command.get(1).asVocab());

			repeats = (unsigned int) abs(argDouble / ctrl->RPC_ROT_UNIT);

			// cannot introduce more commands than the maximum queue size
			if (repeats > ctrl->RPC_QLMT) {
				std::stringstream oss;
				oss << "Asking to introduce more than " << ctrl->RPC_QLMT
						<< " commands in the queue: " << repeats
						<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
				LOG_ALWAYS(oss.str());

				// introduce only the maximum queue size
				repeats = ctrl->RPC_QLMT;
			}

			subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ROT_UNIT);

			// insert sub-commands into the queue
			while (repeats-- > 0)
				ctrl->mainCommandQueue.push_front(subCommand);

			reply.addVocab(RPC_VOCAB_ROT);
			reply.addDouble(argDouble);
			ok = true;
			break;
		}
	}
	break;
	case RPC_VOCAB_STRF:
		rec = true;
		argDouble = command.get(1).asDouble();

		// break down the command in sub-commands
		subCommand.clear();
		subCommand.addVocab(command.get(0).asVocab());

		repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

		// cannot introduce more commands than the maximum queue size
		if (repeats > ctrl->RPC_QLMT) {
			std::stringstream oss;
			oss << "Asking to introduce more than " << ctrl->RPC_QLMT
					<< " commands in the queue: " << repeats
					<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
			LOG_ALWAYS(oss.str());

			// introduce only the maximum queue size
			repeats = ctrl->RPC_QLMT;
		}

		subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

		// insert sub-commands into the queue
		while (repeats-- > 0)
			ctrl->mainCommandQueue.push_back(subCommand);

		reply.addVocab(RPC_VOCAB_STRF);
		reply.addDouble(argDouble);
		ok = true;
		break;
	case RPC_VOCAB_ADV:
		rec = true;
		argDouble = command.get(1).asDouble();

		// break down the command in sub-commands
		subCommand.clear();
		subCommand.addVocab(command.get(0).asVocab());

		repeats = (unsigned int) abs(argDouble / ctrl->RPC_ADV_UNIT);

		// cannot introduce more commands than the maximum queue size
		if (repeats > ctrl->RPC_QLMT) {
			std::stringstream oss;
			oss << "Asking to introduce more than " << ctrl->RPC_QLMT
					<< " commands in the queue: " << repeats
					<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
			LOG_ALWAYS(oss.str());

			// introduce only the maximum queue size
			repeats = ctrl->RPC_QLMT;
		}

		subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ADV_UNIT);

		// insert sub-commands into the queue
		while (repeats-- > 0)
			ctrl->mainCommandQueue.push_back(subCommand);

		reply.addVocab(RPC_VOCAB_ADV);
		reply.addDouble(argDouble);
		ok = true;
		break;
	case RPC_VOCAB_ROT:
		rec = true;
		argDouble = command.get(1).asDouble();

		// break down the command in sub-commands
		subCommand.clear();
		subCommand.addVocab(command.get(0).asVocab());

		repeats = (unsigned int) abs(argDouble / ctrl->RPC_ROT_UNIT);

		// cannot introduce more commands than the maximum queue size
		if (repeats > ctrl->RPC_QLMT) {
			std::stringstream oss;
			oss << "Asking to introduce more than " << ctrl->RPC_QLMT
					<< " commands in the queue: " << repeats
					<< ". Trimming down to: " << ctrl->RPC_QLMT << std::endl;
			LOG_ALWAYS(oss.str());

			// introduce only the maximum queue size
			repeats = ctrl->RPC_QLMT;
		}

		subCommand.addDouble((argDouble > 0 ? 1 : -1) * ctrl->RPC_ROT_UNIT);

		// insert sub-commands into the queue
		while (repeats-- > 0)
			ctrl->mainCommandQueue.push_back(subCommand);

		reply.addVocab(RPC_VOCAB_ROT);
		reply.addDouble(argDouble);
		ok = true;
		break;

	}

	if (!ok) {
		reply.clear();
		reply.addVocab(RPC_VOCAB_FAILED);
	} else {
		reply.addVocab(RPC_VOCAB_OK);

		// log when discarding commands
		if (ctrl->mainCommandQueue.size() > ctrl->RPC_QLMT) {
			std::stringstream oss;
			oss << "More than " << ctrl->RPC_QLMT << " commands in the queue: "
					<< (ctrl->mainCommandQueue.size()) << ". Clearing oldest."
					<< std::endl;
			LOG_ALWAYS(oss.str());

			// if more than RPC_QLMT command in queue, discard the oldest
			do {
				ctrl->mainCommandQueue.pop_front();
			} while (ctrl->mainCommandQueue.size() > ctrl->RPC_QLMT);
			{
				std::stringstream oss;
				oss << "New size of queue: " << ctrl->mainCommandQueue.size()
										<< std::endl;
				LOG_ALWAYS(oss.str());
			}
		}

		// trim commands from remaining sync queues and remove the empty ones
		std::list<std::deque<yarp::os::Bottle> >::iterator qsit =
				ctrl->syncCommandQueues.begin();

		// trim a command from all non-empty synchronous queues,
		while (qsit != ctrl->syncCommandQueues.end()) {
			// remove empty ones
			if (qsit->empty()) {
				// erase empty queue, move forward
				qsit = ctrl->syncCommandQueues.erase(qsit);
			} else {
				// log when discarding commands
				if (qsit->size() > ctrl->RPC_QLMT) {
					std::stringstream oss;
					oss << "More than " << ctrl->RPC_QLMT
							<< " commands in the queue: " << (qsit->size())
							<< ". Clearing oldest." << std::endl;
					LOG_ALWAYS(oss.str());

					// if more than RPC_QLMT command in queue, discard the oldest
					do {
						qsit->pop_front();
					} while (qsit->size() > ctrl->RPC_QLMT);
					// show result of trimming
					{
						std::stringstream oss;
						oss << "New size of queue: " << qsit->size() << std::endl;
						LOG_ALWAYS(oss.str());
					}
				}
				++qsit;
			}
		}
	}

	// release lock on queue
	ctrl->commandQueueMutex.post();
}

} // end of namespace srInterface
