/**
 * @file YarpUtil.cpp
 *
 * This file YarpUtil.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author	Anne C. van Rossum
 * @date	Jul 29, 2010
 * @project	Replicator FP7
 * @company	Almende B.V.
 * @case	YARP
 */


// General files
#include <string>

// Open scene graph
#include <osg/Vec4>
#include <osg/Vec2d>

// Plugin files
#include <srInterface/YarpUtil.h>

using namespace yarp::os;

namespace srInterface {

/* **************************************************************************************
 * Implementation of YarpUtil
 * **************************************************************************************/

YarpUtil_::YarpUtil_() {

}

YarpUtil_::~YarpUtil_() {

}

/**
 * Small helper function to cast strings to new allocated char pointers. Do
 * not forget to deallocate the result after you are done with it.
 */
char *YarpUtil_::StringToChar(std::string str) {
	char * result = new char[str.size()+1];
	std::copy(str.begin(), str.end(), result);
	result[str.size()] = '\0';
	return result;
}

/**
 * There seems to be something wrong with pop(). However, it might also be the case
 * that clone() needs to be used instead of "Value v = b->pop()". Anyway, using this
 * function works.
 */
Value * YarpUtil_::pop_front(Bottle &bottle) {
	Value *result = bottle.get(0).clone();

	if (bottle.size() == 1) {
		bottle.clear();
		return result;
	}

	Bottle temp;
	temp.copy(bottle, 1);
	bottle.clear();
	bottle.copy(temp);
	return result;
}

/**
 * Maps from Bottle to ParameterReturnValue. An offset can be used to skip eventual command headers
 * in the Bottle
 */
srCore::ParameterReturnValue * YarpUtil_::BottletoPRV(const Bottle &bottle, int offset) {
	int index = offset;
	std::string type = bottle.get(index).asString().c_str();
	index++;
	if (type.compare(typeid(float*).name()) == 0) {
		float *value = new float;
		double temp = bottle.get(index).asDouble();
		value = reinterpret_cast<float*> (&temp);
		srCore::ParameterReturnValue *result = new srCore::ParameterReturnValue(type, value);
		return result;
	} else if (type.compare(typeid(osg::Vec4f*).name()) == 0) {
		osg::Vec4f *value = new osg::Vec4f();
		for (int i = 0; i < osg::Vec4f::num_components; i++) value->_v[i] = bottle.get(index+i).asDouble();
		srCore::ParameterReturnValue *result = new srCore::ParameterReturnValue(type, value);
		return result;
	} else if (type.compare(typeid(osg::Vec4*).name()) == 0) { //actually just a typedef of Vec4f
		osg::Vec4 *value = new osg::Vec4();
		for (int i = 0; i < osg::Vec4::num_components; i++) value->_v[i] = bottle.get(index+i).asDouble();
		srCore::ParameterReturnValue *result = new srCore::ParameterReturnValue(type, value);
		return result;
	} else if (type.compare(typeid(osg::Vec3f*).name()) == 0) {
		osg::Vec3f *value = new osg::Vec3f();
		for (int i = 0; i < osg::Vec3f::num_components; i++) value->_v[i] = bottle.get(index+i).asDouble();
		srCore::ParameterReturnValue *result = new srCore::ParameterReturnValue(type, value);
		return result;
	} else if (type.compare(typeid(osg::Vec2d*).name()) == 0) {
		osg::Vec2d *value = new osg::Vec2d();
		for (int i = 0; i < osg::Vec2d::num_components; i++) value->_v[i] = bottle.get(index+i).asDouble();
		srCore::ParameterReturnValue *result = new srCore::ParameterReturnValue(type, value);
		return result;
	}
	return NULL;
}

/**
 * The bottle is not cleared beforehand.
 */
void YarpUtil_::PRVtoBottle(const srCore::ParameterReturnValue &param, Bottle &bottle) {
	std::string type = param.type.c_str();
	bottle.addString(type.c_str());
	if (type.compare(typeid(float*).name()) == 0) {
		float* value = reinterpret_cast<float*> (param.pValue);
		bottle.addDouble(*value);
	} else if (type.compare(typeid(osg::Vec4f*).name()) == 0) {
		osg::Vec4f* value = reinterpret_cast<osg::Vec4f*> (param.pValue);
		for (int i = 0; i < osg::Vec4f::num_components; i++) bottle.addDouble(value->_v[i]);
	} else if (type.compare(typeid(osg::Vec4*).name()) == 0) {
		osg::Vec4* value = reinterpret_cast<osg::Vec4*> (param.pValue);
		for (int i = 0; i < osg::Vec4::num_components; i++) bottle.addDouble(value->_v[i]);
	} else if (type.compare(typeid(osg::Vec3f*).name()) == 0) {
		osg::Vec3f* value = reinterpret_cast<osg::Vec3f*> (param.pValue);
		for (int i = 0; i < osg::Vec3f::num_components; i++) bottle.addDouble(value->_v[i]);
	} else if (type.compare(typeid(osg::Vec2d*).name()) == 0) {
		osg::Vec2d* value = reinterpret_cast<osg::Vec2d*> (param.pValue);
		for (int i = 0; i < osg::Vec2d::num_components; i++) bottle.addDouble(value->_v[i]);
	}

}

}
