/**
 * @file YarpedRobot.cpp
 * @brief 
 *
 * This file YarpedRobot.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license.
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * @author  Anne C. van Rossum
 * @date    Jul 20, 2010
 * @project Replicator FP7
 * @company Almende B.V.
 * @case	YARP
 */

#define SAVE_IMAGES
#undef SAVE_IMAGES

// General files
#include <stdio.h>

// Delta3D files
#include <dtUtil/log.h>

// Robot3D files
#include <srCore/simulationUtils.h>
#include <srCore/robot/robotActorBase.h>
#include <srInterface/YarpedRobot.h>
#include <srInterface/YarpUtil.h>
#include <srInterface/Robot3DVocab.h>

#include <srCore/sensor/cameraSensor.h>

// YARP files
#include <yarp/dev/ControlBoardInterfaces.h>
#ifdef WITH_OPENCV
#include <yarp/sig/Image.h>
#endif

#ifdef SAVE_IMAGES
#pragma message ("Now you have to use -lhighgui in the linker ")
#include <opencv/highgui.h>
#endif

using namespace srCore;

using namespace yarp::os;

namespace srInterface {

/* **************************************************************************************
 * Implementation of YarpedRobot
 * **************************************************************************************/

/**
 * The YARP interface hides a lot of the functionality of the robot class. This is on purpose,
 * on a real robot you will not be able to set its position to a random spot in the world. Of
 * course, someone else might pick up a robot and place it somewhere else (which is called a
 * kidnapping experiment), but the software equivalent would be a Delta3D game component which
 * acts like such a kidnapper and does have this ability. The YarpedRobot is NOT such a
 * component. The component that can actually "play god" is the YarpedSim component.
 *
 * The yarp modules that communicate with YarpedRobot are preferably derived from RFModuleRPC.
 * The YarpedRobot class is not derived from it, to diminish code cluttering, and multiple
 * inheritance. Each YarpedRobot comes with a YarpedRobotThread class that listens for yarp
 * messages in a separate thread.
 *
 * @param name		Name of the to be created robot
 * @param robot		The robot entity itself
 *
 * TODO: Think over YARP device derivation.
 *
 * TODO: It is not possible to derive from a YARP module. Or, at least, it is not clear for me directly
 * from where it would be started then. And anyway, we will need a component that positions the
 * robots anyway.
 */
YarpedRobot::YarpedRobot(const std::string &name, RobotActorBase *robot): Agent(name, robot),
		initialised(false),
		tick_counter(0),
		camera_frequency(10),
		rpcServerPort(),
		sensorPorts(),
		actuatorPorts(),
		camera_selected(),
		camera_enabled(false),
		yarpedRobotThread(NULL)
{
	SetCameraFrequency(10);
	SelectCamera(VOCAB_FRONT);
}

/**
 * Destruction of interface, YARP port, and everything that belongs to an Agent in the Robot3D
 * simulator.
 */
YarpedRobot::~YarpedRobot() { }

/**
 * Sets port name and opens it. This is called from YarpedSim::AddRobot. It appends the prefix
 * name (which includes the name of the simulator and the robot) with e.g. "/camera", or
 * "wheels".
 */
void YarpedRobot::SetPort(const std::string &prefixName) {

	// Create and open command ports (incoming for RPC, but will also be used to send back)
	std::string port_name;
	port_name = prefixName + "/rpc:i";
	rpcServerPort.open(port_name.c_str());

	// Create and open ports for actuators
	std::string port_type = "wheel";
	port_name = prefixName + "/" + port_type;
	BufferedPort<Bottle> *bport = new BufferedPort<Bottle>();
	actuatorPorts.insert(std::pair < std::string, PortVariant > (port_type, bport));
	bport->open(port_name.c_str());

	// Create and open ports for sensors
	SetCameraPort(prefixName);

	// Start thread to handle YARP messages in parallel
	yarpedRobotThread = new YarpedRobotThread(this);
	yarpedRobotThread->start();

	initialised = true;
}

/**
 * There will be only one camera port, because on the real robot it is possible to only
 * read one camera at a time.
 */
void YarpedRobot::SetCameraPort(const std::string &prefixName) {
	std::string camera_name = GetCameraName(camera_selected);
	std::string port_type = "camera";
	std::string port_name = prefixName + "/" + port_type;
	BufferedPort<Image> *imgport = new BufferedPort<Image>();
	sensorPorts.insert(std::pair < std::string, PortVariant > (port_type, imgport));
	imgport->open(port_name.c_str());
}

/**
 * Get configuration and set configuration of the camera.
 * <ul>
 * <li>MinDistance
 * <li>FlareAngle
 * <li>MaxDistance
 * <li>Frustum
 * <li>Resolution
 * <li>BackgroundColor
 * </ul>
 */
bool YarpedRobot::ConfigureCamera(const Bottle& command, Bottle& reply) {
	reply.clear();
	bool ok = false;

	if (command.size() < 4) {
		LOG_ERROR("We at least need three arguments: [CCAM vocab, id, set/get command, parameter]");
		reply.addVocab(VOCAB_FAILED);
		return ok;
	}

	if (command.get(0).asVocab() != RPC_VOCAB_CCAM) {
		reply.addVocab(VOCAB_FAILED);
		return ok;
	}
	//normally this would be an integer, but cameras are different
	int side = command.get(1).asVocab();
	int cmd = command.get(2).asVocab();
	switch (cmd){
	case VOCAB_SET: {
		std::string camera_name = GetCameraName(side);
		dtCore::RefPtr<srCore::SensorBase> sensor;
		if (!GetSensor(camera_name, sensor)) break;
		// We don't need to cast, but we may include some type casting to help the user
		// in understanding why setting a certain parameter doesn't work
		// srCore::CameraSensor *camera = dynamic_cast<srCore::CameraSensor*>(sensor.get());
		std::string parameter_name = command.get(3).asString().c_str();
		ParameterReturnValue *prv = YarpUtil::Instance()->BottletoPRV(command, 4);
		sensor->setParameter(parameter_name, *prv);
		ok = true;
		break;
	}
	case VOCAB_GET: {
		std::string camera_name = GetCameraName(side);
		dtCore::RefPtr<srCore::SensorBase> sensor;
		if (!GetSensor(camera_name, sensor)) break;
		std::string parameter_name = command.get(3).asString().c_str();
		ParameterReturnValue *prv = sensor->getParameter(parameter_name);
		if (prv == NULL) break;
		YarpUtil::Instance()->PRVtoBottle(*prv, reply);
		ok = true;
		break;
	}
	}

	if(!ok)
		reply.addVocab(VOCAB_FAILED);
	else
		reply.addVocab(VOCAB_OK);
	return ok;
}


/**
 * A (new) image will be written to the YARP port with the given frequency.
 *
 * @param freq		frequency in Hertz
 */
void YarpedRobot::SetCameraFrequency(const int &freq) {
	camera_frequency = freq;
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "The camera will be polled with " << freq << " Hz.";
	LOG_INFO(msg.str());
}

/**
 * Returns the name of a camera given the VOCAB integer.
 */
std::string YarpedRobot::GetCameraName(const int &vocab_robotside) {
	std::string result;
	switch (vocab_robotside) {
	case VOCAB_FRONT: result = "Front Camera"; break;
	case VOCAB_BACK: result = "Back Camera"; break;
	case VOCAB_LEFT: result = "Left Camera"; break;
	case VOCAB_RIGHT: result = "Right Camera"; break;
	case VOCAB_NONE: result = ""; break;
	default: result = "";
	}
	return result;
}

/**
 * There are four cameras on the robot. There can be only one in use at the same time.
 * This function selects one of the four indicated by vocabs: four-lettered tuples that
 * correspond to 32-bit integers.
 *
 * @param vocab_robotside	physical location on the robot as a vocab
 */
void YarpedRobot::SelectCamera(const int &vocab_robotside) {
	std::string camera_name = GetCameraName(vocab_robotside);
	if (camera_name.compare("") == 0) {
		LOG_ERROR("Unknown robot side to select camera for. Camera will be turned off.");
		camera_enabled = false;
	}
	if (camera_enabled) {
		camera_selected = vocab_robotside;
		dtCore::RefPtr<srCore::SensorBase> sensor;
		GetSensor(camera_name, sensor);
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "The camera to be used: " << camera_name;
		LOG_INFO(msg.str());
	}
}

/**
 * Sets the sensor, if it returns success.
 */
bool YarpedRobot::GetSensor(const std::string &name, dtCore::RefPtr<srCore::SensorBase> &sensor) {
	sensor = getSensor(name.c_str());
	if (sensor == NULL) {
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "The sensor " << name << " does not exist.";
		LOG_ERROR(msg.str());
		return false;
	}
	return true;
}

/**
 * Gets a bottle and if we have a "vocab ports" command, return all the ports.
 */
bool YarpedRobot::ListPorts(const Bottle& command, Bottle& reply) {
	reply.clear();
	bool ok = false;

	if (command.get(0).asVocab() != RPC_VOCAB_PORT) {
		return ok;
	}

	switch (command.size()) {
	case 2: {
		std::string key = command.get(1).asString().c_str();
		std::vector<std::string> port_names;
		if (key.compare("sensor") == 0) {
			getPorts(SENSOR, port_names);
		} else if (key.compare("actuator") == 0) {
			getPorts(ACTUATOR, port_names);
		} else if (key.compare("all") == 0) {
			getPorts(ALL, port_names);
		} else {
			getPorts(key.c_str(), port_names);
		}
		Robot3DVocab::Instance()->toBottle(port_names, reply);
		ok = true;
		break;
	}
	case 3: {
		std::string key = command.get(1).asString().c_str();
		int id = command.get(2).asInt();
		std::string port_name;
		if (key.compare("sensor") == 0) {
			getPort(SENSOR, id, port_name);
		} else if (key.compare("actuator") == 0) {
			getPort(ACTUATOR, id, port_name);
		} else if (key.compare("all") == 0) {
			getPort(ALL, id, port_name);
		} else {
			getPort(key.c_str(), id, port_name);
		}
		Robot3DVocab::Instance()->toBottle(port_name, reply);
		ok = true;
		break;
	}
	}

	if(!ok)
		reply.addVocab(VOCAB_FAILED);
	else
		reply.addVocab(VOCAB_OK);
	return ok;
}

/**
 * Get all ports of a given type. The type can be "sensor", "actuator", or "all".
 */
bool YarpedRobot::getPorts(const PORT_TYPE &portType, std::vector<std::string> &portNames) {
	bool result = true;
	portNames.clear();
	PortMapIter it;
	switch(portType) {
	case ALL:
	case SENSOR:
		for (it = sensorPorts.begin(); it != sensorPorts.end(); it++) {
			PortVariant portVar = it->second;
			ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
			portNames.push_back(port_name.c_str());
		}
		if (portType != ALL) {
			break;
		}
	case ACTUATOR:
		for (it = actuatorPorts.begin(); it != actuatorPorts.end(); it++) {
			PortVariant portVar = it->second;
			ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
			portNames.push_back(port_name.c_str());
		}
		break;
	default:
		result = false;
	}
	return result;
}

/**
 * Get specific port. The list of ports might be different from run to run, so use
 * getPorts to retrieve this list first, before you make assumptions regarding
 * order. Subsequently the order will be the same except if you add code yourself
 * that removes sensors, actuators, etc on-the-fly.
 */
bool YarpedRobot::getPort(const PORT_TYPE &portType, int index, std::string &portName) {
	bool result = true;
	PortMapIter it;
	assert (index >= 0);
	int counter = 0;
	switch(portType) {
	case ALL:
	case SENSOR:
		for (it = sensorPorts.begin(); it != sensorPorts.end(); it++) {
			if (counter == index) {
				PortVariant portVar = it->second;
				ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
				portName = port_name;
				return result;
			}
			counter++;
		}
		if (portType != ALL) {
			break;
		}
	case ACTUATOR:
		for (it = actuatorPorts.begin(); it != actuatorPorts.end(); it++) {
			if (counter == index) {
				PortVariant portVar = it->second;
				ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
				portName = port_name;
				return result;
			}
			counter++;
		}
		result = false; // no ports found
		break;
	default:
		result = false;
	}
	return result;
}

/**
 * Return list of ports corresponding to a key in the map.
 */
bool YarpedRobot::getPorts(const std::string &key, std::vector<std::string> &port_names) {
	port_names.clear();
	bool success = false;
	PortMapIter it;
	std::pair <PortMapIter, PortMapIter> mapBounds;
	mapBounds = sensorPorts.equal_range(key);
	for (it = mapBounds.first; it != mapBounds.second; it++) {
		PortVariant portVar = it->second;
		ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
		port_names.push_back(port_name.c_str());
		success = true;
	}
	return success;
}

/**
 * Get port of given key and index.
 */
bool YarpedRobot::getPort(const std::string &key, int index, std::string &portName) {
	portName.clear();
	bool success = false;
	PortMapIter it;
	int counter = 0;
	std::pair <PortMapIter, PortMapIter> mapBounds;
	mapBounds = sensorPorts.equal_range(key);
	for (it = mapBounds.first; it != mapBounds.second; it++) {
		if (index == counter) {
			PortVariant portVar = it->second;
			ConstString port_name = boost::apply_visitor(GetPortName(), portVar);
			portName = port_name.c_str();
			success = true;
			return success;
		}
		counter++;
	}
	return success;
}

/**
 * This function polls the chosen camera with a certain frequency. As on the real robot is not
 * possible to query all four cameras at once. One needs to be selected.
 *
 * @param message	game message using the Delta3D messaging architecture
 */
void YarpedRobot::ProcessMessage(const dtGame::Message& message) {

	// Produce (one time) a warning if process message is executed when robot is not initialised
	if (!initialised) {
		if (tick_counter == 0) {
			LOG_WARNING("Message received but not initialised (yet)");
			tick_counter++;
		}
//		LOG_WARNING("Not initialised (yet)");
		return;
	}

	// First process it by the base class, with multiple controllers this will be executed multiple times!
	Agent::ProcessMessage(message);

#ifdef WITH_OPENCV

	// Now continue with sending a message to the YARP port
	if ((message.GetMessageType() == dtGame::MessageType::TICK_LOCAL)) {

		tick_counter++;

		if (!(tick_counter % camera_frequency) && camera_enabled) {
			dtCore::RefPtr<srCore::SensorBase> sensor;
			std::string camera_name = GetCameraName(camera_selected);
			if (!GetSensor(camera_name, sensor)) return;

			osg::Image* tmpImage;
			srCore::ParameterReturnValue prv = srCore::MAKE_PARAMETER_VALUE(&tmpImage);
			sensor->evaluate(prv);
			if (prv.type != typeid(osg::Image**).name()) {
				LOG_ERROR("The data type the camera returns is wrong!");
				return;
			}
			// cast void* to Image**
			osg::Image** osgimg_ptr = reinterpret_cast<osg::Image**> (prv.pValue);
			// cast to osg::Image, use smart pointer
			osg::ref_ptr<osg::Image> osgimg = *osgimg_ptr;
			// convert to IplImage (here we depend on open cv, needs to be removed)
			IplImage* iplimg = convertToOpenCV(osgimg.get(), RGB);

			// send over default camera port
			//			std::map<std::string, BuffPort >::iterator it;
			//			it = sensorPorts.find("camera");
			//			PortVariant portVar = boost::get<1>(it->second);
			//			BufferedPort<Image>* camPort = boost::get< BufferedPort<Image>* >( portVar );

			std::map<std::string, PortVariant >::iterator it = sensorPorts.find("camera");
			PortVariant portVar = it->second;
			BufferedPort<Image>* camPort = boost::get< BufferedPort<Image>* >( portVar );

			Image &return_img = camPort->prepare();
			return_img.resize(osgimg->s(), osgimg->t());
			return_img.wrapIplImage(iplimg);
			camPort->write();

			// deallocate
			cvReleaseImage(&iplimg);
		}
	}

#endif
}

/**
 * Process commands over "rpc" port. Realize that the read routine is blocking, so this
 * function need to be called from a separate thread.
 */
void YarpedRobot::ProcessYARPCommands() {
	// Non-blocking read
	Bottle b;
	LOG_INFO("Wait for new RPC command");
	rpcServerPort.read(b, true);
	if (b == NULL) return;

	LOG_INFO("Received a bottle...");
	if (b.size() <= 1) {
		LOG_ERROR("Received bottle with size <=1. Disregard...");
		return;
	}

	assert (b.size() > 0);

	Bottle result;
	bool success;
	switch (b.get(0).asVocab()) {
	case RPC_VOCAB_PORT:
		success = ListPorts(b, result);
		break;
	case RPC_VOCAB_CCAM:
		success = ConfigureCamera(b, result);
		break;
	default:
		result.addVocab(VOCAB_FAILED);
		success = false;
	}

	if (!success) {
		LOG_ERROR("No success in ListPorts");
		return;
	}


	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Sent bottle of size " << result.size();
	LOG_INFO(msg.str());

	rpcServerPort.reply(result);
}

/**
 * Overloaded function to explain the user that the entity is actually added to the
 * game manager.
 */
void YarpedRobot::OnAddedToGM() {
	Agent::OnAddedToGM();
	LOG_INFO("Yarped Robot added to game manager");
}

#ifdef WITH_OPENCV
/**
 * Convert to IPL or OpenCV format for YARP
 */
IplImage* YarpedRobot::convertToOpenCV(osg::Image *source, IMAGE_TYPE type) {
	// We can only convert pictures with chars
	// assert (source->getDataType() == GL_UNSIGNED_BYTE);

	int d_iWidth = source->s();
	int d_iHeight = source->t();

	CvSize size;
	size.width = d_iWidth;
	size.height = d_iHeight;

	// The YARP Image file only understands pixels of 3-bytes (without transparency channel)
	unsigned char NOF_CHANNELS = 3;
	IplImage* target = cvCreateImage(size, IPL_DEPTH_8U, NOF_CHANNELS);

	// std::string str = "rgb";
	// target->colorModel = str.c_str();

	target->colorModel[0] = 'r';
	target->colorModel[1] = 'g';
	target->colorModel[2] = 'b';
	target->colorModel[3] = '\0';

	if(type == BGR)
	{
		target->channelSeq[0] = 'b';
		target->channelSeq[1] = 'g';
		target->channelSeq[2] = 'r';
		target->channelSeq[3] = '\0';
	}
	else
	{
		target->channelSeq[0] = 'r';
		target->channelSeq[1] = 'g';
		target->channelSeq[2] = 'b';
		target->channelSeq[3] = '\0';
	}

	// Interleaved (default)
	target->dataOrder = IPL_DATA_ORDER_PIXEL;
	// Origin bottom-left (see @c source->getOrigin())
	target->origin = IPL_ORIGIN_BL;
	// Width of image in bytes
	target->widthStep = d_iWidth * NOF_CHANNELS;
	// A 4-byte alignment
	target->align = 3; //IPL_ALIGN_DWORD;
	// Size is height times width (not pixel width, but width in bytes)
	target->imageSize = d_iHeight * target->widthStep;

	for (int x = 0; x < d_iWidth; ++x) {
		for (int y = 0; y < d_iHeight; ++y) {
			osg::Vec4 color = source->getColor(osg::Vec2(x,y));
			// CvScalar is 4 bytes (BGR and 0 for transparency)
			CvScalar imgScalar;
			if(type == BGR)
				imgScalar = cvScalar((unsigned char)(color.b()*255),(unsigned char)(color.g()*255),(unsigned char)(color.r()*255), 0); //color.a());
			else
				imgScalar = cvScalar((unsigned char)(color.r()*255),(unsigned char)(color.g()*255),(unsigned char)(color.b()*255), 0); //color.a());
			//In cvSaveImage the image is upside down but in YARP it is okay
			//cvSet2D(iplimg,d_iHeight-1-y,d_iWidth-1-x,bgr);
			cvSet2D(target,y,x,imgScalar);
		}
	}

	//Highgui knows how to print to a file. This is now working properly!
#ifdef SAVE_IMAGES
	cvSaveImage("YARP_image.jpg", iplimg);
#endif

	// std::ostringstream msg;
	// msg << "Write image of size " << (int)d_iWidth << "x" << (int)d_iHeight << " to camera port";
	// LOG_INFO(msg.str());
	return target;
}

#endif


/**
 * Run the thread for the yarped robot.
 */
void YarpedRobotThread::run() {
	while (isStopping() != true) {
		yarpedRobot->ProcessYARPCommands();
	}
}

} // end of namespace srInterface
