/**
 * @file YarpedSim.cpp
 *
 * This file YarpedSim.cpp is created at Almende B.V. It is open-source software and part
 * of the Common Hybrid Agent Platform (CHAP). A toolbox with a lot of open-source tools.
 * Ranging from thread pools, and TCP/IP components to control architectures and learning
 * algorithms. This software is published under the GNU Lesser General Public license,
 *
 * It is not possible to add usage restrictions to an open-source license. Nevertheless,
 * we personally strongly object against this software used by the military, in the
 * bio-industry, for animal experimentation, or anything that violates the Universal
 * Declaration of Human Rights.
 *
 * Copyright © 2010 Anne van Rossum <anne@almende.com>
 *
 * Author:	Anne C. van Rossum
 * Date:	Aug 2, 2010
 * Project:	Replicator FP7
 * Company:	Almende B.V.
 */


// General files
#include <cstdlib>

// Open scene graph
#include <osg/Vec3f>

// Delta3D files
#include <dtUtil/log.h>
#include <dtGame/messagetype.h>
#include <dtGame/message.h>
#include <dtCore/refptr.h>

// Robot3D files
#include <srInterface/YarpedSim.h>
#include <srInterface/Robot3DVocab.h>

// YARP files
#include <yarp/os/Value.h>

// Robot3D files
#include <srInterface/YarpUtil.h>
#include <srCore/robotCreator.h>
#include <srInterface/YarpedRobot.h>
#include <srInterface/YarpRobotNavigation.h>
#include <srInterface/InterfaceUtil.h>

using namespace yarp::os;

namespace srInterface {

/* **************************************************************************************
 * Implementation of YarpedSim
 * **************************************************************************************/

/*
 * YarpedSim opens a port for the Robot3D simulator. It can be connected to a component
 * that knows when to add and remove robots, etc. This component actually knows how to
 * execute those commands. It is again implemented as exogenous coordination, thus the
 * Simpl3D simulator does not know itself to which party it needs to connect.
 */
YarpedSim::YarpedSim(): GMComponent("YarpedSim"),
		agents (new std::vector<dtCore::RefPtr< srInterface::Agent> >()),
		globalYarpPort(),
		network(),
		yarpedSimThread(NULL),
		time_counter(0),
		pauseMessageProcessed(false) {
	std::string globalYarpPortName = "/robot3d";
	if (!globalYarpPort.open(globalYarpPortName.c_str())) {
		LOG_WARNING("Opening YARP port in YarpedSim constructor fails");
		LOG_INFO("This is no problem, just if you need YARP functionality, start \"yarpserver\" before Robot3D.");
		return;
	}
	agents->clear();
	LOG_INFO("Opening YARP port succeeded, please, connect now...");

	yarpedSimThread = new YarpedSimThread(this);

	yarpedSimThread->start();
}

/**
 * The destructor ~YarpedSim closes the YARP port.
 */
YarpedSim::~YarpedSim() {
	LOG_INFO("Destroy YarpedSim");
	yarpedSimThread->stop();
	globalYarpPort.close();
}

/**
 * Process game message. This routine is not called automatically.
 * Every tick the YARP port, which should have been opened before, is read. This read is
 * non-blocking and expects a message in the Bottle format. When there is nothing to be
 * received, the routine quits. If there is a bottle, a case statement will switch on a
 * so-called VOCAB format. This is an integer disguised as a four-letter word, so that it
 * is easy to recognise. For example, it is perfectly valid C to have a case-statement that
 * says "VOCAB4('r','o','b','+')".
 *
 * The first item of the Bottle will be the type of message and popped off the bottle. The
 * rest of the bottle will be past to dedicated routines like ReadBottleAddRobot. The proper
 * action is subsequently undertaken in this routine, upon which the function quits. The
 * type of bottles that YarpedSim understands is only on the level of the entire simulator.
 * Commands that have to do with individual robots need to be handled in YarpedRobot.
 *
 * Logging might be excessive in this stage...
 */
void YarpedSim::ProcessMessage(const dtGame::Message &message) {

	dtGame::GMComponent::ProcessMessage(message);

	if (message.GetMessageType() == dtGame::MessageType::TICK_LOCAL) {
		LOG_DEBUG("Tick YarpedSim component");
		time_counter++;
	}

	if (message.GetMessageType() == dtGame::MessageType::INFO_PAUSED) {
		pauseMessageProcessed = true;
	}
}

/**
 * Read YARP message from port. This is a separate function from ReadBottle(Bottle*) so that
 * it is possible for the thread that inspects the YARP port, namely YarpedSimThread, to
 * initiate in actually creating the robot and corresponding controllers when the proper
 * message has arrived. The thread is needed not for YARP, but for the Delta3D simulator,
 * see @ReadBottle.
 */
Bottle *YarpedSim::ReadPort() {
	// Non-blocking read
	Bottle *b = globalYarpPort.read(false);
	if (b == NULL) return b;
	LOG_INFO("Received a bottle...");
	if (b->size() <= 2) {
		LOG_ERROR("Received bottle with size <=2. Disregard...");
		return b;
	}
	return b;
}

/**
 * This function can only be called when the simulator has paused. This is because it adds
 * components to the vector of components internal to Delta3D. This interacts with the
 * routine that calls ProcessMessage on each component, namely
 *   GameComponent::DoSendMessageToComponents.
 * As soon as the iterator in that function does not become corrupt by adding components,
 * this needs to be done in a separate thread.
 */
void YarpedSim::ReadBottle(Bottle *b) {
	assert (b->size() > 0);

	// There seems to be something wrong with pop(), so replaced by YarpUtil::Instance()->pop_front()
	Value *v_get = YarpUtil::Instance()->pop_front(*b);
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Pop value \"" << v_get->asString() << "\"";
	LOG_ALWAYS(msg.str());
	Value v(v_get->asString(), true);
	if (!v_get->isVocab()) {
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Wrong message format, should start with vocab (value = " << v.toString() << ")";
		LOG_WARNING(msg.str());
		msg.clear(); msg.str("");
		msg << "For example: " << VOCAB_ROBOT_ADD;
		LOG_INFO(msg.str());
		return;
	}

	switch (v_get->asVocab()) {
	case VOCAB_ROBOT_ADD: {
		LOG_INFO("Got a ROBOT_ADD message");
		std::string name, type;
		int id;
		osg::Vec3f position;
		if (ReadBottleAddRobot(*b, name, type, id, position))
			AddRobot(name, type, id, position);
		break;
	}
	default:
		LOG_WARNING("I don't understand this vocab");
		std::ostringstream msg; msg.clear(); msg.str("");
		msg << "Vocab corresponds to integer: " << (int)v_get->asVocab() << "";
		LOG_INFO(msg.str());
	}
}

/**
 * Add a robot to the simulator. For now the type is hardcoded ("RobotKIT", normally the type would
 * be obtained over the YARP port in a bottle. All the YARP ports that are created on the robot are
 * prepended with the name of the simulator "robot3d" and a robot index "robot_id".
 */
void YarpedSim::AddRobot(const std::string &name, const std::string &type, const int &id,
		const osg::Vec3f &position) {

	dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
	//	dtUtil::Log::LogMessageType loglevel = dtUtil::Log::GetInstance().GetLogLevel();
	//	dtUtil::Log::GetInstance().SetLogLevel(dtUtil::Log::LOG_DEBUG);

	GA->GetGameManager()->DebugStatisticsTurnOn(true, false, 0.1);

	RobotInitData rid;
	std::ostringstream temp; temp.clear();
	temp << name << (int)id;
	std::string robot_name = temp.str();
	rid.name = YarpUtil::Instance()->StringToChar(robot_name);
	rid.type = YarpUtil::Instance()->StringToChar(type.c_str());
	rid.pos = position;
	rid.jointAngles.push_back(0);
	rid.sensors.clear();

	std::ostringstream msg; msg.clear();
	msg << "Add robot " << rid.name << " and type " << rid.type;
	LOG_INFO(msg.str());

	// Add the robot itself to the simulator
	bool success = srCore::SimulationRobotCreator::Instance()->addRobotToGM(id, rid, position.z());

	if (success) {
		// Name of controller
		std::stringstream controller_name; controller_name.clear();
		controller_name << name << "YarpController" << id;
		std::string controller_type = "YarpedRobot";
		ControllerInitData cid;
		cid.name = YarpUtil::Instance()->StringToChar(controller_name.str());
		cid.robot = rid.name;
		cid.type = YarpUtil::Instance()->StringToChar(controller_type);
		cid.parameter = NULL;
		cid.pos = rid.pos; //not necessary

		AddController(cid);

		controller_name.clear(); controller_name.str("");
		controller_name << name << id << "Nav";
		controller_type = "NavigationController";

		cid.name = YarpUtil::Instance()->StringToChar(controller_name.str());
		cid.type = YarpUtil::Instance()->StringToChar(controller_type);

		AddController(cid);

		std::vector<srInterface::Agent*> allagents;
		std::vector<srInterface::Agent*>::iterator iter;
		InterfaceUtil::GetAllAgents(allagents);
		for (iter = allagents.begin(); iter != allagents.end(); iter++) {
			msg.clear(); msg.str("");
			msg << "Now we have the agent belonging to robot " << (*iter)->getRobotID();
			LOG_INFO(msg.str());
		}

	}

	std::vector<dtGame::GMComponent*> components;
	GA->GetGameManager()->GetAllComponents(components);
	for(unsigned int i=0; i < components.size(); i++) {
		msg.clear(); msg.str("");
		msg << "Component \"";
		msg << components[i]->GetName();
		msg << "\" loaded";
		LOG_INFO(msg.str());
	}
}

/**
 * Add controller (or agent) from the fields defined in ControllerInitData, normally created
 * from the config.xml file, but can also be populated over here after a command over a YARP
 * command port.
 */
void YarpedSim::AddController(const ControllerInitData &cid) {

	// Retrieve the robot by name, both as a test and to add a controller to it
	srCore::RobotActorBase *actor = InterfaceUtil::GetRobot(cid.robot);

	// We also need to add the YARP controller as an interface
	std::ostringstream msg; msg.clear(); msg.str("");
	msg << "Create and add yarp controller \"" << cid.name << "\" to robot " << actor->GetUniqueId().ToString();
	LOG_INFO(msg.str());

	dtCore::RefPtr<Agent> agent;
	std::string controllerType = "YarpedRobot";
	if (controllerType.compare(cid.type) == 0) {
		dtCore::RefPtr<YarpedRobot> yarpController = new YarpedRobot(cid.name, actor);
		actor->GetGameActorProxy().GetGameManager()->AddComponent(*yarpController,
				dtGame::GameManager::ComponentPriority::NORMAL);

		// set port
		std::stringstream port_prefix; port_prefix.clear();
		port_prefix << "/robot3d/" << cid.robot;
		yarpController->SetPort(port_prefix.str());

		agent = yarpController;
	}

	controllerType = "NavigationController";
	if (controllerType.compare(cid.type) == 0) {
		dtCore::RefPtr<YarpRobotNavigation> yarpNavigationController = new YarpRobotNavigation(
				cid.name, actor);
		actor->GetGameActorProxy().GetGameManager()->AddComponent(
				*yarpNavigationController, dtGame::GameManager::ComponentPriority::NORMAL);

		// set port
		std::stringstream port_prefix; port_prefix.clear();
		port_prefix << "/robot3d/" << cid.robot;
		yarpNavigationController->SetPort(port_prefix.str());

		agent = yarpNavigationController;

	}

	// Store controller temporarily
	if (agent != NULL) {
		agents->push_back(agent);
	}

}

/**
 * This routine interprets the AddRobot bottles that are sent over YARP to Robot3D.
 *
 * TODO: Create a separate format description and a generic YARP message interpreter.
 */
bool YarpedSim::ReadBottleAddRobot(Bottle &bottle, std::string &name, std::string &type,
		int &id, osg::Vec3f &position) {
	Value *v = YarpUtil::Instance()->pop_front(bottle);

	if (!v->isString()) {
		LOG_WARNING("Wrong message format, should be the robot's name");
		return false;
	}
	name = v->asString();

	v = YarpUtil::Instance()->pop_front(bottle);
	if (!v->isString()) {
		LOG_WARNING("Wrong message format, should be the robot's type");
		return false;
	}
	type = v->asString();

	v = YarpUtil::Instance()->pop_front(bottle);
	if (!v->isInt()) {
		LOG_WARNING("Robot id should be integer.");
		id = std::atoi(v->toString());
		std::ostringstream msg; msg.clear();
		msg << "Convert \"" << v->toString() << "\" to " << id;
		LOG_WARNING(msg.str());
	} else {
		id = v->asInt();
	}

	float x, y, z;
	v = YarpUtil::Instance()->pop_front(bottle);
	if (!v->isDouble()) {
		LOG_WARNING("Wrong message format, should be the robot's x position");
		x = std::strtod(v->toString(), NULL);
	} else {
		x = v->asDouble();
	}

	v = YarpUtil::Instance()->pop_front(bottle);
	if (!v->isDouble()) {
		LOG_WARNING("Wrong message format, should be the robot's y position");
		y = std::strtod(v->toString(), NULL);
	} else {
		y = v->asDouble();
	}

	v = YarpUtil::Instance()->pop_front(bottle);
	if (!v->isDouble()) {
		LOG_WARNING("Wrong message format, should be the robot's z position");
		z = std::strtod(v->asString(), NULL);
	} else {
		z = v->asDouble();
	}
	position.set(x,y,z);
	return true;
}

/**
 * Run the thread for the yarped robot.
 */
void YarpedSimThread::run() {
	while (isStopping() != true) {
		while (time_counter == yarpedSim->time_counter) {
			usleep(10000);
		}
		time_counter = yarpedSim->time_counter;

		Bottle *b = yarpedSim->ReadPort();

		if (b == NULL) continue;

		yarpedSim->pauseMessageProcessed = false;

		dtGame::GameApplication *GA = dtGame::GameApplication::GetInstance(0);
		GA->GetGameManager()->SetPaused(true);

		LOG_INFO("Pausing game manager, wait till we receive our own message...");
		usleep(10000); // Just to be sure, pause message is apparently FOLLOWED by others
		while (!yarpedSim->pauseMessageProcessed) {
			if (isStopping()) return;
			usleep(10000);
		}
		yarpedSim->ReadBottle(b);
		GA->GetGameManager()->SetPaused(false);
	}
}

} // end of namespace srAlmende
