/**
 * @file robotKIT.cpp
 *
 * The Robot3D simulator is created within the European Replicator and Symbrion projects. The
 * Replicator project is funded by the European Commission within the work programme "Cognitive
 * Systems, Interaction, Robotics" under the grant agreement no. 216240. The Symbrion project
 * is funded within the work programme "Future and Emergent Technologies Proactive" under the
 * grant agreement no. 216342. Replicator and Symbrion concerns the design of modular robots
 * in the real world and in 3D simulation.
 *
 * This file is part of the 3D simulator Robot3D which is built on top of the open source game
 * engine Delta3D (copyright (C) 2005, BMH Associates, Inc). This library is free software; you
 * can redistribute it and/or modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version. This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details. You
 * should have received a copy of the GNU Lesser General Public License along with this library;
 * if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * @author Adrian Friebel
 * @author Lutz Winkler
 * @author Anne C. van Rossum
 * @date   2008
 */

// Robot3D files
#include <srRobot/robotKIT.h>
#include <srFactory/bodyFactory.h>
#include <srFactory/actuatorFactory.h>
#include <srFactory/sensorFactory.h>
#include <srCore/simulationEvents.h>
#include <srCore/export.h>

/**
 * TODO: It should not be necessary to know the concrete types of the created entities. That
 * renders the factory design pattern useless...
 */
#include <srActuator/connectorUniKarl.h>
#include <srCore/actuator/hingeActuator.h>
#include <srActuator/screwDrive.h>
#include <srBody/led.h>

/************************************************************************************************
 *                      Simulator settings
 ***********************************************************************************************/

#define TIME_STEP				250

/*
 * Notes about measuring units:
 * Replicator specification parameters (in mm, N, kg, m/s^2)
 * Units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.
 *   1 m/s ^= 20 (5cm)/s
 */
#define MAX_SPEED				15

using namespace srCore;

namespace srRobot {

/************************************************************************************************
 *                      Robot settings
 ***********************************************************************************************/

//! The default game event handler
const std::string RobotKIT::EVENT_HANDLER_NAME("HandleGameEvent");

//! The default name (used in configuration files)
const std::string RobotKIT::DEFAULT_NAME("RobotKIT");

//! Maximum velocity of the robot
// TODO: does not seem to be used
const float MAX_ROBOT_VELOCITY = 15.0f;

/************************************************************************************************
 *                      Robot implementation
 ***********************************************************************************************/

/**
 * The robot is derived from the RobotActorBase, which is a Delta3D game actor. It does not have
 * access to (the internals of) other robots.
 */
RobotKIT::RobotKIT(dtGame::GameActorProxy &proxy, float jointAngle) : RobotActorBase(proxy),
		rotation_event_function(),
		deltaTime(0.01),
		manual (true),
		orientation_quadrant (-1)
{
	jointAngles.push_back(0.0);
	SetName(DEFAULT_NAME);
}

/**
 * Destructor of the Karlsruhe robot.
 */
RobotKIT::~RobotKIT() { }


/**
 * Initialisation of the robot. It adds bodies, actuators, and sensors to the robot.
 */
void RobotKIT::Initialize() {
	// In the osgGroup are all Drawables included.
	// The osgGroup might be useful for creating shadows
	osgGroup = new osg::Group();

	LOG_DEBUG("Initialise a UniKarl robot");

	// First, we'll create the bodies, the robot consists of two main bodies ...
	addBody(BodyFactory::createBody(
			srCore::KIT_1, this,
			0, 0, 0,
			0, 180, 90));

	addBody(BodyFactory::createBody(
			srCore::KIT_2, this,
			0, 0, 0,
			0, 180 + jointAngles[0], 90));			//0, 180, 90

	// Let's create some references to the "back" and the "front" part of the robot
	dtCore::RefPtr<BodyBase> body1 = this->myBodies[0];
	dtCore::RefPtr<BodyBase> body2 = this->myBodies[1];

	// our main hinge ...
	addActuator(srCore::ActuatorFactory::createActuator(
			srCore::HINGE_ACTUATOR,
			body1, body2,
			0, 0, 0,
			1, 0, 0,
			"Main Hinge"));

	dynamic_cast<HingeActuator*>(this->myActuators[0].get())->setZeroAngle(jointAngles[0] * M_PI / 180);

	AddDrive();

	AddConnectors();

	AddLEDs();

	AddSensors();

	dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(0);




	// ... and ready is our robot
	// Now we need to fill the robot with life. Therefore we need to write what the robot should do during the simulation.
	// This will be handled in TickLocal(). TickLocal will be called every simulation step.
}

/**
 * Add the wheels, screw drive or other type of actuator that is need for locomotion.
 */
void RobotKIT::AddDrive() {
	// Let's create some references to the "back" and the "front" part of the robot
	dtCore::RefPtr<BodyBase> body1 = this->myBodies[0];
	dtCore::RefPtr<BodyBase> body2 = this->myBodies[1];

	// With a z-offset of 0.9 the wheels are not low enough, the robot touches the ground
	// on one side
	float WHEEL_Z_OFFSET = 1.2; //0.9
	float WHEEL_OFFSET = 0.45; //0.75 //the lower, the more to the center of the robot
	if (SimulationUtils::Instance()->getWithWheels()) {
		// ... and the four wheels that are simulating the screw drive

		addBody(BodyFactory::createBody(
				srCore::WHEEL_BODY, this,
				-WHEEL_OFFSET, WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				//				0, 90, 90));
				-45, 90, 90));

		addBody(BodyFactory::createBody(
				srCore::WHEEL_BODY, this,
				WHEEL_OFFSET, WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				//				0, 90, 90));
				45, 90, 90));

		addBody(BodyFactory::createBody(
				srCore::WHEEL_BODY, this,
				WHEEL_OFFSET, -WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				//				180, 90, 90));
				-45, 90, 90));
		//				-115, 90, 90));

		addBody(BodyFactory::createBody(
				srCore::WHEEL_BODY, this,
				-WHEEL_OFFSET, -WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				//				180, 90, 90));
				45, 90, 90));
		//				115, 90, 90));
	}

	// The screw drive would consume too many computational resources, so they will be simulated by four wheels
	if (SimulationUtils::Instance()->getWithWheels()) {

		// Attach actuator to "body2", the "back half" of the robot
		dtCore::RefPtr<ActuatorBase> actuator2 = srCore::ActuatorFactory::createActuator(
				srCore::WHEEL_ACTUATOR,
				body2, this->myBodies[2],
				-WHEEL_OFFSET, WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				1, -1, 0,
				"Helper Wheel");

		// Attach actuator to "body1", the "front half" of the robot
		dtCore::RefPtr<ActuatorBase> actuator1 = srCore::ActuatorFactory::createActuator(
				srCore::WHEEL_ACTUATOR,
				body1, this->myBodies[5],
				-WHEEL_OFFSET, -WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				1, 1, 0,
				"Helper Wheel");

		dtCore::RefPtr<ActuatorBase> actuator4 = srCore::ActuatorFactory::createActuator(
				srCore::WHEEL_ACTUATOR,
				body1, this->myBodies[4],
				WHEEL_OFFSET, -WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				-1, 1, 0,
				"Helper Wheel");

		dtCore::RefPtr<ActuatorBase> actuator3 = srCore::ActuatorFactory::createActuator(
				srCore::WHEEL_ACTUATOR,
				body2, this->myBodies[3],
				WHEEL_OFFSET, WHEEL_OFFSET, -WHEEL_Z_OFFSET,
				-1, -1, 0,
				"Helper Wheel");

		LOG_DEBUG("Create the screw drives");
		dtCore::RefPtr<ActuatorBase> leftScrew = new ScrewDrive("LeftScrew");
		dtCore::RefPtr<ActuatorBase> rightScrew = new ScrewDrive("RightScrew");

		LOG_DEBUG("Initialise the screw drives");
		(dynamic_cast<ScrewDrive* >(leftScrew.get()))->init(actuator1, actuator3, dynamic_cast<ScrewDrive*>(rightScrew.get()), false);
		(dynamic_cast<ScrewDrive* >(rightScrew.get()))->init(actuator2, actuator4, dynamic_cast<ScrewDrive*>(leftScrew.get()), true);

		addActuator(leftScrew);
		addActuator(rightScrew);
	}
}

/**
 * Add the connectors that enable a robot to dock to another robot.
 */
void RobotKIT::AddConnectors() {
	addConnector(new myConnector(this->myBodies[0],  0,  1.1,  0, -90, 0,   0));
	addConnector(new myConnector(this->myBodies[1],  0, -1.1,  0, -90, 0, 180));
	addConnector(new myConnector(this->myBodies[1],  0,  0,  1.1, -90, 0,  90));
	addConnector(new myConnector(this->myBodies[0],  0,  0, -1.1, -90, 0, -90));
}

/**
 * A specific type of actuator, the LED.
 */
void RobotKIT::AddLEDs() {

#ifdef WITH_LED
	dtCore::RefPtr<Led> led = new Led(
			myBodies[0],
			osg::Vec3(0, 0, WHEEL_Z_OFFSET + 0.1),
			osg::Vec3(0, 0, 0));
	myObjects.push_back(led.get());

	char rgb_r = rand() % 2;
	char rgb_g = rand() % 2;
	osg::Vec4f color(rgb_r, rgb_g,1,0.5);
	led->setColor(color);
#endif

}

/**
 * Add cameras, microphone, infrared sensors and whatever is needed for perception.
 */
void RobotKIT::AddSensors() {
	/**
	 * TODO: We have to check the locations of the camera. Be aware that the orientations of myBodies[0]
	 * can be different from the orientation of myBodies[1]. The first vector is the position, the second
	 * is the orientation, the last vector is the resolution. All the camera are at z-position 2. We have
	 * to check if it is actually the case that the other cameras are turned off and do not consume
	 * resource.
	 */
	mySensors.push_back(SensorFactory::createSensor(CAMERA_SENSOR,
			"Back Camera",
			myBodies[0],
			osg::Vec3f(4.5,0,2),
			osg::Vec3f(0,0,-90),
			1.0,
			osg::Vec2(256,256)
	));
	//
	//
	//	mySensors.push_back(SensorFactory::createSensor(CAMERA_SENSOR,
	//			"Front Camera",
	//			myBodies[1],
	//			osg::Vec3f(-2.5,-1,2),
	//			osg::Vec3f(-90,0,0),
	//			1.0,
	//			osg::Vec2(256,256)
	//	));
	//
	mySensors.push_back(SensorFactory::createSensor(CAMERA_SENSOR,
			"Left Camera",
			myBodies[0],
			osg::Vec3f(0,4.5,2),
			osg::Vec3f(180,0,0),
			1.0,
			osg::Vec2(256,256)
	));

	//
	//	mySensors.push_back(SensorFactory::createSensor(DEPTH_IMAGE_SENSOR,
	//			"Depth Image Camera",
	//			myBodies[0],
	//			osg::Vec3f(4.5,0,2),
	//			osg::Vec3f(90,0,0),
	//			1.0,
	//			osg::Vec2(256,256)
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(LIGHT_SENSOR,
	//			"Light Sensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,-2,2),
	//			osg::Vec3f(90,0,0),
	//			0.5
	//	));
	//
	//a laser scanner
	//	mySensors.push_back(SensorFactory::createSensor(LASER_SCANNER,
	//			"LaserScanner",
	//			myBodies[0],
	//			osg::Vec3f(4.5,2,2),
	//			osg::Vec3f(90,0,0),
	//			1.0,
	//			osg::Vec2(3,3)									//(256,256)
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(SAMPLE_SENSOR,
	//			"TemperatureSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,-2,2),
	//			osg::Vec3f(90,0,0)
	//	));
	//



	//and a few distance sensors -on every side three
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Back Left DistanceSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5, -2, 2),
	//			osg::Vec3f(90,   0, 0),
	//			0.5
	//	));



	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Back Right DistanceSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,2,2),
	//			osg::Vec3f(90,0,0),
	//			0.5
	//	));
	//
	////	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////			"Back DistanceSensor",
	////			myBodies[0],
	////			osg::Vec3f(4.5,0,2),
	////			osg::Vec3f(90,0,0),
	////			0.5,
	////	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Front Left DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-4.5,-2,2),
	//			osg::Vec3f(-90,0,0),
	//			0.5
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Front Right DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-4.5,2,2),
	//			osg::Vec3f(-90,0,0),
	//			0.5
	//	));
	//
	////	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////			"Front DistanceSensor",
	////			myBodies[1],
	////			osg::Vec3f(-4.5,0,2),
	////			osg::Vec3f(-90,0,0),
	////			0.5,
	////	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Right Front DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-2,4.5,2),
	//			osg::Vec3f(180,0,0),
	//			0.5
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Right Back DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(2,4.5,2),
	//			osg::Vec3f(180,0,0),
	//			0.5
	//	));
	//
	////		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////				"Right DistanceSensor",
	////				myBodies[1],
	////				osg::Vec3f(0,4.5,2),
	////				osg::Vec3f(180,0,0),
	////				0.5,
	////
	////		));
	//
	//		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//				"Left Front DistanceSensor",
	//				myBodies[0],
	//				osg::Vec3f(-2,-4.5,2),
	//				osg::Vec3f(0,0,0),
	//				0.5
	//		));
	//
	//		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//				"Left Back DistanceSensor",
	//				myBodies[0],
	//				osg::Vec3f(2,-4.5,2),
	//				osg::Vec3f(0,0,0),
	//				0.5
	//		));
	//
	////		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////				"Left DistanceSensor",
	////				myBodies[0],
	////				osg::Vec3f(0,-4.5,2),
	////				osg::Vec3f(0,0,0),
	////				0.5,
	////
	////		));


}

/**
 * Just sets the internal callback function. Take care not to use it Callback
 * because that is used by OpenSceneGraph too.
 */
void RobotKIT::SetRotationEventFunction(RotationCallback func) {
	rotation_event_function = func;
}

/**
 * Calls the function with the proper parameter, namely the robot on which the
 * connection / disconnection event takes place.
 */
void RobotKIT::ExecuteRotationEventFunction() {
	if (rotation_event_function != NULL) {
		RobotActorBase &super = dynamic_cast<RobotActorBase&>(*this);
		rotation_event_function(super);
	}
}

/**
 * Trigger for a change in rotation.
 */
void RobotKIT::RotationTrigger() {
	osg::Vec3f rot = getRotation();
	int rotation =(int) rot.x();

	rotation += 45;
	assert (rotation >= -360);
	rotation = (rotation + 360) % 360;
	assert (rotation >= 0);

	char orientation_quadrant = -1;
	if (rotation < 90) orientation_quadrant = 0;
	else if (rotation < 180) orientation_quadrant = 1;
	else if (rotation < 270) orientation_quadrant = 2;
	else if (rotation <= 360) orientation_quadrant = 3;
	assert (orientation_quadrant >= 0);

	// Callback if orientation changed
	if (orientation_quadrant != this->orientation_quadrant) {
		ExecuteRotationEventFunction();
		this->orientation_quadrant = orientation_quadrant;
	}
}

void RobotKIT::TickLocal(const dtGame::Message &tickMessage)
{
	//#pragma omp parallel <-- does not work

	// get delta time of the last simulation step
	const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(tickMessage);
	RobotActorBase::TickLocal(tick);
	float deltaSimTime = tick.GetDeltaSimTime();
	deltaSimTime++;
	deltaSimTime--;
	//	std::vector<dtCore::RefPtr<srCore::ConnectorBase> > mConns = this->getConnectors();
	//
	//	for (int i=0; i<mConns.size(); i++) {
	//
	//		mConns[i]->SetScale(osg::Vec3(2,2,2));
	//
	//	}

	// get translation and rotation of the robot
	myPos = myBodies[0]->getTranslation();
	myRot = myBodies[0]->getRotation();

	RotationTrigger();

	//	for (unsigned int i=0; i<myConnectors.size(); i++) {
	////		myConnectors[i]->setScale(osg::Vec3(2, 2, 2));
	//
	//		dtCore::Transform xform;
	//		myConnectors[i]->GetTransform(xform);
	//
	//		std::cout << "dtCore: conns of robot " << GetName() << ": " << xform.GetTranslation() << std::endl;
	//		std::cout << "srCore: conns of robot " << GetName() << ": " << myConnectors[i]->getTranslation() << std::endl;
	//	}

	//	std::cout << "num parents: " << myBodies[0]->GetOSGNode()->getNumParents() << std::endl;
	//	std::cout << "body node mask: " << myBodies[0]->GetOSGNode()->getNodeMask() << ", coll node mask: " << myBodies[0]->getCollisionModel()->getNodeMask() << std::endl;


	// Anne audio

#ifdef AUDIO_PRINT
	dtCore::Transform xform;
	dynamic_cast<SymbrionBodyI*>(myBodies[0])->bodySound->GetTransform(xform);

	printf("sound pos: %f %f %f\n", xform.GetTranslation().x(), xform.GetTranslation().y(),
			xform.GetTranslation().z());

	dtAudio::AudioManager::GetListener()->GetTransform(xform);
	printf("listener pos: %f %f %f\n", xform.GetTranslation().x(), xform.GetTranslation().y(),
			xform.GetTranslation().z());
#endif


}

// Now we need to process messages that we'll get from components or other robot actors
// Components will be the GUI component, the Input component and controller components
///////////////////////////////////////////////////////////////////////////////
void RobotKIT::ProcessMessage(const dtGame::Message &message)
{
	//jointUpTorque = 0;
	//jointSideTorque = 0;

	if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT)
	{
		//get the game message and handle them accordingly
		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);
		//		printf("Robot %d:%s\n", myID, eventMsg.GetGameEvent()->GetName().c_str());

		if (eventMsg.GetGameEvent() != NULL)
		{


			if (eventMsg.GetGameEvent()->GetName() == SWITCH_CONTROL)
			{
				manual ^= 1;
				if (manual) {LOG_INFO("Switch control to manual.");}
				else {
					LOG_INFO("Switch control to Cortex.");
					myActuators[1]->setODEParameter(dParamVel, 0);
					myActuators[2]->setODEParameter(dParamVel, 0);
					myActuators[3]->setODEParameter(dParamVel, 0);
					myActuators[4]->setODEParameter(dParamVel, 0);
				}
			}

			if (this->GetUniqueId() == eventMsg.GetAboutActorId()) {

				// Connect
				if (eventMsg.GetGameEvent()->GetName() == CONNECT_ALL)
				{

					for (unsigned int i=0; i<myConnectors.size(); i++)
						myConnectors[i]->connect();

				}

				else if (eventMsg.GetGameEvent()->GetName() == DISCONNECT_ALL)
				{

					LOG_INFO("Got DisconnectAll command");

					// try to connect with every connector
					for (unsigned int i=0; i<myConnectors.size(); i++)
						if (i%2 == 0)
							myConnectors[i]->disconnect();


				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_UP)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					//rotating around local axis
					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0, 0, 0.01);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_DOWN)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);
					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0, 0,-0.01);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_LEFT)
				{

					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0.01, 0, 0);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_RIGHT)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(-0.01, 0, 0);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				/**
				 * This is one of the wheels. The other wheel is ACTUATOR3.
				 */
				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR1_FWD)
				{

					float value = myActuators[2]->getVel(deltaTime);

					value += 0.5;
					if (value > MAX_SPEED) value = MAX_SPEED;

					myActuators[2]->setVel(value, deltaTime);

				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR1_BWD)
				{
					float value = myActuators[2]->getVel(deltaTime);

					value -= 0.5;
					if (value < -MAX_SPEED) value = -MAX_SPEED;

					myActuators[2]->setVel(value, deltaTime);
				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR2_FWD)
				{
					//					float value;
					//					myActuators[0]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
					//
					//					if (value < 50) value += 5;
					//
					//					myActuators[0]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

					float value;
					myActuators[0]->getParameter("Angle", MAKE_PARAMETER_VALUE(&value));

					float PI = 3.1415;
					if (value < PI/2) value += 0.2;
					else value = PI/2;

					dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(value);

				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR2_BWD)
				{
					//					float value;
					//					myActuators[0]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
					//
					//					if (value > -50) value -= 5;
					//
					//					myActuators[0]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

					float value;
					myActuators[0]->getParameter("Angle", MAKE_PARAMETER_VALUE(&value));

					float PI = 3.1415;
					if (value > -PI/2) value -= 0.2;
					else value = -PI/2;

					dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(value);

				}

				/**
				 * The MOVE_ACTUATOR3_FWD command belongs to the mAccelerate event in simulationInputComponent.
				 * It tells the wheel or screw drive to set it to a higher speed.
				 */
				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR3_FWD)
				{
					float value = myActuators[1]->getVel(deltaTime);

					value += 0.5;
					if (value > MAX_SPEED) value = MAX_SPEED;

					std::ostringstream msg;
					msg << "Set velocity of " << myActuators[1]->GetUniqueId() << " on robot " << getID();
					msg << " to "  << value;
					LOG_INFO(msg.str());
					myActuators[1]->setVel(value, deltaTime);

					value = myActuators[2]->getVel(deltaTime);
					value += 0.5;
					if (value > MAX_SPEED) value = MAX_SPEED;
					msg.clear(); msg.str("");
					msg << "Set velocity of " << myActuators[2]->GetUniqueId() << " on robot " << getID();
					msg << " to "  << value;
					LOG_INFO(msg.str());
					myActuators[2]->setVel(value, deltaTime);

				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR3_BWD)
				{
					float value = myActuators[1]->getVel(deltaTime);
					value -= 0.5;
					if (value < -MAX_SPEED) value = -MAX_SPEED;
					myActuators[1]->setVel(value, deltaTime);

					value = myActuators[2]->getVel(deltaTime);
					value -= 0.5;
					if (value < -MAX_SPEED) value = -MAX_SPEED;
					myActuators[2]->setVel(value, deltaTime);
				}

			}

		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Initialize our robot then enter the world
void RobotKIT::OnEnteredWorld()
{

	Initialize();

	RobotActorBase::OnEnteredWorld();			//needs to be called after the initialization

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


//todo: program RobotActorProxy so that it can be used within STAGE
RobotKITProxy::RobotKITProxy()
{
	SetClassName("RobotKIT");
}

///////////////////////////////////////////////////////////////////////////////
void RobotKITProxy::BuildPropertyMap()
{

	// add id, position and orientation properties:
	RobotActorBaseProxy::BuildPropertyMap();

	//for adding other properties, do it like this:

	//	const std::string GROUP = "Robot";
	//
	//	dtGame::GameActorProxy::BuildPropertyMap();
	//	RobotActor &actor = static_cast<RobotRamon&>(GetGameActor());
	//

	//	dtDAL::GroupActorProperty* gap = 		new dtDAL::GroupActorProperty("connectors","connectors",
	//			dtDAL::MakeFunctor(actor,&RobotActor::setConnectors),
	//			dtDAL::MakeFunctorRet(actor,&RobotActor::getConnectors),
	//			"gets rotation and position of the robots connectors.",GROUP);
	//
	//
	//	AddProperty( gap );



}

///////////////////////////////////////////////////////////////////////////////
RobotKITProxy::~RobotKITProxy()
{
}

///////////////////////////////////////////////////////////////////////////////
void RobotKITProxy::CreateActor()
{
	SetActor(*new RobotKIT(*this));
}

///////////////////////////////////////////////////////////////////////////////
void RobotKITProxy::OnEnteredWorld()
{
	// Register an invokable for Game Events and
	// register an invokable for tick messages. Local or Remote only, not both!  --  will be done in the base class:

	RobotActorBaseProxy::OnEnteredWorld();

}

}

