/*
 * Robot3D Physical realistic 3D simulator for robots
 *
 * @license GNU Lesser General Public License
 *
 * @author Adrian Friebel
 * @author Lutz Winkler
 * @author Anne C. van Rossum
 */

// Robot3D files
#include <srRobot/robotKIT2.h>
#include <srFactory/bodyFactory.h>
#include <srFactory/actuatorFactory.h>
#include <srFactory/sensorFactory.h>
#include <srBody/led.h>
#include <srActuator/connectorUniKarl.h>
#include <srCore/simulationEvents.h>

#include <srCore/actuator/hingeActuator.h>
#include <srActuator/screwDrive2.h>

#define TIME_STEP				250

// 1 m/s ^= 20 (5cm)/s
#define MAX_SPEED				15

// Notes about measuring units:
// Replicator specification parameters (in mm, N, kg, m/s^2, )
// units need to be around one to make the ode most stable. Hereby 1 length union is equal to 5cm.

using namespace srCore;

namespace srRobot {

///////////////////////////////////////////////////////////////////////////////
const std::string RobotKIT2::EVENT_HANDLER_NAME("HandleGameEvent");
const float MAX_ROBOT_VELOCITY = 15.0f;

//! The default name (used in configuration files)
//const std::string RobotKIT::DEFAULT_NAME("RobotKIT");

///////////////////////////////////////////////////////////////////////////////
RobotKIT2::RobotKIT2(dtGame::GameActorProxy &proxy, float jointAngle) :
					//dtGame::GameActor(proxy),
					RobotActorBase(proxy),
					deltaTime(0.01),
					manual (true)
{
	jointAngles.push_back(0.0);
	SetName("RobotKIT2");
}

///////////////////////////////////////////////////////////////////////////////
RobotKIT2::~RobotKIT2()
{

}


void RobotKIT2::Initialize()
{

	// In the osgGroup are all Drawables included.
	// The osgGroup might be useful for creating shadows
	osgGroup = new osg::Group();

	LOG_INFO("Building UniKarl robot of type RobotKIT2");
	// create replicon or Robot3D

	// First, we'll create the bodies, the robot consists of
	// two main bodies ...
	addBody(BodyFactory::createBody(
			srCore::KIT_1, this,
			0, 0, 0,
			0, 180, 90));

	addBody(BodyFactory::createBody(
			srCore::KIT_2, this,
			0, 0, 0,
			0, 180 + jointAngles[0], 90));			//0, 180, 90

//	float WHEEL_Z_OFFSET = 1.0;
//	float WHEEL_OFFSET = 0.75;

	////	myObjects.push_back(new Led(
	////				myBodies[0],
	////				osg::Vec3(0, 0, WHEEL_Z_OFFSET + 0.1),
	////				osg::Vec3(0, 0, 0)));
	//
	//
	//	// Now we need to create some hinge joints to be able later to move the robot
	//
	//
	dtCore::RefPtr<BodyBase> body1 = this->myBodies[0];
	dtCore::RefPtr<BodyBase> body2 = this->myBodies[1];


	// our main hinge ...
	addActuator(srCore::ActuatorFactory::createActuator(
			srCore::HINGE_ACTUATOR,
			body1, body2,
			0, 0, 0,
			1, 0, 0,
			"Main Hinge"));

	dynamic_cast<HingeActuator*>(this->myActuators[0].get())->setZeroAngle(jointAngles[0] * M_PI / 180);
	//
	//
	//	// Now the screw drive:
	//	// As it consumes too much computational resources, the screws will be simulated by four wheels
	//	// The axes for our wheels:
	//
	//
	
	srCore::BodyBase* screw1 = BodyFactory::createBody(
				srCore::KIT_SCREW, this,
				0, -1, -1,
				180, 180, 90);
				
	srCore::BodyBase* screw2 = BodyFactory::createBody(
				srCore::KIT_SCREW, this,
				0, 1, -1,
				0, 180, 90);
			
	body1->mergeBody(screw1);
 	body1->ClearCollisionGeometry();
 	body1->SetCollisionMesh(body1->getCollisionModel());
	
	
	body2->mergeBody(screw2);
 	body2->ClearCollisionGeometry();
 	body2->SetCollisionMesh(body2->getCollisionModel());
	
	screw1->SetModelTranslation(osg::Vec3(42,20,0));
	screw2->SetModelTranslation(osg::Vec3(42,20,0));
	
	dtCore::RefPtr<ActuatorBase> leftScrew = new ScrewDrive2("LeftScrew");
	dtCore::RefPtr<ActuatorBase> rightScrew = new ScrewDrive2("RightScrew");

	(dynamic_cast<ScrewDrive2* >(leftScrew.get()))->init(body1, dynamic_cast<ScrewDrive2*>(rightScrew.get()), false);
	(dynamic_cast<ScrewDrive2* >(rightScrew.get()))->init(body2, dynamic_cast<ScrewDrive2*>(leftScrew.get()), true);


	addActuator(leftScrew);
	addActuator(rightScrew);

	dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(0);


	// robots should be able to connect to eachother. Therefore we need some connectors
	// On each side one.

	addConnector(new myConnector(this->myBodies[0],  0,  1.1  , 0, -90, 0, 0));
	addConnector(new myConnector(this->myBodies[1],  0,-1.1  , 0,  -90, 0, 180));
	addConnector(new myConnector(this->myBodies[1],  0,  0, 1.1, -90, 0, 90));
	addConnector(new myConnector(this->myBodies[0],  0, 0, -1.1,  -90, 0, -90));


	// reduce the size of our drawable objects that make the connectors visible
	//	for (unsigned int i=0; i<myConnectors.size(); i++)
	//		myConnectors[i]->setScale(osg::Vec3(2, 2, 2));





	//	dtCore::Transform tx(-10.0f,-10.0f,7.0f,0.0f,5.0f,2.0f);

	// a robot without sensors doesn't make much sense. So we put a few sensors to the robot as well:

	// a camera
	mySensors.push_back(SensorFactory::createSensor(CAMERA_SENSOR,
			"Back Camera",
			myBodies[0],
			osg::Vec3f(4.5,0,2),
			osg::Vec3f(0,0,-90),
			1.0,
			osg::Vec2(256,256)
	));

	//
	//	mySensors.push_back(SensorFactory::createSensor(DEPTH_IMAGE_SENSOR,
	//			"Depth Image Camera",
	//			myBodies[0],
	//			osg::Vec3f(4.5,0,2),
	//			osg::Vec3f(90,0,0),
	//			1.0,
	//			osg::Vec2(256,256)
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(LIGHT_SENSOR,
	//			"Light Sensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,-2,2),
	//			osg::Vec3f(90,0,0),
	//			0.5
	//	));
	//
	//a laser scanner
	//	mySensors.push_back(SensorFactory::createSensor(LASER_SCANNER,
	//			"LaserScanner",
	//			myBodies[0],
	//			osg::Vec3f(4.5,2,2),
	//			osg::Vec3f(90,0,0),
	//			1.0,
	//			osg::Vec2(3,3)									//(256,256)
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(SAMPLE_SENSOR,
	//			"TemperatureSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,-2,2),
	//			osg::Vec3f(90,0,0)
	//	));
	//



	//and a few distance sensors -on every side three
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Back Left DistanceSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5, -2, 2),
	//			osg::Vec3f(90,   0, 0),
	//			0.5
	//	));



	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Back Right DistanceSensor",
	//			myBodies[0],
	//			osg::Vec3f(4.5,2,2),
	//			osg::Vec3f(90,0,0),
	//			0.5
	//	));
	//
	////	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////			"Back DistanceSensor",
	////			myBodies[0],
	////			osg::Vec3f(4.5,0,2),
	////			osg::Vec3f(90,0,0),
	////			0.5,
	////	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Front Left DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-4.5,-2,2),
	//			osg::Vec3f(-90,0,0),
	//			0.5
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Front Right DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-4.5,2,2),
	//			osg::Vec3f(-90,0,0),
	//			0.5
	//	));
	//
	////	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////			"Front DistanceSensor",
	////			myBodies[1],
	////			osg::Vec3f(-4.5,0,2),
	////			osg::Vec3f(-90,0,0),
	////			0.5,
	////	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Right Front DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(-2,4.5,2),
	//			osg::Vec3f(180,0,0),
	//			0.5
	//	));
	//
	//	mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//			"Right Back DistanceSensor",
	//			myBodies[1],
	//			osg::Vec3f(2,4.5,2),
	//			osg::Vec3f(180,0,0),
	//			0.5
	//	));
	//
	////		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////				"Right DistanceSensor",
	////				myBodies[1],
	////				osg::Vec3f(0,4.5,2),
	////				osg::Vec3f(180,0,0),
	////				0.5,
	////
	////		));
	//
	//		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//				"Left Front DistanceSensor",
	//				myBodies[0],
	//				osg::Vec3f(-2,-4.5,2),
	//				osg::Vec3f(0,0,0),
	//				0.5
	//		));
	//
	//		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	//				"Left Back DistanceSensor",
	//				myBodies[0],
	//				osg::Vec3f(2,-4.5,2),
	//				osg::Vec3f(0,0,0),
	//				0.5
	//		));
	//
	////		mySensors.push_back(SensorFactory::createSensor(DISTANCE_SENSOR,
	////				"Left DistanceSensor",
	////				myBodies[0],
	////				osg::Vec3f(0,-4.5,2),
	////				osg::Vec3f(0,0,0),
	////				0.5,
	////
	////		));


	// ... and ready is our robot
	// Now we need to fill the robot with life. Therefore we need to write what the robot should do during the simulation.
	// This will be handled in TickLocal(). TickLocal will be called every simulation step.


}


///////////////////////////////////////////////////////////////////////////////
void RobotKIT2::TickLocal(const dtGame::Message &tickMessage)
{
	//#pragma omp parallel <-- does not work

	// get delta time of the last simulation step
	const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(tickMessage);
	RobotActorBase::TickLocal(tick);
	float deltaSimTime = tick.GetDeltaSimTime();
	deltaSimTime++;
	deltaSimTime--;
	//	std::vector<dtCore::RefPtr<srCore::ConnectorBase> > mConns = this->getConnectors();
	//
	//	for (int i=0; i<mConns.size(); i++) {
	//
	//		mConns[i]->SetScale(osg::Vec3(2,2,2));
	//
	//	}

	// get translation and rotation of the robot
	myPos = myBodies[0]->getTranslation();
	myRot = myBodies[0]->getRotation();


	//	for (unsigned int i=0; i<myConnectors.size(); i++) {
	////		myConnectors[i]->setScale(osg::Vec3(2, 2, 2));
	//
	//		dtCore::Transform xform;
	//		myConnectors[i]->GetTransform(xform);
	//
	//		std::cout << "dtCore: conns of robot " << GetName() << ": " << xform.GetTranslation() << std::endl;
	//		std::cout << "srCore: conns of robot " << GetName() << ": " << myConnectors[i]->getTranslation() << std::endl;
	//	}

	//	std::cout << "num parents: " << myBodies[0]->GetOSGNode()->getNumParents() << std::endl;
	//	std::cout << "body node mask: " << myBodies[0]->GetOSGNode()->getNodeMask() << ", coll node mask: " << myBodies[0]->getCollisionModel()->getNodeMask() << std::endl;


	// Anne audio

#ifdef AUDIO_PRINT
	dtCore::Transform xform;
	dynamic_cast<SymbrionBodyI*>(myBodies[0])->bodySound->GetTransform(xform);

	printf("sound pos: %f %f %f\n", xform.GetTranslation().x(), xform.GetTranslation().y(),
			xform.GetTranslation().z());

	dtAudio::AudioManager::GetListener()->GetTransform(xform);
	printf("listener pos: %f %f %f\n", xform.GetTranslation().x(), xform.GetTranslation().y(),
			xform.GetTranslation().z());
#endif


}

///////////////////////////////////////////////////////////////////////////////

// standard remote stuff is handled by the base class, overwrite it if necessary
//void RobotActor::TickRemote(const dtGame::Message &tickMessage)

//{
//	const dtGame::TickMessage &tick = static_cast<const dtGame::TickMessage&>(tickMessage);
//	float deltaSimTime = tick.GetDeltaSimTime();
//
//	// do NOT recompute velocity and turn rate since we don't own this tank!
//
//	//MoveTheRobot(deltaSimTime);
//}


// Now we need to process messages that we'll get from components or other robot actors
// Components will be the GUI component, the Input component and controller components
///////////////////////////////////////////////////////////////////////////////
void RobotKIT2::ProcessMessage(const dtGame::Message &message)
{
	//jointUpTorque = 0;
	//jointSideTorque = 0;

	if (message.GetMessageType() == dtGame::MessageType::INFO_GAME_EVENT)
	{
		//get the game message and handle them accordingly
		const dtGame::GameEventMessage &eventMsg = static_cast<const dtGame::GameEventMessage&>(message);
		//		printf("Robot %d:%s\n", myID, eventMsg.GetGameEvent()->GetName().c_str());

		if (eventMsg.GetGameEvent() != NULL)
		{


			if (eventMsg.GetGameEvent()->GetName() == SWITCH_CONTROL)
			{
				manual ^= 1;
				if (manual) {LOG_INFO("Switch control to manual.");}
				else {
					LOG_INFO("Switch control to Cortex.");
					myActuators[1]->setODEParameter(dParamVel, 0);
					myActuators[2]->setODEParameter(dParamVel, 0);
					myActuators[3]->setODEParameter(dParamVel, 0);
					myActuators[4]->setODEParameter(dParamVel, 0);
				}
			}

			if (this->GetUniqueId() == eventMsg.GetAboutActorId()) {

				// Connect
				if (eventMsg.GetGameEvent()->GetName() == CONNECT_ALL)
				{

					for (unsigned int i=0; i<myConnectors.size(); i++)
						myConnectors[i]->connect();

				}

				else if (eventMsg.GetGameEvent()->GetName() == DISCONNECT_ALL)
				{

					std::cout << "got disconnect all command" << std::endl;

					// try to connect with every connector
					for (unsigned int i=0; i<myConnectors.size(); i++)
						if (i%2 == 0)
							myConnectors[i]->disconnect();


				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_UP)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					//rotating around local axis
					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0, 0, 0.01);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_DOWN)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);
					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0, 0,-0.01);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_LEFT)
				{

					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(0.01, 0, 0);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == CAMERA1_MOVE_RIGHT)
				{
					osg::Vec3f data = osg::Vec3f(0,0,0);

					if (!mySensors.empty())
						mySensors[0]->getParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);

					dtUtil::Coordinates* coord = new dtUtil::Coordinates();
					//					coord->SetOriginRotation(data.x(), data.y(), data.z());
					data = coord->ConvertToLocalRotation(-0.01, 0, 0);

					if (!mySensors.empty())
						mySensors[0]->setParameter(
								"RelOrientation",
								MAKE_PARAMETER_VALUE(&data)
						);
				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR1_FWD)
				{

					float value = myActuators[2]->getVel(deltaTime);

					
					value += 5;
					//if (value > MAX_SPEED) value = MAX_SPEED;
std::cout << value << std::endl;
					myActuators[2]->setVel(value, deltaTime);

				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR1_BWD)
				{
					float value = myActuators[2]->getVel(deltaTime);

					value -= 5;
					//if (value < -MAX_SPEED) value = -MAX_SPEED;
std::cout << value << std::endl;
					myActuators[2]->setVel(value, deltaTime);
				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR2_FWD)
				{
					//					float value;
					//					myActuators[0]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
					//
					//					if (value < 50) value += 5;
					//
					//					myActuators[0]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

					float value;
					myActuators[0]->getParameter("Angle", MAKE_PARAMETER_VALUE(&value));

					float PI = 3.1415;
					if (value < PI/2) value += 0.2;
					else value = PI/2;

					dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(value);

				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR2_BWD)
				{
					//					float value;
					//					myActuators[0]->getParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));
					//
					//					if (value > -50) value -= 5;
					//
					//					myActuators[0]->setParameter("DesiredVelocity", MAKE_PARAMETER_VALUE(&value));

					float value;
					myActuators[0]->getParameter("Angle", MAKE_PARAMETER_VALUE(&value));

					float PI = 3.1415;
					if (value > -PI/2) value -= 0.2;
					else value = -PI/2;

					dynamic_cast<HingeActuator*>(myActuators[0].get())->gotoAngle(value);

				}


				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR3_FWD)
				{
					float value = myActuators[1]->getVel(deltaTime);

					value += 5;
					//if (value > MAX_SPEED) value = MAX_SPEED;
std::cout << value << std::endl;
					myActuators[1]->setVel(value, deltaTime);
				}

				else if (eventMsg.GetGameEvent()->GetName() == MOVE_ACTUATOR3_BWD)
				{
					float value = myActuators[1]->getVel(deltaTime);

					value -= 5;
					//if (value < -MAX_SPEED) value = -MAX_SPEED;
std::cout << value << std::endl;
					myActuators[1]->setVel(value, deltaTime);
				}

			}

		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Initialize our robot then enter the world
void RobotKIT2::OnEnteredWorld()
{

	Initialize();

	RobotActorBase::OnEnteredWorld();			//needs to be called after the initialization

}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


//todo: program RobotActorProxy so that it can be used within STAGE
RobotKIT2Proxy::RobotKIT2Proxy()
{
	SetClassName("RobotKIT2");
}

///////////////////////////////////////////////////////////////////////////////
void RobotKIT2Proxy::BuildPropertyMap()
{

	// add id, position and orientation properties:
	RobotActorBaseProxy::BuildPropertyMap();

	//for adding other properties, do it like this:

	//	const std::string GROUP = "Robot";
	//
	//	dtGame::GameActorProxy::BuildPropertyMap();
	//	RobotActor &actor = static_cast<RobotRamon&>(GetGameActor());
	//

	//	dtDAL::GroupActorProperty* gap = 		new dtDAL::GroupActorProperty("connectors","connectors",
	//			dtDAL::MakeFunctor(actor,&RobotActor::setConnectors),
	//			dtDAL::MakeFunctorRet(actor,&RobotActor::getConnectors),
	//			"gets rotation and position of the robots connectors.",GROUP);
	//
	//
	//	AddProperty( gap );



}

///////////////////////////////////////////////////////////////////////////////
RobotKIT2Proxy::~RobotKIT2Proxy()
{
}

///////////////////////////////////////////////////////////////////////////////
void RobotKIT2Proxy::CreateActor()
{
	SetActor(*new RobotKIT2(*this));
}

///////////////////////////////////////////////////////////////////////////////
void RobotKIT2Proxy::OnEnteredWorld()
{
	// Register an invokable for Game Events and
	// register an invokable for tick messages. Local or Remote only, not both!  --  will be done in the base class:

	RobotActorBaseProxy::OnEnteredWorld();

}

}

