#ifdef AUDIO_ON
/*!
 *
 * Implementation of the microphoneSensor
 * To use the microphone sensor in the simulator, the sound from the
 * simulator needs to be redirected back to an input device, from
 * which alcCaptureOpenDevice can capture sound.
 *
 * Redirection can be done using pulseaudio (linux)
 *
 *
 * @author Ted Schmidt
 *
 */

#include "microphoneSensor.h"

namespace srCore{

	MicrophoneSensor::MicrophoneSensor(BodyBase* attachedTo,osg::Vec3f pos, osg::Vec3f orient,
										const std::string &name)
	:
		SensorBase(attachedTo, pos, orient, name),
		freq(15625), 		// Audio frequency
		channels(2),		// Stereo or mono
		bits(16), 			// Bit rate, also change AL_FORMAT_... in alcCaptureOpenDevice
		bufferSize(7824), 	// 50% of the frequency needs to be big enough to hold data in between a tick_local
		newData(false)		// Flag for new data recorded
	{
			sensorReturnValueType = typeid(char*).name();
	}

	void MicrophoneSensor::setup()
	{
		// Init Capture
		if(bits == 8){
			if(channels == 2)
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_STEREO8, bufferSize);
			else
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_MONO8, bufferSize);
		}
		else
		{
			if(channels == 2)
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_STEREO16, bufferSize);
			else
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_MONO16, bufferSize);
		}

	   //Get the Listener from the Audio Manager and attach it to the Camera:
	   dtAudio::Listener* globalEar = dtAudio::AudioManager::GetInstance().GetListener();
	   this->AddChild(globalEar);
	   alcCaptureStart(mydevice);
	}

	void MicrophoneSensor::requestData()
	{}

	void MicrophoneSensor::resetMicrophone()
	{
		alcCaptureStop(mydevice);
		alcCaptureCloseDevice(mydevice);

		// Init Capture
		if(bits == 8){
			if(channels == 2)
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_STEREO8, bufferSize);
			else
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_MONO8, bufferSize);
		}
		else
		{
			if(channels == 2)
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_STEREO16, bufferSize);
			else
				mydevice = alcCaptureOpenDevice(NULL, freq, AL_FORMAT_MONO16, bufferSize);
		}
		 alcCaptureStart(mydevice);

	}
	/* After setting a parameter the microphone needs to be reset
	 *
	 */
	void MicrophoneSensor::getParameter(const std::string &name, ParameterReturnValue  parameter)
	{
		if (name == "bufferSize")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			*data = bufferSize;
		}
		else if (name == "channels")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			*data = channels;
		}
		else if (name == "freq")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			*data = freq;
		}
		else if (name == "bits")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			*data = bits;
		}
		else if (name == "newData")
		{
			//check parameter type
			if (parameter.type != typeid(bool*).name())
				printf("Parameter type needs to be bool*. Casting could lead to errors.  \n");
			bool *data =  reinterpret_cast<bool*> (parameter.pValue);
			*data = newData;
		}
		else SensorBase::getParameter(name, parameter);
	}

	void MicrophoneSensor::setParameter(const std::string &name, ParameterReturnValue  parameter)
	{
		if (name == "bufferSize")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			bufferSize = *data;
		}
		else if (name == "channels")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			channels = *data;
		}
		else if (name == "freq")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			freq = *data;
		}
		else if (name == "bits")
		{
			//check parameter type
			if (parameter.type != typeid(int*).name())
				printf("Parameter type needs to be int*. Casting could lead to errors.  \n");
			int *data =  reinterpret_cast<int*> (parameter.pValue);
			bits = *data;
		}
		else
			SensorBase::setParameter(name, parameter);
	}

	void MicrophoneSensor::evaluate(ParameterReturnValue srv)
	{
		newData = false;

		ALchar mybuf[bufferSize];
		ALint samples;

		int chunk = bufferSize / (channels * (bits / 8));

		if (srv.type!=typeid(char*).name())
				printf("Parameter type needs to be char*. Casting could lead to errors.  \n "
					"Parameter given: %s \n", srv.type.data());

		// Copy the samples from OpenAL to our buffer
		alcGetIntegerv(mydevice, ALC_CAPTURE_SAMPLES, 1, &samples);
		if (samples >  chunk)
		{
		   alcCaptureSamples(mydevice, (ALCvoid *)mybuf, chunk);

		   for(int i = 0; i < bufferSize; i++)
			   (reinterpret_cast<char*>(srv.pValue))[i] = mybuf[i];
		   newData = true;
		}
	}

	/*
	 * This method is useless for this sensor
	 */
	void MicrophoneSensor::getSensorValue(std::ostream& theStream) const
	{
		theStream << newData;
	}

	MicrophoneSensor::~MicrophoneSensor()
	{
		alcCaptureStop(mydevice);
		alcCaptureCloseDevice(mydevice);
	}
}
#endif
